#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "LDrawModel.h"
#include "LDrawSubModel.h"
#include "ModelMacros.h"
#include "TGLShape.h"
#include "TGLTorus.h"
#include "TGLTorusQ.h"
#include "TGLEighthSphere.h"
#include "TGLStudLogo.h"
#include "VectorKey.h"
#include <TCFoundation/TCDictionary.h>
#include <TCFoundation/TCStringArray.h>
#include <TCFoundation/mystring.h>
//#include <TCFoundation/TCStringObjectTree.h>

#ifdef WIN32
#pragma warning(push)
#pragma warning(disable:4100)
#endif // WIN32
#include <map>
#ifdef WIN32
#pragma warning(pop)
#endif // WIN32

const int LDM_QUAD_STRIPS = 0;
const int LDM_QUADS = 1;
const int LDM_TRIANGLES = 2;
const int LDM_LINES = 3;
const int LDM_LINE_STRIPS = 4;
const int LDM_TRANS_LINES = 5;
const int LDM_TRANS_LINE_STRIPS = 6;

static const int NUM_SHAPE_TYPES = 7;

static const int LO_NUM_SEGMENTS = 8;
static const int HI_NUM_SEGMENTS = 16;
//static const int HI_NUM_SEGMENTS = 48;

static const float POLYGON_OFFSET_FACTOR = 0.75f;
static const float POLYGON_OFFSET_UNITS = 0.25f;
static const int UNUSED_COLOR = -2;

#define GOLD_SCALE (255.0f / 240.0f * 2.0f)

//int				LDrawModel::lastColor = UNUSED_COLOR;
TCDictionary*	LDrawModel::loadedModels = NULL;
float			LDrawModel::seamWidth = 0.0f;
int				LDrawModel::usePolygonOffset = 1;
int				LDrawModel::useLighting = 1;
int				LDrawModel::drawWireframe = 0;
char*			LDrawModel::systemLDrawDir = NULL;
LDMSpecialColorSettings LDrawModel::specialColors[4] = {
	{255, {0.075f, 0.075f, 0.075f, -100.0f}, 15.0f},	// Black rubber
	{
		334, {
			240.0f / 255.0f * GOLD_SCALE,
			176.0f / 255.0f * GOLD_SCALE,
			51.0f / 255.0f * GOLD_SCALE
		},
		5.0f
	},													// Gold
	{383, {0.9f, 1.2f, 1.5f, -100.0f}, 5.0f},			// Chrome
	{494, {0.9f, 0.9f, 1.5f, -100.0f}, 5.0f}			// Electrical contacts
};
void (APIENTRY *LDrawModel::polygonOffsetFunc)(GLfloat, GLfloat) =
	glPolygonOffset;
//bool			LDrawModel::gotPolygonOffsetFunc = NO;

int totalModels = 0;

#ifdef _LOG_PERFORMANCE
LONGLONG parseDelta = 0;
LONGLONG parseTriangleDelta = 0;
LONGLONG parseQuadDelta = 0;
#endif // _LOG_PERFORMANCE

LDrawModel::LDrawModelCleanup LDrawModel::modelCleanup;

LDrawModel::LDrawModelCleanup::~LDrawModelCleanup(void)
{
	delete LDrawModel::systemLDrawDir;
	LDrawModel::systemLDrawDir = NULL;
	if (LDrawModel::loadedModels)
	{
		LDrawModel::loadedModels->release();
		LDrawModel::loadedModels = NULL;
	}
}

void LDrawModel::commonInit(void)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "LDrawModel");
#endif
	flags.cullBackFaces = 0;
	flags.compiled = 0;
//	flags.childrenComputed = 0;
	flags.allowLowResStuds = true;
	flags.drawHighlights = 0;
	flags.subModelIsPart = 0;
	flags.haveMainModel = 0;
	flags.isPart = 0;
	flags.isForcedPart = 0;
	flags.normalsComputed = 0;
	flags.allowPrimitiveSubstitution = 1;
	flags.useStipple = 0;
	flags.readCanceled = 0;
	flags.sortTransparent = 1;
	flags.blackHighlights = false;
	flags.cutawayDraw = false;
	flags.drawConditionalHighlights = false;
	flags.performSmoothing = true;
	flags.lineSmoothing = false;
	flags.compiling = false;
	totalModels++;
	topParent = this;
	initShapes();
}

void LDrawModel::initShapes(void)
{
	for (int i = 0; i < NUM_SHAPE_TYPES; i++)
	{
		TGLShapeArray* theseShapes = new TGLShapeArray;

		shapes->addObject(theseShapes);
		theseShapes->release();
	}
}

LDrawModel::LDrawModel(void)
		   :parent(NULL),
		    shapes(new TCObjectArrayArray(NUM_SHAPE_TYPES)),
			newSubModels(new LDrawSubModelArray),
//			subModels(NULL),
//			transformationMatrices(NULL),
//			colors(NULL),
//			colorNumbers(NULL),
//			subModelCount(0),
//			subModelsAllocated(0),
			fileLines(NULL),
			name(NULL),
			filename(NULL),
			progressCallback(NULL),
			errorCallback(NULL),
			currentFileLine(NULL),
			curveQuality(2),
			defaultColorNumber(7)
{
	commonInit();
}

LDrawModel::LDrawModel(FILE* file)
		   :LDrawModelFile(file),
		    parent(NULL),
		    shapes(new TCObjectArrayArray(NUM_SHAPE_TYPES)),
			newSubModels(new LDrawSubModelArray),
//			subModels(NULL),
//			transformationMatrices(NULL),
//			colors(NULL),
//			colorNumbers(NULL),
//			subModelCount(0),
//			subModelsAllocated(0),
			fileLines(NULL),
			name(NULL),
			filename(NULL),
			progressCallback(NULL),
			errorCallback(NULL),
			currentFileLine(NULL),
			curveQuality(2),
			defaultColorNumber(7)
{
	commonInit();
	read(file, 1);
}

LDrawModel::LDrawModel(LDrawModel& other)
		   :parent(other.parent),
		    topParent(other.topParent),
		    shapes((TCObjectArrayArray*)(other.shapes->copy())),
			newSubModels((LDrawSubModelArray*)(other.newSubModels->copy())),
//			subModels(new LDrawModel*[other.subModelCount]),
//			transformationMatrices(new float*[other.subModelCount]),
//			colors(new int[other.subModelCount]),
//			colorNumbers(new int[other.subModelCount]),
//			subModelCount(other.subModelCount),
//			subModelsAllocated(other.subModelCount),
			fileLines((LDMFileLineArray*)other.fileLines->retain()),
			name(copyString(other.name)),
			filename(copyString(other.filename)),
			progressCallback(other.progressCallback),
			progressUserData(other.progressUserData),
			errorCallback(other.errorCallback),
			errorUserData(other.errorUserData),
			currentFileLine(NULL),
			curveQuality(other.curveQuality),
			defaultColorNumber(other.defaultColorNumber),
			flags(other.flags)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "LDrawModel");
#endif
/*
	int i;

	for (i = 0; i < subModelCount; i++)
	{
		subModels[i] = new LDrawModel(*(other.subModels[i]));
		transformationMatrices[i] = new float[16];
		memcpy(transformationMatrices[i], other.transformationMatrices[i],
			16 * sizeof(float));
		colors[i] = other.colors[i];
		colorNumbers[i] = other.colorNumbers[i];
	}
*/
//	shapes = (TCObjectArrayArray*)(other.shapes->copy());
/*
	for (i = 0; i < NUM_SHAPE_TYPES; i++)
	{
		int j;
		TGLShapeArray* theseShapes = new TGLShapeArray;
		TGLShapeArray* otherShapes = (*other.shapes)[i];
		int count = otherShapes->getCount();

		shapes->addObject(theseShapes);
		theseShapes->release();
		for (j = 0; j < count; j++)
		{
			TGLShape *otherShape = (*otherShapes)[j]->copyShape();
			theseShapes->addObject(otherShape);
			otherShape->release();
		}
	}
*/
	totalModels++;
}

TCObject *LDrawModel::copy(void)
{
	return new LDrawModel(*this);
}

TGLShape* LDrawModel::getShape(int type, int offset)
{
	return (TGLShape*)(*getShapes(type))[offset];
}

TGLShapeArray* LDrawModel::getShapes(int type)
{
	return (*shapes)[type];
}

void LDrawModel::addShape(int type, TGLShape* shape)
{
	getShapes(type)->addObject(shape);
	shape->release();
}

LDrawModel::~LDrawModel(void)
{
}

int LDrawModel::getNumCircleSegments(void)
{
	return topParent->curveQuality * LO_NUM_SEGMENTS;
}

void LDrawModel::dealloc(void)
{
//	printf("LDrawModel::dealloc for %s\n", name ? name : "<null>");
//	checkError("LDrawModel::dealloc 1");
	if (shapes)
	{
		shapes->release();
	}
	if (newSubModels)
	{
		newSubModels->release();
	}
	//deleteSubModels();
	if (fileLines)
	{
		fileLines->release();
	}
	if (!parent/*name*/ && loadedModels)
	{
		loadedModels->release();
		loadedModels = NULL;
	}
	delete name;
	delete filename;
	if (flags.compiled)
	{
		glDeleteLists(listId, 1);
	}
	totalModels--;

//	checkError("LDrawModel::dealloc n");
	LDrawModelFile::dealloc();
}

void LDrawModel::setLDrawDir(const char* value)
{
	delete systemLDrawDir;
	systemLDrawDir = copyString(value);
}

bool LDrawModel::isColorNumberRGB(int colorNumber, TCByte r, TCByte g,
				  TCByte b)
{
	int pr, pg, pb, pa;

	if (colorNumber == 16 || colorNumber == 24)
	{
		return false;
	}
	getRGBA(colorNumber, pr, pg, pb, pa);
	if (pr == r && pg == g && pb == b)
	{
		return true;
	}
	else
	{
		return false;
	}
}

int LDrawModel::getColorNumberForRGB(TCByte r, TCByte g, TCByte b)
{
	if (r != 255 || g != 128 || b != 0)
	{
		int i;

		for (i = 0; i <= 27; i++)
		{
			if (isColorNumberRGB(i, r, g, b))
			{
				return i;
			}
		}
	}
	return 0x2000000 | ((unsigned long)r << 16) | ((unsigned long)g << 8) |
		(unsigned long)b;
}

void LDrawModel::setDefaultRGB(TCByte r, TCByte g, TCByte b)
{
	topParent->defaultColorNumber = getColorNumberForRGB(r, g, b);
}

void LDrawModel::setDefaultColorNumber(int value)
{
	topParent->defaultColorNumber = value;
}

/*
void LDrawModel::computeChildren(void)
{
	float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
								0.0f, 1.0f, 0.0f, 0.0f,
								0.0f, 0.0f, 1.0f, 0.0f,
								0.0f, 0.0f, 0.0f, 1.0f};

	computeChildren(identityMatrix);
}
*/

/*
void LDrawModel::computeChildren(float* matrix)
{
	int i;

	for (i = 0; i < subModelCount; i++)
	{
		float newMatrix[16];

		subModels[i] = new LDrawModel(*(subModels[i]));
		TGLShape::multMatrix(matrix, transformationMatrices[i], newMatrix);
		subModels[i]->compute(newMatrix);
	}
	flags.childrenComputed = 1;
}
*/

/*
void LDrawModel::computeShapes(TCObjectArray* shapes, float* matrix)
{
	int i;
	int count = getArrayCount(shapes);

	for (i = 0; i < count; i++)
	{
		((TGLShape*)(*shapes)[i])->transform(matrix);
	}
}
*/

/*
void LDrawModel::compute(float* matrix)
{
	computeShapes(quadStrips, matrix);
	computeShapes(quads, matrix);
	computeShapes(triangles, matrix);
	computeShapes(lines, matrix);
	computeChildren(matrix);
}
*/

/*
int LDrawModel::getLastColor(void)
{
	return lastColor;
}

void LDrawModel::setLastColor(int c)
{
	lastColor = c;
}
*/

void LDrawModel::setParent(LDrawModel* p)
{
	parent = p;
	setTopParent(parent->getTopParent());
}

void LDrawModel::setTopParent(LDrawModel* p)
{
	topParent = p;
}

bool LDrawModel::isTorusO(char *filename)
{
	return strlen(filename) == 12 && toupper(filename[0]) == 'T' &&
		isdigit(filename[1]) && isdigit(filename[2]) &&
		toupper(filename[3]) == 'O' && isdigit(filename[4]) &&
		isdigit(filename[7]) && isdigit(filename[6]) &&
		isdigit(filename[7]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::isTorusI(char *filename)
{
	return strlen(filename) == 12 && toupper(filename[0]) == 'T' &&
		isdigit(filename[1]) && isdigit(filename[2]) &&
		toupper(filename[3]) == 'I' && isdigit(filename[4]) &&
		isdigit(filename[7]) && isdigit(filename[6]) &&
		isdigit(filename[7]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::startsWithFraction(char *filename)
{
	return isdigit(filename[0]) && filename[1] == '-' && isdigit(filename[2]);
}

bool LDrawModel::isCyli(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "cyli.dat");
}

bool LDrawModel::isCyls(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "cyls.dat");
}

bool LDrawModel::isCyls2(char *filename)
{
	return strlen(filename) == 12 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "cyls2.dat");
}

bool LDrawModel::isDisc(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "disc.dat");
}

bool LDrawModel::isNdis(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "ndis.dat");
}

bool LDrawModel::isEdge(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "edge.dat");
}

bool LDrawModel::isCcyli(char *filename)
{
	return strlen(filename) == 12 && startsWithFraction(filename) &&
		stringHasCaseInsensitiveSuffix(filename, "ccyli.dat");
}

bool LDrawModel::is1DigitCon(char *filename)
{
	return strlen(filename) == 11 && startsWithFraction(filename) &&
		stringHasCaseInsensitivePrefix(filename + 3, "con") &&
		isdigit(filename[6]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::is2DigitCon(char *filename)
{
	return strlen(filename) == 12 && startsWithFraction(filename) &&
		stringHasCaseInsensitivePrefix(filename + 3, "con") &&
		isdigit(filename[6]) && isdigit(filename[7]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::isRing(char *filename)
{
	int len = strlen(filename);

	if (len >= 9 && len <= 12 &&
		stringHasCaseInsensitivePrefix(filename, "ring") &&
		stringHasCaseInsensitiveSuffix(filename, ".dat"))
	{
		int i;

		for (i = 4; i < len - 5; i++)
		{
			if (!isdigit(filename[i]))
			{
				return false;
			}
		}
		return true;
	}
	else
	{
		return false;
	}
}

bool LDrawModel::isFracRing(char *filename)
{
	return strlen(filename) == 12 && startsWithFraction(filename) &&
		stringHasCaseInsensitivePrefix(filename + 3, "ring") &&
		isdigit(filename[7]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::isFracRin(char *filename)
{
	return strlen(filename) == 12 && startsWithFraction(filename) &&
		stringHasCaseInsensitivePrefix(filename + 3, "rin") &&
		isdigit(filename[6]) && isdigit(filename[7]) &&
		stringHasCaseInsensitiveSuffix(filename, ".dat");
}

bool LDrawModel::isCon(char *filename)
{
	return is1DigitCon(filename) || is2DigitCon(filename);
}

float LDrawModel::startingFraction(char *filename)
{
	int top;
	int bottom;

	sscanf(filename, "%d", &top);
	sscanf(filename + 2, "%d", &bottom);
	return (float)top / (float)bottom;
}

void LDrawModel::performPrimitiveSubstitution(void)
{
	if ((flags.allowLowResStuds && strcasecmp(name, "stu2.dat") == 0))
	{
		convertStu2();
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu22.dat") == 0)
	{
		convertStu22();
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu22a.dat") == 0)
	{
		convertStu22(YES);
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu23.dat") == 0)
	{
		convertStu23();
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu23a.dat") == 0)
	{
		convertStu23(YES);
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu24.dat") == 0)
	{
		convertStu24();
	}
	else if (flags.allowLowResStuds && strcasecmp(name, "stu24a.dat") == 0)
	{
		convertStu24(YES);
	}
	else if (strcasecmp(name, "stud.dat") == 0)
	{
		convertStud(getNumCircleSegments());
	}
	else if (strcasecmp(name, "1-8sphe.dat") == 0)
	{
		convertEighthSphere();
	}
	else if (isCyli(name))
	{
		convertCylinder(startingFraction(name));
	}
	else if (isCyls(name))
	{
		convertCyls(startingFraction(name));
	}
	else if (isCyls2(name))
	{
		convertCyls2(startingFraction(name));
	}
	else if (isDisc(name))
	{
		convertDisk(startingFraction(name));
	}
	else if (isNdis(name))
	{
		convertNdis(startingFraction(name));
	}
	else if (isEdge(name))
	{
		convertEdge(startingFraction(name));
	}
	else if (isCcyli(name))
	{
		convertCcyli(startingFraction(name));
	}
	else if (strcasecmp(name, "4-4con12.dat") == 0)
	{
		// Special case screwy filename.
		convertCone(1.0f, 1);
	}
	else if (isCon(name))
	{
		int size;

		sscanf(name + 6, "%d", &size);
		convertCone(startingFraction(name), size);
	}
	else if (isRing(name))
	{
		int size;

		sscanf(name + 4, "%d", &size);
		convertRing(1.0f, size);
	}
	else if (isFracRing(name))
	{
		int size;

		sscanf(name + 7, "%d", &size);
		convertRing(startingFraction(name), size);
	}
	else if (isFracRin(name))
	{
		int size;

		sscanf(name + 6, "%d", &size);
		convertRing(startingFraction(name), size);
	}
	else if (isTorusO(name))
	{
		int numSegments;
		int size;

		sscanf(name + 1, "%d", &numSegments);
		sscanf(name + 4, "%d", &size);
		numSegments = getNumCircleSegments() * numSegments / HI_NUM_SEGMENTS;
		convertTorusO(numSegments, (float)size / 10000.0f);
	}
	else if (isTorusI(name))
	{
		int numSegments;
		int size;

		sscanf(name + 1, "%d", &numSegments);
		sscanf(name + 4, "%d", &size);
		numSegments = getNumCircleSegments() * numSegments / HI_NUM_SEGMENTS;
		convertTorusI(numSegments, (float)size / 10000.0f);
	}
}

void LDrawModel::setFilename(char* value)
{
	delete filename;
	filename = copyString(value);
	replaceStringCharacter(filename, '/', '\\');
}

void LDrawModel::setName(char* value)
{
	delete name;
	name = copyString(value);
	if (flags.allowPrimitiveSubstitution)
	{
		performPrimitiveSubstitution();
	}
}

Vector LDrawModel::rearrangeVector(const Vector& v, int x, int y, int z)
{
	Vector newVector = Vector(v.get(x), v.get(y), v.get(z));
	return newVector;
}

int LDrawModel::checkLinePoint(const Vector& linePoint, const Vector& point)
{
	Vector circlePoint = linePoint / linePoint.length();

	if (fEq(circlePoint.get(0), point.get(0)))
	{
		return 0;
	}
	else if (circlePoint.get(0) < point.get(0))
	{
		return -1;
	}
	return 1;
}

Vector LDrawModel::findLinePoint(const Vector& point)
{
	float delta = 0.5f;
	float current = 0.0f;
	Vector newPoint;

	while (1)
	{
		int result;

		newPoint[0] = 1.0f - current;
		newPoint[2] = 1.0f - newPoint[0];
		result = checkLinePoint(newPoint, point);
		if (result > 0)
		{
			if (delta < 0.0f)
			{
				delta *= -0.5f;
			}
		}
		else if (result < 0)
		{
			if (delta > 0.0f)
			{
				delta *= -0.5;
			}
		}
		else
		{
			return newPoint;
		}
		current += delta;
	}
}

Vector LDrawModel::calcIntersection(int i, int j, int num, Vector* zeroXPoints,
									Vector* zeroYPoints, Vector* zeroZPoints)
{
	Vector temp1, temp2, temp3, temp4, temp5, temp6;
	Vector p1, p2, p3;

	if (i + j == num)
	{
		return zeroXPoints[j];
	}
	else if (i == 0)
	{
		return zeroZPoints[num - j];
	}
	else if (j == 0)
	{
		return zeroYPoints[i];
	}
	temp1 = zeroYPoints[i];
	temp2 = zeroXPoints[num - i];
	temp3 = zeroZPoints[num - j];
	temp4 = zeroXPoints[j];
	temp5 = zeroYPoints[i + j];
	temp6 = zeroZPoints[num - i - j];
	return (temp1 + temp2 + temp3 + temp4 + temp5 + temp6 -
		zeroXPoints[0] - zeroYPoints[0] - zeroZPoints[0]) / 9.0f;
//	return Vector(1.0f, 1.0f, 1.0f);
}

void LDrawModel::addEighthSphere(const Vector& /*center*/, float /*radius*/,
								 int numSegments)
{
	Vector* zeroXPoints;
	Vector* zeroYPoints;
	Vector* zeroZPoints;
	int usedSegments = numSegments / 4;
	int i, j;
	Vector p1, p2, p3, p4;
	Vector circlePoint;
	bool useTriangles = false;
	TGLEighthSphere* eighthSphere = NULL;
	int numMainPoints = (usedSegments + 1) * (usedSegments + 1) - 1;
	int numExtraPoints = usedSegments * 3;
	int mainSpot = 0;
	int extraSpot = numMainPoints;

	zeroXPoints = new Vector[usedSegments + 1];
	zeroYPoints = new Vector[usedSegments + 1];
	zeroZPoints = new Vector[usedSegments + 1];
	for (i = 0; i <= usedSegments; i++)
	{
		float angle = 2.0f * (float)M_PI / numSegments * i;

		circlePoint[0] = (float)cos(angle);
		circlePoint[2] = (float)sin(angle);
		zeroYPoints[i][0] = 1.0f / ((float)tan(angle) + 1);
		zeroYPoints[i][1] = 0.0f;
		zeroYPoints[i][2] = 1.0f - zeroYPoints[i][0];
		zeroZPoints[i] = rearrangeVector(zeroYPoints[i], 2, 0, 1);
		zeroXPoints[i] = rearrangeVector(zeroYPoints[i], 1, 2, 0);
	}
	if (!useTriangles)
	{
		eighthSphere = new TGLEighthSphere;
		eighthSphere->setGlMode(GL_TRIANGLE_STRIP);
		eighthSphere->setNumPoints(numMainPoints + numExtraPoints);
//		eighthSphere->setEighthSphere();
		eighthSphere->setParent(this);
	}
	for (j = 0; j < usedSegments; j++)
	{
		int stripCount = usedSegments - j;
		
		for (i = 0; i < stripCount; i++)
		{
			if (i == 0 || useTriangles)
			{
				p1 = calcIntersection(i, j, usedSegments, zeroXPoints,
					zeroYPoints, zeroZPoints);
				p1 *= 1.0f / p1.length();
				if (!useTriangles)
				{
					eighthSphere->setPoint(mainSpot++, p1);
				}
			}
			p2 = calcIntersection(i + 1, j, usedSegments, zeroXPoints,
				zeroYPoints, zeroZPoints);
			p2 *= 1.0f / p2.length();
			if (!useTriangles)
			{
				eighthSphere->setPoint(mainSpot + 1/*2 * i + 2*/, p2);
/*
				if (i == 0 && j == 0)
				{
					p2[0] = 1.0f - p2[0];
					eighthSphere->setPoint(extraSpot++, p2);
				}
*/
			}

			p3 = calcIntersection(i, j + 1, usedSegments, zeroXPoints,
				zeroYPoints, zeroZPoints);
			p3 *= 1.0f / p3.length();
			if (!useTriangles)
			{
				eighthSphere->setPoint(mainSpot/*2 * i + 1*/, p3);
				mainSpot += 2;
				if (j == 0)
				{
					if (i == 0)
					{
						p2[2] = -p2[2];
						eighthSphere->setPoint(extraSpot++, p2);
					}
					p3[1] = -p3[1];
					eighthSphere->setPoint(extraSpot++, p3);
				}
			}
			if (useTriangles)
			{
				newTriangle(16, p1, p2, p3);
				if (i < usedSegments - j - 1)
				{
					p4 = calcIntersection(i + 1, j + 1, usedSegments,
						zeroXPoints, zeroYPoints, zeroZPoints);
					p4 *= 1.0f / p4.length();
					newTriangle(16, p2, p4, p3);
				}
			}
		}
	}
	if (!useTriangles)
	{
		for (i = 0; i < usedSegments; i++)
		{
			p1 = eighthSphere->getPoints()[numMainPoints + i];
			p1 = rearrangeVector(p1, 1, 2, 0);
			eighthSphere->setPoint(numMainPoints + usedSegments + i, p1);
			p1 = rearrangeVector(p1, 1, 2, 0);
			eighthSphere->setPoint(numMainPoints + usedSegments * 2 + i,
				p1);
		}
		addQuadStrip(eighthSphere);
		if (shouldLoadConditionalLines())
		{
			genEighthSphereConditionals(eighthSphere, numSegments);
		}
	}
//	writeToFile();
	delete[] zeroXPoints;
	delete[] zeroYPoints;
	delete[] zeroZPoints;
}

void LDrawModel::genEighthSphereConditionals(TGLEighthSphere *eighthSphere,
											 int numSegments)
{
	int usedSegments = numSegments / 4;
	int i, j;
	Vector p1, p2, p3, p4;
	Vector circlePoint;
	int numMainPoints = (usedSegments + 1) * (usedSegments + 1) - 1;
	int mainSpot = 0;
	Vector *points = eighthSphere->getPoints();

	for (j = 0; j < usedSegments; j++)
	{
		int stripCount = usedSegments - j;
		
		for (i = 0; i < stripCount; i++)
		{
			if (i > 0)
			{
				p3 = points[mainSpot - 1];
			}
			else
			{
				if (j > 0)
				{
					p3 = points[numMainPoints + 3 * usedSegments - j];
					p3 = points[mainSpot];
					p3[2] -= 0.1f;
				}
				else
				{
					p3 = points[mainSpot];
					p3[2] -= 0.1f;
//					p3 = points[numMainPoints];
				}
			}
			p4 = points[mainSpot + 2];
			p1 = points[mainSpot];
			p2 = points[mainSpot + 1];
			newConditionalLine(24, p1, p2, p3, p4);
			p3 = p1;
			p1 = p2;
			p2 = p4;
			if (i < stripCount - 1)
			{
				p4 = points[mainSpot + 3];
			}
			else
			{
//				p4 = points[numMainPoints + usedSegments + j + 1];
				p4 = points[mainSpot + 1];
				p4[0] -= 0.1f;
			}
			newConditionalLine(24, p1, p2, p3, p4);
			p1 = points[mainSpot];
			p2 = points[mainSpot + 2];
			p3 = points[mainSpot + 1];
			if (j == 0)
			{
//				p4 = points[numMainPoints + i + 1];
				p4 = points[mainSpot];
				p4[1] -= 0.1f;
			}
			else
			{
				p4 = points[eighthSphere->sphereIndex(i * 2 + 2, j - 1,
					usedSegments)];
//				p4 = points[eighthSphere->sphereIndex(i / 2 + 2, j - 1,
//					usedSegments)];
//				p4 = points[0];
			}
			newConditionalLine(24, p1, p2, p3, p4);
			mainSpot += 2;
		}
		mainSpot++;
	}
}

void LDrawModel::addCcyli(const Vector& center, float yRadius, float xzRadius,
						  int numSegments, int usedSegments)
{
	int i, j;
	Vector p1, p2;
	Vector top = center;
	Vector zeroVector = Vector(0.0f, 0.0f, 0.0f);
	TGLTorus* torus = new TGLTorus;

	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	torus->setNumPoints((numSegments * 2 + 6) * (usedSegments + 2));
	torus->setGlMode(GL_QUAD_STRIP);
//	torus->setTorus();
	torus->setParent(this);
	for (i = -1; i < usedSegments + 1; i++)
	{
		float yAngle1, yAngle2;

		yAngle1 = 2.0f * (float)M_PI / numSegments * i;
		yAngle2 = 2.0f * (float)M_PI / numSegments * (i + 1);
		for (j = -1; j <= numSegments + 1; j++)
		{
			float xzAngle;
			float currentRadius;

			xzAngle = 2.0f * (float)M_PI / numSegments * j;
			top[1] = xzRadius * (float)sin(xzAngle) + center.get(1);
			currentRadius = xzRadius * (float)cos(xzAngle) + yRadius;
			setCirclePoint(yAngle1, currentRadius, top, p1);
			setCirclePoint(yAngle2, currentRadius, top, p2);
			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2, p1);
			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2 + 1,
				p2);
		}
	}
	addQuadStrip(torus);
	if (shouldLoadConditionalLines())
	{
		genTorusConditionals(torus, numSegments, usedSegments);
	}
}

void LDrawModel::genTorusConditionals(TGLTorus *torus, int numSegments,
									  int usedSegments)
{
	int i, j;
	Vector p1, p2, p3, p4;
	Vector *points = torus->getPoints();

	for (i = 0; i <= usedSegments; i++)
	{
		for (j = 0; j < numSegments; j++)
		{
			p1 = points[(i + 1) * (numSegments + 3) * 2 + (j + 1) * 2];
			p2 = points[(i + 1) * (numSegments + 3) * 2 + (j + 2) * 2];
			p3 = points[(i + 0) * (numSegments + 3) * 2 + (j + 1) * 2];
			p4 = points[(i + 1) * (numSegments + 3) * 2 + (j + 1) * 2 + 1];
			if (i == 0)
			{
				p3 = p1;
				p3[2] -= 0.1f;
			}
			else if (i == usedSegments)
			{
				p4 = p1;
				p4[0] -= 0.1f;
			}
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
	for (i = 0; i < usedSegments; i++)
	{
		for (j = 0; j <= numSegments; j++)
		{
			p1 = points[(i + 1) * (numSegments + 3) * 2 + (j + 1) * 2];
			p2 = points[(i + 1) * (numSegments + 3) * 2 + (j + 1) * 2 + 1];
			p3 = points[(i + 1) * (numSegments + 3) * 2 + (j + 0) * 2];
			p4 = points[(i + 1) * (numSegments + 3) * 2 + (j + 2) * 2];
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
}

void LDrawModel::addTorusO(const Vector& center, float yRadius, float xzRadius,
						   int numSegments, int usedSegments)
{
	int i, j;
	Vector p1, p2;
	Vector top = center;
	Vector zeroVector = Vector(0.0f, 0.0f, 0.0f);
	TGLTorusQ* torus = new TGLTorusQ;

	torus->setNumPoints((usedSegments * 2 + 6) * (usedSegments + 2));
	torus->setGlMode(GL_QUAD_STRIP);
//	torus->setTorusQ();
	torus->setParent(this);
	for (i = -1; i < usedSegments + 1; i++)
	{
		float yAngle1, yAngle2;

		yAngle1 = 2.0f * (float)M_PI / numSegments * i;
		yAngle2 = 2.0f * (float)M_PI / numSegments * (i + 1);
//		for (j = usedSegments - 1; j <= (usedSegments * 2) + 1; j++)
		for (j = -1; j <= usedSegments + 1; j++)
		{
			float xzAngle;
			float currentRadius;

			xzAngle = 2.0f * (float)M_PI / numSegments * j;
			top[1] = xzRadius * (float)sin(xzAngle) + center.get(1);
			currentRadius = xzRadius * (float)cos(xzAngle) + yRadius;
			setCirclePoint(yAngle1, currentRadius, top, p1);
			setCirclePoint(yAngle2, currentRadius, top, p2);
//			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2, p1);
//			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2 + 1,
//				p2);
			torus->setPoint((i + 1) * (usedSegments + 3) * 2 + (j + 1) * 2, p1);
			torus->setPoint((i + 1) * (usedSegments + 3) * 2 + (j + 1) * 2 + 1,
				p2);
		}
	}
	addQuadStrip(torus);
	if (shouldLoadConditionalLines())
	{
		genTorusOConditionals(torus, usedSegments, center, xzRadius);
	}
}

void LDrawModel::genTorusOConditionals(TGLTorusQ *torus, int usedSegments,
									   const Vector& center, float height)
{
	int i, j;
	Vector p1, p2, p3, p4;
	Vector *points = torus->getPoints();

	for (i = 1; i <= usedSegments + 1; i++)
	{
		for (j = 1; j <= usedSegments; j++)
		{
			p1 = points[i * (usedSegments + 3) * 2 + j * 2];
			p2 = points[i * (usedSegments + 3) * 2 + (j + 1) * 2];
			p3 = points[(i - 1) * (usedSegments + 3) * 2 + j * 2];
			p4 = points[i * (usedSegments + 3) * 2 + j * 2 + 1];
			if (i == 1)
			{
				p3 = p1;
				p3[2] -= 0.1f;
			}
			else if (i == usedSegments + 1)
			{
				p4 = p1;
				p4[0] -= 0.1f;
			}
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
	for (i = 1; i <= usedSegments; i++)
	{
		for (j = 1; j <= usedSegments + 1; j++)
		{
			p1 = points[i * (usedSegments + 3) * 2 + j * 2];
			p2 = points[i * (usedSegments + 3) * 2 + j * 2 + 1];
			p3 = points[i * (usedSegments + 3) * 2 + (j - 1) * 2];
			p4 = points[i * (usedSegments + 3) * 2 + (j + 1) * 2];
			if (j == 1)
			{
				p3 = p1;
				p3[1] -= 0.1f;
			}
			else if (j == usedSegments + 1)
			{
				p4 = center;
				p4[1] = height;
			}
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
}

void LDrawModel::addTorusI(const Vector& center, float yRadius, float xzRadius,
						   int numSegments, int usedSegments)
{
	int i, j;
	Vector p1, p2;
	Vector top = center;
	Vector zeroVector = Vector(0.0f, 0.0f, 0.0f);
	TGLTorusQ* torus = new TGLTorusQ;

	torus->setNumPoints((usedSegments * 2 + 6) * (usedSegments + 2));
	torus->setGlMode(GL_QUAD_STRIP);
//	torus->setTorusQ();
	torus->setParent(this);
	for (i = -1; i < usedSegments + 1; i++)
	{
		float yAngle1, yAngle2;

		yAngle1 = 2.0f * (float)M_PI / numSegments * i;
		yAngle2 = 2.0f * (float)M_PI / numSegments * (i + 1);
//		for (j = usedSegments - 1; j <= (usedSegments * 2) + 1; j++)
		for (j = -1; j <= usedSegments + 1; j++)
		{
			float xzAngle;
			float currentRadius;

			xzAngle = 2.0f * (float)M_PI / numSegments * j + (float)M_PI / 2.0f;
			top[1] = xzRadius * (float)sin(xzAngle) + center.get(1);
			currentRadius = xzRadius * (float)cos(xzAngle) + yRadius;
			setCirclePoint(yAngle1, currentRadius, top, p1);
			setCirclePoint(yAngle2, currentRadius, top, p2);
//			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2, p1);
//			torus->setPoint((i + 1) * (numSegments + 3) * 2 + (j + 1) * 2 + 1,
//				p2);
			torus->setPoint((i + 1) * (usedSegments + 3) * 2 + (j + 1) * 2, p1);
			torus->setPoint((i + 1) * (usedSegments + 3) * 2 + (j + 1) * 2 + 1,
				p2);
		}
	}
	addQuadStrip(torus);
	if (shouldLoadConditionalLines())
	{
		genTorusIConditionals(torus, numSegments, usedSegments, center, yRadius,
			xzRadius);
	}
}

void LDrawModel::genTorusIConditionals(TGLTorusQ *torus, int numSegments,
									   int usedSegments, const Vector& center,
									   float radius, float height)
{
	int i, j;
	Vector p1, p2, p3, p4;
	Vector *points = torus->getPoints();
	Vector top = center;
	top[1] = height;

	for (i = 1; i <= usedSegments + 1; i++)
	{
		for (j = 1; j <= usedSegments; j++)
		{
			p1 = points[i * (usedSegments + 3) * 2 + j * 2];
			p2 = points[i * (usedSegments + 3) * 2 + (j + 1) * 2];
			p3 = points[(i - 1) * (usedSegments + 3) * 2 + j * 2];
			p4 = points[i * (usedSegments + 3) * 2 + j * 2 + 1];
			if (i == 1)
			{
				p3 = p1;
				p3[2] -= 0.1f;
			}
			else if (i == usedSegments + 1)
			{
				p4 = p1;
				p4[0] -= 0.1f;
			}
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
	for (i = 1; i <= usedSegments; i++)
	{
		for (j = 1; j <= usedSegments + 1; j++)
		{
			p1 = points[i * (usedSegments + 3) * 2 + j * 2];
			p2 = points[i * (usedSegments + 3) * 2 + j * 2 + 1];
			p3 = points[i * (usedSegments + 3) * 2 + (j - 1) * 2];
			p4 = points[i * (usedSegments + 3) * 2 + (j + 1) * 2];
			if (j == 1)
			{
				float angle = 2.0f * (float)M_PI / numSegments * (i - 1);

				setCirclePoint(angle, radius + 0.1f, top, p3);
				p3[1] = height;
			}
			else if (j == usedSegments + 1)
			{
				p4 = p1;
				p4[1] -= 0.1f;
			}
			newConditionalLine(24, p1, p2, p3, p4);
		}
	}
}

void LDrawModel::addCylinder(const Vector& center, float radius, float height,
							 int numSegments, int usedSegments)
{
	addCone(center, radius, radius, height, numSegments, usedSegments);
}

void LDrawModel::setCirclePoint(float angle, float radius, const Vector& center,
								Vector& point)
{
	float x1, z1;

	x1 = radius * (float)cos(angle);
	z1 = radius * (float)sin(angle);
	point[0] = center.get(0) + x1;
	point[1] = center.get(1);
	point[2] = center.get(2) + z1;
}

void LDrawModel::addCyls(const Vector& center, float radius, float height,
						 int numSegments, int usedSegments)
{
	TGLShape* cone = new TGLShape;
	int i;
	Vector p1, p2, p3, p4;
	Vector top = center;

	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	cone->setNumPoints(usedSegments * 2 + 2);
	cone->setGlMode(GL_QUAD_STRIP);
	cone->setParent(this);
	cone->setNumControlPoints(4);
	for (i = -1; i <= usedSegments + 1; i++)
	{
		float angle;

		angle = 2.0f * (float)M_PI / numSegments * i;
		setCirclePoint(angle, radius, center, p1);
		top[1] = center.get(1) + height - ((height / radius) * p1[0]);
		setCirclePoint(angle, radius, top, p2);
		if (i < 0)
		{
			cone->setControlPoint(0, p1);
			cone->setControlPoint(1, p2);
		}
		else if (i > usedSegments)
		{
			cone->setControlPoint(2, p1);
			cone->setControlPoint(3, p2);
		}
		else
		{
			cone->setPoint(i * 2, p1);
			cone->setPoint(i * 2 + 1, p2);
		}
	}
	addQuadStrip(cone);
	if (shouldLoadConditionalLines() && !fEq(height, 0.0f))
	{
		genCylsConditionals(cone, numSegments, usedSegments);
	}
}

void LDrawModel::genCylsConditionals(TGLShape *cone, int numSegments,
									 int usedSegments)
{
	Vector p1, p2, p3, p4;
	int i;
	Vector *points = cone->getPoints();
	int axis2 = 0;

	if (numSegments == usedSegments * 2)
	{
		axis2 = 2;
	}
	for (i = 1; i <= usedSegments; i++)
	{
		p1 = points[i * 2];
		p2 = points[i * 2 + 1];
		p3 = points[(i - 1) * 2];
		if (i == usedSegments)
		{
			if (numSegments == usedSegments)
			{
				p4 = cone->getControlPoints()[2];
			}
			else
			{
				p4 = p1;
				if (usedSegments * 8 == numSegments)
				{
					p4[0] -= 0.1f;
					p4[2] += 0.1f;
				}
				else if (usedSegments * 8 == numSegments * 3)
				{
					p4[0] -= 0.1f;
					p4[2] -= 0.1f;
				}
				else if (usedSegments * 4 == numSegments * 3)
				{
					p4[0] += 0.1f;
				}
				else
				{
					p4[axis2] -= 0.1f;
				}
			}
		}
		else
		{
			p4 = points[(i + 1) * 2];
		}
		newConditionalLine(24, p1, p2, p3, p4);
	}
}

void LDrawModel::addCyls2(const Vector& center, float radius, float height,
						  int numSegments, int usedSegments)
{
	TGLShape* cone = new TGLShape;
	int i;
	Vector p1, p2, p3, p4;
	Vector top = center;

	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	cone->setNumPoints(usedSegments * 2 + 2);
	cone->setGlMode(GL_QUAD_STRIP);
	cone->setParent(this);
	cone->setNumControlPoints(4);
	for (i = -1; i <= usedSegments + 1; i++)
	{
		float angle;

		angle = 2.0f * (float)M_PI / numSegments * i + (float)M_PI / 2.0f;
		setCirclePoint(angle, radius, center, p1);
		top[1] = abs(p1[0]);
		setCirclePoint(angle, radius, top, p2);
		if (i < 0)
		{
			cone->setControlPoint(0, p1);
			cone->setControlPoint(1, p2);
		}
		else if (i > usedSegments)
		{
			cone->setControlPoint(2, p1);
			cone->setControlPoint(3, p2);
		}
		else
		{
			cone->setPoint(i * 2, p1);
			cone->setPoint(i * 2 + 1, p2);
		}
	}
	addQuadStrip(cone);
	if (shouldLoadConditionalLines() && !fEq(height, 0.0f))
	{
		genCyls2Conditionals(cone, numSegments, usedSegments);
	}
}

void LDrawModel::genCyls2Conditionals(TGLShape *cone, int numSegments,
									  int usedSegments)
{
	Vector p1, p2, p3, p4;
	int i;
	Vector *points = cone->getPoints();
	int axis2 = 2;

	if (numSegments == usedSegments * 2)
	{
		axis2 = 0;
	}
	for (i = 1; i <= usedSegments; i++)
	{
		p1 = points[i * 2];
		p2 = points[i * 2 + 1];
		p3 = points[(i - 1) * 2];
		if (i == usedSegments)
		{
			if (numSegments == usedSegments)
			{
				p4 = cone->getControlPoints()[2];
			}
			else
			{
				p4 = p1;
				if (usedSegments * 8 == numSegments)
				{
					p4[0] += 0.1f;
					p4[2] -= 0.1f;
				}
				else if (usedSegments * 8 == numSegments * 3)
				{
					p4[0] -= 0.1f;
					p4[2] -= 0.1f;
				}
				else if (usedSegments * 4 == numSegments * 3)
				{
					p4[2] += 0.1f;
				}
				else
				{
					p4[axis2] -= 0.1f;
				}
			}
		}
		else
		{
			p4 = points[(i + 1) * 2];
		}
		newConditionalLine(24, p1, p2, p3, p4);
	}
}

void LDrawModel::addRing(const Vector& center, float radius1, float radius2,
						 int numSegments, int usedSegments)
{
	addCone(center, radius1, radius2, 0.0f, numSegments, usedSegments);
}

void LDrawModel::addCone(const Vector& center, float radius1, float radius2,
						 float height, int numSegments, int usedSegments)
{
	TGLShape* cone = new TGLShape;
	int i;
	Vector p1, p2, p3, p4;
	Vector top = center;

	top[1] += height;
	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	cone->setNumPoints(usedSegments * 2 + 2);
	cone->setGlMode(GL_QUAD_STRIP);
	cone->setParent(this);
	cone->setNumControlPoints(4);
	for (i = -1; i <= usedSegments + 1; i++)
	{
		float angle;

		angle = 2.0f * (float)M_PI / numSegments * i;
		setCirclePoint(angle, radius2, top, p1);
		setCirclePoint(angle, radius1, center, p2);
		if (i < 0)
		{
			cone->setControlPoint(0, p1);
			cone->setControlPoint(1, p2);
		}
		else if (i > usedSegments)
		{
			cone->setControlPoint(2, p1);
			cone->setControlPoint(3, p2);
		}
		else
		{
			cone->setPoint(i * 2, p1);
			cone->setPoint(i * 2 + 1, p2);
		}
	}
	addQuadStrip(cone);
	if (shouldLoadConditionalLines() && !fEq(height, 0.0f))
	{
		genConeConditionals(cone, numSegments, usedSegments);
	}
}

void LDrawModel::genConeConditionals(TGLShape *cone, int numSegments,
									 int usedSegments)
{
	Vector p1, p2, p3, p4;
	int i;
	Vector *points = cone->getPoints();
	int axis1 = 2;
	int axis2 = 0;

	if (numSegments == usedSegments * 2)
	{
		axis2 = 2;
	}
	for (i = 0; i <= usedSegments; i++)
	{
		p1 = points[i * 2];
		p2 = points[i * 2 + 1];
		if (i == 0)
		{
			if (numSegments == usedSegments)
			{
				p3 = cone->getControlPoints()[0];
			}
			else
			{
				p3 = p1;
				p3[axis1] -= 0.1f;
			}
		}
		else
		{
			p3 = points[(i - 1) * 2];
		}
		if (i == usedSegments)
		{
			if (numSegments == usedSegments)
			{
				p4 = cone->getControlPoints()[2];
			}
			else
			{
				p4 = p1;
				if (usedSegments * 8 == numSegments)
				{
					p4[0] -= 0.1f;
					p4[2] += 0.1f;
				}
				else if (usedSegments * 8 == numSegments * 3)
				{
					p4[0] -= 0.1f;
					p4[2] -= 0.1f;
				}
				else if (usedSegments * 4 == numSegments * 3)
				{
					p4[0] += 0.1f;
				}
				else
				{
					p4[axis2] -= 0.1f;
				}
			}
		}
		else
		{
			p4 = points[(i + 1) * 2];
		}
		newConditionalLine(24, p1, p2, p3, p4);
	}
}

void LDrawModel::addDisk(const Vector& center, float radius, int numSegments,
						 int usedSegments)
{
	TGLShape* disk = new TGLShape;
	int i;
	Vector p1;

	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	disk->setNumPoints(usedSegments + 2);
	disk->setGlMode(GL_TRIANGLE_FAN);
	disk->setPoint(0, center);
	disk->setParent(this);
	for (i = 0; i <= usedSegments; i++)
	{
		float x, z;
		float angle;

		angle = 2.0f * (float)M_PI / numSegments * i;
		x = radius * (float)cos(angle);
		z = radius * (float)sin(angle);
		p1[0] = center.get(0) + x;
		p1[2] = center.get(2) + z;
		p1[1] = center.get(1);
		disk->setPoint(i + 1, p1);
	}
	addQuadStrip(disk);
}

void LDrawModel::addNotDisk(const Vector& center, float radius, int numSegments,
							int usedSegments)
{
	int quarter = numSegments / 4;
	int numQuarters;
	int i;
	Vector p1;

	if (usedSegments == -1)
	{
		usedSegments = numSegments;
	}
	numQuarters = (usedSegments + quarter - 1) / quarter;
	for (i = 0; i < numQuarters; i++)
	{
		TGLShape* ndisk = new TGLShape;
		int j;
		float zMult = 1.0f;
		float xMult = 1.0f;
		int quarterSegments = quarter;

		if (i >= 2)
		{
			zMult = -1.0f;
		}
		if (i == 1 || i == 2)
		{
			xMult = -1.0f;
		}
		if (i == numQuarters - 1)
		{
			quarterSegments = usedSegments % quarter;
			if (!quarterSegments)
			{
				quarterSegments = quarter;
			}
		}
		ndisk->setNumPoints(quarterSegments + 2);
		ndisk->setGlMode(GL_TRIANGLE_FAN);
		ndisk->setPoint(0, center + Vector(xMult * radius, 0.0f,
			zMult * radius));
		ndisk->setParent(this);
		for (j = 0; j <= quarterSegments; j++)
		{
			float x, z;
			float angle;

			angle = 2.0f * (float)M_PI / numSegments *
				(j + i * quarterSegments);
			x = radius * (float)cos(angle);
			z = radius * (float)sin(angle);
			p1[0] = center.get(0) + x;
			p1[2] = center.get(2) + z;
			p1[1] = center.get(1);
			ndisk->setPoint(quarterSegments - j + 1, p1);
		}
		addQuadStrip(ndisk);
	}
}

void LDrawModel::addEdge(const Vector& center, float radius, int numSegments,
						 int usedSegments)
{
	if (flags.drawHighlights)
	{
		TGLShape* edge = new TGLShape;
		int i;
		Vector p1;

		if (usedSegments == -1)
		{
			usedSegments = numSegments;
		}
		edge->setNumPoints(usedSegments + 1);
		edge->setGlMode(GL_LINE_STRIP);
		edge->setColorNumber(24);
		edge->setParent(this);
		for (i = 0; i <= usedSegments; i++)
		{
			float x, z;
			float angle;

			angle = 2.0f * (float)M_PI / numSegments * i;
			x = radius * (float)cos(angle);
			z = radius * (float)sin(angle);
			p1[0] = center.get(0) + x;
			p1[2] = center.get(2) + z;
			p1[1] = center.get(1);
			edge->setPoint(i, p1);
		}
		addLineStrip(edge);
	}
}

void LDrawModel::clearModel(void)
{
	if (shapes)
	{
		shapes->release();
		shapes = new TCObjectArrayArray(NUM_SHAPE_TYPES);
		initShapes();
	}
	if (newSubModels)
	{
		newSubModels->removeAll();
	}
}

int LDrawModel::numCylinderSegments(void)
{
	if (flags.allowLowResStuds)
	{
		return 8;
	}
	else
	{
		return getNumCircleSegments();
	}

}

void LDrawModel::convertStu2(void)
{
	convertStud(LO_NUM_SEGMENTS);
}

void LDrawModel::convertStud(int numSegments)
{
	TGLStudLogo *studLogo = new TGLStudLogo;

	clearModel();
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 6.0f, 4.0f, numSegments);
	studLogo->setParent(this);
	studLogo->setDisk(Vector(0.0f, -4.0f, 0.0f), 6.0f, numSegments);
	addQuadStrip(studLogo);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 6.0f, numSegments);
	addEdge(Vector(0.0f, 0.0f, 0.0f), 6.0f, numSegments);
}

void LDrawModel::convertStu22(bool isA)
{
	int numSegments = LO_NUM_SEGMENTS;

	clearModel();
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 4.0f, 4.0f, numSegments);
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 6.0f, 4.0f, numSegments);
	addCone(Vector(0.0f, -4.0f, 0.0f), 6.0f, 4.0f, 0.0f, numSegments);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 4.0f, numSegments);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 6.0f, numSegments);
	if (!isA)
	{
		addEdge(Vector(0.0f, 0.0f, 0.0f), 4.0f, numSegments);
		addEdge(Vector(0.0f, 0.0f, 0.0f), 6.0f, numSegments);
	}
}

void LDrawModel::convertStu23(bool isA)
{
	int numSegments = LO_NUM_SEGMENTS;

	clearModel();
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 4.0f, 4.0f, numSegments);
	addDisk(Vector(0.0f, -4.0f, 0.0f), 4.0f, numSegments);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 4.0f, numSegments);
	if (!isA)
	{
		addEdge(Vector(0.0f, 0.0f, 0.0f), 4.0f, numSegments);
	}
}

void LDrawModel::convertStu24(bool isA)
{
	int numSegments = LO_NUM_SEGMENTS;

	clearModel();
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 6.0f, 4.0f, numSegments);
	addCylinder(Vector(0.0f, -4.0f, 0.0f), 8.0f, 4.0f, numSegments);
	addCone(Vector(0.0f, -4.0f, 0.0f), 8.0f, 6.0f, 0.0f, numSegments);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 6.0f, numSegments);
	addEdge(Vector(0.0f, -4.0f, 0.0f), 8.0f, numSegments);
	if (!isA)
	{
		addEdge(Vector(0.0f, 0.0f, 0.0f), 6.0f, numSegments);
		addEdge(Vector(0.0f, 0.0f, 0.0f), 8.0f, numSegments);
	}
}

void LDrawModel::convertEighthSphere(void)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addEighthSphere(Vector(0.0f, 0.0f, 0.0f), 1.0f, numSegments);
}

void LDrawModel::convertCyls(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addCyls(Vector(0.0f, 0.0f, 0.0f), 1.0f, 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertCyls2(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addCyls2(Vector(0.0f, 0.0f, 0.0f), 1.0f, 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertCcyli(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addCcyli(Vector(0.0f, 0.0f, 0.0f), 2.5f, 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertTorusO(int numSegments, float size)
{
	clearModel();
	addTorusO(Vector(0.0f, 0.0f, 0.0f), 1.0f, size, numSegments * 4,
		numSegments);
}

void LDrawModel::convertTorusI(int numSegments, float size)
{
	clearModel();
	addTorusI(Vector(0.0f, 0.0f, 0.0f), 1.0f, size, numSegments * 4,
		numSegments);
}

void LDrawModel::convertCylinder(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addCylinder(Vector(0.0f, 0.0f, 0.0f), 1.0f, 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertDisk(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addDisk(Vector(0.0f, 0.0f, 0.0f), 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertNdis(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addNotDisk(Vector(0.0f, 0.0f, 0.0f), 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertEdge(float fraction)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addEdge(Vector(0.0f, 0.0f, 0.0f), 1.0f, numSegments,
		(int)(numSegments * fraction));
}

void LDrawModel::convertCone(float fraction, int size)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addCone(Vector(0.0f, 0.0f, 0.0f), (float)size + 1, (float)size, 1.0f,
		numSegments, (int)(numSegments * fraction));
}

void LDrawModel::convertRing(float fraction, int size)
{
	int numSegments = getNumCircleSegments();

	clearModel();
	addRing(Vector(0.0f, 0.0f, 0.0f), (float)size + 1, (float)size,
		numSegments, (int)(numSegments * fraction));
}

bool LDrawModel::readFile(bool closeAfterRead)
{
	char buf[MODEL_BUF_SIZE];
	int subModelsRead = 0;
	int quadsRead = 0;
	int trianglesRead = 0;
	int linesRead = 0;

	if (!reportProgress("Reading model.", 0.0f))
	{
		return false;
	}
	while (1)
	{
		char* line;
		char subModelName[1024];
		bool skippingSubModel = false;

		if (!reportProgress(NULL, (float)offset / filesize))
		{
			return false;
		}
		if (getLine(buf))
		{
			break;
		}
		line = new char[strlen(buf) + 1];
		strcpy(line, buf);
		addFileLine(line);
		// Have to check for submodels first, because they are the only thing
		// that can be done while we are skipping a submodel.
		if (sscanf(line, "0 FILE %s", subModelName) == 1)
		{
			if (flags.haveMainModel)
			{
				TCDictionary* subModelDict = getLoadedModels();
				LDrawModel* subModel = (LDrawModel*)
					(subModelDict->objectForKey(subModelName));
				
				if (subModel)
				{
					skippingSubModel = true;
				}
				else
				{
					subModel = new LDrawModel;

					skippingSubModel = false;
					subModel->setFilename(filename);
					subModel->currentLineNumber = currentLineNumber;
					subModel->flags.haveMainModel = 1;
					if (!initializeNewSubModel(subModel, subModelName,
						file, false))
					{
						subModel->release();
						delete line;
						return false;
					}
					subModel->release();
					currentLineNumber = -subModel->currentLineNumber;
					if (subModel->flags.isForcedPart)
					{
						TCDictionary* subModelDict = getLoadedModels();
						char dictName[1024];

						subModel = (LDrawModel*)subModel->copy();
						sprintf(dictName, "LDV-UNSCALED:%s",
							subModel->getName());
						subModelDict->setObjectForKey(subModel, dictName);
						subModel->release();
					}
				}
			}
			else
			{
				flags.haveMainModel = YES;
			}
		}
		else if (!skippingSubModel)
		{
			switch(line[0])
			{
				case '0': // Comment
					break;
				case '1': // Sub-model
					subModelsRead++;
					break;
				case '2': // Line
					linesRead++;
					break;
				case '3': // Triangle
					trianglesRead++;
					break;
				case '4': // Quad
					quadsRead++;
					break;
				case '5': // Conditional Line
					linesRead++;
					break;
				default: // Unknown line
					break;
			}
		}
		delete line;
	}
	if (!reportProgress(NULL, 1.0f))
	{
		return false;
	}
	if (subModelsRead)
	{
		newSubModels->setCapacity(subModelsRead);
//		subModels = new LDrawModel*[subModelsAllocated];
//		transformationMatrices = new float*[subModelsAllocated];
//		colors = new int[subModelsAllocated];
//		colorNumbers = new int[subModelsAllocated];
//		memset(transformationMatrices, 0, subModelsAllocated * sizeof(float*));
	}
	if (linesRead)
	{
		getShapes(LDM_LINES)->setCapacity(linesRead);
//		lines = new TCObjectArray(linesRead);
	}
	if (trianglesRead)
	{
		getShapes(LDM_TRIANGLES)->setCapacity(trianglesRead);
//		triangles = new TCObjectArray(trianglesRead);
	}
	if (quadsRead)
	{
		getShapes(LDM_QUADS)->setCapacity(quadsRead);
//		quads = new TCObjectArray(quadsRead);
	}
	if (closeAfterRead)
	{
		closeFile();
	}
	else
	{
		file = NULL;
	}
	return true;
}

/*
void LDrawModel::compShapeNormals(TCObjectArray* shapes)
{
	if (shapes)
	{
		int i;
		int count = shapes->getCount();

		for (i = 0; i < count; i++)
		{
			((TGLShape*)(*shapes)[i])->compNormal(0);
		}
	}
}

void LDrawModel::compNormals(void)
{
	if (!flags.normalsComputed)
	{
		if (subModelCount)
		{
			int i;
			int count = subModelCount;

			for (i = 0; i < count; i++)
			{
				subModels[i]->compNormals();
			}
		}
		compShapeNormals(quads);
		compShapeNormals(triangles);
	}
}
*/

bool LDrawModel::parseLines(void)
{
	int i;
	int fileLineCount = getArrayCount(fileLines);

#ifdef _LOG_PERFORMANCE
	LARGE_INTEGER startCounter;
	LARGE_INTEGER stopCounter;

	QueryPerformanceCounter(&startCounter);
#endif // _LOG_PERFORMANCE
	if (!reportProgress("Parsing model.", 0))
	{
		return false;
	}
	for (i = 0; i < fileLineCount; i++)
	{
		char* fileLine;
		currentFileLine = (*fileLines)[i];
		
		fileLine = currentFileLine->getStrippedLine();
		switch (fileLine[0])
		{
			case '0':
				readComment(fileLine);
				break;
			case '1':
#ifdef _LOG_PERFORMANCE
				QueryPerformanceCounter(&stopCounter);
				parseDelta += stopCounter.QuadPart - startCounter.QuadPart;
#endif // _LOG_PERFORMANCE
				if (!readSubModel(fileLine))
				{
					return false;
				}
#ifdef _LOG_PERFORMANCE
				QueryPerformanceCounter(&startCounter);
#endif // _LOG_PERFORMANCE
				break;
			case '2':
				readLine(fileLine);
				break;
			case '3':
				readTriangle(fileLine);
				break;
			case '4':
				readQuad(fileLine);
				break;
			case '5':
				readConditionalLine(fileLine);
				break;
			default:
				if (strlen(fileLine))
				{
					reportError("Could not parse line.", LDMEParse);
				}
				break;
		}
		if (!reportProgress(NULL, (float)(i + 1) / fileLineCount))
		{
			return false;
		}
	}
	currentFileLine = NULL;
#ifdef _LOG_PERFORMANCE
	QueryPerformanceCounter(&stopCounter);
	parseDelta += stopCounter.QuadPart - startCounter.QuadPart;
#endif // _LOG_PERFORMANCE
	return true;
}

int LDrawModel::reportProgress(char* message, float progress)
{
	if (progressCallback)
	{
		return progressCallback(message, progress, progressUserData);
	}
	else
	{
		if (topParent && topParent->progressCallback)
		{
			return topParent->reportProgress(NULL, -1.0f);
		}
		else
		{
			return 1;
		}
	}
}

int LDrawModel::reportError(LDMError* error)
{
	if (errorCallback)
	{
		return errorCallback(error, errorUserData);
	}
	else if (topParent && topParent->errorCallback)
	{
		return topParent->reportError(error);
	}
	else
	{
		return 1;
	}
}

int LDrawModel::reportError(char* format, LDMErrorType type, ...)
{
	va_list argPtr;
	char message[1024];
	LDMError* error;
	int retVal;
	char* line = NULL;
	int lineNumber = -1;
	char** components;
	int componentCount;

	va_start(argPtr, type);
	vsprintf(message, format, argPtr);
	va_end(argPtr);
	if (currentFileLine)
	{
		line = currentFileLine->getOriginalLine();
		lineNumber = currentFileLine->getLineNumber();
	}
	stripCRLF(message);
	components = componentsSeparatedByString(message, "\n", componentCount);
	if (componentCount > 1)
	{
		int i;
		TCStringArray *extraInfo = new TCStringArray(componentCount - 1);

		*strchr(message, '\n') = 0;
		for (i = 1; i < componentCount; i++)
		{
			extraInfo->addString(components[i]);
		}
		error = new LDMError(type, message, filename, line, lineNumber,
			extraInfo);
		extraInfo->release();
	}
	else
	{
		error = new LDMError(type, message, filename, line, lineNumber);
	}
	deleteStringArray(components, componentCount);
	retVal = reportError(error);
	error->release();
	return retVal;
}

void LDrawModel::addTransparentShape(int type, TGLShape* shape)
{
	(*shapes)[type + NUM_SHAPE_TYPES]->addObject(shape);
	shape->calcCentroid();
}

void LDrawModel::propagateColors(void)
{
	float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
								0.0f, 1.0f, 0.0f, 0.0f,
								0.0f, 0.0f, 1.0f, 0.0f,
								0.0f, 0.0f, 0.0f, 1.0f};

	initShapes();
/*
	for (int i = 0; i < NUM_SHAPE_TYPES; i++)
	{
		shapes->addObject(new TGLShapeArray);
	}
*/
	propagateColors(identityMatrix, 16);
}

void LDrawModel::propagateColors(float* matrix, int colorNumber)
{
	int i, j;
	int r, g, b, a;
	int rh, gh, bh, ah;
	int highlightColorNumber;

	if (colorNumber == 16)
	{
		colorNumber = topParent->defaultColorNumber;
	}
	highlightColorNumber = getHighlightColor(colorNumber);
	getRGBA(colorNumber, r, g, b, a);
	getRGBA(highlightColorNumber, rh, gh, bh, ah);
	for (i = 0; i < NUM_SHAPE_TYPES; i++)
	{
		TGLShapeArray* theseShapes = getShapes(i);

		// Process in reverse order so that removals won't hurt us.
		for (j = theseShapes->getCount() - 1; j >= 0; j--)
		{
			TGLShape* shape = (*theseShapes)[j];

			if (shape->getColorNumber() == 16)
			{
				shape->setColorNumber(colorNumber);
				shape->setColor(r, g, b, a);
			}
			else if (shape->getColorNumber() == 24)
			{
				shape->setColorNumber(highlightColorNumber);
				shape->setColor(rh, gh, bh, ah);
			}
			if (i == LDM_LINES || i == LDM_LINE_STRIPS)
			{
				int r, g, b, a;

				shape->getRGBA(r, g, b, a);
				if (a < 255)
				{
					shape->setColor(r, g, b, 255);
					if (i == LDM_LINES)
					{
						addShape(LDM_TRANS_LINES, shape);
						shape->retain();
						theseShapes->removeObject(j);
						continue;
					}
					else
					{
						addShape(LDM_TRANS_LINE_STRIPS, shape);
						shape->retain();
						theseShapes->removeObject(j);
						continue;
					}
				}
			}
			if (shouldSortTransparent() && shape->isTransparent())
			{
				shape->transform(matrix);
				shape->setParent(topParent);
				topParent->addTransparentShape(i < LDM_LINES ? 0 : LDM_LINES,
					shape);
				theseShapes->removeObject(j);
			}
		}
	}
//	count = newSubModels->getCount();
	for (i = 0; i < newSubModels->getCount(); i++)
	{
		float newMatrix[16];
		LDrawSubModel *subModel = newSubModels->objectAtIndex(i);
		int subColorNumber = subModel->getColorNumber();

		TGLShape::multMatrix(matrix, subModel->getTransformationMatrix(),
			newMatrix);
		if (subColorNumber == 16)
		{
			subModel->setColorNumber(colorNumber);
			subModel->setColor(TGLShape::colorForRGBA(r, g, b, a));
			subColorNumber = colorNumber;
		}
		else if (subColorNumber == 24)
		{
			subModel->setColorNumber(highlightColorNumber);
			subModel->setColor(TGLShape::colorForRGBA(rh, gh, bh, ah));
			subColorNumber = highlightColorNumber;
		}
		subModel->getModel()->propagateColors(newMatrix, subColorNumber);
	}
/*
	for (i = 0; i < subModelCount; i++)
	{
		float newMatrix[16];

		TGLShape::multMatrix(matrix, transformationMatrices[i], newMatrix);
		if (colorNumbers[i] == 16)
		{
			colorNumbers[i] = colorNumber;
			setColor(i, r, g, b, a);
		}
		else if (colorNumbers[i] == 24)
		{
			colorNumbers[i] = highlightColorNumber;
			setColor(i, rh, gh, bh, ah);
		}
		subModels[i]->propagateColors(newMatrix, colorNumbers[i]);
	}
*/
}

int LDrawModel::read(FILE* f, bool closeAfterRead, int subFileSize)
{
	int retValue = -1;

	setFile(f, subFileSize);
	if (readFile(closeAfterRead))
	{
		bool parseResult;

		parseResult = parseLines();
		if (closeAfterRead)
		{
//			compNormals();
		}
		if (parseResult)
		{
			if (!parent)
			{
				int i;
				int count;

				if (flags.isPart)
				{
					finishPart();
				}
				count = newSubModels->getCount();
				for (i = 0; i < count; i++)
				{
					LDrawSubModel *subModel = newSubModels->objectAtIndex(i);

					subModel->setModel(new LDrawModel(*(subModel->getModel())));
					subModel->getModel()->release();
				}
/*
				for (i = 0; i < subModelCount; i++)
				{
					LDrawModel* subModel = subModels[i];

					subModels[i] = new LDrawModel(*subModel);
					subModel->release();
				}
*/
				propagateColors();
			}
			retValue = 0;
		}
	}
	if (retValue == -1)
	{
		reportProgress(NULL, 2.0f);
	}
#ifdef _LOG_PERFORMANCE
	if (!closeAfterRead)
	{
		debugPrintf("parseDelta: %I64d\n", parseDelta);
		debugPrintf("parseTriangleDelta: %I64d\n", parseTriangleDelta);
		debugPrintf("parseQuadDelta: %I64d\n", parseQuadDelta);
	}
#endif // _LOG_PERFORMANCE
	return retValue;
}

void LDrawModel::readComment(char* buf)
{
	if (stringHasCaseInsensitivePrefix(buf, "0 unofficial ") ||
		stringHasCaseInsensitivePrefix(buf, "0 un-official ") ||
		stringHasCaseInsensitivePrefix(buf, "0 official lcad ") ||
		stringHasCaseInsensitivePrefix(buf, "0 ldraw_org "))
	{
		if (strcasestr(buf, "part") || strcasestr(buf, "element"))
		{
			if (!flags.isPart)
			{
				flags.isPart = true;
				flags.isForcedPart = true;
			}
		}
	}
}

const char* LDrawModel::lDrawDir(void)
{
	if (!systemLDrawDir)
	{
		systemLDrawDir = copyString(getenv("LDRAWDIR"));
		if (!systemLDrawDir)
		{
			systemLDrawDir = copyString("C:\\LDRAW");
		}
	}
	return systemLDrawDir;
}

bool LDrawModel::isPartOrSubpart(void)
{
	if (flags.isPart)
	{
		return YES;
	}
	else if (parent)
	{
		return parent->isPartOrSubpart();
	}
	else
	{
		return NO;
	}
}

FILE* LDrawModel::openSubModelNamed(char* subModelName, char* subModelPath)
{
	FILE* subModelFile;
//	char filename[1024];

	strcpy(subModelPath, subModelName);
	if ((subModelFile = fopen(subModelPath, "rb")) != NULL)
	{
		return subModelFile;
	}
	sprintf(subModelPath, "%s/P/%s", lDrawDir(), subModelName);
	if ((subModelFile = fopen(subModelPath, "rb")) != NULL)
	{
		return subModelFile;
	}
	sprintf(subModelPath, "%s/PARTS/%s", lDrawDir(), subModelName);
	if ((subModelFile = fopen(subModelPath, "rb")) != NULL)
	{
		flags.subModelIsPart = YES;
		return subModelFile;
	}
	sprintf(subModelPath, "%s/MODELS/%s", lDrawDir(), subModelName);
	if ((subModelFile = fopen(subModelPath, "rb")) != NULL)
	{
		return subModelFile;
	}
	reportError("Error opening submodel %s\n", LDMEFileNotFound, subModelName);
//	printf("Error opening submodel %s\n", subModelName);
	return NULL;
}

TCDictionary* LDrawModel::getLoadedModels(void)
{
	if (!loadedModels)
	{
		loadedModels = new TCDictionary(0);
	}
	return loadedModels;
}

bool LDrawModel::initializeNewSubModel(LDrawModel* subModel,
									   char* subModelName,
									   FILE* subModelFile,
									   bool closeAfterRead,
									   bool allowScale)
{
	TCDictionary* subModelDict = getLoadedModels();
	char dictName[1024];
	int subFileSize = -1;

	if (flags.subModelIsPart)
	{
		subModel->flags.isPart = true;
		flags.subModelIsPart = 0;
	}
	subModel->setParent(this);
	if (allowScale)
	{
		strcpy(dictName, subModelName);
	}
	else
	{
		sprintf(dictName, "LDV-UNSCALED:%s", subModelName);
	}
	subModelDict->setObjectForKey(subModel, dictName);
	subModel->flags.drawHighlights = flags.drawHighlights;
	subModel->flags.allowLowResStuds = flags.allowLowResStuds;
	subModel->flags.allowPrimitiveSubstitution =
		flags.allowPrimitiveSubstitution;
	subModel->flags.drawConditionalHighlights = flags.drawConditionalHighlights;
	subModel->flags.performSmoothing = flags.performSmoothing;
	subModel->flags.lineSmoothing = flags.lineSmoothing;
	if (subModelFile == file)
	{
		subFileSize = filesize;
	}
	if (subModel->read(subModelFile, closeAfterRead, subFileSize) != 0)
	{
		return false;
	}
	subModel->setName(dictName);
	if (subModel->flags.isPart)
	{
		subModel->finishPart(allowScale);
	}
	return true;
//	subModel->release();
}

void LDrawModel::finishPart(bool allowScale)
{
	flatten();
	if ((!parent || !parent->isPartOrSubpart()))
	{
		if (allowScale && !fEq(seamWidth, 0.0f))
		{
			shrink();
		}
		if (flags.performSmoothing)
		{
			smooth(true);
		}
	}
}

/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/

void LDrawModel::addConditionalPoint(VectorShapeArrayTree *conditionalTree,
									 TGLShape *line, int pointIndex)
{
	VectorKey pointKey(line->getPoints()[pointIndex]);
//	char pointString[1024];
	TGLShapeArray *conditionalArray;

//	line->getPoints()[pointIndex].print(pointString);
	conditionalArray =
		(TGLShapeArray*)conditionalTree->objectForKey(&pointKey);
	if (!conditionalArray)
	{
		TGLShape *normalShape = new TGLShape;

		normalShape->setNumPoints(1);
		conditionalArray = new TGLShapeArray;
		conditionalTree->setObjectForKey(conditionalArray, &pointKey);
		conditionalArray->release();
		conditionalArray->addObject(normalShape);
		normalShape->release();
	}
	conditionalArray->addObject(line);
}

void LDrawModel::fillConditionalTree(VectorShapeArrayTree *conditionalTree)
{
	TGLShapeArray* lines = shapes->objectAtIndex(LDM_LINES);
	int i;
	int count = lines->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *line = lines->objectAtIndex(i);

		if (line->isConditional())
		{
			addConditionalPoint(conditionalTree, line, 0);
			addConditionalPoint(conditionalTree, line, 1);
		}
	}
}

void LDrawModel::processSmoothEdge(VectorShapeArrayTree *conditionalTree,
								   TGLShape *shape, int index0, int index1,
								   const Vector &normal)
{
	VectorKey pointKey(shape->getPoints()[index0]);
	TGLShapeArray *lines0;

	lines0 = (TGLShapeArray *)conditionalTree->objectForKey(&pointKey);
	if (lines0)
	{
		TGLShapeArray *lines1;

		pointKey = shape->getPoints()[index1];
		lines1 = (TGLShapeArray *)conditionalTree->objectForKey(&pointKey);
		if (lines1)
		{
			int i, j;
			int count0 = lines0->getCount();
			int count1 = lines1->getCount();
			bool found = false;

			for (i = 1; i < count0 && !found; i++)
			{
				for (j = 1; j < count1; j++)
				{
					TGLShape *line0 = lines0->objectAtIndex(i);
					TGLShape *line1 = lines1->objectAtIndex(j);

					if (line0 == line1)
					{
						found = true;
					}
				}
			}
			if (found)
			{
				TGLShape *normalShape = lines0->objectAtIndex(0);

				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					normalShape->getPoints()[0] -= normal;
				}
				else
				{
					normalShape->getPoints()[0] += normal;
				}
				normalShape = lines1->objectAtIndex(0);
				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					normalShape->getPoints()[0] -= normal;
				}
				else
				{
					normalShape->getPoints()[0] += normal;
				}
				shape->setUserData((void*)1);
			}
		}
	}
}

void LDrawModel::calcShapeNormals(VectorShapeArrayTree *conditionalTree,
								  int shapeIndex)
{
	TGLShapeArray* theseShapes = shapes->objectAtIndex(shapeIndex);
	int i;
	int count = theseShapes->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *shape = theseShapes->objectAtIndex(i);
		Vector normal = shape->calcNormal(/*true*/false);
		int j;
		int count = shape->getNumPoints();

		for (j = 0; j < count; j++)
		{
			processSmoothEdge(conditionalTree, shape, j, (j + 1) % count,
				normal);
		}
	}
}

void LDrawModel::normalizeFunc(const void * /*key*/, TCObject *value)
{
	TGLShapeArray *shapeArray = (TGLShapeArray *)value;

	shapeArray->objectAtIndex(0)->getPoints()[0].normalize();
}

void LDrawModel::normalizeNormals(VectorShapeArrayTree *conditionalTree)
{
	conditionalTree->traverseTree(normalizeFunc);
}

void LDrawModel::applySmoothEdge(VectorShapeArrayTree *conditionalTree,
								 TGLShape *shape, int index0, int index1,
								 const Vector &normal)
{
	VectorKey pointKey(shape->getPoints()[index0]);
	TGLShapeArray *lines0;

	lines0 = (TGLShapeArray *)conditionalTree->objectForKey(&pointKey);
	if (lines0)
	{
		TGLShapeArray *lines1;

		pointKey = shape->getPoints()[index1];
		lines1 = (TGLShapeArray *)conditionalTree->objectForKey(&pointKey);
		if (lines1)
		{
			int i, j;
			int count0 = lines0->getCount();
			int count1 = lines1->getCount();
			bool found = false;

			for (i = 1; i < count0 && !found; i++)
			{
				for (j = 1; j < count1; j++)
				{
					TGLShape *line0 = lines0->objectAtIndex(i);
					TGLShape *line1 = lines1->objectAtIndex(j);

					if (line0 == line1)
					{
						found = true;
					}
				}
			}
			if (found)
			{
				TGLShape *normalShape = lines0->objectAtIndex(0);

				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					shape->setNormal(index0, -normalShape->getPoints()[0]);
				}
				else
				{
					shape->setNormal(index0, normalShape->getPoints()[0]);
				}
				normalShape = lines1->objectAtIndex(0);
				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					shape->setNormal(index1, -normalShape->getPoints()[0]);
				}
				else
				{
					shape->setNormal(index1, normalShape->getPoints()[0]);
				}
			}
		}
	}
}

void LDrawModel::applyShapeNormals(VectorShapeArrayTree *conditionalTree,
								   int shapeIndex)
{
	TGLShapeArray* theseShapes = shapes->objectAtIndex(shapeIndex);
	int i;
	int count = theseShapes->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *shape = theseShapes->objectAtIndex(i);

		if (shape->getUserData())
		{
			Vector normal = shape->calcNormal(true);
			int j;
			int count = shape->getNumPoints();

			if (shape->getMirrored())
			{
				normal *= -1.0f;
			}
			for (j = 0; j < count; j++)
			{
				shape->setNormal(j, normal);
			}
			for (j = 0; j < count; j++)
			{
				applySmoothEdge(conditionalTree, shape, j, (j + 1) % count,
					normal);
			}
		}
	}
}

/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/

void LDrawModel::addConditionalPoint(TGLShapeArrayDictionary *conditionalDict,
									 TGLShape *line, int pointIndex)
{
	char pointString[1024];
	TGLShapeArray *conditionalArray;

	line->getPoints()[pointIndex].print(pointString);
	conditionalArray = conditionalDict->objectForKey(pointString);
	if (!conditionalArray)
	{
		TGLShape *normalShape = new TGLShape;

		normalShape->setNumPoints(1);
		conditionalArray = new TGLShapeArray;
		conditionalDict->setObjectForKey(conditionalArray, pointString);
		conditionalArray->release();
		conditionalArray->addObject(normalShape);
		normalShape->release();
	}
	conditionalArray->addObject(line);
}

void LDrawModel::fillConditionalDict(TGLShapeArrayDictionary *conditionalDict)
{
	TGLShapeArray* lines = shapes->objectAtIndex(LDM_LINES);
	int i;
	int count = lines->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *line = lines->objectAtIndex(i);

		if (line->isConditional())
		{
			addConditionalPoint(conditionalDict, line, 0);
			addConditionalPoint(conditionalDict, line, 1);
		}
	}
}

bool LDrawModel::shouldFlipNormal(const Vector &normal1, const Vector &normal2)
{
	if (normal1.get(0) == 0.0f && normal1.get(1) == 0.0f &&
		normal1.get(2) == 0.0f)
	{
		return false;
	}
	else
	{
		Vector normal3 = normal1 / normal1.length();
		Vector normal4 = normal2 / normal2.length();
		float dotProduct = normal3.dot(normal4);

		if (!fEq(dotProduct, 0.0f) && dotProduct < 0.0f)
		{
			return true;
		}
		else
		{
			return false;
		}
	}
}

void LDrawModel::processSmoothEdge(TGLShapeArrayDictionary *conditionalDict,
								   TGLShape *shape, int index0, int index1,
								   const Vector &normal)
{
	char pointString[1024];
	TGLShapeArray *lines0;

	shape->getPoints()[index0].print(pointString);
	lines0 = conditionalDict->objectForKey(pointString);
	if (lines0)
	{
		TGLShapeArray *lines1;

		shape->getPoints()[index1].print(pointString);
		lines1 = conditionalDict->objectForKey(pointString);
		if (lines1)
		{
			int i, j;
			int count0 = lines0->getCount();
			int count1 = lines1->getCount();
			bool found = false;

			for (i = 1; i < count0 && !found; i++)
			{
				for (j = 1; j < count1; j++)
				{
					TGLShape *line0 = lines0->objectAtIndex(i);
					TGLShape *line1 = lines1->objectAtIndex(j);

					if (line0 == line1)
					{
						found = true;
					}
				}
			}
			if (found)
			{
				TGLShape *normalShape = lines0->objectAtIndex(0);

				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					normalShape->getPoints()[0] -= normal;
				}
				else
				{
					normalShape->getPoints()[0] += normal;
				}
				normalShape = lines1->objectAtIndex(0);
				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					normalShape->getPoints()[0] -= normal;
				}
				else
				{
					normalShape->getPoints()[0] += normal;
				}
				shape->setUserData((void*)1);
			}
		}
	}
}

void LDrawModel::calcShapeNormals(TGLShapeArrayDictionary *conditionalDict,
								  int shapeIndex)
{
	TGLShapeArray* theseShapes = shapes->objectAtIndex(shapeIndex);
	int i;
	int count = theseShapes->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *shape = theseShapes->objectAtIndex(i);
		Vector normal = shape->calcNormal(/*true*/false);
		int j;
		int count = shape->getNumPoints();

		for (j = 0; j < count; j++)
		{
			processSmoothEdge(conditionalDict, shape, j, (j + 1) % count,
				normal);
		}
	}
}

void LDrawModel::normalizeNormals(TGLShapeArrayDictionary *conditionalDict)
{
	TCTypedObjectArray<TGLShapeArray> *normalArrays =
		conditionalDict->allObjects();
	int i;
	int count = normalArrays->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShapeArray* shapeArray = normalArrays->objectAtIndex(i);

		shapeArray->objectAtIndex(0)->getPoints()[0].normalize();
	}
}

void LDrawModel::applySmoothEdge(TGLShapeArrayDictionary *conditionalDict,
								 TGLShape *shape, int index0, int index1,
								 const Vector &normal)
{
	char pointString[1024];
	TGLShapeArray *lines0;

	shape->getPoints()[index0].print(pointString);
	lines0 = conditionalDict->objectForKey(pointString);
	if (lines0)
	{
		TGLShapeArray *lines1;

		shape->getPoints()[index1].print(pointString);
		lines1 = conditionalDict->objectForKey(pointString);
		if (lines1)
		{
			int i, j;
			int count0 = lines0->getCount();
			int count1 = lines1->getCount();
			bool found = false;

			for (i = 1; i < count0 && !found; i++)
			{
				for (j = 1; j < count1; j++)
				{
					TGLShape *line0 = lines0->objectAtIndex(i);
					TGLShape *line1 = lines1->objectAtIndex(j);

					if (line0 == line1)
					{
						found = true;
					}
				}
			}
			if (found)
			{
				TGLShape *normalShape = lines0->objectAtIndex(0);

				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					shape->setNormal(index0, -normalShape->getPoints()[0]);
				}
				else
				{
					shape->setNormal(index0, normalShape->getPoints()[0]);
				}
				normalShape = lines1->objectAtIndex(0);
				if (shouldFlipNormal(normalShape->getPoints()[0], normal))
				{
					shape->setNormal(index1, -normalShape->getPoints()[0]);
				}
				else
				{
					shape->setNormal(index1, normalShape->getPoints()[0]);
				}
			}
		}
	}
}

void LDrawModel::applyShapeNormals(TGLShapeArrayDictionary *conditionalDict,
								   int shapeIndex)
{
	TGLShapeArray* theseShapes = shapes->objectAtIndex(shapeIndex);
	int i;
	int count = theseShapes->getCount();

	for (i = 0; i < count; i++)
	{
		TGLShape *shape = theseShapes->objectAtIndex(i);

		if (shape->getUserData())
		{
			Vector normal = shape->calcNormal(true);
			int j;
			int count = shape->getNumPoints();

			if (shape->getMirrored())
			{
				normal *= -1.0f;
			}
			for (j = 0; j < count; j++)
			{
				shape->setNormal(j, normal);
			}
			for (j = 0; j < count; j++)
			{
				applySmoothEdge(conditionalDict, shape, j, (j + 1) % count,
					normal);
			}
		}
	}
}

void LDrawModel::smooth(bool useTree)
{
	if (useTree)
	{
		VectorShapeArrayTree *conditionalTree = new VectorShapeArrayTree;

		fillConditionalTree(conditionalTree);
//		conditionalTree->balance();
		calcShapeNormals(conditionalTree, LDM_TRIANGLES);
		calcShapeNormals(conditionalTree, LDM_QUADS);
		normalizeNormals(conditionalTree);
		applyShapeNormals(conditionalTree, LDM_TRIANGLES);
		applyShapeNormals(conditionalTree, LDM_QUADS);
		conditionalTree->release();
	}
	else
	{
		TGLShapeArrayDictionary *conditionalDict = new TGLShapeArrayDictionary;

		fillConditionalDict(conditionalDict);
		calcShapeNormals(conditionalDict, LDM_TRIANGLES);
		calcShapeNormals(conditionalDict, LDM_QUADS);
		normalizeNormals(conditionalDict);
		applyShapeNormals(conditionalDict, LDM_TRIANGLES);
		applyShapeNormals(conditionalDict, LDM_QUADS);
		conditionalDict->release();
	}
}

LDrawModel* LDrawModel::subModelNamed(char* subModelName, bool allowSwap,
									  bool allowScale)
{
	TCDictionary* subModelDict = getLoadedModels();
	LDrawModel* subModel;
	int swappedStud = 0;
	char dictName[1024];

	replaceStringCharacter(subModelName, '\\', '/');
	if (allowSwap && stringHasCaseInsensitivePrefix(subModelName, "stud"))
	{
		subModelName[3] = '2';
		swappedStud = 1;
	}
	if (allowScale)
	{
		strcpy(dictName, subModelName);
	}
	else
	{
		sprintf(dictName, "LDV-UNSCALED:%s", subModelName);
	}
	subModel = (LDrawModel*)(subModelDict->objectForKey(dictName));
	if (subModel)
	{
		return subModel;
	}
	else
	{
		FILE* subModelFile;
		char subModelPath[1024];

		if ((subModelFile = openSubModelNamed(subModelName, subModelPath))
			== NULL)
		{
			if (swappedStud)
			{
				subModelName[3] = 'd';
				return subModelNamed(subModelName, 0);
			}
			else
			{
				return NULL;
			}
		}
		subModel = new LDrawModel;
		subModel->setFilename(subModelPath);
		if (!initializeNewSubModel(subModel, subModelName, subModelFile, true,
			allowScale))
		{
			subModel->release();
			flags.readCanceled = 1;
			return NULL;
		}
		subModel->release();
		return subModel;
	}
}

void LDrawModel::shrink(void)
{
	Vector min, max;
	Vector delta;
	Vector scale;
	Vector center;
	int i;

	getMinMax(min, max);
	delta = max - min;
	center = (min + max) / 2.0f;
	scale[0] = (delta[0] - seamWidth) / delta[0];
	scale[1] = (delta[1] - seamWidth) / delta[1];
	scale[2] = (delta[2] - seamWidth) / delta[2];
	for (i = 0; i < NUM_SHAPE_TYPES; i++)
	{
		int j;
		TGLShapeArray* theseShapes = /*(TCObjectArray*)*/(*shapes)[i];
		int count = theseShapes->getCount();

		for (j = 0; j < count; j++)
		{
			((TGLShape*)(*theseShapes)[j])->scale(scale, center);
		}
	}
/*
	if (quadStrips)
	{
		for (i = 0; i < getQuadStripCount(); i++)
		{
			((TGLShape*)(*quadStrips)[i])->scale(scale, center);
		}
	}
	if (quads)
	{
		for (i = 0; i < getQuadCount(); i++)
		{
			((TGLShape*)(*quads)[i])->scale(scale, center);
		}
	}
	if (triangles)
	{
		for (i = 0; i < getTriangleCount(); i++)
		{
			((TGLShape*)(*triangles)[i])->scale(scale, center);
		}
	}
	if (lines)
	{
		for (i = 0; i < getLineCount(); i++)
		{
			((TGLShape*)(*lines)[i])->scale(scale, center);
		}
	}
	if (lineStrips)
	{
		for (i = 0; i < getLineStripCount(); i++)
		{
			((TGLShape*)(*lineStrips)[i])->scale(scale, center);
		}
	}
*/
}

void LDrawModel::countSubShapes(LDrawModel* subModel, int& subLineCount,
							   int& subLineStripCount, int& subTriangleCount,
							   int& subQuadCount, int& subQuadStripCount)
{
	int i;
	int count = subModel->newSubModels->getCount();

	for (i = 0; i < count; i++)
	{
		LDrawModel* newSubModel =
			subModel->newSubModels->objectAtIndex(i)->getModel();

		subLineCount += newSubModel->getLineCount();
		subLineStripCount += newSubModel->getLineStripCount();
		subTriangleCount += newSubModel->getTriangleCount();
		subQuadCount += newSubModel->getQuadCount();
		subQuadStripCount += newSubModel->getQuadStripCount();
		countSubShapes(newSubModel, subLineCount, subLineStripCount,
			subTriangleCount, subQuadCount, subQuadStripCount);
	}
/*
	for (i = 0; i < subModel->subModelCount; i++)
	{
		LDrawModel* newSubModel = subModel->subModels[i];

		subLineCount += newSubModel->getLineCount();
		subLineStripCount += newSubModel->getLineStripCount();
		subTriangleCount += newSubModel->getTriangleCount();
		subQuadCount += newSubModel->getQuadCount();
		subQuadStripCount += newSubModel->getQuadStripCount();
		countSubShapes(newSubModel, subLineCount, subLineStripCount,
			subTriangleCount, subQuadCount, subQuadStripCount);
	}
*/
}

void LDrawModel::countSubShapes(LDrawModel* subModel, int* counts)
{
	int i;
	TCObjectArrayArray* subShapes = subModel->shapes;
	int count = subShapes->getCount();

	for (i = 0; i < count; i++)
	{
		counts[i] += (*subShapes)[i]->getCount();
	}
	count = subModel->newSubModels->getCount();
	for (i = 0; i < count; i++)
	{
		countSubShapes(subModel->newSubModels->objectAtIndex(i)->getModel(),
			counts);
	}
/*
	for (i = 0; i < subModel->subModelCount; i++)
	{
		countSubShapes(subModel->subModels[i], counts);
	}
*/
}

void LDrawModel::flatten(void)
{
	if (flags.isPart || !parent/*!parent || parent->flags.isPart*/)
//	if (!parent || !parent->parent || !parent->parent->flags.subModelIsPart)
	{
//		int subLineCount = 0;
//		int subLineStripCount = 0;
//		int subTriangleCount = 0;
//		int subQuadStripCount = 0;
//		int subQuadCount = 0;
		int counts[NUM_SHAPE_TYPES * 2];
		float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
									0.0f, 1.0f, 0.0f, 0.0f,
									0.0f, 0.0f, 1.0f, 0.0f,
									0.0f, 0.0f, 0.0f, 1.0f};
		int count = shapes->getCount();

/*
		countSubShapes(this, subLineCount, subLineStripCount, subTriangleCount,
			subQuadCount, subQuadStripCount);
		if (lines)
		{
			lines->setCapacity(getLineCount() + subLineCount);
		}
		if (lineStrips)
		{
			lineStrips->setCapacity(getLineStripCount() + subLineStripCount);
		}
		if (triangles)
		{
			triangles->setCapacity(getTriangleCount() + subTriangleCount);
		}
		if (quads)
		{
			quads->setCapacity(getQuadCount() + subQuadCount);
		}
		if (quadStrips)
		{
			quadStrips->setCapacity(getQuadStripCount() + subQuadStripCount);
		}
*/
		memset(counts, 0, sizeof(counts));
		countSubShapes(this, counts);
		for (int i = 0; i < count; i++)
		{
			// @todo: Set proper capacities.
			(*shapes)[i]->setCapacity(counts[i]);
		}
		flatten(this, identityMatrix);
		newSubModels->removeAll();
		//deleteSubModels();
	}
}

/*
TCObject* LDrawModel::retain(void)
{
	return LDrawModelFile::retain();
}
*/

/*
void LDrawModel::deleteSubModels(void)
{
	if (subModelCount)
	{
		int i;

		for (i = 0; i < subModelCount; i++)
		{
			subModels[i]->release();
			delete transformationMatrices[i];
		}
		delete subModels;
		subModels = NULL;
		subModelsAllocated = 0;
		delete transformationMatrices;
		transformationMatrices = NULL;
		delete colorNumbers;
		colorNumbers = NULL;
		delete colors;
		colors = NULL;
		subModelCount = 0;
	}
}
*/

void LDrawModel::flattenShape(TGLShape* newShape, float* matrix,
							  int colorNumber)
{
	newShape->transform(matrix);
	newShape->setParent(this);
	if (colorNumber != 16 && colorNumber != 24)
	{
		int shapeColorNumber = newShape->getColorNumber();
		int r, g, b, a;

		if (shapeColorNumber == 16)
		{
			shapeColorNumber = colorNumber;
		}
		else if (shapeColorNumber == 24)
		{
			shapeColorNumber = getHighlightColor(colorNumber);
		}
		newShape->setColorNumber(shapeColorNumber);
		getRGBA(shapeColorNumber, r, g, b, a);
		newShape->setColor(r, g, b, a);
	}
}

void LDrawModel::flatten(LDrawModel* subModel, float* matrix, int colorNumber)
{
	int i, j;
	int subCount = subModel->newSubModels->getCount();

	for (i = 0; i < subCount; i++)
	{
		LDrawSubModel *newSub = subModel->newSubModels->objectAtIndex(i);
		LDrawModel* newSubModel = newSub->getModel();;
		int newColorNumber = newSub->getColorNumber();
		float newMatrix[16];

		if (newColorNumber == 16 && colorNumber != 16 && colorNumber != 24)
		{
			newColorNumber = colorNumber;
		}
		TGLShape::multMatrix(matrix, newSub->getTransformationMatrix(),
			newMatrix);

		for (j = 0; j < NUM_SHAPE_TYPES; j++)
		{
			TGLShapeArray* theseShapes = newSubModel->getShapes(j);
			int count = theseShapes->getCount();

			for (int k = 0; k < count; k++)
			{
				TGLShape *newShape = (*theseShapes)[k]->copyShape();

				flattenShape(newShape, newMatrix, newColorNumber);
				if (j == LDM_LINES || j == LDM_LINE_STRIPS)
				{
					int r, g, b, a;

					newShape->getRGBA(r, g, b, a);
					if (a == 255)
					{
						addShape(j, newShape);
					}
					else
					{
						newShape->setColor(r, g, b, 255);
						if (j == LDM_LINES)
						{
							addShape(LDM_TRANS_LINES, newShape);
						}
						else
						{
							addShape(LDM_TRANS_LINE_STRIPS, newShape);
						}
					}
				}
				else
				{
					addShape(j, newShape);
				}
			}
		}
		flatten(newSubModel, newMatrix, newColorNumber);
	}
/*
	for (i = 0; i < subModel->subModelCount; i++)
	{
		LDrawModel* newSubModel = subModel->subModels[i];
		int newColorNumber = subModel->colorNumbers[i];
		float newMatrix[16];

		if (newColorNumber == 16 && colorNumber != 16 && colorNumber != 24)
		{
			newColorNumber = colorNumber;
		}
		TGLShape::multMatrix(matrix, subModel->transformationMatrices[i],
			newMatrix);

		for (j = 0; j < NUM_SHAPE_TYPES; j++)
		{
			TGLShapeArray* theseShapes = newSubModel->getShapes(j);
			int count = theseShapes->getCount();

			for (int k = 0; k < count; k++)
			{
				TGLShape *newShape = (*theseShapes)[k]->copyShape();

				flattenShape(newShape, newMatrix, newColorNumber);
				addShape(j, newShape);
			}
		}
		flatten(newSubModel, newMatrix, newColorNumber);
	}
*/
}

void LDrawModel::setProgressCallback(LDMProgressCallback callback,
									 void* userData)
{
	progressCallback = callback;
	progressUserData = userData;
}

void LDrawModel::setErrorCallback(LDMErrorCallback callback, void* userData)
{
	errorCallback = callback;
	errorUserData = userData;
}

bool LDrawModel::readSubModel(char* buf)
{
	int colorNumber;
	float x, y, z;
	float a, b, c, d, e, f, g, h, i;
	char subModelName[1024];

	if (sscanf(buf, "1 %i %f %f %f %f %f %f %f %f %f %f %f %f %s",
		&colorNumber, &x, &y, &z, &a, &b, &c, &d, &e, &f,
		&g, &h, &i, subModelName) == 14)
	{
		LDrawModel* subModel = subModelNamed(subModelName,
			flags.allowLowResStuds);
		int red, green, blue, alpha;
		LDrawSubModel *newSub = new LDrawSubModel;

		if (!subModel)
		{
			newSub->release();
			if (flags.readCanceled)
			{
				flags.readCanceled = 0;
				return false;
			}
			else
			{
				return true;
			}
		}
		getRGBA(colorNumber, red, green, blue, alpha);
		newSubModels->addObject(newSub);
		newSub->release();
		newSub->setModel(subModel);
		setTransformation(newSub, x, y, z, a, b, c, d, e, f, g, h, i);
		newSub->setColorNumber(colorNumber);
		newSub->setColor(TGLShape::colorForRGBA(red, green, blue, alpha));
		if (!fEq(seamWidth, 0.0f) && subModel->flags.isPart)
		{
			float determinant =
				TGLShape::determinant(newSub->getTransformationMatrix());

			if (!fEq2(determinant, 1.0f, 0.05f) &&
				!fEq2(determinant, -1.0f, 0.05f))
			{
				// If the determinant is not either 1 or -1, they applied a
				// non-uniform scale, and we don't want to shrink the part,
				// so grab an unshrunk version and replace the shrunk version
				// the unshrunk one.  Note that we are being EXTREMELY loose
				// with this "equality" check (within 0.05).
				subModel = subModelNamed(subModelName, flags.allowLowResStuds,
					false);
				newSub->setModel(subModel);
				reportError("Part transformed non-uniformly.",
					LDMEPartDeterminant);
//				printf("determinant: %f\n", determinant);
			}
		}
	}
	return true;
}

void LDrawModel::newLine(int colorNumber, const Vector& p1, const Vector& p2)
{
	TGLShape* line = new TGLShape;
	int r, g, b, a;

	getRGBA(colorNumber, r, g, b, a);
	line->setPoints(p1, p2);
	line->setColorNumber(colorNumber);
	line->setColor(r, g, b, a);
	line->setParent(this);
	addLine(line);
}

void LDrawModel::readLine(char* buf)
{
	int colorNumber;
	float x1, y1, z1;
	float x2, y2, z2;

	if (sscanf(buf, "2 %i %f %f %f %f %f %f", &colorNumber,
		&x1, &y1, &z1, &x2, &y2, &z2) == 7)
	{
		if (flags.drawHighlights || colorNumber != 24)
		{
			Vector p1(x1, y1, z1);
			Vector p2(x2, y2, z2);

			newLine(colorNumber, p1, p2);
		}
	}
	else
	{
		reportError("Error parsing line line.", LDMEParse);
//		printf("Error reading line line:\n%s\n", buf);
	}
}

bool LDrawModel::isTransparent(int index)
{
	int colorNumber = newSubModels->objectAtIndex(index)->getColorNumber();
//	int colorNumber = colorNumbers[index];

	return colorNumberIsTransparent(colorNumber);
//	return colorNumber >= 32 && colorNumber < 48;
}

bool LDrawModel::colorNumberIsTransparent(int colorNumber)
{
	return TGLShape::colorNumberIsTransparent(colorNumber);
}

int LDrawModel::getHighlightColor(int colorNumber)
{
	if ((topParent && topParent->flags.blackHighlights) ||
		(!topParent && flags.blackHighlights))
	{
		if (colorNumberIsTransparent(colorNumber))
		{
			return 0x3333333;	// Color 32 is transparent dark gray, and has a
								// different RGB value than black, so return
								// the extended color number for the same RGB
								// value as black, but transparent.
		}
		else
		{
			return 0;
		}
	}
	switch (colorNumber)
	{
		case 0:
			return 8;
			break;
		case 1:
			return 9;
			break;
		case 2:
			return 10;
			break;
		case 3:
			return 11;
			break;
		case 4:
			return 12;
			break;
		case 5:
			return 13;
			break;
		case 6:
			return 0;
			break;
		case 7:
			return 8;
			break;
		case 8:
			return 0;
			break;
		case 9:
			return 1;
			break;
		case 10:
			return 2;
			break;
		case 11:
			return 3;
			break;
		case 12:
			return 4;
			break;
		case 13:
			return 5;
			break;
		case 14:
			return 8;
			break;
		case 15:
			return 8;
			break;
		case 16:
			if (defaultColorNumber != 16)
			{
				return getHighlightColor(defaultColorNumber);
			}
			else
			{
				return 16;
			}
			break;
		case 32:
			return 40;
			break;
		case 33:
			return 41;
			break;
		case 34:
			return 42;
			break;
		case 35:
			return 43;
			break;
		case 36:
			return 44;
			break;
		case 37:
			return 45;
			break;
		case 38:
			return 32;
			break;
		case 39:
			return 40;
			break;
		case 40:
			return 32;
			break;
		case 41:
			return 33;
			break;
		case 42:
			return 34;
			break;
		case 43:
			return 35;
			break;
		case 44:
			return 36;
			break;
		case 45:
			return 37;
			break;
		case 46:
			return 40;
			break;
		case 47:
			return 40;
			break;
		case 57:
			return 40;
			break;
		default:
			return 0;
			break;
	}
}

void LDrawModel::getRGBA(int colorNumber, int& r, int& g, int& b, int& a)
{
	if (colorNumberIsTransparent(colorNumber))
	{
		a = 110;
	}
	else
	{
		a = 255;
	}
	switch (colorNumber)
	{
		case 0: // Black
			r = 51;		//48;
			g = 51;		//48;
			b = 51;		//48;
			break;
		case 1: // Blue
			r = 0;
			g = 51;		//0;
			b = 178;	//128;
			break;
		case 2: // Green
			r = 0;
			g = 127;	//128;
			b = 51;		//0;
			break;
		case 3: // Dark Cyan
			r = 0;
			g = 181;	//128;
			b = 166;	//128;
			break;
		case 4: // Red
			r = 204;	//128;
			g = 0;
			b = 0;
			break;
		case 5: // Magenta
		case 26: // Dark Pink
			r = 255;	//128;
			g = 51;		//0;
			b = 153;	//128;
			break;
		case 6: // Brown
			r = 102;	//112;
			g = 51;		//71;
			b = 0;		//44;
			break;
		case 7: // Grey
			r = 153;	//128;
			g = 153;	//128;
			b = 153;	//128;
			break;
		case 8: // Dark Grey
			r = 102;	//96;
			g = 102;	//96;
			b = 88;		//96;
			break;
		case 9: // Light Blue
			r = 0;
			g = 128;	//0;
			b = 255;
			break;
		case 10: // Light Green
			r = 51;		//0;
			g = 255;
			b = 102;	//0;
			break;
		case 11: // Cyan
			r = 171;	//0;
			g = 253;	//255;
			b = 249;	//255;
			break;
		case 12: // Light Red
			r = 255;
			g = 0;
			b = 0;
			break;
		case 13: // Pink
			r = 255;
			g = 176;	//128;
			b = 204;	//128;
			break;
		case 14: // Yellow
			r = 255;
			g = 229;	//255;
			b = 0;
			break;
		case 15: // White
			r = 255;
			g = 255;
			b = 255;
			break;
		case 17: // Pastel Green
			r = 102;
			g = 240;
			b = 153;
			break;
		case 18: // Light Yellow
			r = 255;
			g = 255;
			b = 128;
			break;
		case 19: // Tan
		case 28: // Tan
		case 382: // Tan
			r = 204;
			g = 170;
			b = 102;
			break;
		case 20: // Light Purple
			r = 224;
			g = 204;
			b = 240;
			break;
		case 21: // Glow in the dark
			r = 224;
			g = 255;
			b = 176;
			break;
		case 22: // Purple
			r = 153;
			g = 51;
			b = 153;
			break;
		case 23: // Purple Blue
			r = 76;
			g = 0;
			b = 204;
			break;
		case 25: // Orange
			r = 255;
			g = 102;
			b = 0;
			break;
		case 27: // Lime Green
			r = 173;
			g = 221;
			b = 80;
			break;
		case 32: // Transparent Dark Grey
			r = 102;
			g = 102;
			b = 102;
			break;
		case 33: // Transparent Dark Blue
			r = 0;
			g = 0;
			b = 153;
			break;
		case 34: // Transparent Green
			r = 0;
			g = 80;
			b = 24;
			break;
		case 35: // Transparent Dark Cyan
			r = 0;
			g = 181;
			b = 166;
			break;
		case 36: // Transparent Red
			r = 204;
			g = 0;
			b = 0;
			break;
		case 37: // Transparent Magenta
			r = 255;
			g = 51;
			b = 153;
			break;
		case 38: // Transparent Brown (RIIIIIIGHT)
			r = 102;
			g = 51;
			b = 0;
			break;
		case 39: // Transparent Grey
			r = 153;
			g = 153;
			b = 153;
			break;
		case 40: // Transparent Dark Grey
			r = 102;
			g = 102;
			b = 88;
			break;
		case 41: // Transparent Light Blue
			r = 153;
			g = 192;
			b = 240;
			break;
		case 42: // Transparent Green-Yellow
			r = 204;
			g = 255;
			b = 0;
			break;
		case 43: // Transparent Cyan
			r = 171;
			g = 253;
			b = 249;
			break;
		case 44: // Transparent Light Red
			r = 255;
			g = 0;
			b = 0;
			break;
		case 45: // Transparent Pink
			r = 255;
			g = 176;
			b = 204;
			break;
		case 46: // Transparent Yellow
			r = 240;
			g = 196;
			b = 0;
			break;
		case 47: // Clear
			r = 255;
			g = 255;
			b = 255;
			break;
		case 57: // Transparent Orange
			r = 255;
			g = 102;
			b = 0;
			break;
		case 334: // Gold
			r = 240;
			g = 176;
			b = 51;
			break;
		case 383: // Chrome
			r = 204;
			g = 204;
			b = 204;
			break;
		case 494: // Electrical Contacts
			r = 204;
			g = 204;
			b = 204;
			break;
		default:
			if (colorNumber >= 256 && colorNumber < 512)
			{
				int colorNumber1;
				int colorNumber2;
				int r1, g1, b1;
				int r2, g2, b2;

				colorNumber -= 256;
				colorNumber1 = colorNumber / 16;
				colorNumber2 = colorNumber % 16;
				getRGBA(colorNumber1, r1, g1, b1, a);
				getRGBA(colorNumber2, r2, g2, b2, a);
				r = (r1 + r2) / 2;
				g = (g1 + g2) / 2;
				b = (b1 + b2) / 2;
			}
			else if (colorNumber >= 0x2000000 && colorNumber < 0x4000000)
			{
				r = (colorNumber & 0xFF0000) >> 16;
				g = (colorNumber & 0xFF00) >> 8;
				b = (colorNumber & 0xFF);
			}
			else if (colorNumber >= 0x4000000 && colorNumber < 0x5000000)
			{
				r = (((colorNumber & 0xF00000) >> 20) * 17 +
					((colorNumber & 0xF00) >> 8) * 17) / 2;
				g = (((colorNumber & 0xF0000) >> 16) * 17 +
					((colorNumber & 0xF0) >> 4) * 17) / 2;
				b = (((colorNumber & 0xF000) >> 12) * 17 +
					(colorNumber & 0xF) * 17) / 2;
			}
			else if ((colorNumber >= 0x5000000 && colorNumber < 0x6000000) ||
				(colorNumber >= 0x7000000 && colorNumber < 0x8000000))
			{
				r = ((colorNumber & 0xF00000) >> 20) * 17;
				g = ((colorNumber & 0xF0000) >> 16) * 17;
				b = ((colorNumber & 0xF000) >> 12) * 17;
				if (colorNumber >= 0x7000000 && colorNumber < 0x8000000)
				{
					a = 0;
				}
			}
			else if (colorNumber >= 0x6000000 && colorNumber < 0x7000000)
			{
				r = ((colorNumber & 0xF00) >> 8) * 17;
				g = ((colorNumber & 0xF0) >> 4) * 17;
				b = (colorNumber & 0xF) * 17;
			}
			else
			{
				r = 255;
				g = 128;
				b = 0;
			}
			break;
	}
}

void LDrawModel::newTriangle(int colorNumber, const Vector& p1,
							 const Vector& p2, const Vector& p3,
							 int *returnIndex)
{
	TGLShape* triangle = new TGLShape;
	int r, g, b, a;
	int index = colinearIndex(p1, p2, p3);
	char pointBuf[64] = "";

	if (returnIndex)
	{
		*returnIndex = index;
	}
	switch (index)
	{
	case 0:
		triangle->setPoints(p2, p3);
		p1.print(pointBuf);
		break;
	case 1:
		triangle->setPoints(p1, p3);
		p2.print(pointBuf);
		break;
	case 2:
		triangle->setPoints(p1, p2);
		p3.print(pointBuf);
		break;
	default:
		triangle->setPoints(p1, p2, p3);
		break;
	}
	getRGBA(colorNumber, r, g, b, a);
	triangle->setColorNumber(colorNumber);
	triangle->setColor(r, g, b, a);
	triangle->setParent(this);
	if (index >= 0)
	{
		addLine(triangle);
		if (!returnIndex)
		{
			reportError("Triangle contains co-linear points.\n"
				"Point %d <%s> removed.\n", LDMEColinear, index + 1, pointBuf);
		}
	}
	else
	{
		addTriangle(triangle);
	}
}

void LDrawModel::readTriangle(char* buf)
{
	int colorNumber;
	float x1, y1, z1;
	float x2, y2, z2;
	float x3, y3, z3;

#ifdef _LOG_PERFORMANCE
	LARGE_INTEGER startCounter;
	LARGE_INTEGER stopCounter;

	QueryPerformanceCounter(&startCounter);
#endif // _LOG_PERFORMANCE
	if (sscanf(buf, "3 %i %f %f %f %f %f %f %f %f %f", &colorNumber,
		&x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3) == 10)
	{
		Vector p1(x1, y1, z1);
		Vector p2(x2, y2, z2);
		Vector p3(x3, y3, z3);

		newTriangle(colorNumber, p1, p2, p3);
	}
	else
	{
		reportError("Error parsing triangle line.", LDMEParse);
//		printf("Error reading triangle line:\n%s\n", buf);
	}
#ifdef _LOG_PERFORMANCE
	QueryPerformanceCounter(&stopCounter);
	parseTriangleDelta += stopCounter.QuadPart - startCounter.QuadPart;
#endif // _LOG_PERFORMANCE
}

void LDrawModel::uncompile(void)
{
	if (flags.compiled)
	{
		checkError("Before uncompile");
		flags.compiled = 0;
		glDeleteLists(listId, 1);
		checkError("After uncompile");
	}
}

bool LDrawModel::compile(int execute)
{
//	float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
//								0.0f, 1.0f, 0.0f, 0.0f,
//								0.0f, 0.0f, 1.0f, 0.0f,
//								0.0f, 0.0f, 0.0f, 1.0f};
	bool drawResult;

	flags.compiling = true;
	checkError("Before compile");
	if (!reportProgress("Compiling Model.", 0.0f))
	{
		flags.compiling = false;
		return false;
	}
	if (flags.compiled)
	{
		flags.compiled = 0;
		glDeleteLists(listId, 1);
	}
	else
	{
		listId = glGenLists(1);
	}
	if (execute)
	{
		glNewList(listId, GL_COMPILE_AND_EXECUTE);
	}
	else
	{
		glNewList(listId, GL_COMPILE);
	}
	drawResult = draw();
//	draw(identityMatrix);
	glEndList();
	checkError("After compile");

	if (drawResult)
	{
		if (reportProgress("Compiling Model.", 1.0f))
		{
			flags.compiled = 1;
			reportProgress(NULL, 2.0f);
			flags.compiling = false;
			return true;
		}
	}
	glDeleteLists(listId, 1);
	flags.compiled = false;
	flags.compiling = false;
	return false;
}

int LDrawModel::middleIndex(const Vector& p0, const Vector& p1, const Vector& p2)
{
	Vector min = p0;
	Vector max = p0;
	int i;

	for (i = 0; i < 3; i++)
	{
		if (p1.get(i) < min.get(i))
		{
			min[i] = p1.get(i);
		}
		if (p1.get(i) > max.get(i))
		{
			max[i] = p1.get(i);
		}
	}
	for (i = 0; i < 3; i++)
	{
		if (p2.get(i) < min.get(i))
		{
			min[i] = p2.get(i);
		}
		if (p2.get(i) > max.get(i))
		{
			max[i] = p2.get(i);
		}
	}
	if (p0 != min && p0 != max)
	{
		return 0;
	}
	else if (p1 != min && p1 != max)
	{
		return 1;
	}
	else
	{
		return 2;
	}
}

int LDrawModel::colinearIndex(const Vector& p0, const Vector& p1,
	const Vector& p2, const Vector& p3)
{
	Vector normal = (p0 - p3) * (p0 - p1);

	if (normal.length() == 0.0f/*fEq(normal.length(), 0.0f)*/)
	{
		int index = middleIndex(p0, p1, p3);

		switch (index)
		{
		case 0:
			return 0;
		case 1:
			return 1;
		case 2:
			return 3;
		default:
			return 0;
		}
	}
	normal = (p1 - p0) * (p1 - p2);
	if (normal.length() == 0.0f/*fEq(normal.length(), 0.0f)*/)
	{
		return middleIndex(p0, p1, p2);
	}
	normal = (p2 - p1) * (p2 - p3);
	if (normal.length() == 0.0f/*fEq(normal.length(), 0.0f)*/)
	{
		return middleIndex(p1, p2, p3) + 1;
	}
	normal = (p3 - p2) * (p3 - p0);
	if (normal.length() == 0.0f/*fEq(normal.length(), 0.0f)*/)
	{
		int index = middleIndex(p0, p2, p3);

		switch (index)
		{
		case 0:
			return 0;
		case 1:
			return 2;
		case 2:
			return 3;
		default:
			return 0;
		}
	}
	return -1;
}

int LDrawModel::colinearIndex(const Vector& p0, const Vector& p1,
	const Vector& p2)
{
	Vector normal = (p0 - p2) * (p0 - p1);

	if (fEq(normal.length(), 0.0f))
	{
		return middleIndex(p0, p1, p2);
	}
/*
	normal = (p1 - p0) * (p1 - p2);
	if (fEq(normal.length(), 0.0f))
	{
		return middleIndex(p0, p1, p2);
	}
	normal = (p2 - p1) * (p2 - p0);
	if (fEq(normal.length(), 0.0f))
	{
		return middleIndex(p0, p1, p2);
	}
*/
	return -1;
}

int LDrawModel::swapNeeded(const Vector& p0, const Vector& p1, const Vector& p2,
						   const Vector& p3)
{
	Vector normal1 = (p0 - p3) * (p0 - p1);
	Vector normal2 = (p1 - p0) * (p1 - p2);
	Vector normal3 = (p2 - p1) * (p2 - p3);
	Vector normal4 = (p3 - p2) * (p3 - p0);
	float length1 = normal1.length();
	float length2 = normal2.length();
	float length3 = normal3.length();
	float length4 = normal4.length();

	if (fEq(length1, 0.0) || fEq(length2, 0.0) ||
		fEq(length3, 0.0) || fEq(length4, 0.0))
	{
		return 0;
	}
	normal1 /= length1;
	normal2 /= length2;
	normal3 /= length3;
	normal4 /= length4;
	return (normal1.dot(normal2) <= 0.0 ||
			normal1.dot(normal3) <= 0.0 ||
			normal1.dot(normal4) <= 0.0 ||
			normal2.dot(normal3) <= 0.0 ||
			normal2.dot(normal4) <= 0.0 ||
			normal3.dot(normal4) <= 0.0);
}

void LDrawModel::swapPoints(Vector& p0, Vector& p1)
{
	Vector tmp = p0;

	p0 = p1;
	p1 = tmp;
}

int LDrawModel::swapPointsIfNeeded(Vector& p0, Vector& p1, Vector& p2,
								   Vector& p3)
{
	if (swapNeeded(p0, p1, p2, p3))
	{
		if (!swapNeeded(p0, p1, p3, p2))
		{
			swapPoints(p2, p3);
		}
		else if (!swapNeeded(p0, p2, p1, p3))
		{
			swapPoints(p1, p2);
		}
		else if (!swapNeeded(p0, p2, p3, p1))
		{
			Vector tmp = p1;

			p1 = p2;
			p2 = p3;
			p3 = tmp;
		}
		else if (!swapNeeded(p0, p3, p1, p2))
		{
			Vector tmp = p1;

			p1 = p3;
			p3 = p2;
			p2 = tmp;
		}
		else if (!swapNeeded(p0, p3, p2, p1))
		{
			swapPoints(p1, p3);
		}
		else
		{
			return 0;
		}
		return 1;
	}
	else
	{
		return 1;
	}
}

int LDrawModel::reportQuadSplit(const Vector& q1, const Vector& q2,
								const Vector& q3, const Vector& q4,
								const Vector& t1, const Vector& t2,
								const Vector& t3, const Vector& t4,
								const Vector& t5, const Vector& t6)
{
	char q1Buf[64], q2Buf[64], q3Buf[64], q4Buf[64];
	char t1Buf[64], t2Buf[64], t3Buf[64], t4Buf[64], t5Buf[64], t6Buf[64];

	q1.print(q1Buf);
	q2.print(q2Buf);
	q3.print(q3Buf);
	q4.print(q4Buf);
	t1.print(t1Buf);
	t2.print(t2Buf);
	t3.print(t3Buf);
	t4.print(t4Buf);
	t5.print(t5Buf);
	t6.print(t6Buf);
	return reportError("Concave quad split into two triangles.\n"
		"Original Quad: <%s> <%s> <%s> <%s>\n"
		"Triangle 1: <%s> <%s> <%s>\n"
		"Triangle 2: <%s> <%s> <%s>\n", LDMEConcaveQuad,
		q1Buf, q2Buf, q3Buf, q4Buf, t1Buf, t2Buf, t3Buf, t4Buf, t5Buf, t6Buf);
}



bool LDrawModel::splitConcavePoints(int colorNumber, const Vector& p1,
									const Vector& p2, const Vector& p3,
									const Vector& p4)
{
	TGLShape* triangle1 = new TGLShape;
	TGLShape* triangle2 = new TGLShape;
	int r, g, b, a;
	Vector normal1 = (p1 - p4) * (p1 - p2);
	Vector normal2 = (p2 - p1) * (p2 - p3);
	Vector normal3 = (p3 - p2) * (p3 - p4);
	Vector normal4 = (p4 - p3) * (p4 - p1);
	float length1 = normal1.length();
	float length2 = normal2.length();
	float length3 = normal3.length();
	float length4 = normal4.length();

	if (fEq(length1, 0.0) || fEq(length2, 0.0) ||
		fEq(length3, 0.0) || fEq(length4, 0.0))
	{
		// Ack!
		return NO;
	}
	normal1 /= length1;
	normal2 /= length2;
	normal3 /= length3;
	normal4 /= length4;
	if (normal1.dot(normal2) <= 0.0)
	{
		triangle1->setPoints(p1, p2, p4);
		triangle2->setPoints(p2, p3, p4);
		reportQuadSplit(p1, p2, p3, p4, p1, p2, p4, p2, p3, p4);
	}
	else if (normal2.dot(normal3) <= 0.0)
	{
		triangle1->setPoints(p1, p2, p3);
		triangle2->setPoints(p1, p3, p4);
		reportQuadSplit(p1, p2, p3, p4, p1, p2, p3, p1, p3, p4);
	}
	else if (normal3.dot(normal4) <= 0.0)
	{
		triangle1->setPoints(p1, p2, p4);
		triangle2->setPoints(p2, p3, p4);
		reportQuadSplit(p1, p2, p3, p4, p1, p2, p4, p2, p3, p4);
	}
	else if (normal4.dot(normal1) <= 0.0)
	{
		triangle1->setPoints(p1, p2, p3);
		triangle2->setPoints(p1, p3, p4);
		reportQuadSplit(p1, p2, p3, p4, p1, p2, p3, p1, p3, p4);
	}
	else
	{
		// Ack!
		return NO;
	}
	getRGBA(colorNumber, r, g, b, a);
	triangle1->setColorNumber(colorNumber);
	triangle1->setColor(r, g, b, a);
	triangle1->setParent(this);
	addTriangle(triangle1);
	triangle2->setColorNumber(colorNumber);
	triangle2->setColor(r, g, b, a);
	triangle2->setParent(this);
	addTriangle(triangle2);
	return YES;
}

void LDrawModel::newQuad(int colorNumber, Vector& p1, Vector& p2, Vector& p3,
						 Vector& p4)
{
	if (swapPointsIfNeeded(p1, p2, p3, p4))
	{
		int index1 = colinearIndex(p1, p2, p3, p4);

		if (index1 >= 0)
		{
			char pointBuf1[64] = "";
			char pointBuf2[64] = "";
			int index2 = -1;

			switch (index1)
			{
			case 0:
				newTriangle(colorNumber, p2, p3, p4, &index2);
				p1.print(pointBuf1);
				if (index2 >= 0)
				{
					index2++;
				}
				break;
			case 1:
				newTriangle(colorNumber, p1, p3, p4, &index2);
				p2.print(pointBuf1);
				if (index2 > 0)
				{
					index2++;
				}
				break;
			case 2:
				newTriangle(colorNumber, p1, p2, p4, &index2);
				p3.print(pointBuf1);
				if (index2 > 1)
				{
					index2++;
				}
				break;
			case 3:
				newTriangle(colorNumber, p1, p2, p3, &index2);
				p4.print(pointBuf1);
				break;
			default:
				break;
			}
			if (index2 >= 0)
			{
				switch (index2)
				{
				case 0:
					p1.print(pointBuf2);
					break;
				case 1:
					p2.print(pointBuf2);
					break;
				case 2:
					p3.print(pointBuf2);
					break;
				case 3:
					p4.print(pointBuf2);
					break;
				default:
					break;
				}
				if (index1 < index2)
				{
					reportError("Quad contains co-linear points.\n"
						"Point %d <%s> and point %d <%s> removed.\n",
						LDMEColinear, index1 + 1, pointBuf1, index2 + 1,
						pointBuf2);
				}
				else
				{
					reportError("Quad contains co-linear points.\n"
						"Point %d <%s> and point %d <%s> removed.\n",
						LDMEColinear, index2 + 1, pointBuf2, index1 + 1,
						pointBuf1);
				}
			}
			else
			{
				reportError("Quad contains co-linear points.\n"
					"Point %d <%s> removed.\n", LDMEColinear, index1 + 1,
					pointBuf1);
			}
		}
		else
		{
			TGLShape* quad = new TGLShape;
			int r, g, b, a;

			getRGBA(colorNumber, r, g, b, a);
			quad->setPoints(p1, p2, p3, p4);
			quad->setColorNumber(colorNumber);
			quad->setColor(r, g, b, a);
			quad->setParent(this);
			addQuad(quad);
		}
	}
	else
	{
		if (!splitConcavePoints(colorNumber, p1, p2, p3, p4) &&
			!splitConcavePoints(colorNumber, p1, p2, p4, p3) &&
			!splitConcavePoints(colorNumber, p1, p3, p2, p4) &&
			!splitConcavePoints(colorNumber, p1, p3, p4, p2) &&
			!splitConcavePoints(colorNumber, p1, p4, p2, p3) &&
			!splitConcavePoints(colorNumber, p1, p4, p3, p2))
		{
			reportError("Unable to determine split for concave quad.\n",
				LDMEConcaveQuadSplit);
//			printf("Unable to split concave quad!\n");
			// Ack!
		}
	}
}

void LDrawModel::readQuad(char* buf)
{
	int colorNumber;
	float x1, y1, z1;
	float x2, y2, z2;
	float x3, y3, z3;
	float x4, y4, z4;

#ifdef _LOG_PERFORMANCE
	LARGE_INTEGER startCounter;
	LARGE_INTEGER stopCounter;

	QueryPerformanceCounter(&startCounter);
#endif // _LOG_PERFORMANCE
	if (sscanf(buf, "4 %i %f %f %f %f %f %f %f %f %f %f %f %f",
		&colorNumber, &x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3,
		&x4, &y4, &z4) == 13)
	{
		Vector p1(x1, y1, z1);
		Vector p2(x2, y2, z2);
		Vector p3(x3, y3, z3);
		Vector p4(x4, y4, z4);

		if (p1 == p2)
		{
			newTriangle(colorNumber, p1, p3, p4);
		}
		else if (p1 == p3 || p2 == p3)
		{
			newTriangle(colorNumber, p1, p2, p4);
		}
		else if (p1 == p4 || p2 == p4 || p3 == p4)
		{
			newTriangle(colorNumber, p1, p2, p3);
		}
		else
		{
			newQuad(colorNumber, p1, p2, p3, p4);
		}
	}
	else
	{
		reportError("Error parsing quad line.", LDMEParse);
//		printf("Error reading quad line:\n%s\n", buf);
	}
#ifdef _LOG_PERFORMANCE
	QueryPerformanceCounter(&stopCounter);
	parseQuadDelta += stopCounter.QuadPart - startCounter.QuadPart;
#endif // _LOG_PERFORMANCE
}

void LDrawModel::newConditionalLine(int colorNumber, const Vector& p1,
									const Vector& p2, const Vector& p3,
									const Vector& p4)
{
	TGLShape* line = new TGLShape;
	int r, g, b, a;

	getRGBA(colorNumber, r, g, b, a);
	line->setPoints(p1, p2);
	line->setNumControlPoints(2);
	line->setControlPoint(0, p3);
	line->setControlPoint(1, p4);
	line->setColorNumber(colorNumber);
	line->setColor(r, g, b, a);
	line->setParent(this);
	addLine(line);
}

bool LDrawModel::shouldLoadConditionalLines(void)
{
	return (flags.drawHighlights && flags.drawConditionalHighlights) ||
		flags.performSmoothing;
}

void LDrawModel::readConditionalLine(char* buf)
{
	int colorNumber;
	float x1, y1, z1;
	float x2, y2, z2;
	float x3, y3, z3;
	float x4, y4, z4;

	if (sscanf(buf, "5 %i %f %f %f %f %f %f %f %f %f %f %f %f", &colorNumber,
		&x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3, &x4, &y4, &z4) == 13)
	{
		if (shouldLoadConditionalLines() || colorNumber != 24)
		{
			Vector p1(x1, y1, z1);
			Vector p2(x2, y2, z2);
			Vector p3(x3, y3, z3);
			Vector p4(x4, y4, z4);

			newConditionalLine(colorNumber, p1, p2, p3, p4);
		}
	}
	else
	{
		reportError("Error parsing conditional line line.", LDMEParse);
//		printf("Error reading line line:\n%s\n", buf);
	}
}

void LDrawModel::addFileLine(char* line)
{
	LDMFileLine *fileLine = new LDMFileLine(line, currentLine,
		currentLineNumber);

	if (!fileLines)
	{
		fileLines = new LDMFileLineArray;
	}
	fileLines->addObject(fileLine);
	fileLine->release();
}

/*
void LDrawModel::addShape(TCObjectArray*& shapes, TGLShape* shape)
{
	if (!shapes)
	{
		shapes = new TCObjectArray;
	}
	shapes->addObject(shape);
	shape->release();
}
*/

void LDrawModel::addQuadStrip(TGLShape* quadStrip)
{
	addShape(LDM_QUAD_STRIPS, quadStrip);
//	addShape(quadStrips, quadStrip);
}

void LDrawModel::addQuad(TGLShape* quad)
{
	addShape(LDM_QUADS, quad);
//	addShape(quads, quad);
}

void LDrawModel::addTriangle(TGLShape* triangle)
{
	addShape(LDM_TRIANGLES, triangle);
//	addShape(triangles, triangle);
}

void LDrawModel::addLine(TGLShape* line)
{
	int r, g, b, a;

	line->getRGBA(r, g, b, a);
	if (a == 255)
	{
		addShape(LDM_LINES, line);
	}
	else
	{
		line->setColor(r, g, b, 255);
		addShape(LDM_TRANS_LINES, line);
	}
//	addShape(lines, line);
}

void LDrawModel::addLineStrip(TGLShape* lineStrip)
{
	int r, g, b, a;

	lineStrip->getRGBA(r, g, b, a);
	if (a == 255)
	{
		addShape(LDM_LINE_STRIPS, lineStrip);
	}
	else
	{
		lineStrip->setColor(r, g, b, 255);
		addShape(LDM_TRANS_LINE_STRIPS, lineStrip);
	}
}

/*
void LDrawModel::addSubModel(LDrawModel* subModel)
{
	if (subModelCount >= subModelsAllocated)
	{
		LDrawModel** temp = subModels;
		float** oldMatrices = transformationMatrices;
		int* oldColors = colors;
		int* oldColorNumbers = colorNumbers;

		subModelsAllocated *= 2;
		if (subModelsAllocated == 0)
		{
			subModelsAllocated = 4;
		}
		subModels = new LDrawModel*[subModelsAllocated];
		transformationMatrices = new float*[subModelsAllocated];
		colors = new int[subModelsAllocated];
		colorNumbers = new int[subModelsAllocated];
		memcpy(subModels, temp, subModelCount * sizeof(LDrawModel*));
		memcpy(transformationMatrices, oldMatrices,
			subModelCount * sizeof(float*));
		memset(transformationMatrices + subModelCount, 0,
			(subModelsAllocated - subModelCount) * sizeof(float*));
		memcpy(colors, oldColors, subModelCount * sizeof(int));
		memcpy(colorNumbers, oldColorNumbers, subModelCount * sizeof(int));
		delete temp; // Don't array delete, since moved to new array.
		delete oldMatrices; // Ditto.
		delete oldColors; // Ditto.
		delete oldColorNumbers; // Ditto.
	}
	subModels[subModelCount] = subModel;
	subModelCount++;
}
*/

void LDrawModel::checkError(char* msg)
{
	GLenum errorCode;
	
	while ((errorCode = glGetError()) != GL_NO_ERROR)
	{
		char* errorString;

		switch (errorCode)
		{
			case GL_INVALID_ENUM:
				errorString = "GL_INVALID_ENUM";
				break;
			case GL_INVALID_VALUE:
				errorString = "GL_INVALID_VALUE";
				break;
			case GL_INVALID_OPERATION:
				errorString = "GL_INVALID_OPERATION";
				break;
			case GL_STACK_OVERFLOW:
				errorString = "GL_STACK_OVERFLOW";
				break;
			case GL_STACK_UNDERFLOW:
				errorString = "GL_STACK_UNDERFLOW";
				break;
			case GL_OUT_OF_MEMORY:
				errorString = "GL_OUT_OF_MEMORY";
				break;
			default:
				errorString = "Unknown Error";
				break;
		}
		debugPrintf("%s: %s\n", msg, errorString);
		reportError("OpenGL error:\n%s: %s\n", LDMEOpenGL, msg, errorString);
#ifdef _DEBUG
//		_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "%s: %s\n", msg, errorString);
#endif
	}
}

void LDrawModel::setCullBackFaces(int value)
{
	if (value != flags.cullBackFaces)
	{
		int i;
		int count = newSubModels->getCount();

		flags.cullBackFaces = value;
		for (i = 0; i < count; i++)
		{
			newSubModels->objectAtIndex(i)->getModel()->setCullBackFaces(value);
		}
/*
		for (i = 0; i < subModelCount; i++)
		{
			subModels[i]->setCullBackFaces(value);
		}
*/
	}
}

void LDrawModel::applySpecialColorSettings(int colorNumber, int /*color*/)
{
	int lastColorNumber = TGLShape::getLastColorNumber();

	if (colorNumber != lastColorNumber)
	{
		int i;
		bool found = false;

		for (i = 0; i < NUM_SPECIAL_COLORS && !found; i++)
		{
			if (colorNumber == specialColors[i].colorNumber)
			{
				float *specular = specialColors[i].specular;
				TGLShape::setSpecular(specular[0], specular[1], specular[2],
					specular[3], specialColors[i].shininess);
				found = true;
//				TGLShape::setSpecular(0.075f, 0.075f, 0.075f, -1.0f, 15.0f);
			}
		}
/*
		if (colorNumber == 256)
		{
			// Black rubber
			TGLShape::setSpecular(0.075f, 0.075f, 0.075f, -1.0f, 15.0f);
		}
		else if (colorNumber == 334)
		{
			// Gold
			float scale = 255.0f / 240.0f * 2.0f;

			TGLShape::setSpecular(240.0f / 255.0f * scale,
				176.0f / 255.0f * scale,
				51.0f / 255.0f * scale, -1.0f, 5.0f);
		}
		else if (colorNumber == 383 || colorNumber == 494)
		{
			// Chrome or Electrical Contacts (rgb=204)
			float scale = 1.0f;

			TGLShape::setSpecular(scale, scale, 1.5f, -1.0f, 5.0f);
		}
*/
		if (!found && (lastColorNumber == 256 || lastColorNumber == 334 ||
			lastColorNumber == 383 || lastColorNumber == 494))
		{
			TGLShape::setToNormalSpecular();
		}
	}
}

void LDrawModel::drawShapes(TGLShapeArray* shapes, float* matrix,
							int colorNumber, int transparent)
{
	int count = getArrayCount(shapes);

	if (count)
	{
		int i;
		int begun = 0;
		GLdouble modelMatrix[4*4];
		GLdouble projectionMatrix[4*4];
		GLint viewMatrix[4];
		double comboMatrix[16];

		if (transparent == 5)
		{
			glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
			glGetDoublev(GL_PROJECTION_MATRIX, projectionMatrix);
			glGetIntegerv(GL_VIEWPORT, viewMatrix);
			TGLShape::multMatrixd(projectionMatrix, modelMatrix, comboMatrix);
		}
		for (i = 0; i < count; i++)
		{
			TGLShape* shape = (TGLShape*)((*shapes)[i]);
			int shapeColorNumber = shape->getColorNumber();
			int r, g, b, a;
			int color;

			if (shapeColorNumber == 16)
			{
				shapeColorNumber = colorNumber;
			}
			else if (shapeColorNumber == 24)
			{
				shapeColorNumber = getHighlightColor(colorNumber);
			}
			if (shape->isConditional())
			{
				if (transparent == 5)
				{
					// if processing conditional lines
					if (shape->skipConditionalLine(matrix, comboMatrix))
					{
						// conditional line not visible so skip it.
						continue;
					}
					// else it's visible, so drawPoints() below.
				}
				else
				{
					// The shape is conditional, but we aren't processing
					// conditional lines right now, so skip.
					continue;
				}
			}
			else if (transparent == 5)
			{
				 // do NOT draw normal bits when processing contitional line.
				continue;
			}
			else if (transparent)
			{
				//If doing transparent and the shape is NOT, then skip it.
				if (!shape->isTransparent())
				{
					continue;
				}
			}
			else
			{
				//else if NOT doing transparent and the shape is, then skip it.
				if (shape->isTransparent())
				{
					continue;
				}
			}
			getRGBA(shapeColorNumber, r, g, b, a);
			if (shape->getGlMode() == GL_LINES ||
				shape->getGlMode() == GL_LINE_STRIP)
			{
				a = 255;
			}
			color = TGLShape::colorForRGBA(r, g, b, a);
			applySpecialColorSettings(shapeColorNumber, color);
			TGLShape::setGlColor(color, shapeColorNumber);
//			if (color != lastColor)
//			{
//				glColor4ubv((GLubyte*)&(color));
//				lastColor = color;
//			}
			if (!begun)
			{
				shape->begin();
				begun = 1;
			}
			shape->drawPoints(matrix);
			if (transparent || shape->getGlMode() == GL_QUAD_STRIP ||
				shape->getGlMode() == GL_TRIANGLE_FAN ||
				shape->getGlMode() == GL_TRIANGLE_STRIP ||
				shape->getGlMode() == GL_LINE_STRIP)
			{
				shape->end();
				begun = 0;
			}
		}
		if (begun)
		{
			((TGLShape*)(*shapes)[0])->end();
		}
	}
}

void LDrawModel::writePointToFile(Vector &point, FILE *outputFile)
{
	if (fEq(point[0], 0.0f))
	{
		point[0] = 0.0f;
	}
	if (fEq(point[1], 0.0f))
	{
		point[1] = 0.0f;
	}
	if (fEq(point[2], 0.0f))
	{
		point[2] = 0.0f;
	}
	fprintf(outputFile, " %.4g %.4g %.4g", point[0], point[1],
		point[2]);
}

void LDrawModel::writeShapeToFile(TGLShape* shape, FILE* outputFile,
								  float* matrix)
{
	int i;
	int count = shape->getNumPoints();

	if (shape->isConditional())
	{
		fprintf(outputFile, "5 24");
	}
	else
	{
		fprintf(outputFile, "%d 16", count);
	}
	for (i = 0; i < count; i++)
	{
		Vector point = shape->transformPoint(shape->getPoints()[i],
			matrix);

		writePointToFile(point, outputFile);
	}
	if (shape->isConditional())
	{
		writePointToFile(shape->getControlPoints()[0], outputFile);
		writePointToFile(shape->getControlPoints()[1], outputFile);
	}
	fprintf(outputFile, "\n");
}

void LDrawModel::writeToFile(FILE* outputFile, float* matrix)
{
	int i;
	int count;

	count = getLineCount();
	for (i = 0; i < count; i++)
	{
		writeShapeToFile((TGLShape*)(*getShapes(LDM_LINES))[i], outputFile,
			matrix);
	}
	count = getTriangleCount();
	for (i = 0; i < count; i++)
	{
		writeShapeToFile((TGLShape*)(*getShapes(LDM_TRIANGLES))[i], outputFile,
			matrix);
	}
	count = getQuadCount();
	for (i = 0; i < count; i++)
	{
		writeShapeToFile((TGLShape*)(*getShapes(LDM_QUADS))[i], outputFile,
			matrix);
	}
	count = newSubModels->getCount();
	for (i = 0; i < count; i++)
	{
		LDrawSubModel *subModel = newSubModels->objectAtIndex(i);
		float newMatrix[16];

		TGLShape::multMatrix(matrix, subModel->getTransformationMatrix(),
			newMatrix);
		subModel->getModel()->writeToFile(outputFile, newMatrix);
	}
/*
	for (i = 0; i < subModelCount; i++)
	{
		float newMatrix[16];

		TGLShape::multMatrix(matrix, transformationMatrices[i], newMatrix);
		subModels[i]->writeToFile(outputFile, newMatrix);
	}
*/
}

void LDrawModel::writeToFile(void)
{
	FILE* outputFile = fopen("ldview.dat", "w");

	if (outputFile)
	{
		float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
									0.0f, 1.0f, 0.0f, 0.0f,
									0.0f, 0.0f, 1.0f, 0.0f,
									0.0f, 0.0f, 0.0f, 1.0f};

		writeToFile(outputFile, identityMatrix);
		fclose(outputFile);
	}
}

//static float currentMatrix[16];

int LDrawModel::shapeSortFunction(const void* left, const void* right)
{
	TGLShape* leftShape = *(TGLShape**)left;
	TGLShape* rightShape = *(TGLShape**)right;

	if (leftShape->getNumPoints() == 0)
	{
		if (rightShape->getNumPoints() == 0)
		{
			return 0;
		}
		else
		{
			return -1;
		}
	}
	else if (rightShape->getNumPoints() == 0)
	{
		return 1;
	}
	else
	{
		Vector leftPoint = leftShape->getCentroids()[1];
		Vector rightPoint = rightShape->getCentroids()[1];
		float leftSize;
		float rightSize;

		leftSize = leftPoint.length();
		rightSize = rightPoint.length();
		if (leftSize > rightSize)
		{
			return -1;
		}
		else if (leftSize == rightSize)
		{
			return 0;
		}
		return 1;
	}
}

void LDrawModel::prepareForTransparentSort(float* matrix)
{
	TGLShapeArray* theseShapes = (*shapes)[NUM_SHAPE_TYPES];
	int i;

	for (i = theseShapes->getCount() - 1; i >= 0; i--)
	{
		(*theseShapes)[i]->calcCentroid(matrix);
	}
/*
	theseShapes = (*shapes)[NUM_SHAPE_TYPES + LDM_LINES];
	for (i = theseShapes->getCount() - 1; i >= 0; i--)
	{
		(*theseShapes)[i]->calcCentroid(matrix);
	}
*/
}

void LDrawModel::sortTransparentShapes(void)
{
	float currentMatrix[16];

	glGetFloatv(GL_MODELVIEW_MATRIX, currentMatrix);
	prepareForTransparentSort(currentMatrix);
	(*shapes)[NUM_SHAPE_TYPES]->sortUsingFunction(shapeSortFunction);
//	(*shapes)[NUM_SHAPE_TYPES + LDM_LINES]->sortUsingFunction(shapeSortFunction);
}

bool LDrawModel::draw(void)
{
	float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
								0.0f, 1.0f, 0.0f, 0.0f,
								0.0f, 0.0f, 1.0f, 0.0f,
								0.0f, 0.0f, 0.0f, 1.0f};

//	checkError("LDrawModel::draw 1");
	TGLStudLogo::attachTexture();
	TGLShape::forceGlColor();
	TGLShape::resetNormalSpecular();
//	lastColor = UNUSED_COLOR;
	if (flags.compiled)
	{
		glCallList(listId);
		if (flags.drawConditionalHighlights && flags.drawHighlights)
		{
			bool drawResult;

			drawResult = draw(identityMatrix, topParent->defaultColorNumber, 5);
		}
		if (shouldSortTransparent())
		{
			if (!flags.cutawayDraw)
			{
				sortTransparentShapes();
			}
			return drawTransparent(identityMatrix);
		}
		else
		{
			return true;
		}
	}
	else
	{
		bool drawResult;

		TGLShape::forceGlColor();
//		lastColor = UNUSED_COLOR;
		TGLShape::resetCounts();
		drawResult = draw(identityMatrix);
		TGLShape::printCounts();
		return drawResult;
	}
}

void LDrawModel::enableLighting(void)
{
	if (useLighting)
	{
		glEnable(GL_LIGHTING);
	}
	else
	{
		glDisable(GL_LIGHTING);
	}
}

void LDrawModel::disableLighting(void)
{
	glDisable(GL_LIGHTING);
//	checkError("disableLighting");
}

bool LDrawModel::drawTransparent(float* matrix)
{
	bool drawResult;
	float specular[] = {0.75f, 0.75f, 0.75f, 1.0f};

	TGLShape::forceGlColor();
//	lastColor = UNUSED_COLOR;
	if (flags.useStipple)
	{
		glEnable(GL_POLYGON_STIPPLE);
	}
	else if (!flags.cutawayDraw)
	{
//		glDepthMask(0);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_FALSE);
	}
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.0f);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	if (polygonOffsetFunc && !drawWireframe)
	{
		if (usePolygonOffset && flags.drawHighlights)
		{
			(*polygonOffsetFunc)(0.0, 0.0);
		}
		else
		{
			glEnable(GL_POLYGON_OFFSET_FILL);
			(*polygonOffsetFunc)(-POLYGON_OFFSET_FACTOR, -POLYGON_OFFSET_UNITS);
//			(*polygonOffsetFunc)(-1.0, -0.5);
		}
	}
	drawResult = draw(matrix, 7, 1);
	if (polygonOffsetFunc)
	{
		glDisable(GL_POLYGON_OFFSET_FILL);
	}
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, TGLShape::getNormalSpecular());
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, TGLShape::getNormalShininess());
	if (flags.useStipple)
	{
		glDisable(GL_POLYGON_STIPPLE);
	}
	else if (!flags.cutawayDraw)
	{
		glDisable(GL_BLEND);
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
//		glDepthMask(1);
	}
	return drawResult;
}

bool LDrawModel::draw(float* matrix)
{
	bool drawResult;

//	lastColor = UNUSED_COLOR;
	if (usePolygonOffset && flags.drawHighlights && polygonOffsetFunc &&
		!drawWireframe)
	{
		glEnable(GL_POLYGON_OFFSET_FILL);
		(*polygonOffsetFunc)(POLYGON_OFFSET_FACTOR, POLYGON_OFFSET_UNITS);
	}
	drawResult = draw(matrix, topParent->defaultColorNumber, 0);
	if (!drawResult)
	{
		return false;
	}

	if (shouldSortTransparent())
	{
		if (polygonOffsetFunc)
		{
			glDisable(GL_POLYGON_OFFSET_FILL);
		}
		return true;
	}
	else
	{
		return drawTransparent(matrix);
	}
}

bool LDrawModel::draw(int shapeIndex, float* matrix, int colorNumber,
					  int transparent, float progressFraction,
					  float progressStartFraction)
{
	int i;
	int count;

	drawShapes(getShapes(shapeIndex), matrix, colorNumber, transparent);
	count = newSubModels->getCount();
	for (i = 0; i < count; i++)
	{
		LDrawSubModel *subModel = newSubModels->objectAtIndex(i);
		float newMatrix[16];

		TGLShape::multMatrix(matrix, subModel->getTransformationMatrix(),
			newMatrix);
		if (subModel->getColorNumber() == 16 && colorNumber != 16)
		{
			if (!subModel->getModel()->draw(shapeIndex, newMatrix, colorNumber,
				transparent, progressFraction, progressStartFraction))
			{
				return false;
			}
		}
		else
		{
			if (!subModel->getModel()->draw(shapeIndex, newMatrix,
				subModel->getColorNumber(), transparent, progressFraction,
				progressStartFraction))
			{
				return false;
			}
		}
		if (!transparent)
		{
			if (!reportProgress(NULL, progressStartFraction +
				(float)i / (count + 1) * progressFraction))
			{
				return false;
			}
		}
	}
	if (transparent && transparent != 5 && !shouldSortTransparent())
	{
		if (!reportProgress(NULL, progressStartFraction +
			(float)count / (count + 1) * progressFraction))
		{
			return false;
		}
	}
	return true;
}

bool LDrawModel::draw(float* matrix, int colorNumber, int transparent)
{
	int i;
	int start;
	int end;
	bool retValue = true;
	int progressNumerator = 0;
	int progressDenom = NUM_SHAPE_TYPES;

	if (!shouldSortTransparent())
	{
		progressDenom *= 2;
	}
	if (transparent && transparent != 5)
	{
		progressNumerator = 5;
	}
	// Process QUAD_STRIPS, QUADS, and TRIANGLES with lighting enabled.
	enableLighting();
	start = LDM_QUAD_STRIPS;
	end = LDM_TRIANGLES;
	if (transparent && transparent != 5 && shouldSortTransparent())
	{
		start += NUM_SHAPE_TYPES;
		end += NUM_SHAPE_TYPES;
	}
	for (i = start; i <= end; i++)
	{
		if (!draw(i, matrix, colorNumber, transparent,
			1.0f / progressDenom, (float)progressNumerator / progressDenom))
		{
			return false;
		}
		progressNumerator++;
	}
	// Process LINES and LINE_STRIPS with lighting disabled
	disableLighting();
	start = LDM_LINES;
	end = LDM_LINE_STRIPS;
	if (transparent && transparent != 5 && shouldSortTransparent())
	{
		start += NUM_SHAPE_TYPES;
		end += NUM_SHAPE_TYPES;
	}
	if (flags.lineSmoothing && !transparent)
	{
		glPushAttrib(GL_ENABLE_BIT);
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	for (i = start; i <= end && retValue; i++)
	{
		retValue = draw(i, matrix, colorNumber, transparent,
			1.0f / progressDenom, (float)progressNumerator / progressDenom);
		progressNumerator++;
	}
	if (flags.lineSmoothing && !transparent)
	{
		glPopAttrib();
	}
	start = LDM_TRANS_LINES;
	end = LDM_TRANS_LINE_STRIPS;
	for (i = start; i <= end && retValue; i++)
	{
		retValue = draw(i, matrix, colorNumber, transparent,
			1.0f / progressDenom, (float)progressNumerator / progressDenom);
		progressNumerator++;
	}
	return retValue;
}

void LDrawModel::getMinMax(TGLShapeArray* shapes, Vector& min, Vector& max,
						   float* matrix)
{
	int count = getArrayCount(shapes);
	int i;

	for (i = 0; i < count; i++)
	{
		(/*(TGLShape*)*/(*shapes)[i])->getMinMax(min, max, matrix);
	}
}

void LDrawModel::getMinMax(Vector& min, Vector& max, float* matrix)
{
	int i;
	int count = shapes->getCount();
	int subCount = newSubModels->getCount();

/*
	getMinMax(quadStrips, min, max, matrix);
	getMinMax(quads, min, max, matrix);
	getMinMax(triangles, min, max, matrix);
	getMinMax(lines, min, max, matrix);
	getMinMax(lineStrips, min, max, matrix);
*/
	for (i = 0; i < count; i++)
	{
		getMinMax(getShapes(i), min, max, matrix);
	}
	for (i = 0; i < subCount; i++)
	{
		LDrawSubModel *subModel = newSubModels->objectAtIndex(i);
		float newMatrix[16];

		TGLShape::multMatrix(matrix, subModel->getTransformationMatrix(),
			newMatrix);
		subModel->getModel()->getMinMax(min, max, newMatrix);
	}
/*
	if (subModelCount)
	{
//		glMatrixMode(GL_MODELVIEW);
//		glPushMatrix();
		for (i = 0; i < subModelCount; i++)
		{
			float newMatrix[16];

			TGLShape::multMatrix(matrix, transformationMatrices[i], newMatrix);
			subModels[i]->getMinMax(min, max, newMatrix);
		}
//		glPopMatrix();
	}
*/
}

void LDrawModel::getMinMax(Vector& min, Vector& max)
{
	float identityMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
								0.0f, 1.0f, 0.0f, 0.0f,
								0.0f, 0.0f, 1.0f, 0.0f,
								0.0f, 0.0f, 0.0f, 1.0f};
	min[0] = 9999999.0f;
	min[1] = 9999999.0f;
	min[2] = 9999999.0f;
	max[0] = -9999999.0f;
	max[1] = -9999999.0f;
	max[2] = -9999999.0f;
	getMinMax(min, max, identityMatrix);
}

double absvec( int n, double vec[] )
{
	double mag=0.0;
	int i;

	for(i=0; i<n; i++)
	{
		mag += vec[i]*vec[i];
	}

	if(mag < 0.000001)
	{
		vec[0] = 1.0;
		return 0.0;
	}

	mag = sqrt(mag);
	for(i=0; i<n; i++)
	{
		vec[i] /= mag;
	}

	return mag;
}

/*
void LDrawModel::setTransformation(int index, float x, float y, float z,
								   float a, float b, float c,
								   float d, float e, float f,
								   float g, float h, float i)
{
	if (!transformationMatrices[index])
	{
		transformationMatrices[index] = new float[16];
		transformationMatrices[index][3] = 0.0f;
		transformationMatrices[index][7] = 0.0f;
		transformationMatrices[index][11] = 0.0f;
		transformationMatrices[index][15] = 1.0f;
	}
	transformationMatrices[index][0] = a;
	transformationMatrices[index][4] = b;
	transformationMatrices[index][8] = c;
	transformationMatrices[index][1] = d;
	transformationMatrices[index][5] = e;
	transformationMatrices[index][9] = f;
	transformationMatrices[index][2] = g;
	transformationMatrices[index][6] = h;
	transformationMatrices[index][10] = i;
	transformationMatrices[index][12] = x;
	transformationMatrices[index][13] = y;
	transformationMatrices[index][14] = z;
}
*/

void LDrawModel::setTransformation(LDrawSubModel *subModel,
								   float x, float y, float z,
								   float a, float b, float c,
								   float d, float e, float f,
								   float g, float h, float i)
{
	float transformationMatrix[16];

	transformationMatrix[3] = 0.0f;
	transformationMatrix[7] = 0.0f;
	transformationMatrix[11] = 0.0f;
	transformationMatrix[15] = 1.0f;
	transformationMatrix[0] = a;
	transformationMatrix[4] = b;
	transformationMatrix[8] = c;
	transformationMatrix[1] = d;
	transformationMatrix[5] = e;
	transformationMatrix[9] = f;
	transformationMatrix[2] = g;
	transformationMatrix[6] = h;
	transformationMatrix[10] = i;
	transformationMatrix[12] = x;
	transformationMatrix[13] = y;
	transformationMatrix[14] = z;
	subModel->setTransformationMatrix(transformationMatrix);
}

/*
void LDrawModel::setGlColor(int index)
{
	int colorNumber = newSubModels->objectAtIndex(index)->getColorNumber();
//	int colorNumber = colorNumbers[index];

	if (colorNumber == 16)
	{
	}
	else if (colorNumber == 24)
	{
	}
	else
	{
		int color = newSubModels->objectAtIndex(index)->getColor();
//		int color = colors[index];

//		glColor4ubv((GLubyte*)&(color));
		TGLShape::setGlColor(color, colorNumber);
//		if (color != lastColor)
//		{
//			glColor4ubv((GLubyte*)&(color));
//			lastColor = color;
//		}
	}
}
*/

void LDrawModel::setColor(int index, int r, int g, int b, int a)
{
	newSubModels->objectAtIndex(index)->
		setColor(TGLShape::colorForRGBA(r, g, b, a));
//	colors[index] = TGLShape::colorForRGBA(r, g, b, a);
}

void LDrawModel::setColor(int index, char* c)
{
	int r, g, b;
	int num;

	if (strlen(c) == 5)
	{ // 12 bit color spec: 0xRGB
		sscanf(c, "%i", &num);
		r = (num >> 8) * 17;
		g = ((num >> 4) & 0xF) * 17;
		b = (num & 0xF) * 17;
		setColor(index, r, g, b, 255);
	}
	else if (strlen(c) == 8)
	{ // 24 bit color spec: 0xRRGGBB
		sscanf(c, "%i", &num);
		r = num >> 16;
		g = (num >> 8) & 0xFF;
		b = num & 0xFF;
		setColor(index, r, g, b, 255);
	}
	else
	{
		reportError("Unknown color spec: %s\n", LDMEColor, c);
//		printf("Error:  unknown color spec: %s\n", c);
		exit(1);
	}
}

