#ifndef __LDRAWMODELVIEWER_H__
#define __LDRAWMODELVIEWER_H__

#include <TCFoundation/TCObject.h>
#include <LDLib/LDrawModel.h>
#include <LDLib/TGLCamera.h>

typedef enum
{
	LDVStereoNone,
	LDVStereoHardware,
	LDVStereoCrossEyed,
	LDVStereoParallel
} LDVStereoMode;

typedef enum
{
	LDVCutawayNormal,
	LDVCutawayWireframe,
	LDVCutawayStencil
} LDVCutawayMode;

class TCImage;

class LDrawModelViewer: public TCObject
{
	public:
		LDrawModelViewer(int, int);
		virtual void update(void);
		virtual void perspectiveView(void);
		virtual void perspectiveView(bool resetViewport);
		void setQualityLighting(bool value) { flags.qualityLighting = value; }
		bool getQualityLighting(void) { return flags.qualityLighting; }
		virtual void setSubduedLighting(bool value);
		bool getSubduedLighting(void) { return flags.subduedLighting; }
		virtual void setShowsHighlightLines(bool value);
		bool getShowsHighlightLines(void) { return flags.showsHighlightLines; }
		virtual void setDrawConditionalHighlights(bool value);
		bool getDrawConditionalHighlights(void)
		{
			return flags.drawConditionalHighlights;
		}
		virtual void setPerformSmoothing(bool value);
		bool getPerformSmoothing(void)
		{
			return flags.performSmoothing;
		}
		void setConstrainZoom(bool value) { flags.constrainZoom = value; }
		bool getConstrainZoom(void) { return flags.constrainZoom; }
		virtual void setLineSmoothing(bool value);
		bool getLineSmoothing(void)
		{
			return flags.lineSmoothing;
		}
		virtual void setQualityStuds(bool value);
/*
		{
			flags.qualityStuds = value;
			flags.needsReload = true;
		}
*/
		bool getQualityStuds(void) { return flags.qualityStuds; }
		virtual void setAllowPrimitiveSubstitution(bool value);
/*
		{
			flags.allowPrimitiveSubstitution = value;
			flags.needsReload = true;
		}
*/
		bool getAllowPrimitiveSubstitution(void)
		{
			return flags.allowPrimitiveSubstitution;
		}
		void setUsesFlatShading(bool value) { flags.usesFlatShading = value; }
		bool getUsesFlatShading(void) { return flags.usesFlatShading; }
		virtual void setUsesSpecular(bool value);
		bool getUsesSpecular(void) { return flags.usesSpecular; }
		virtual void setOneLight(bool value);
		bool getOneLight(void) { return flags.oneLight; }
		virtual void setWidth(int value);
		virtual void setHeight(int value);
		int getWidth(void) { return width; }
		int getHeight(void) { return height; }
		void setXRotate(float value) { xRotate = value; }
		void setYRotate(float value) { yRotate = value; }
		void setZRotate(float value) { zRotate = value; }
		float getXRotate(void) { return xRotate; }
		float getYRotate(void) { return yRotate; }
		float getZRotate(void) { return zRotate; }
		void panXY(int xValue, int yValue);
		void setRotationSpeed(float value);
		void setCameraXRotate(float value) { cameraXRotate = value; }
		void setCameraYRotate(float value) { cameraYRotate = value; }
		void setCameraZRotate(float value) { cameraZRotate = value; }
		float getCameraXRotate(void) { return cameraXRotate; }
		float getCameraYRotate(void) { return cameraYRotate; }
		float getCameraZRotate(void) { return cameraZRotate; }
		void setCameraMotion(const Vector &value) { cameraMotion = value; }
		Vector getCameraMotion(void) { return cameraMotion; }
		virtual void setZoomSpeed(float value);
		float getZoomSpeed(void) { return zoomSpeed; }
		virtual void zoom(float);
		void setClipZoom(bool value) { clipZoom = value; }
		bool getClipZoom(void) { return clipZoom; }
		virtual void setFilename(const char*);
		virtual void setProgramPath(const char *value);
		virtual void setFileIsPart(bool);
		char* getFilename(void) { return filename; }
		virtual int loadModel(bool = true);
		virtual void drawFPS(float);
		virtual void drawBoundingBox(void);
		virtual void setup(void);
		virtual void setBackgroundRGB(int r, int g, int b);
		virtual void setBackgroundRGBA(int r, int g, int b, int a);
		int getBackgroundR(void) { return (int)(backgroundR * 255); }
		int getBackgroundG(void) { return (int)(backgroundG * 255); }
		int getBackgroundB(void) { return (int)(backgroundB * 255); }
		int getBackgroundA(void) { return (int)(backgroundA * 255); }
		virtual void setDefaultRGB(TCByte r, TCByte g, TCByte b);
		virtual void getDefaultRGB(TCByte &r, TCByte &g, TCByte &b);
		virtual void setDefaultColorNumber(int value);
		int getDefaultColorNumber(void) { return defaultColorNumber; }
		virtual void setSeamWidth(float);
		float getSeamWidth(void) { return LDrawModel::getSeamWidth(); }
		virtual void setDrawWireframe(bool);
		bool getDrawWireframe(void) { return flags.drawWireframe; }
		virtual void setUseWireframeFog(bool);
		bool getUseWireframeFog(void) { return flags.useWireframeFog; }
		virtual void setUsePolygonOffset(bool);
		bool getUsePolygonOffset(void) { return flags.usePolygonOffset; }
		virtual void setUseLighting(bool);
		bool getUseLighting(void) { return flags.useLighting; }
		virtual void setUseStipple(bool);
		bool getUseStipple(void) { return flags.useStipple; }
		virtual void setSortTransparent(bool);
		bool getSortTransparent(void) { return flags.sortTransparent; }
		virtual void setHighlightLineWidth(float value);
		float getHighlightLineWidth(void) { return highlightLineWidth; }
		virtual void setWireframeLineWidth(float value);
		float getWireframeLineWidth(void) { return wireframeLineWidth; }
		virtual bool recompile(void);
		virtual void uncompile(void);
		virtual void reload(void);
		virtual void setProgressCallback(LDMProgressCallback callback,
			void* userData);
		virtual void setErrorCallback(LDMErrorCallback callback,
			void* userData);
		virtual void clear(void);
		virtual void resetView(void);
		virtual void pause(void);
		virtual void unpause(void);
		bool getPaused(void) { return flags.paused; }
		virtual void setXTile(int value);
		int getXTile(void) { return xTile; }
		virtual void setYTile(int value);
		int getYTile(void) { return yTile; }
		virtual void setNumXTiles(int value);
		int getNumXTiles(void) { return numXTiles; }
		virtual void setNumYTiles(int value);
		int getNumYTiles(void) { return numYTiles; }
		virtual void setStereoMode(LDVStereoMode mode);
		LDVStereoMode getStereoMode(void) { return stereoMode; }
		void setStereoEyeSpacing(GLfloat spacing)
		{
			stereoEyeSpacing = spacing;
		}
		GLfloat getStereoEyeSpacing(void) { return stereoEyeSpacing; }
		virtual void setCutawayMode(LDVCutawayMode mode);
		LDVCutawayMode getCutawayMode(void) { return cutawayMode; }
		virtual void setCutawayAlpha(GLfloat value);
		GLfloat getCutawayAlpha(void) { return cutawayAlpha; }
		virtual void setCutawayLineWidth(GLfloat value);
		GLfloat getCutawayLineWidth(void) { return cutawayLineWidth; }
		void setSlowClear(bool value) { flags.slowClear = value; }
		bool getSlowClear(void) { return flags.slowClear; }
		virtual void setBlackHighlights(bool value);
		bool getBlackHighlights(void) { return flags.blackHighlights; }
		void setZoomMax(float value) { zoomMax = value; }
		float getZoomMax(void) { return zoomMax; }
		void setShowAllConditionalLines(bool value)
		{
			flags.showAllConditionalLines = value;
		}
		bool getShowAllConditionalLines(void)
		{
			return flags.showAllConditionalLines;
		}
		void setShowConditionalControlPoints(bool value)
		{
			flags.showConditionalControlPoints = value;
		}
		bool getShowConditionalControlPoints(void)
		{
			return flags.showConditionalControlPoints;
		}
		bool getNeedsReload(void) { return flags.needsReload != 0; }
		void setCurveQuality(int value);
		int getCurveQuality(void) { return curveQuality; }
		void setTextureStuds(bool value);
		bool getTextureStuds(void) { return flags.textureStuds; }
		void setTextureFilterType(int value);
		int getTextureFilterType(void) { return textureFilterType; }
		LDrawModel* getLDrawModel(void) { return lDrawModel; }
		bool getCompiled(void)
		{
			if (lDrawModel)
			{
				return lDrawModel->getCompiled() ? true : false;
			}
			else
			{
				return false;
			}
		}
		void setPixelAspectRatio(float value) { pixelAspectRatio = value; }
		float getPixelAspectRatio(void) { return pixelAspectRatio; }
		bool getLDrawCommandLineMatrix(char *matrixString, int bufferLength);
		bool getLDrawCommandLine(char *shortFilename, char *commandString,
			int bufferLength);
		bool getLDGLiteCommandLine(char *commandString, int bufferLength);
		void setDistanceMultiplier(GLfloat value)
		{
			distanceMultiplier = value;
		}
		GLfloat getDistanceMultiplier(void) { return distanceMultiplier; }
		virtual void clearBackground(void);
		virtual void setFontData(TCByte *fontData, long length);

		static void setPolygonOffsetFunc(GLPolygonOffsetFunc func)
		{
			LDrawModel::setPolygonOffsetFunc(func);
		}
		static GLPolygonOffsetFunc getPolygonOffsetFunc(void)
		{
			return LDrawModel::getPolygonOffsetFunc();
		}
		static char *getOpenGLDriverInfo(int &numExtensions);
	protected:
		~LDrawModelViewer(void);
		void dealloc(void);
		virtual void drawSetup(GLfloat eyeXOffset = 0.0f);
		virtual void drawModel(GLfloat eyeXOffset = 0.0f);
		virtual void setFieldOfView(double, float, float);
		virtual void setupMaterial(void);
		virtual void setupLight(GLenum);
		virtual void orthoView(void);
		void drawLight(GLenum, float, float, float);
		virtual void setupLighting(void);
		virtual void setupTextures(void);
		void drawLights(void);
		virtual void drawToClipPlane(GLfloat eyeXOffset);
		virtual void drawToClipPlaneUsingStencil(GLfloat eyeXOffset);
//		virtual void drawToClipPlaneUsingAccum(GLfloat eyeXOffset);
		virtual void drawToClipPlaneUsingDestinationAlpha(GLfloat eyeXOffset);
		virtual void drawToClipPlaneUsingNoEffect(GLfloat eyeXOffset);
		virtual void perspectiveViewToClipPlane(void);
		virtual void applyTile(void);
		virtual void drawString(GLfloat xPos, GLfloat yPos, char* string);
		virtual void setupFont(char *fontFilename);
		virtual void loadVGAFont(char *fontFilename);

		LDrawModel* lDrawModel;
		char* filename;
		char* programPath;
		int width;
		int height;
		float pixelAspectRatio;
		float size;
		Vector center;
		Vector min;
		Vector max;
		int lastFrameTimeLength;
		int cullBackFaces;
		float xRotate;
		float yRotate;
		float zRotate;
		float rotationSpeed;
		float cameraXRotate;
		float cameraYRotate;
		float cameraZRotate;
		Vector cameraMotion;
		float zoomSpeed;
//		float distance;
		float xPan;
		float yPan;
		float* rotationMatrix;
		float clipAmount;
		float highlightLineWidth;
		float wireframeLineWidth;
		bool clipZoom;
		GLuint fontListBase;
		GLclampf backgroundR;
		GLclampf backgroundG;
		GLclampf backgroundB;
		GLclampf backgroundA;
		TCByte defaultR;
		TCByte defaultG;
		TCByte defaultB;
		int defaultColorNumber;
		LDMProgressCallback progressCallback;
		void* progressUserData;
		LDMErrorCallback errorCallback;
		void* errorUserData;
		int xTile;
		int yTile;
		int numXTiles;
		int numYTiles;
		LDVStereoMode stereoMode;
		GLfloat stereoEyeSpacing;
		LDVCutawayMode cutawayMode;
		GLfloat cutawayAlpha;
		GLfloat cutawayLineWidth;
		float zoomMax;
		int curveQuality;
		int textureFilterType;
		GLfloat distanceMultiplier;
		TCImage *fontImage;
		GLuint fontTextureID;
		TGLCamera camera;
		float aspectRatio;
		float currentFov;
		struct
		{
			bool qualityLighting:1;
			bool showsHighlightLines:1;
			bool qualityStuds:1;
			bool usesFlatShading:1;
			bool usesSpecular:1;
			bool drawWireframe:1;
			bool useWireframeFog:1;
			bool usePolygonOffset:1;
			bool useLighting:1;
			bool subduedLighting:1;
			bool allowPrimitiveSubstitution:1;
			bool useStipple:1;
			bool fileIsPart:1;
			bool sortTransparent:1;
			bool needsSetup:1;
			bool needsTextureSetup:1;
			bool needsMaterialSetup:1;
			bool needsLightingSetup:1;
			bool needsReload:1;
			bool needsRecompile:1;
			bool needsResize:1;
			bool paused:1;
			bool slowClear:1;
			bool blackHighlights:1;
			bool textureStuds:1;
			bool oneLight:1;
			bool drawConditionalHighlights:1;
			bool showAllConditionalLines:1;
			bool showConditionalControlPoints:1;
			bool performSmoothing:1;
			bool lineSmoothing:1;
			bool constrainZoom:1;
//			int _reserved:0;
		} flags;
};

#endif // __LDRAWMODELVIEWER_H__
