#include <stdlib.h>
#include <string.h>
#include <TCFoundation/mystring.h>

#include "ModelFile.h"
#include "ModelMacros.h"

// ModelFile::ModelFile(void) -- Default Constructor
ModelFile::ModelFile(void)
	 :file(NULL),
	  outFile(stdout),
	  comment(NULL),
	  currentLine(NULL),
	  currentLineNumber(0)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "ModelFile");
#endif
}

// ModelFile::ModelFile(FILE*) -- Constructor
// Expects:
//		file	: File to read from.
ModelFile::ModelFile(FILE* file)
	 :offset(ftell(file)),
	  file(file),
	  outFile(stdout),
	  comment(NULL),
	  currentLine(NULL),
	  currentLineNumber(0)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "ModelFile");
#endif
}

ModelFile::~ModelFile(void)
{
}

void ModelFile::dealloc(void)
{
	if (file != NULL)
	{
		closeFile();
	}
	// Don't close outFile (stdout)
	delete comment;
	delete currentLine;
	TCObject::dealloc();
}

void ModelFile::closeFile(void)
{
	fclose(file);
	file = NULL;
}

// ModelFile::setFile(FILE*) -- Member Function
// Expects:
//		f	: File to read from.
void ModelFile::setFile(FILE* f, int size)
{
	file = f;
	offset = ftell(file);
	if (size == -1)
	{
		// Since we can load multiple models out of one file, we might have
		// already known our size; no need to re-discover it.
		if (fseek(file, 0, SEEK_END) == 0)
		{
			filesize = ftell(file);
			fseek(file, offset, SEEK_SET);
		}
		else
		{
			filesize = -1;
		}
	}
	else
	{
		filesize = size;
	}
}

int ModelFile::stripLine(char* line)
{
	char* spot;

	while (strchr(line, '\t') != NULL)
		*strchr(line, '\t') = ' ';
	if (strchr(line, '\n') != NULL)
		*strchr(line, '\n') = 0;
	if (strchr(line, '\r') != NULL)
		*strchr(line, '\r') = 0;
	if (strstr(line, "//") != NULL)
	{
		spot = strstr(line, "//") + 2;
		delete comment;
		comment = new char[strlen(spot)+1];
		strcpy(comment, spot);
		*strstr(line, "//") = 0;
	}
#if DEBUG_VR > 1
	printf("<<%s>>\n", line);
#endif
	for (spot = line; *spot == ' '; spot++)
		;
	if (*spot)
	{
		return 0;
	}
	else
	{
		return 1;
	}
}

// ModelFile::getLine(char*) -- Member Function
// Expects:
//		line	: Buffer to store the line in.
// Results:
//		Reads a non-blank (after the removal of comments and white space) line
//		from the file and stores it in "line".
int ModelFile::getLine(char* line)
{
	offset = ftell(file);
	debugPut("Entered ModelFile::getLine()", 3);
	while (1)
	{
		delete currentLine;
		if (fgets(line, MODEL_BUF_SIZE, file) == NULL)
		{
			currentLine = NULL;
			currentLineNumber = -currentLineNumber;
			return 1;
		}
		currentLine = copyString(line);
		currentLineNumber++;
#if DEBUG_VR > 0
		printf(">>%s", line);
#endif
		if (!stripLine(line))
		{
			break;
		}
	}
#if DEBUG_VR > 1
	printf("About to exit ModelFile::getLine()\n");
#endif
	return 0;
}

// ModelFile::unGetLine(void) -- Member Function
// Results:
//		Backs up in the file by one line.  This will only work once between calls
//		to ModelFile::getLine().
int ModelFile::unGetLine(void)
{
	if (ftell(file) == offset)
		return 0;
	fseek(file, offset, SEEK_SET);
	return 1;
}
