#include <string.h>
#include "TGLCamera.h"
#ifdef WIN32
#include <windows.h>
#endif // WIN32
#include "ModelMacros.h"
#include <TCFoundation/mystring.h>
#include <GL/gl.h>

TGLCamera::TGLCamera(void)
	:name(NULL)
{
}

TGLCamera::~TGLCamera(void)
{
	delete name;
}

TGLCamera& TGLCamera::duplicate(const TGLCamera& copyFrom)
{
	position = copyFrom.position;
	facing = copyFrom.facing;
	setName(copyFrom.name);
	return *this;
}

void TGLCamera::setName(char* n)
{
	delete name;
	name = copyString(n);
}


void TGLCamera::project(const Vector &distance)
{
	float inverseMatrix[16];
	Vector center = -distance;

	facing.getInverseMatrix(inverseMatrix);
	center = -distance.mult(inverseMatrix) - position;
	glMultMatrixf(inverseMatrix);
	glTranslatef(center[0], center[1], center[2]);
}

void TGLCamera::rotate(const Vector &rotation)
{
	float inverseMatrix[16];

	facing.getInverseMatrix(inverseMatrix);
	if (!fEq(rotation.get(1), 0.0))
	{
		facing = facing +
			TGLFacing(Vector(-1,0,0).mult(inverseMatrix).normalize(),
			rotation.get(1));
	}
	if (!fEq(rotation.get(0), 0.0))
	{
		facing = facing +
			TGLFacing(Vector(0,1,0).mult(inverseMatrix).normalize(),
			rotation.get(0));
	}
	if (!fEq(rotation.get(2), 0.0))
	{
		facing = facing +
			TGLFacing(Vector(0,0,1).mult(inverseMatrix).normalize(),
			rotation.get(2));
	}
}

void TGLCamera::move(const Vector &distance)
{
	float inverseMatrix[16];
//	float *matrix;
	Vector v = distance;

//	matrix = facing.getMatrix();
	facing.getInverseMatrix(inverseMatrix);
	position += v.mult(inverseMatrix);
}
