#include <stdlib.h>
#include <string.h>

#include "TGLEighthSphere.h"

#include "LDrawModel.h"
#include "ModelMacros.h"

TGLEighthSphere::TGLEighthSphere(void)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLEighthSphere");
#endif
}

TGLEighthSphere::TGLEighthSphere(TGLEighthSphere &other)
				:TGLCustomShape(other)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLEighthSphere");
#endif
}

TGLEighthSphere::~TGLEighthSphere(void)
{
}

void TGLEighthSphere::dealloc(void)
{
	TGLCustomShape::dealloc();
}

TCObject *TGLEighthSphere::copy(void)
{
	return new TGLEighthSphere(*this);
}

int TGLEighthSphere::sphereIndex(int i, int j, int usedSegments)
{
	int retVal = 0;
	int k;

	for (k = 0; k < j; k++)
	{
		int rowSize = usedSegments - k;

		retVal += rowSize * 2 + 1;
	}
	return retVal + i;
}

void TGLEighthSphere::drawPoints(float* matrix)
{
	int i, j;
	int usedSegments = parent->getNumCircleSegments() / 4;
	Vector p1, p2, p3, p4, p5, p6, p7;
	Vector* newPoints = new Vector[numPoints];
	Vector normal;
	Vector lastNormal = Vector(0.0f, 0.0f, 0.0f);
	int numMainPoints = (usedSegments + 1) * (usedSegments + 1) - 1;

	triangleCount += usedSegments * usedSegments;
	triangleStripCount += usedSegments;
	for (i = 0; i < numPoints; i++)
	{
		newPoints[i] = transformPoint(points[i], matrix);
	}
	for (j = 0; j < usedSegments; j++)
	{
		int rowSize = usedSegments - j;

		glBegin(GL_TRIANGLE_STRIP);
		for (i = 0; i < rowSize * 2 + 1; i++)
		{
			int mainIndex = sphereIndex(i, j, usedSegments);

			p1 = newPoints[mainIndex];
			if (i < rowSize * 2)
			{
				p2 = newPoints[mainIndex + 1];
				if (i < rowSize * 2 - 1)
				{
					p3 = newPoints[mainIndex + 2];
				}
				else
				{
					p3 = newPoints[numMainPoints + usedSegments + j + 1];
				}
			}
			else
			{
				p2 = newPoints[numMainPoints + usedSegments + j + 1];
				p3 = newPoints[numMainPoints + usedSegments + j];
			}
			if (i > 0)
			{
				p4 = newPoints[mainIndex - 1];
				if (i > 1)
				{
					p5 = newPoints[mainIndex - 2];
				}
				else
				{
					if (j > 0)
					{
						p5 = newPoints[numMainPoints + 3 * usedSegments - j];
					}
					else
					{
						p5 = newPoints[numMainPoints];
					}
				}
			}
			else
			{
				if (j > 0)
				{
					p4 = newPoints[numMainPoints + 3 * usedSegments - j];
					if (j > 1)
					{
						p5 = newPoints[numMainPoints + 3 * usedSegments -
							(j - 1)];
					}
					else
					{
						p5 = newPoints[numMainPoints];
					}
				}
				else
				{
					p4 = newPoints[numMainPoints];
					p5 = newPoints[numMainPoints + 1];
				}
			}
			if (i % 2)
			{
				normal = ((p1 - p3) * (p1 - p2)).normalize();
				normal += ((p1 - p2) * (p1 - p4)).normalize();
				normal += ((p1 - p4) * (p1 - p5)).normalize();
				if (j < usedSegments - 1)
				{
					if (i > 1)
					{
						p6 = newPoints[sphereIndex(i - 2, j + 1, usedSegments)];
					}
					else
					{
						p6 = newPoints[numMainPoints + usedSegments * 3 - j -
							1];
					}
					if (i < (rowSize * 2) - 1)
					{
						p7 = newPoints[sphereIndex(i, j + 1, usedSegments)];
					}
					else
					{
						p7 = newPoints[numMainPoints + usedSegments + j + 2];
					}
					normal += ((p1 - p5) * (p1 - p6)).normalize();
					normal += ((p1 - p6) * (p1 - p7)).normalize();
					normal += ((p1 - p7) * (p1 - p3)).normalize();
				}
				else
				{
					p6 = newPoints[numMainPoints + usedSegments * 3 - j - 1];
					normal += ((p1 - p5) * (p1 - p6)).normalize();
				}
			}
			else
			{
				normal = ((p1 - p2) * (p1 - p3)).normalize();
				normal += ((p1 - p4) * (p1 - p2)).normalize();
				normal += ((p1 - p5) * (p1 - p4)).normalize();
				if (j > 0)
				{
					p6 = newPoints[sphereIndex(i, j - 1, usedSegments)];
					p7 = newPoints[sphereIndex(i + 2, j - 1, usedSegments)];
					normal += ((p1 - p6) * (p1 - p5)).normalize();
					normal += ((p1 - p7) * (p1 - p6)).normalize();
					normal += ((p1 - p3) * (p1 - p7)).normalize();
				}
				else
				{
					if (i > 0)
					{
						p6 = newPoints[numMainPoints + (i / 2)];
						normal += ((p1 - p6) * (p1 - p5)).normalize();
						if (i < rowSize * 2)
						{
							p7 = newPoints[numMainPoints + (i / 2) + 1];
							normal += ((p1 - p7) * (p1 - p6)).normalize();
							normal += ((p1 - p3) * (p1 - p7)).normalize();
						}
					}
					else
					{
						normal += ((p1 - p3) * (p1 - p5)).normalize();
					}
				}
			}
			normal.normalize();
			if (normal != lastNormal)
			{
				glNormal3fv(normal);
				lastNormal = normal;
			}
//			p2 = normal - p1;
			glVertex3fv(p1);
			vertexCount++;
		}
		glEnd();
	}
	delete[] newPoints;
}
