#include <stdlib.h>
#include <string.h>
//#include <png.h>

#include "TGLShape.h"

#include "Vector.h"
#include "LDrawModel.h"
#include "ModelMacros.h"
#include <TCFoundation/mystring.h>

// Need gluProject()
#include <GL/glu.h>

#ifndef WIN32
#include <netinet/in.h>
#endif // WIN32

int TGLShape::triangleCount = 0;
int TGLShape::triangleStripCount = 0;
int TGLShape::triangleFanCount = 0;
int TGLShape::quadCount = 0;
int TGLShape::quadStripCount = 0;
int TGLShape::lineCount = 0;
int TGLShape::vertexCount = 0;
int TGLShape::totalShapes = 0;
bool TGLShape::showConditionalControlPoints = true;
bool TGLShape::showAllConditionalLines = true;
int TGLShape::lastColor = 0;
int TGLShape::lastColorNumber = 0;
bool TGLShape::forceColor = true;
//bool TGLShape::wasRubber = NO;
float TGLShape::normalSpecular[4] = {0.0f, 0.0f, 0.0f, 0.0f};
float TGLShape::normalShininess = 128.0f;

TGLShape::TGLShape(void)
		 :points(NULL),
		  normals(NULL),
		  transformedPoints(NULL),
		  transformedNormals(NULL),
		  controlPoints(NULL),
		  centroids(NULL),
		  color(-1),
		  colorNumber(16),
		  parent(NULL),
		  numPoints(0),
		  numControlPoints(0),
		  userData(NULL),
		  mirrored(false)
//		  specialType(0)
//		  normal(NULL),
//		  backNormal(NULL),
//		  transformed(0)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLShape");
#endif
	totalShapes++;
}

TGLShape::TGLShape(TGLShape& other)
		 :points(NULL),
		  normals(NULL),
		  transformedPoints(NULL),
		  transformedNormals(NULL),
		  controlPoints(NULL),
		  centroids(NULL),
		  color(other.color),
		  colorNumber(other.colorNumber),
		  parent(other.parent),
//		  normal(other.normal? new Vector(*(other.normal)) : NULL),
//		  backNormal(other.backNormal ? new Vector(*(other.backNormal)) : NULL),
		  numPoints(other.numPoints),
		  numControlPoints(other.numControlPoints),
		  userData(other.userData),
		  glMode(other.glMode),
		  mirrored(other.mirrored)
//		  specialType(other.specialType)
//		  transformed(other.transformed)
{
	int i;

#ifdef _LEAK_DEBUG
	strcpy(className, "TGLShape");
#endif
	if (numPoints)
	{
		points = new Vector[numPoints];
		if (other.normals)
		{
			normals = new Vector[numPoints];
		}
		for (i = 0; i < numPoints; i++)
		{
			points[i] = other.points[i];
			if (normals)
			{
				normals[i] = other.normals[i];
			}
		}
	}
/*
	else
	{
		printf("Empty shape.");
	}
*/
	if (other.controlPoints)
	{
		controlPoints = new Vector[numControlPoints];
		for (i = 0; i < numControlPoints; i++)
		{
			controlPoints[i] = other.controlPoints[i];
		}
	}
	if (other.centroids)
	{
		centroids = new Vector[2];
		centroids[0] = other.centroids[0];
		centroids[1] = other.centroids[1];
	}
	totalShapes++;
}

TGLShape::~TGLShape(void)
{
}

void TGLShape::dealloc(void)
{
	delete[] points;
	delete[] normals;
	delete[] transformedPoints;
	delete[] transformedNormals;
	delete[] controlPoints;
	if (centroids)
	{
		delete[] centroids;
	}
//	delete normal;
//	delete backNormal;
	totalShapes--;
/*
	if (totalShapes < 10 || totalShapes % 100 == 0)
	{
		printf("TGLShape::dealloc: %d\n", totalShapes);
	}
*/
	TCObject::dealloc();
}

TCObject *TGLShape::copy(void)
{
	return new TGLShape(*this);
}

void TGLShape::setPoints(const Vector& p1, const Vector& p2)
{
	delete[] points;
	points = new Vector[2];
	points[0] = p1;
	points[1] = p2;
	numPoints = 2;
	glMode = GL_LINES;
}

void TGLShape::resetCounts(void)
{
	triangleCount = 0;
	triangleStripCount = 0;
	quadCount = 0;
	quadStripCount = 0;
	vertexCount = 0;
}

void TGLShape::setPoints(const Vector& p1, const Vector& p2, const Vector& p3)
{
	delete[] points;
	points = new Vector[3];
	points[0] = p1;
	points[1] = p2;
	points[2] = p3;
	numPoints = 3;
	glMode = GL_TRIANGLES;
//	compNormal();
}

void TGLShape::setPoints(const Vector& p1, const Vector& p2, const Vector& p3,
						 const Vector& p4)
{
	delete[] points;
	points = new Vector[4];
	points[0] = p1;
	points[1] = p2;
	points[2] = p3;
	points[3] = p4;
	numPoints = 4;
	glMode = GL_QUADS;
//	compNormal();
}

void TGLShape::setPoint(int index, const Vector& p)
{
	points[index] = p;
}

void TGLShape::setNormal(int index, const Vector& n)
{
	if (!normals)
	{
		normals = new Vector[numPoints];
	}
	normals[index] = n;
}

void TGLShape::setNumControlPoints(int value)
{
	if (value != numControlPoints)
	{
		delete[] controlPoints;
		numControlPoints = value;
		controlPoints = new Vector[numControlPoints];
	}
}

void TGLShape::setControlPoint(int index, const Vector& p)
{
	controlPoints[index] = p;
}

void TGLShape::setNumPoints(int num)
{
	numPoints = num;
	delete[] points;
	points = new Vector[numPoints];
	if (normals)
	{
		delete[] normals;
		normals = new Vector[numPoints];
	}
}

int TGLShape::colorForRGBA(int r, int g, int b, int a)
{
	return htonl((r & 0xFF) << 24 | (g & 0xFF) << 16 |
		(b & 0xFF) << 8 | (a & 0xFF)); // GL 24 bit color spec: 0xBBGGRRAA
}

void TGLShape::getRGBA(int &r, int &g, int &b, int &a)
{
	r = htonl(color) >> 24;
	g = (htonl(color) >> 16) & 0xFF;
	b = (htonl(color) >> 8) & 0xFF;
	a = htonl(color) & 0xFF;
}

void TGLShape::setColor(int r, int g, int b, int a)
{
	color = colorForRGBA(r, g, b, a);
}

bool TGLShape::isTransparent(void)
{
	return glMode != GL_LINES && glMode != GL_LINE_STRIP &&
		colorNumberIsTransparent(colorNumber);
}

bool TGLShape::colorNumberIsTransparent(int colorNumber)
{
	return (colorNumber >= 32 && colorNumber < 48) ||
		(colorNumber >= 0x3000000 && colorNumber < 0x4000000) ||
		(colorNumber >= 0x5000000 && colorNumber < 0x7000000) ||
		(colorNumber >= 0x7000000 && colorNumber < 0x8000000) ||
		colorNumber == 57;
}

void TGLShape::setColor(char* c)
{
	int r, g, b;
	int num;

	if (strlen(c) == 5)
	{ // 12 bit color spec: 0xRGB
		sscanf(c, "%i", &num);
		r = (num >> 8) * 17;
		g = ((num >> 4) & 0xF) * 17;
		b = (num & 0xF) * 17;
		setColor(r, g, b, 255);
	}
	else if (strlen(c) == 8)
	{ // 24 bit color spec: 0xRRGGBB
		sscanf(c, "%i", &num);
		r = num >> 16;
		g = (num >> 8) & 0xFF;
		b = num & 0xFF;
		setColor(r, g, b, 255);
	}
/*
	else
	{
		printf("Error:  unknown color spec: %s\n", c);
		exit(1);
	}
*/
}

Vector TGLShape::calcNormal(bool normalize)
{
	Vector normal = (points[numPoints-2] - points[numPoints-1]) *
		(points[numPoints-2] - points[numPoints-3]);
	if (normalize)
	{
		normal.normalize();
	}
	return normal;
}

/*
void TGLShape::compNormal(bool recompute)
{
	float len;

	if (normal)
	{
		if (!recompute)
		{
			return;
		}
	}
	{
		normal = new Vector;
	}
//	if (!backNormal)
//	{
//		backNormal = new Vector;
//	}
//	*backNormal = (points[1] - points[0]) * (points[1] - points[2]);
	*normal = (points[numPoints-2] - points[numPoints-1]) *
		(points[numPoints-2] - points[numPoints-3]);
	len = normal->length();
	if (len == 0.0f)
	{
		OutputDebugString("bleck\n");
	}
	else
	{
		*normal /= len;
	}
//	len = backNormal->length();
//	if (len == 0.0f)
//	{
//		OutputDebugString("bleck\n");
//	}
//	else
//	{
//		*backNormal /= len;
//	}
}
*/

void TGLShape::setParent(LDrawModel* p)
{
	parent = p;
}

#define TEXTURE_SIZE 0x80
//static GLubyte *g_texture = NULL;

/*
static GLubyte *loadTexture(void)
{
	GLubyte *texture = NULL;
	FILE *pngFile = fopen("C:\\TEMP\\test\\LDView\\LEGO.png", "rb");

	if (pngFile)
	{
		BYTE header[8];

		if (fread(header, 1, 8, pngFile) == 8)
		{
			fseek(pngFile, 0, SEEK_SET);
			if (png_sig_cmp(header, 0, 8) == 0)
			{
				png_structp png_ptr = png_create_read_struct(
					PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

				if (png_ptr)
				{
					png_infop info_ptr = png_create_info_struct(png_ptr);

					if (info_ptr)
					{
						png_infop end_info = png_create_info_struct(png_ptr);

						if (end_info)
						{
#pragma warning( push )
#pragma warning( disable : 4611 )
							if (!setjmp(png_jmpbuf(png_ptr)))
#pragma warning( pop )
							{
								png_bytep row_pointer;
								unsigned long imageWidth;
								unsigned long imageHeight;
								int bitDepth;
								int colorType;
								unsigned int i;

								png_init_io(png_ptr, pngFile);
								png_read_info(png_ptr, info_ptr);
//								png_read_png(png_ptr, info_ptr, 0, NULL);
								png_get_IHDR(png_ptr, info_ptr, &imageWidth,
									&imageHeight, &bitDepth, &colorType, NULL,
									NULL, NULL);
								row_pointer = new BYTE[imageWidth * 4];
								texture = new GLubyte[imageWidth * imageHeight *
									4];
								for (i = 0; i < imageHeight; i++)
								{
									png_read_row(png_ptr, row_pointer, NULL);
									memcpy(texture + (imageHeight - i - 1) *
										imageWidth * 4, row_pointer,
										imageWidth * 4);
								}
								delete row_pointer;
							}
							png_destroy_read_struct(&png_ptr, &info_ptr,
								&end_info);
						}
						else
						{
							png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
						}
					}
					else
					{
						png_destroy_read_struct(&png_ptr, NULL, NULL);
					}
				}
			}
		}
		fclose(pngFile);
	}
	return texture;
}
*/

/*
static GLubyte *genTexture(void)
{
	int i, j;
	int c;
	GLubyte *texture = NULL;//= loadTexture();
	
	if (texture)
	{
		return texture;
	}
	texture = new GLubyte[TEXTURE_SIZE * TEXTURE_SIZE * 4];
	memset(texture, 0, TEXTURE_SIZE * TEXTURE_SIZE * 4);
	for (i = 0; i < TEXTURE_SIZE; i++)
	{
		for (j = 0; j < TEXTURE_SIZE; j++)
		{
			c = (((i & 0x8) == 0) ^ ((j & 0x8) == 0)) * 255;
			texture[i * TEXTURE_SIZE * 4 + j * 4] = (GLubyte)c;
			texture[i * TEXTURE_SIZE * 4 + j * 4 + 1] = (GLubyte)c;
			texture[i * TEXTURE_SIZE * 4 + j * 4 + 2] = (GLubyte)c;
//			texture[i * TEXTURE_SIZE * 4 + j * 4 + 3] = 127;
			texture[i * TEXTURE_SIZE * 4 + j * 4 + 3] = c ? (GLubyte)0 :
				(GLubyte)64;
		}
	}
	return texture;
}
*/

void TGLShape::begin(void)
{
	if (glMode == GL_TRIANGLE_FAN)
	{
/*
		GLint oldMatrixMode;

		if (!g_texture)
		{
			g_texture = genTexture();
			glTexImage2D(GL_TEXTURE_2D, 0, 4, TEXTURE_SIZE, TEXTURE_SIZE, 0,
				GL_RGBA, GL_UNSIGNED_BYTE, g_texture);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
//			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		}
		glGetIntegerv(GL_MATRIX_MODE, &oldMatrixMode);
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
//		glScalef(2.0f, 2.0f, 2.0f);
		glMatrixMode(oldMatrixMode);
//		glEnable(GL_TEXTURE_2D);
*/
	}
//	if (specialType == 0)
	{
		glBegin(glMode);
	}
//	glBegin(GL_LINE_LOOP);
}

void TGLShape::end(void)
{
//	if (specialType == 0)
	{
		glEnd();
	}
	if (glMode == GL_TRIANGLE_FAN)
	{
//		glDisable(GL_TEXTURE_2D);
	}
}

/*
void TGLShape::draw(unsigned char negative)
{
	begin();
	drawPoints(negative);
	end();
}
*/

void TGLShape::draw(float* matrix)
{
	begin();
	drawPoints(matrix);
	end();
}

/*
void TGLShape::drawBack(unsigned char negative)
{
	begin();
	drawPointsBack(negative);
	end();
}
*/

/*
void TGLShape::drawPointsBack(unsigned char negative)
{
	int i;

	setGlColor();
	if (negative && !transformed)
	{
		if (normal)
		{
			glNormal3fv(*normal);
		}
	}
	else
	{
		if (backNormal)
		{
			glNormal3fv(*backNormal);
		}
	}
	for (i = numPoints - 1; i >= 0; i--)
	{
		glVertex3fv(points[i]);
	}
}
*/

float TGLShape::determinant(float* matrix)
{
	float det;

	det = matrix[0] * (matrix[5] * matrix[10] - matrix[6] * matrix[9]);
	det -= matrix[4] * (matrix[1] * matrix[10] - matrix[2] * matrix[9]);
	det += matrix[8] * (matrix[1] * matrix[6] - matrix[2] * matrix[5]);
	return det;
/*
	float det;

	det =  matrix[0] * matrix[5] * matrix[10];
	det += matrix[1] * matrix[6] * matrix[8];
	det += matrix[2] * matrix[4] * matrix[9];
	det -= matrix[2] * matrix[5] * matrix[8];
	det -= matrix[1] * matrix[4] * matrix[10];
	det -= matrix[0] * matrix[6] * matrix[9];

	return det;
*/
}

void TGLShape::invertMatrix(float* matrix, float* inverseMatrix)
{
	float det = determinant(matrix);

	if (fEq(det, 0.0f))
	{
		debugPrintf("Error: Matrix inversion failed.\n");
	}
	else
	{
		det = 1.0f / det;

		inverseMatrix[0]  =  (matrix[5] * matrix[10] - matrix[6] * matrix[9]) *
			det;
		inverseMatrix[1]  = -(matrix[1] * matrix[10] - matrix[2] * matrix[9]) *
			det;
		inverseMatrix[2]  =  (matrix[1] * matrix[6]  - matrix[2] * matrix[5]) *
			det;
		inverseMatrix[4]  = -(matrix[4] * matrix[10] - matrix[6] * matrix[8]) *
			det;
		inverseMatrix[5]  =  (matrix[0] * matrix[10] - matrix[2] * matrix[8]) *
			det;
		inverseMatrix[6]  = -(matrix[0] * matrix[6]  - matrix[2] * matrix[4]) *
			det;
		inverseMatrix[8]  =  (matrix[4] * matrix[9]  - matrix[5] * matrix[8]) *
			det;
		inverseMatrix[9]  = -(matrix[0] * matrix[9]  - matrix[1] * matrix[8]) *
			det;
		inverseMatrix[10] =  (matrix[0] * matrix[5]  - matrix[1] * matrix[4]) *
			det;

		inverseMatrix[12] = -(matrix[12] * matrix[0] + matrix[13] * matrix[4] +
							matrix[14] * matrix[8]);
		inverseMatrix[13] = -(matrix[12] * matrix[1] + matrix[13] * matrix[5] +
							matrix[14] * matrix[9]);
		inverseMatrix[14] = -(matrix[12] * matrix[2] + matrix[13] * matrix[6] +
							matrix[14] * matrix[10]);

		inverseMatrix[3] = inverseMatrix[7] = inverseMatrix[11] = 0.0;
		inverseMatrix[15] = 1.0;
	}
}

void TGLShape::multMatrix(float* left, float* right, float* result)
{
	result[0] = left[0] * right[0] + left[4] * right[1] +
		left[8] * right[2] + left[12] * right[3];
	result[1] = left[1] * right[0] + left[5] * right[1] +
		left[9] * right[2] + left[13] * right[3];
	result[2] = left[2] * right[0] + left[6] * right[1] +
		left[10] * right[2] + left[14] * right[3];
	result[3] = left[3] * right[0] + left[7] * right[1] +
		left[11] * right[2] + left[15] * right[3];
	result[4] = left[0] * right[4] + left[4] * right[5] +
		left[8] * right[6] + left[12] * right[7];
	result[5] = left[1] * right[4] + left[5] * right[5] +
		left[9] * right[6] + left[13] * right[7];
	result[6] = left[2] * right[4] + left[6] * right[5] +
		left[10] * right[6] + left[14] * right[7];
	result[7] = left[3] * right[4] + left[7] * right[5] +
		left[11] * right[6] + left[15] * right[7];
	result[8] = left[0] * right[8] + left[4] * right[9] +
		left[8] * right[10] + left[12] * right[11];
	result[9] = left[1] * right[8] + left[5] * right[9] +
		left[9] * right[10] + left[13] * right[11];
	result[10] = left[2] * right[8] + left[6] * right[9] +
		left[10] * right[10] + left[14] * right[11];
	result[11] = left[3] * right[8] + left[7] * right[9] +
		left[11] * right[10] + left[15] * right[11];
	result[12] = left[0] * right[12] + left[4] * right[13] +
		left[8] * right[14] + left[12] * right[15];
	result[13] = left[1] * right[12] + left[5] * right[13] +
		left[9] * right[14] + left[13] * right[15];
	result[14] = left[2] * right[12] + left[6] * right[13] +
		left[10] * right[14] + left[14] * right[15];
	result[15] = left[3] * right[12] + left[7] * right[13] +
		left[11] * right[14] + left[15] * right[15];
}

void TGLShape::multMatrixd(double* left, double* right, double* result)
{
	result[0] = left[0] * right[0] + left[4] * right[1] +
		left[8] * right[2] + left[12] * right[3];
	result[1] = left[1] * right[0] + left[5] * right[1] +
		left[9] * right[2] + left[13] * right[3];
	result[2] = left[2] * right[0] + left[6] * right[1] +
		left[10] * right[2] + left[14] * right[3];
	result[3] = left[3] * right[0] + left[7] * right[1] +
		left[11] * right[2] + left[15] * right[3];
	result[4] = left[0] * right[4] + left[4] * right[5] +
		left[8] * right[6] + left[12] * right[7];
	result[5] = left[1] * right[4] + left[5] * right[5] +
		left[9] * right[6] + left[13] * right[7];
	result[6] = left[2] * right[4] + left[6] * right[5] +
		left[10] * right[6] + left[14] * right[7];
	result[7] = left[3] * right[4] + left[7] * right[5] +
		left[11] * right[6] + left[15] * right[7];
	result[8] = left[0] * right[8] + left[4] * right[9] +
		left[8] * right[10] + left[12] * right[11];
	result[9] = left[1] * right[8] + left[5] * right[9] +
		left[9] * right[10] + left[13] * right[11];
	result[10] = left[2] * right[8] + left[6] * right[9] +
		left[10] * right[10] + left[14] * right[11];
	result[11] = left[3] * right[8] + left[7] * right[9] +
		left[11] * right[10] + left[15] * right[11];
	result[12] = left[0] * right[12] + left[4] * right[13] +
		left[8] * right[14] + left[12] * right[15];
	result[13] = left[1] * right[12] + left[5] * right[13] +
		left[9] * right[14] + left[13] * right[15];
	result[14] = left[2] * right[12] + left[6] * right[13] +
		left[10] * right[14] + left[14] * right[15];
	result[15] = left[3] * right[12] + left[7] * right[13] +
		left[11] * right[14] + left[15] * right[15];
}

/*
void TGLShape::drawPoints(unsigned char negative)
{
	int i;

	setGlColor();
	if (negative && !transformed)
	{
		if (backNormal)
		{
			glNormal3fv(*backNormal);
		}
	}
	else
	{
		if (normal)
		{
			glNormal3fv(*normal);
		}
	}
	for (i = 0; i < numPoints; i++)
	{
		glVertex3fv(points[i]);
	}
}
*/

void TGLShape::initTransformedPoints(float *matrix)
{
	int i;

	updateMirrored(matrix);
	transformedPoints = new Vector[numPoints];
	for (i = 0; i < numPoints; i++)
	{
		transformPoint(points[i], matrix, transformedPoints[i]);
	}
}

void TGLShape::computeQuadStrip(float *matrix)
{
	int i;
	int count = numPoints / 2 - 1;
	Vector normal1, normal2;
	Vector p1, p2, p3, p4, p5, p6;

	initTransformedPoints(matrix);
	transformedNormals = new Vector[numPoints];
	for (i = 0; i <= count; i++)
	{
		p1 = transformedPoints[i * 2];
		p2 = transformedPoints[i * 2 + 1];
		if (i == 0)
		{
			if (controlPoints)
			{
				transformPoint(controlPoints[0], matrix, p3);
				transformPoint(controlPoints[1], matrix, p4);
			}
			else
			{
				p3 = transformedPoints[numPoints - 4];
				p4 = transformedPoints[numPoints - 3];
			}
		}
		else
		{
			p3 = transformedPoints[i * 2 - 2];
			p4 = transformedPoints[i * 2 - 1];
		}
		if (i == count)
		{
			if (controlPoints)
			{
				transformPoint(controlPoints[2], matrix, p5);
				transformPoint(controlPoints[3], matrix, p6);
			}
			else
			{
				p5 = transformedPoints[2];
				p6 = transformedPoints[3];
			}
		}
		else
		{
			p5 = transformedPoints[i * 2 + 2];
			p6 = transformedPoints[i * 2 + 3];
		}
		if (p1 == p2)
		{
			normal1 = ((p1 - p3) * (p1 - p4)).normalize() +
				((p1 - p6) * (p1 - p5)).normalize();
			normal2 = ((p2 - p3) * (p2 - p4)).normalize()
				+ ((p2 - p6) * (p2 - p5)).normalize();
		}
		else
		{
			normal1 = ((p1 - p3) * (p1 - p2)).normalize() +
				((p1 - p2) * (p1 - p5)).normalize();
			normal2 = ((p2 - p1) * (p2 - p4)).normalize()
				+ ((p2 - p6) * (p2 - p1)).normalize();
		}
		normal1.normalize();
		normal2.normalize();
		transformedNormals[i * 2] = normal1;
		transformedNormals[i * 2 + 1] = normal2;
	}
}

void TGLShape::computeTriangleFan(float *matrix)
{
	initTransformedPoints(matrix);
	if (numPoints > 2)
	{
		transformedNormals = new Vector[1];
		transformedNormals[0] =
			(transformedPoints[1] - transformedPoints[2]) *
			(transformedPoints[1] - transformedPoints[0]);
		transformedNormals[0].normalize();
	}
}

void TGLShape::computeTriangleStrip(float *matrix)
{
	int i;
	Vector normal;
	Vector p1, p2, p3, p4, p5;

	initTransformedPoints(matrix);
	transformedNormals = new Vector[numPoints];
	for (i = 0; i < numPoints; i++)
	{
		normal = Vector(0.0f, 0.0f, 0.0f);
		p1 = transformedPoints[i];
		if (i < numPoints - 1)
		{
			p2 = transformedPoints[i + 1];
			if (i < numPoints - 2)
			{
				p3 = transformedPoints[i + 2];
				if (i % 2)
				{
					normal = ((p1 - p3) * (p1 - p2)).normalize();
				}
				else
				{
					normal = ((p1 - p2) * (p1 - p3)).normalize();
				}
			}
		}
		if (i > 0)
		{
			p4 = transformedPoints[i - 1];
			if (i > 1)
			{
				p5 = transformedPoints[i - 2];
			}
		}
		if (i < numPoints - 1 && i > 0)
		{
			if (i % 2)
			{
				normal += ((p1 - p2) * (p1 - p4)).normalize();
			}
			else
			{
				normal += ((p1 - p4) * (p1 - p2)).normalize();
			}
		}
		if (i > 1)
		{
			if (i % 2)
			{
				normal += ((p1 - p4) * (p1 - p5)).normalize();
			}
			else
			{
				normal += ((p1 - p5) * (p1 - p4)).normalize();
			}
		}
		normal.normalize();
		transformedNormals[i] = normal;
	}
}

void TGLShape::drawQuadStrip(float* matrix)
{
	int i;
	Vector lastNormal = Vector(0.0f, 0.0f, 0.0f);

	if (!transformedPoints)
	{
		computeQuadStrip(matrix);
	}
	for (i = 0; i < numPoints; i++)
	{
		if (transformedNormals[i] != lastNormal)
		{
			glNormal3fv(transformedNormals[i]);
			lastNormal = transformedNormals[i];
		}
		glVertex3fv(transformedPoints[i]);
	}
}

void TGLShape::drawTriangleFan(float *matrix)
{
	int i;

	if (!transformedPoints)
	{
		computeTriangleFan(matrix);
	}
	if (numPoints > 2)
	{
		glNormal3fv(transformedNormals[0]);
	}
	for (i = 0; i < numPoints; i++)
	{
		glVertex3fv(transformedPoints[i]);
	}
}

// Triangle strips aren't used anywhere, so this hasn't been tested.
void TGLShape::drawTriangleStrip(float* matrix)
{
	int i;
	Vector lastNormal = Vector(0.0f, 0.0f, 0.0f);

	if (!transformedPoints)
	{
		computeTriangleStrip(matrix);
	}
	for (i = 0; i < numPoints; i++)
	{
		if (transformedNormals[i] != lastNormal)
		{
			glNormal3fv(transformedNormals[i]);
			lastNormal = transformedNormals[i];
		}
		glVertex3fv(transformedPoints[i]);
	}
}

void TGLShape::drawLineStrip(float* matrix)
{
	int i;

	if (!transformedPoints)
	{
		initTransformedPoints(matrix);
	}
	for (i = 0; i < numPoints; i++)
	{
		glVertex3fv(transformedPoints[i]);
	}
}

void TGLShape::drawNormal(float* matrix)
{
	if (numPoints > 2)
	{
		Vector point1;
		Vector point2;
		Vector point3;
		Vector newNormal;

		transformPoint(points[0], matrix, point1);
		transformPoint(points[1], matrix, point2);
		transformPoint(points[2], matrix, point3);
		newNormal = (point2 - point3) * (point2 - point1);
		newNormal /= newNormal.length();
		glBegin(GL_LINES);
		glVertex3fv(point1);
		glVertex3fv(point1 + newNormal);
		glEnd();
	}
}

void TGLShape::drawNormals(void)
{
	if (transformedNormals)
	{
		int i;
		int count = numPoints;

		if (!normals)
		{
			count = 1;
		}
		for (i = 0; i < count; i++)
		{
			glBegin(GL_LINES);
			glVertex3fv(transformedPoints[i]);
			glVertex3fv(transformedPoints[i] + (transformedNormals[i] * 3.0f));
			glEnd();
		}
	}
}

/************************************************************************/
bool TGLShape::isConditional(void)
{
	return glMode == GL_LINES && controlPoints;
}

/************************************************************************/
// Return 1 if the v2 bends left of v1, -1 if right, 0 if straight ahead.
int turn_vector(double vx1, double vy1, double vx2, double vy2)
{
	/* Pos for left bend, 0 = linear */
	double vec_product = (vx1 * vy2) - (vy1 * vx2);

	if (vec_product > 0.0) return(1);
	if (vec_product < 0.0) return(-1);
	return(0);
}

void TGLShape::computeConditionalLine(float *matrix)
{
	initTransformedPoints(matrix);
	transformPoint(controlPoints[0], matrix, controlPoints[0]);
	transformPoint(controlPoints[1], matrix, controlPoints[1]);
	lineCount++;
}

/************************************************************************/
bool TGLShape::skipConditionalLine(float* matrix, GLdouble *comb_mat)
{
	if (showAllConditionalLines)
	{
		return false;
	}
	// Use comb_mat--which contains a combination of the projection and the
	// model-view matrix--to calculate coords in the plane of the screen, so
	// we can test optional lines.
	GLdouble s1x, s1y;
	GLdouble s2x, s2y;
	GLdouble s3x, s3y;
	GLdouble s4x, s4y;

	if (!transformedPoints)
	{
		// matrix never changes from one call to the next in LDView; all the
		// duplicate shapes with different matrices are copied.
		computeConditionalLine(matrix);
	}

	// Only draw optional line p1-p2 if p3 and p4 are on the same side of p1-p2.
	// Note that we don't actually adjust for the window size, because it
	// doesn't effect the calculation.  Also, we don't care what the z value is,
	// so we don't bother to comput it.
	transformPoint(transformedPoints[0], comb_mat, &s1x, &s1y);
	transformPoint(transformedPoints[1], comb_mat, &s2x, &s2y);
	transformPoint(controlPoints[0], comb_mat, &s3x, &s3y);
	transformPoint(controlPoints[1], comb_mat, &s4x, &s4y);

	// If we do not turn the same direction \_/ for both test points
	// then they're on opposite sides of segment p1-p2 and we should
	// skip drawing this conditional line.
	if (turn_vector(s2x-s1x, s2y-s1y, s3x-s2x, s3y-s2y) != 
		turn_vector(s2x-s1x, s2y-s1y, s4x-s2x, s4y-s2y))
	{
		return true;  // Skip it.
	}
	else
	{
		return false; // Draw it
	}
}

void TGLShape::drawPoints(float* matrix)
{
//	setGlColor();
/*
	if (specialType == TGL_TORUS)
	{
		drawTorus(matrix);
	}
	else if (specialType == TGL_TORUS_Q)
	{
		drawTorusQ(matrix);
	}
	else if (specialType == TGL_QUARTER_SPHERE)
	{
		drawEighthSphere(matrix);
	}
*/
	if (glMode == GL_QUAD_STRIP)
	{
		quadStripCount++;
		triangleCount += numPoints - 2;
		vertexCount += numPoints;
		drawQuadStrip(matrix);
	}
	else if (glMode == GL_TRIANGLE_STRIP)
	{
		triangleStripCount++;
		triangleCount += numPoints - 2;
		vertexCount += numPoints;
		drawTriangleStrip(matrix);
	}
	else if (glMode == GL_LINE_STRIP)
	{
		lineCount += numPoints - 1;
		drawLineStrip(matrix);
		vertexCount += numPoints;
	}
	else if (glMode == GL_TRIANGLE_FAN)
	{
		triangleFanCount++;
		triangleCount += numPoints - 2;
		vertexCount += numPoints;
		drawTriangleFan(matrix);
	}
	else if (isConditional())
	{
		if (!transformedPoints)
		{
			computeConditionalLine(matrix);
		}
		glVertex3fv(transformedPoints[0]);
		glVertex3fv(transformedPoints[1]);
		if (showConditionalControlPoints)
		{
			glVertex3fv(transformedPoints[0]);
			glVertex3fv(controlPoints[0]);
			glVertex3fv(transformedPoints[0]);
			glVertex3fv(controlPoints[1]);
		}
		return;
	}
	else
	{
		int i;

		if (!transformedPoints)
		{
			initTransformedPoints(matrix);
			if (numPoints > 2) // normal
			{
				if (normals)
				{
					transformedNormals = new Vector[numPoints];
					int i;

					for (i = 0; i < numPoints; i++)
					{
						if (mirrored)
						{
							normals[i] *= -1.0f;
						}
						transformNormal(normals[i], matrix,
							transformedNormals[i]);
					}
				}
				else
				{
					transformedNormals = new Vector[1];
					transformedNormals[0] = (transformedPoints[numPoints - 2] -
						transformedPoints[numPoints - 1]) *
						(transformedPoints[numPoints - 2] -
						transformedPoints[numPoints - 3]);
					float length = transformedNormals[0].length();

					if (fEq(length, 0.0f))
					{
						debugPrintf("Zero-length normal.\n");
						transformedNormals[0] = Vector(1.0f, 0.0f, 0.0f);
					}
					else
					{
						transformedNormals[0] /= length;
					}
				}
			}
		}
		if (numPoints > 2 && !normals) // normal
		{
			glNormal3fv(transformedNormals[0]);
		}
		for (i = 0; i < numPoints; i++)
		{
			if (normals)
			{
				glNormal3fv(transformedNormals[i]);
			}
			glVertex3fv(transformedPoints[i]);
		}
/*
		glEnd();
		drawNormals();
		glBegin(glMode);
*/
		vertexCount += numPoints;
		if (glMode == GL_LINES)
		{
			lineCount++;
		}
		else
		{
			triangleCount += numPoints - 2;
			if (glMode == GL_QUADS)
			{
				quadCount++;
				triangleCount += 2;
			}
		}
	}
}

/*
bool TGLShape::getWasRubber(void)
{
	return wasRubber;
}

void TGLShape::setWasRubber(bool flag)
{
	wasRubber = flag;
}
*/

void TGLShape::resetNormalSpecular(void)
{
	glGetMaterialfv(GL_FRONT, GL_SPECULAR, normalSpecular);
	glGetMaterialfv(GL_FRONT, GL_SHININESS, &normalShininess);
}

void TGLShape::setSpecular(float r, float g, float b, float a, float shininess)
{
	float specular[4];

	if (r <= -100.0f)
	{
		r = normalSpecular[0];
	}
	if (g <= -100.0f)
	{
		g = normalSpecular[1];
	}
	if (b <= -100.0f)
	{
		b = normalSpecular[2];
	}
	if (a <= -100.0f)
	{
		a = normalSpecular[3];
	}
	if (shininess <= -100.0f)
	{
		shininess = normalShininess;
	}
	specular[0] = r;
	specular[1] = g;
	specular[2] = b;
	specular[3] = a;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
}

void TGLShape::setToNormalSpecular(void)
{
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, normalSpecular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, normalShininess);
}

/*
void TGLShape::switchToBlackRubber(void)
{
	float rubberSpecular[4];

	rubberSpecular[0] = normalSpecular[0] * 0.15f;
	rubberSpecular[1] = normalSpecular[1] * 0.15f;
	rubberSpecular[2] = normalSpecular[2] * 0.15f;
	rubberSpecular[3] = normalSpecular[3];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, rubberSpecular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 15.0);
	wasRubber = YES;
}

void TGLShape::switchFromBlackRubber(void)
{
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, normalSpecular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, normalShininess);
	wasRubber = NO;
}
*/

void TGLShape::setGlColor(int glColor, int colorNumber)
{
	if (forceColor || glColor != lastColor || colorNumber != lastColorNumber)
	{
		glColor4ubv((GLubyte*)&(glColor));
		forceColor = false;
		lastColor = glColor;
		lastColorNumber = colorNumber;
	}
}

/*
void TGLShape::setGlColor(void)
{
	if (colorNumber == 16)
	{
	}
	else if (colorNumber == 24)
	{
	}
	else
	{
//		glColor4ubv((GLubyte*)&(color));
		if (colorNumber == 256 && !wasRubber)
		{
			switchToBlackRubber();
		}
		else if (colorNumber != 256 && wasRubber)
		{
			switchFromBlackRubber();
		}
		setGlColor(color, colorNumber);
	}
}
*/

void TGLShape::transformPoint(Vector& point, double* matrix, double* tx,
							  double* ty)
{
	double x = point[0];
	double y = point[1];
	double z = point[2];
	double tw;

//	x' = a*x + b*y + c*z + X
//	y' = d*x + e*y + f*z + Y
//	z' = g*x + h*y + i*z + Z
	tw = matrix[3]*x + matrix[7]*y + matrix[11]*z + matrix[15];
	*tx = (matrix[0]*x + matrix[4]*y + matrix[8]*z + matrix[12]) / tw;
	*ty = (matrix[1]*x + matrix[5]*y + matrix[9]*z + matrix[13]) / tw;
}

void TGLShape::transformPoint(Vector& point, float* matrix, Vector& newPoint)
{
	float x = point[0];
	float y = point[1];
	float z = point[2];

//	x' = a*x + b*y + c*z + X
//	y' = d*x + e*y + f*z + Y
//	z' = g*x + h*y + i*z + Z
	newPoint[0] = matrix[0]*x + matrix[4]*y + matrix[8]*z + matrix[12];
	newPoint[1] = matrix[1]*x + matrix[5]*y + matrix[9]*z + matrix[13];
	newPoint[2] = matrix[2]*x + matrix[6]*y + matrix[10]*z + matrix[14];
}

Vector TGLShape::transformPoint(Vector& point, float* matrix)
{
	Vector newPoint;

	transformPoint(point, matrix, newPoint);
	return newPoint;
}

void TGLShape::transformNormal(Vector& point, float* matrix, Vector& newNormal)
{
	float inverseMatrix[16];
	float x = point[0];
	float y = point[1];
	float z = point[2];

	invertMatrix(matrix, inverseMatrix);
//	x' = a*x + b*y + c*z + X
//	y' = d*x + e*y + f*z + Y
//	z' = g*x + h*y + i*z + Z
//	newNormal[0] = inverseMatrix[0]*x + inverseMatrix[4]*y + inverseMatrix[8]*z;
//	newNormal[1] = inverseMatrix[1]*x + inverseMatrix[5]*y + inverseMatrix[9]*z;
//	newNormal[2] = inverseMatrix[2]*x + inverseMatrix[6]*y + inverseMatrix[10]*z;
	newNormal[0] = inverseMatrix[0]*x + inverseMatrix[1]*y + inverseMatrix[2]*z;
	newNormal[1] = inverseMatrix[4]*x + inverseMatrix[5]*y + inverseMatrix[6]*z;
	newNormal[2] = inverseMatrix[8]*x + inverseMatrix[9]*y + inverseMatrix[10]*z;
	newNormal.normalize();
}

Vector TGLShape::transformNormal(Vector& point, float* matrix)
{
	Vector newNormal;

	transformNormal(point, matrix, newNormal);
	return newNormal;
}

bool TGLShape::isMirrorMatrix(float *matrix)
{
	float det = determinant(matrix);

	if (!fEq(det, 0.0f) && det < 0.0f)
	{
		return true;
	}
	else
	{
		return false;
	}
}

void TGLShape::updateMirrored(float *matrix)
{
	if (isMirrorMatrix(matrix))
	{
		mirrored = !mirrored;
	}
}

void TGLShape::transform(float* matrix)
{
	int i;

	updateMirrored(matrix);
	for (i = 0; i < numPoints; i++)
	{
		transformPoint(points[i], matrix, points[i]);
	}
	if (controlPoints)
	{
		for (i = 0; i < numControlPoints; i++)
		{
			transformPoint(controlPoints[i], matrix, controlPoints[i]);
		}
	}
	if (normals)
	{
		for (i = 0; i < numPoints; i++)
		{
			Vector newNormal = transformNormal(normals[i], matrix);

			normals[i] = newNormal;
		}
	}
//	compNormal();
//	transformed = 1;
}

void TGLShape::getMinMax(Vector& min, Vector& max, float* matrix)
{
	int i;

	for (i = 0; i < numPoints; i++)
	{
		Vector point;
		
		transformPoint(points[i], matrix, point);
		if (point[0] < min[0])
		{
			min[0] = point[0];
		}
		if (point[1] < min[1])
		{
			min[1] = point[1];
		}
		if (point[2] < min[2])
		{
			min[2] = point[2];
		}
		if (point[0] > max[0])
		{
			max[0] = point[0];
		}
		if (point[1] > max[1])
		{
			max[1] = point[1];
		}
		if (point[2] > max[2])
		{
			max[2] = point[2];
		}
	}
/*
	if (normal)
	{
		float newMatrix[16];
		int i;
		Vector transformedNormal;
			
		for (i = 0; i < 16; i++)
		{
			newMatrix[i] = matrix[i];
		}
		newMatrix[12] = 0.0f;
		newMatrix[13] = 0.0f;
		newMatrix[14] = 0.0f;
		transformedNormal = transformPoint(*normal, newMatrix);

		if (fEq(transformedNormal.length(), 0.0f))
		{
			printf("oops!\n");
		}
		transformedNormal = transformPoint(*backNormal, newMatrix);
		if (fEq(transformedNormal.length(), 0.0f))
		{
			printf("oops2!\n");
		}
	}
*/
}

void TGLShape::scale(Vector& scaleFactor, Vector& center)
{
	int i;

	for (i = 0; i < numPoints; i++)
	{
		Vector distance = center - points[i];

		distance[0] *= scaleFactor[0];
		distance[1] *= scaleFactor[1];
		distance[2] *= scaleFactor[2];
		points[i][0] = center[0] - distance[0];
		points[i][1] = center[1] - distance[1];
		points[i][2] = center[2] - distance[2];
//		points[i][0] = points[i][0] * scaleFactor[0];
//		points[i][1] = points[i][1] * scaleFactor[1];
//		points[i][2] = points[i][2] * scaleFactor[2];
	}
	if (controlPoints)
	{
		for (i = 0; i < numControlPoints; i++)
		{
			Vector distance = center - controlPoints[i];

			distance[0] *= scaleFactor[0];
			distance[1] *= scaleFactor[1];
			distance[2] *= scaleFactor[2];
			controlPoints[i][0] = center[0] - distance[0];
			controlPoints[i][1] = center[1] - distance[1];
			controlPoints[i][2] = center[2] - distance[2];
		}
	}
}

void TGLShape::printCounts(void)
{
	debugPrintf("vertices: %d\n", vertexCount);
	debugPrintf("triangles: %d\n", triangleCount);
	debugPrintf("triangle strips: %d\n", triangleStripCount);
	debugPrintf("quads: %d\n", quadCount);
	debugPrintf("quadStrips: %d\n", quadStripCount);
	debugPrintf("lines: %d\n", lineCount);
}

void TGLShape::calcCentroid(void)
{
	delete[] centroids;
	centroids = new Vector[2];
	for (int i = 0; i < numPoints; i++)
	{
		centroids[0] += points[i];
	}
	centroids[0] /= (float)numPoints;
}

void TGLShape::calcCentroid(float* matrix)
{
	if (!centroids)
	{
		calcCentroid();
	}
	transformPoint(centroids[0], matrix, centroids[1]);
}
