#ifndef __TGLSHAPE_H__
#define __TGLSHAPE_H__

#include <stdio.h>
//#include <windows.h>

#ifdef WIN32
#include <windows.h>
#endif // WIN32
#if defined(__APPLE__)
#	include <OpenGL/OpenGL.h>
#	include <GLUT/GLUT.h>
#else
#	include <GL/gl.h>
#endif

#include <LDLib/Vector.h>
#include <TCFoundation/TCObject.h>

//#define TGL_TORUS 1
//#define TGL_TORUS_Q 2
//#define TGL_QUARTER_SPHERE 3

class LDrawModel;

class TGLShape : public TCObject
{
	public:
		TGLShape(void);
		TGLShape(TGLShape&);

		virtual TCObject *copy(void);
		TGLShape *copyShape(void) { return (TGLShape *)copy(); }
		virtual void setPoints(const Vector&, const Vector&);
		virtual void setPoints(const Vector&, const Vector&, const Vector&);
		virtual void setPoints(const Vector&, const Vector&, const Vector&,
			const Vector&);
		virtual void setPoint(int, const Vector&);
		virtual void setNormal(int index, const Vector& n);
		virtual void setControlPoint(int, const Vector&);
		virtual void setNumPoints(int);
		virtual void setNumControlPoints(int value);
		int getNumPoints(void) { return numPoints; }
		void setGlMode(GLenum mode) { glMode = mode; }
		virtual void setColor(char*);
		virtual void setColor(int, int, int, int);
		int getColor(void) { return color; }
		virtual void getRGBA(int &r, int &g, int &b, int &a);
		void setColorNumber(int num) { colorNumber = num; }
		int getColorNumber(void) { return colorNumber; }
		virtual void setParent(LDrawModel*);
//		virtual void draw(unsigned char);
		virtual void draw(float*);
//		virtual void drawBack(unsigned char);
//		virtual void drawPoints(unsigned char);
		virtual void drawPoints(float*);
		virtual void drawNormal(float*);
		virtual void drawNormals(void);
		virtual void drawQuadStrip(float*);
		virtual void drawTriangleStrip(float*);
		virtual void drawTriangleFan(float *matrix);
//		virtual void drawTorus(float*);
//		virtual void drawTorusQ(float*);
//		virtual void drawEighthSphere(float*);
		virtual void drawLineStrip(float*);
//		virtual void drawPointsBack(unsigned char);
		virtual void begin(void);
		virtual void end(void);
		virtual bool isTransparent(void);
		virtual void getMinMax(Vector&, Vector&, float*);
		virtual void transform(float*);
		static Vector transformPoint(Vector&, float*);
		static void transformPoint(Vector&, float*, Vector&);
		static void transformPoint(Vector&, double*, double*, double*);
		static void transformNormal(Vector&, float*, Vector&);
		static Vector transformNormal(Vector&, float*);
		static void multMatrix(float*, float*, float*);
		static void multMatrixd(double*, double*, double*);
		static void invertMatrix(float*, float*);
		static float determinant(float*);
		static int colorForRGBA(int, int, int, int);
		virtual void scale(Vector&, Vector&);
		GLenum getGlMode(void) { return glMode; }
//		void setTorus(void) { specialType = TGL_TORUS; }
//		void setTorusQ(void) { specialType = TGL_TORUS_Q; }
//		void setEighthSphere(void) { specialType = TGL_QUARTER_SPHERE; }
		Vector* getPoints(void) { return points; }
		Vector* getNormals(void) { return normals; }
		Vector* getControlPoints(void) { return controlPoints; }
		void calcCentroid(void);
		void calcCentroid(float* matrix);
		Vector* getCentroids(void) { return centroids; }
		virtual bool isConditional(void);
        virtual bool skipConditionalLine(float* matrix, GLdouble *comb_mat);
		virtual Vector calcNormal(bool normalize = true);
		void setUserData(void *value) { userData = value; }
		void *getUserData(void) { return userData; }
		bool getMirrored(void) { return mirrored; }

		static void resetCounts(void);
		static void printCounts(void);
		static bool colorNumberIsTransparent(int colorNumber);
		static void setShowConditionalControlPoints(bool value)
		{
			showConditionalControlPoints = value;
		}
		static void setShowAllConditionalLines(bool value)
		{
			showAllConditionalLines = value;
		}
		static void forceGlColor(void) { forceColor = true; }
		static void setGlColor(int glColor, int colorNumber);
//		static bool getWasRubber(void);
//		static void setWasRubber(bool value);
		static float* getNormalSpecular(void) { return normalSpecular; }
		static float getNormalShininess(void) { return normalShininess; }
		static void resetNormalSpecular(void);
		static void setSpecular(float r, float g, float b, float a,
			float shininess);
		static void setToNormalSpecular(void);
//		static void switchToBlackRubber(void);
//		static void switchFromBlackRubber(void);
		static int getLastColor(void) { return lastColor; }
		static int getLastColorNumber(void) { return lastColorNumber; }
	protected:
		virtual ~TGLShape(void);
		virtual void dealloc(void);
		virtual void initTransformedPoints(float *matrix);
		virtual void computeTriangleFan(float *matrix);
		virtual void computeQuadStrip(float *matrix);
		virtual void computeTriangleStrip(float *matrix);
		virtual void computeConditionalLine(float *matrix);
		virtual bool isMirrorMatrix(float *matrix);
		virtual void updateMirrored(float *matrix);

		Vector* points;
		Vector* normals;
		Vector* transformedPoints;
		Vector* transformedNormals;
		Vector* controlPoints;
		Vector* centroids;
		int color;
		int colorNumber;
		LDrawModel* parent;
//		Vector* normal;
//		Vector* backNormal;
		int numPoints;
		int numControlPoints;
		void *userData;
		GLenum glMode;
		bool mirrored;

//		int specialType;
		static int triangleCount;
		static int triangleStripCount;
		static int triangleFanCount;
		static int quadCount;
		static int quadStripCount;
		static int lineCount;
		static int totalShapes;
		static int vertexCount;
		static bool showConditionalControlPoints;
		static bool showAllConditionalLines;
		static int lastColor;
		static int lastColorNumber;
		static bool forceColor;
//		static bool wasRubber;
		static float normalSpecular[4];
		static float normalShininess;
//		int transformed;

//		virtual void setGlColor(void);
};

#endif // __TGLSHAPE_H__
