#include <stdlib.h>
#include <string.h>

#include "TGLStudLogo.h"

#include "LDrawModel.h"
#include "ModelMacros.h"
#include <GL/glu.h>
#include <GL/glext.h>
#include <TCFoundation/TCImage.h>

TCImageArray *TGLStudLogo::textures = NULL;
bool TGLStudLogo::textureStuds = true;
GLint TGLStudLogo::textureFilterType = GL_LINEAR_MIPMAP_LINEAR;
GLuint TGLStudLogo::textureID;
Vector *TGLStudLogo::textureCoords = NULL;
Vector *TGLStudLogo::mirroredTextureCoords = NULL;

TGLStudLogo::TGLStudLogoCleanup TGLStudLogo::studLogoCleanup;

TGLStudLogo::TGLStudLogoCleanup::~TGLStudLogoCleanup(void)
{
	if (textures)
	{
		textures->release();
		textures = NULL;
	}
	delete[] textureCoords;
	textureCoords = NULL;
	delete[] mirroredTextureCoords;
	mirroredTextureCoords = NULL;
}

TGLStudLogo::TGLStudLogo(void)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLStudLogo");
#endif
}

TGLStudLogo::TGLStudLogo(TGLStudLogo &other)
			:TGLCustomShape(other)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLStudLogo");
#endif
	if (transformedPoints)
	{
		delete[] transformedPoints;
		transformedPoints = NULL;
		delete[] transformedNormals;
		transformedNormals = NULL;
	}
}

TGLStudLogo::~TGLStudLogo(void)
{
}

void TGLStudLogo::dealloc(void)
{
	TGLCustomShape::dealloc();
}

TCObject *TGLStudLogo::copy(void)
{
	return new TGLStudLogo(*this);
}

void TGLStudLogo::resetTextureCoords(void)
{
	delete[] textureCoords;
	textureCoords = NULL;
	delete[] mirroredTextureCoords;
	mirroredTextureCoords = NULL;
}

void TGLStudLogo::attachTexture(void)
{
	glBindTexture(GL_TEXTURE_2D, textureID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	if (textureFilterType == GL_NEAREST_MIPMAP_NEAREST ||
		textureFilterType == GL_NEAREST_MIPMAP_LINEAR ||
		textureFilterType == GL_NEAREST)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		textureFilterType);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
}

void TGLStudLogo::setTextureData(TCByte *data, long length)
{
	if (!textures)
	{
		TCImage *mainImage = new TCImage;

		mainImage->setFlipped(true);
		mainImage->setLineAlignment(4);
		if (mainImage->loadData(data, length))
		{
			loadMipTextures(mainImage);
		}
		mainImage->release();
	}
}

void TGLStudLogo::setRawTextureData(TCByte *data, long length)
{
	if (!textures)
	{
		TCImage *mainImage = new TCImage;
		int rowSize;

		mainImage->setFlipped(true);
		mainImage->setLineAlignment(4);
		mainImage->setDataFormat(TCRgba8);
		mainImage->setSize(128, 255);
		mainImage->allocateImageData();
		rowSize = mainImage->getRowSize();
		if (length == rowSize * 255)
		{
			TCByte *imageData = mainImage->getImageData();
			int i;

			for (i = 0; i < 255; i++)
			{
				memcpy(imageData + rowSize * (254 - i), data + rowSize * i, rowSize);
			}
			loadMipTextures(mainImage);
		}
		mainImage->release();
	}
}

void TGLStudLogo::loadMipTextures(TCImage *mainImage)
{
	int i;
	int yPosition = 0;
	int thisSize = mainImage->getWidth();
	int imageHeight = mainImage->getHeight();

	textures = new TCImageArray(8);
	for (i = 0; thisSize > 0 && yPosition + thisSize <= imageHeight; i++)
	{
		TCImage *texture = mainImage->createSubImage(0, yPosition,
			thisSize, thisSize);
		textures->addObject(texture);
		texture->release();
		yPosition += thisSize;
		thisSize /= 2;
	}
}

void TGLStudLogo::loadTexture(char *filename)
{
	int i;
	int count;

	glGenTextures(1, &textureID);
	glBindTexture(GL_TEXTURE_2D, textureID);
	if (!textures)
	{
		TCImage *mainImage = new TCImage;

		textures = new TCImageArray(8);
		mainImage->setFlipped(true);
		mainImage->setLineAlignment(4);
		if (mainImage->loadFile(filename))
		{
			loadMipTextures(mainImage);
		}
		mainImage->release();
	}
	count = textures->getCount();
	for (i = 0; i < count; i++)
	{
		TCImage *texture = (*textures)[i];

		if (texture)
		{
			int textureSize = texture->getWidth();

			glTexImage2D(GL_TEXTURE_2D, i, 4, textureSize, textureSize, 0,
				GL_RGBA, GL_UNSIGNED_BYTE, texture->getImageData());
		}
	}
}

void TGLStudLogo::genTextureCoords(void)
{
	if (!textureCoords)
	{
		int i;
		Vector p1;
		Vector offset = Vector(0.5f, 0.5f, 0.0f);
		int numSegments = numPoints - 2;

		textureCoords = new Vector[numPoints];
		mirroredTextureCoords = new Vector[numPoints];
		textureCoords[0] = Vector(0.5f, 0.5f, 0.0f);
		mirroredTextureCoords[0] = textureCoords[0];
		p1[2] = 0.0f;
		for (i = 1; i < numPoints; i++)
		{
			float x, z;
			float angle;

			angle = 2.0f * (float)M_PI / numSegments * (i - 1);
			x = (float)cos(angle) * 0.5f;
			z = (float)sin(angle) * 0.5f;
			p1[0] = z;
			p1[1] = x;
			p1 += offset;
			mirroredTextureCoords[i] = p1;
			p1[1] = 1.0f - p1[1];
			textureCoords[i] = p1;
		}
	}
}

void TGLStudLogo::setDisk(const Vector& center, float radius, int numSegments)
{
	int i;
	Vector p1;

	setNumPoints(numSegments + 2);
	setGlMode(GL_TRIANGLE_FAN);
	setPoint(0, center);
	for (i = 1; i < numPoints; i++)
	{
		float x, z;
		float angle;

		angle = 2.0f * (float)M_PI / numSegments * (i - 1);
		x = radius * (float)cos(angle);
		z = radius * (float)sin(angle);
		p1[0] = center.get(0) + x;
		p1[2] = center.get(2) + z;
		p1[1] = center.get(1);
		setPoint(i, p1);
	}
	genTextureCoords();
}

void TGLStudLogo::drawPoints(float* matrix)
{
	int i;
	Vector *myTextureCoords = textureCoords;

	if (!transformedPoints)
	{
		initTransformedPoints(matrix);
		if (numPoints > 2)
		{
			if (!transformedNormals)
			{
				transformedNormals = new Vector[1];

				transformedNormals[0] = (transformedPoints[1] -
					transformedPoints[2]) * (transformedPoints[1] -
					transformedPoints[0]);
				transformedNormals[0].normalize();
			}
		}
		triangleCount += numPoints - 2;
		triangleFanCount++;
		vertexCount += numPoints;
	}
	if (mirrored)
	{
		myTextureCoords = mirroredTextureCoords;
	}
	if (numPoints > 2)
	{
		glNormal3fv(transformedNormals[0]);
	}
	for (i = 0; i < numPoints; i++)
	{
		glTexCoord2fv(myTextureCoords[i]);
		glVertex3fv(transformedPoints[i]);
	}
}

void TGLStudLogo::begin(void)
{
	if (textureStuds && textures && (*textures)[0])
	{
		glEnable(GL_TEXTURE_2D);
	}
	glBegin(GL_TRIANGLE_FAN);
}

void TGLStudLogo::end(void)
{
	glEnd();
	if (textureStuds && textures && (*textures)[0])
	{
		glDisable(GL_TEXTURE_2D);
	}
}
