#include <stdlib.h>
#include <string.h>

#include "TGLTorusQ.h"

//#include "LDrawModel.h"
#include "ModelMacros.h"

TGLTorusQ::TGLTorusQ(void)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLTorusQ");
#endif
}

TGLTorusQ::TGLTorusQ(TGLTorusQ &other)
		  :TGLCustomShape(other)
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TGLTorusQ");
#endif
}

TGLTorusQ::~TGLTorusQ(void)
{
}

void TGLTorusQ::dealloc(void)
{
	TGLCustomShape::dealloc();
}

TCObject *TGLTorusQ::copy(void)
{
	return new TGLTorusQ(*this);
}

void TGLTorusQ::drawPoints(float* matrix)
{
	int i, j;
	Vector* newPoints = new Vector[numPoints];
	Vector* normals = new Vector[numPoints];
	Vector p1, p2, p3, p4, p5, p6, p7, p8;
	Vector normal1, normal2, normal3, normal4;
	Vector lastNormal = Vector(0.0f, 0.0f, 0.0f);
	int c = 12 - numPoints;
	float x1 = float((-10.0 + sqrt(100.0 - 8.0 * c)) / 4.0);
	float x2 = float((-10.0 - sqrt(100.0 - 8.0 * c)) / 4.0);
	int numSegments = (int)(x1 > 0 ? x1 : x2);
	int numQuads;

//	parent->checkError("TGLTorusQ::drawPoints 1");
	for (i = 0; i < numPoints; i++)
	{
		newPoints[i] = transformPoint(points[i], matrix);
	}
	for (i = 1; i <= numSegments; i++)
	{
		for (j = 1; j <= numSegments + 1; j++)
		{
			p1 = newPoints[i * (numSegments + 3) * 2 + j * 2];
			p2 = newPoints[i * (numSegments + 3) * 2 + j * 2 + 1];
			p3 = newPoints[(i - 1) * (numSegments + 3) * 2 + j * 2];
			p4 = newPoints[(i + 1) * (numSegments + 3) * 2 + j * 2 + 1];
			p5 = newPoints[i * (numSegments + 3) * 2 + (j - 1) * 2];
			p6 = newPoints[i * (numSegments + 3) * 2 + (j + 1) * 2];
			p7 = newPoints[i * (numSegments + 3) * 2 + (j - 1) * 2 + 1];
			p8 = newPoints[i * (numSegments + 3) * 2 + (j + 1) * 2 + 1];
			normal1 = ((p1 - p3) * (p1 - p5)).normalize();
			normal2 = ((p1 - p6) * (p1 - p3)).normalize();
			normal3 = ((p1 - p2) * (p1 - p6)).normalize();
			normal4 = ((p1 - p5) * (p1 - p2)).normalize();
			normals[i * (numSegments + 3) * 2 + j * 2] =
				(normal1 + normal2 + normal3 + normal4).normalize();
			normal1 = ((p2 - p1) * (p2 - p7)).normalize();
			normal2 = ((p2 - p8) * (p2 - p1)).normalize();
			normal3 = ((p2 - p4) * (p2 - p8)).normalize();
			normal4 = ((p2 - p7) * (p2 - p4)).normalize();
			normals[i * (numSegments + 3) * 2 + j * 2 + 1] =
				(normal1 + normal2 + normal3 + normal4).normalize();
		}
	}
	numQuads = numSegments * (numSegments - 2);
	quadCount += numQuads;
	quadStripCount += numSegments;
	triangleCount += numQuads * 2;
	vertexCount += numSegments * numSegments;
	for (i = 1; i <= numSegments; i++)
	{
		glBegin(GL_QUAD_STRIP);
		for (j = 1; j <= numSegments + 1; j++)
		{
			p1 = newPoints[i * (numSegments + 3) * 2 + j * 2];
			p2 = newPoints[i * (numSegments + 3) * 2 + j * 2 + 1];
			normal1 = normals[i * (numSegments + 3) * 2 + j * 2];
			normal2 = normals[i * (numSegments + 3) * 2 + j * 2 + 1];
			if (normal1 != lastNormal)
			{
				glNormal3fv(normal1);
				lastNormal = normal1;
			}
			glVertex3fv(p1);
			if (normal2 != lastNormal)
			{
				glNormal3fv(normal2);
				lastNormal = normal2;
			}
			glVertex3fv(p2);
		}
		glEnd();
	}
	delete[] newPoints;
	delete[] normals;
//	parent->checkError("TGLTorusQ::drawPoints n");
}
