//#include <math.h>
//#include <stdio.h>

#ifdef WIN32
#include <windows.h>
#endif // WIN32

#include "Vector.h"
#include "ModelMacros.h"
//#include "simMacros.h"

//static int vectorCount = 0;

// Vector::Vector(void) -- Default Constructor
Vector::Vector(void)
{
#ifdef _LEAK_DEBUG
	className[0] = 'V';
	className[1] = 'e';
	className[2] = 'c';
	className[3] = 0;
#endif
	vector[0] = (float)0.0;
	vector[1] = (float)0.0;
	vector[2] = (float)0.0;
//	vectorCount++;
}

// Vector::Vector(float, float, float) -- Constructor
// Expects:
//		x, y, z	: The coordinates to store in the Vector
Vector::Vector(float x, float y, float z)
{
#ifdef _LEAK_DEBUG
	className[0] = 'V';
	className[1] = 'e';
	className[2] = 'c';
	className[3] = 0;
#endif
	vector[0] = x;
	vector[1] = y;
	vector[2] = z;
//	vectorCount++;
}

// Vector::Vector(const Vector&) -- Copy Constructor
// Expects:
//		v	: The Vector to copy.
Vector::Vector(const Vector& v)
{
#ifdef _LEAK_DEBUG
	className[0] = 'V';
	className[1] = 'e';
	className[2] = 'c';
	className[3] = 0;
#endif
	vector[0] = v.vector[0];
	vector[1] = v.vector[1];
	vector[2] = v.vector[2];
//	vectorCount++;
}

Vector::~Vector(void)
{
//	vectorCount--;
//	if (vectorCount < 10 || vectorCount % 1000 == 0)
//	{
//		printf("~Vector: %d\n", vectorCount);
//	}
}

// Vector::length(void) -- Member Function
// Returns:
//		The length of the Vector.
float Vector::length(void) const
{
	return (float)sqrt(sqr(vector[0]) + sqr(vector[1]) + sqr(vector[2]));
}

// Vector::dot(const Vector&) -- Member Function
// Expects:
//		right	: The right hand side of the dot product.
// Returns:
//		The dot product of "*this" dot "right".
float Vector::dot(const Vector& right) const
{
	return vector[0] * right.vector[0] + vector[1] * right.vector[1] +
			 vector[2] * right.vector[2];
}

// Vector::operator*(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the cross product.
// Returns:
//		The cross product of "*this" cross "right".
Vector Vector::operator*(const Vector& right) const
{
	// <a, b, c> x <d, e, f> = <bf-ce, cd-af, ae-bd>
	// <x1, y1, z1> x <x2, y2, z2> = <y1*z2-y2*z1, z1*x2-z2*x1, x1*y2-x2*y1>
	return Vector(vector[1] * right.vector[2] - vector[2] * right.vector[1],
					  vector[2] * right.vector[0] - vector[0] * right.vector[2],
					  vector[0] * right.vector[1] - vector[1] * right.vector[0]);
}

// Vector::operator*(float) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the scalar multiply.
// Returns:
//		The Vector multiplied by the scalar "right".
Vector Vector::operator*(float right) const
{
	return Vector(vector[0]*right, vector[1]*right, vector[2]*right);
}

// Vector::operator/(float) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the scalar divide.
// Returns:
//		The Vector divided by the scalar "right".
Vector Vector::operator/(float right) const
{
	float mult = 1.0f / right;

	return Vector(vector[0]*mult, vector[1]*mult, vector[2]*mult);
}

// Vector::operator+(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the addition
//	Returns:
//		The Vector added value "*this" + "right".
Vector Vector::operator+(const Vector& right) const
{
	return Vector(vector[0] + right.vector[0], vector[1] + right.vector[1],
					  vector[2] + right.vector[2]);
}

// Vector::operator-(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the subtraction
//	Returns:
//		The Vector subtracted value "*this" - "right".
Vector Vector::operator-(const Vector& right) const
{
	return Vector(vector[0] - right.vector[0], vector[1] - right.vector[1],
					  vector[2] - right.vector[2]);
}

// Vector::operator-(void) -- Overloaded Operator
// Returns:
//		The Vector -"*this".
// Results:
//		This is the unary minus operator
Vector Vector::operator-(void) const
{
	return Vector(-vector[0], -vector[1], -vector[2]);
}

// Vector::operator*=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the cross product *=
// Returns:
//		*this, after it has been set to be the cross product of "*this" cross
//		"right"
Vector& Vector::operator*=(const Vector& right)
{
	vector[0] = vector[1] * right.vector[2] - vector[2] * right.vector[1];
	vector[1] = vector[2] * right.vector[0] - vector[0] * right.vector[2];
	vector[2] = vector[0] * right.vector[1] - vector[1] * right.vector[0];
	return *this;
}

// Vector::operator+=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the +=
// Returns:
//		*this, after "right" has been added to it.
Vector& Vector::operator+=(const Vector& right)
{
	vector[0] += right.vector[0];
	vector[1] += right.vector[1];
	vector[2] += right.vector[2];
	return *this;
}

// Vector::operator-=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the -=
// Returns:
//		*this, after "right" has been subtracted from it.
Vector& Vector::operator-=(const Vector& right)
{
	vector[0] -= right.vector[0];
	vector[1] -= right.vector[1];
	vector[2] -= right.vector[2];
	return *this;
}

// Vector::operator*=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the scalar *=
// Returns:
//		*this, after it has been multiplied by the scalar "right".
Vector& Vector::operator*=(float right)
{
	vector[0] *= right;
	vector[1] *= right;
	vector[2] *= right;
	return *this;
}

// Vector::operator/=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the scalar /=
// Returns:
//		*this, after it has been divided by the scalar "right".
Vector& Vector::operator/=(float right)
{
	return *this *= 1.0f / right;
/*
	vector[0] /= right;
	vector[1] /= right;
	vector[2] /= right;
	return *this;
*/
}

// Vector::operator=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The vector copy the value of.
// Returns:
//		*this, which has the value of "right".
Vector& Vector::operator=(const Vector& right)
{
	vector[0] = right.vector[0];
	vector[1] = right.vector[1];
	vector[2] = right.vector[2];
	return *this;
}

// Vector::operator==(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the equality test.
// Returns:
//		If the contents of "*this" equal the contents of "right", then 1.
//		Otherwise, 0.  The contents are defined to be equal if they are within a
//		very small distance from each other.
int Vector::operator==(const Vector& right) const
{
	return fEq(vector[0], right.vector[0]) &&
			 fEq(vector[1], right.vector[1]) &&
			 fEq(vector[2], right.vector[2]);
}

// Vector::operator!=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the inequality test.
// Returns:
//		If the contents of "*this" do not equal the contents of "right", then 1.
//		Otherwise, 0.  The contents are defined to be inequal if they are within
//		a very small distance from each other.
int Vector::operator!=(const Vector& right) const
{
	return !(*this == right);
}

// Vector::operator<(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the < test.
// Returns:
//		If the contents of "*this" < the contents of "right", then 1.  Otherwise,
//		0.
int Vector::operator<(const Vector& right) const
{
	return length() < right.length();
}

// Vector::operator>(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the > test.
// Returns:
//		If the contents of "*this" > the contents of "right", then 1.  Otherwise,
//		0.
int Vector::operator>(const Vector& right) const
{
	return length() > right.length();
}

// Vector::operator<=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the <= test.
// Returns:
//		If the contents of "*this" <= the contents of "right", then 1.
//		Otherwise, 0.
int Vector::operator<=(const Vector& right) const
{
	return length() <= right.length();
}

// Vector::operator>=(const Vector&) -- Overloaded Operator
// Expects:
//		right	: The right hand side of the >= test.
// Returns:
//		If the contents of "*this" >= the contents of "right", then 1.
//		Otherwise, 0.
int Vector::operator>=(const Vector& right) const
{
	return length() >= right.length();
}

// Vector::print(FILE*) -- Member Function
// Expects:
//		outFile	: File to output to.
// Results:
//		Prints the Vector.  Does NOT print a trailing line feed.
void Vector::print(FILE* outFile) const
{
	fprintf(outFile, "%f %f %f", vector[0], vector[1], vector[2]);
}

void Vector::print(char* buffer) const
{
	sprintf(buffer, "%.3g %.3g %.3g", vector[0], vector[1], vector[2]);
}

// operator*(float, Vector&) -- Overloaded Non-Member Operator
// Expects:
//		left	: The left hand side of the scalar multiply.
//		right	: The right hand side of the scalar multiply.
// Returns:
//		The Vector resulting from "right" being multiplied by the scalar "left".
Vector operator*(float left, const Vector& right)
{
	return right * left;
}

Vector Vector::mult2(float* matrix) const
{
	Vector tempVec;

	tempVec[0] = vector[0]*matrix[0] + vector[0]*matrix[1] +
	 vector[0]*matrix[2] + matrix[3];
	tempVec[1] = vector[1]*matrix[4] + vector[1]*matrix[5] +
	 vector[1]*matrix[6] + matrix[7];
	tempVec[2] = vector[2]*matrix[8] + vector[2]*matrix[9] +
	 vector[2]*matrix[10] + matrix[11];
	return tempVec;
}


Vector Vector::mult(float* matrix) const
{
	Vector tempVec;

	tempVec[0] = vector[0]*matrix[0] + vector[1]*matrix[1] +
	 vector[2]*matrix[2] + matrix[3];
	tempVec[1] = vector[0]*matrix[4] + vector[1]*matrix[5] +
	 vector[2]*matrix[6] + matrix[7];
	tempVec[2] = vector[0]*matrix[8] + vector[1]*matrix[9] +
	 vector[2]*matrix[10] + matrix[11];
	return tempVec;
}
