/*
 ******************************************************************************
 * SimStarIV
 * Travis Cobbs, Chris Connley, Rusty Larner
 * CSC 480
 *
 * FILE:  Vector.h
 ******************************************************************************
 */
#ifndef __VECTOR_H__
#define __VECTOR_H__

#include <stdio.h>
//#include <windows.h>
//#include <GL/gl.h>

// All overload * operators with 2 vectors are cross product.
// dot product is not an overloaded operator, simply dot.

typedef float* GlPt;

class Vector
{
	public:
		// Constructors
		Vector(void);
		Vector(float, float, float);
		Vector(const Vector&);

		// Destructor
		~Vector(void);

		// Member Functions
		void print(FILE* = stdout) const;
		void print(char* buffer) const;
		float length(void) const;
		float dot(const Vector&) const;
		Vector& normalize(void) {return *this *= 1.0f/length();}
		float get(int i) const { return vector[i];}

		// Overloaded Operators
		Vector operator*(const Vector&) const;
		Vector operator*(float) const;
		Vector operator/(float) const;
		Vector operator+(const Vector&) const;
		Vector operator-(const Vector&) const;
		Vector operator-(void) const;
		Vector& operator*=(const Vector&);
		Vector& operator*=(float);
		Vector& operator/=(float);
		Vector& operator+=(const Vector&);
		Vector& operator-=(const Vector&);
		Vector& operator=(const Vector&);
		int operator==(const Vector&) const;
		int operator!=(const Vector&) const;
		int operator<(const Vector& right) const;
		int operator>(const Vector& right) const;
		int operator<=(const Vector& right) const;
		int operator>=(const Vector& right) const;
		float& operator[](int i) {return vector[i];}
		operator GlPt(void) {return vector;}
		Vector mult(float* matrix) const;
		Vector mult2(float* matrix) const;

	protected:
#ifdef _LEAK_DEBUG
		char className[4];
#endif
		float vector[3];
};

// Overloaded Operator with non-Vector as first argument
Vector operator*(float, const Vector&);

#endif
