#include "LDVExtensionsSetup.h"
#include <TCFoundation/mystring.h>

PFNWGLGETPIXELFORMATATTRIBIVEXTPROC
	LDVExtensionsSetup::wglGetPixelFormatAttribivARB = NULL;
PFNWGLGETPIXELFORMATATTRIBFVEXTPROC
	LDVExtensionsSetup::wglGetPixelFormatAttribfvARB = NULL;
PFNWGLCHOOSEPIXELFORMATEXTPROC
	LDVExtensionsSetup::wglChoosePixelFormatARB = NULL;
PFNWGLGETEXTENSIONSSTRINGARBPROC
	LDVExtensionsSetup::wglGetExtensionsStringARB = NULL;

PFNWGLCREATEPBUFFERARBPROC LDVExtensionsSetup::wglCreatePbufferARB = NULL;
PFNWGLGETPBUFFERDCARBPROC LDVExtensionsSetup::wglGetPbufferDCARB = NULL;
PFNWGLRELEASEPBUFFERDCARBPROC LDVExtensionsSetup::wglReleasePbufferDCARB = NULL;
PFNWGLDESTROYPBUFFERARBPROC LDVExtensionsSetup::wglDestroyPbufferARB = NULL;
PFNWGLQUERYPBUFFERARBPROC LDVExtensionsSetup::wglQueryPbufferARB = NULL;

char *LDVExtensionsSetup::wglExtensions = NULL;
char *LDVExtensionsSetup::glExtensions = NULL;
bool LDVExtensionsSetup::performedInitialSetup = false;
bool LDVExtensionsSetup::stencilPresent = false;
bool LDVExtensionsSetup::alphaPresent = false;

LDVExtensionsSetup::LDVExtensionsSetupCleanup
	LDVExtensionsSetup::extensionsSetupCleanup;

LDVExtensionsSetup::LDVExtensionsSetupCleanup::~LDVExtensionsSetupCleanup(void)
{
	delete LDVExtensionsSetup::wglExtensions;
	LDVExtensionsSetup::wglExtensions = NULL;
	delete LDVExtensionsSetup::glExtensions;
	LDVExtensionsSetup::glExtensions = NULL;
}


LDVExtensionsSetup::LDVExtensionsSetup(HINSTANCE hInstance):
		   CUIOGLWindow("", hInstance, 100, 100, 100, 100)
{
}

LDVExtensionsSetup::~LDVExtensionsSetup(void)
{
}

BOOL LDVExtensionsSetup::initWindow(void)
{
	if (CUIOGLWindow::initWindow())
	{
		GLint intValue;

		if (!wglExtensions)
		{
			wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC)
				wglGetProcAddress("wglGetExtensionsStringARB");

			if (!wglGetExtensionsStringARB)
			{
				return FALSE;
			}
			wglExtensions = copyString(wglGetExtensionsStringARB(hdc));
			glExtensions = copyString((char*)glGetString(GL_EXTENSIONS));
		}
		if (checkForWGLExtension("WGL_ARB_pixel_format"))
		{
			wglGetPixelFormatAttribivARB = (PFNWGLGETPIXELFORMATATTRIBIVEXTPROC)
				wglGetProcAddress("wglGetPixelFormatAttribivARB");
			wglGetPixelFormatAttribfvARB = (PFNWGLGETPIXELFORMATATTRIBFVEXTPROC)
				wglGetProcAddress("wglGetPixelFormatAttribfvARB");
			wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATEXTPROC)
				wglGetProcAddress("wglChoosePixelFormatARB");
		}
		if (havePixelBufferExtension())
		{
			wglCreatePbufferARB = (PFNWGLCREATEPBUFFERARBPROC)
				wglGetProcAddress("wglCreatePbufferARB");
			wglGetPbufferDCARB = (PFNWGLGETPBUFFERDCARBPROC)
				wglGetProcAddress("wglGetPbufferDCARB");
			wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC)
				wglGetProcAddress("wglReleasePbufferDCARB");
			wglDestroyPbufferARB = (PFNWGLDESTROYPBUFFERARBPROC)
				wglGetProcAddress("wglDestroyPbufferARB");
			wglQueryPbufferARB = (PFNWGLQUERYPBUFFERARBPROC)
				wglGetProcAddress("wglQueryPbufferARB");
		}
		glGetIntegerv(GL_STENCIL_BITS, &intValue);
		if (intValue)
		{
			stencilPresent = true;
		}
		glGetIntegerv(GL_ALPHA_BITS, &intValue);
		if (intValue)
		{
			alphaPresent = true;
		}
		return TRUE;
	}
	return FALSE;
}

bool LDVExtensionsSetup::haveMultisampleExtension(void)
{
	return checkForWGLExtension("WGL_ARB_multisample");
}

bool LDVExtensionsSetup::havePixelBufferExtension(void)
{
	return checkForWGLExtension("WGL_ARB_pbuffer");
}

bool LDVExtensionsSetup::haveQuincunxExtension(void)
{
	return checkForExtension("GL_NV_multisample_filter_hint");
}

bool LDVExtensionsSetup::checkForExtension(char* extensionsString,
										   char* extension)
{
	if (extensionsString)
	{
		int extensionLen = strlen(extension);
		char* extensions = extensionsString;

		while (extensions)
		{
			if (strcmp(extensions, extension) == 0 ||
				(strncmp(extensions, extension, extensionLen) == 0 &&
				extensions[extensionLen] == ' ') &&
				(extensions == extensionsString || extensions[-1] == ' '))
			{
				return true;
			}
			extensions = strstr(extensions, extension);
		}
	}
	return false;
}

bool LDVExtensionsSetup::checkForExtension(char* extension)
{
	return checkForExtension(glExtensions, extension);
}

bool LDVExtensionsSetup::checkForWGLExtension(char* extension)
{
	return checkForExtension(wglExtensions, extension);
/*
	if (wglExtensions)
	{
		int extensionLen = strlen(extension);
		char* extensions = wglExtensions;

		while (extensions)
		{
			if (strcmp(extensions, extension) == 0 ||
				(strncmp(extensions, extension, extensionLen) == 0 &&
				extensions[extensionLen] == ' ') &&
				(extensions == wglExtensions || extensions[-1] == ' '))
			{
				return true;
			}
			extensions = strstr(extensions, extension);
		}
	}
	return false;
*/
}

void LDVExtensionsSetup::closeWindow(void)
{
	if (hWindow)
	{
		ShowWindow(hWindow, SW_HIDE);
	}
	CUIOGLWindow::closeWindow();
}

void LDVExtensionsSetup::setup(HINSTANCE hInstance)
{
	if (!performedInitialSetup)
	{
		LDVExtensionsSetup *extensionsSetup = new LDVExtensionsSetup(hInstance);

		extensionsSetup->initWindow();
		extensionsSetup->closeWindow();
		extensionsSetup->release();
		performedInitialSetup = true;
	}
}