#include "LDViewPreferences.h"
#include "LDrawModelViewer.h"
#include "AppResources.h"

#include <TCFoundation/TCUserDefaults.h>
#include <Commctrl.h>
#include <stdio.h>

#define QUALITY_LIGHTING_KEY "UseQualityLighting"
#define SHOW_FPS_KEY "ShowFPS"
#define SHOWS_HIGHLIGHT_LINES_KEY "ShowHighlightLines"
#define QUALITY_STUDS_KEY "UseQualityStuds"
#define PRIMITIVE_SUBSTITUTION_KEY "AllowPrimitiveSubstitution"
#define FLAT_SHADING_KEY "UseFlatShading"
#define SPECULAR_KEY "UseSpecular"
#define BACKGROUND_COLOR_KEY "BackgroundColor"
#define CUSTOM_COLORS_KEY "CustomColors"
#define SEAMS_KEY "Seams"
#define SEAM_WIDTH_KEY "SeamWidth"
#define WIREFRAME_KEY "Wireframe"
#define WIREFRAME_FOG_KEY "WireframeFog"
#define POLYGON_OFFSET_KEY "PolygonOffset"
#define LIGHTING_KEY "Lighting"
#define SUBDUED_LIGHTING_KEY "SubduedLighting"
#define FULLSCREEN_REFRESH_KEY "FullscreenRefresh"
#define STIPPLE_KEY "UseStipple"
#define SORT_KEY "SortTransparent"
#define POLL_KEY "Poll"
#define SHOW_ERRORS_KEY "ShowErrors"

LDViewPreferences::LDViewPreferences(LDrawModelViewer* modelViewer)
				  :modelViewer(modelViewer ?
					   ((LDrawModelViewer*)modelViewer->retain()) : NULL),
				   hDialog(NULL),
				   qualityLighting(TCUserDefaults::longForKey(QUALITY_LIGHTING_KEY)),
				   showsFPS(TCUserDefaults::longForKey(SHOW_FPS_KEY)),
				   showsHighlightLines(TCUserDefaults::
					   longForKey(SHOWS_HIGHLIGHT_LINES_KEY)),
				   qualityStuds(TCUserDefaults::longForKey(QUALITY_STUDS_KEY)),
				   allowPrimitiveSubstitution(TCUserDefaults::longForKey(
					   PRIMITIVE_SUBSTITUTION_KEY, 1)),
				   usesFlatShading(TCUserDefaults::longForKey(FLAT_SHADING_KEY)),
				   showErrors(TCUserDefaults::longForKey(SHOW_ERRORS_KEY, 1)),
				   usesSpecular(TCUserDefaults::longForKey(SPECULAR_KEY, 1)),
				   backgroundColor((COLORREF)TCUserDefaults::
					   longForKey(BACKGROUND_COLOR_KEY)),
				   useSeams(TCUserDefaults::longForKey(SEAMS_KEY, -1)),
				   seamWidth(TCUserDefaults::longForKey(SEAM_WIDTH_KEY, 50)),
				   drawWireframe(TCUserDefaults::longForKey(WIREFRAME_KEY)),
				   useWireframeFog(TCUserDefaults::longForKey(WIREFRAME_FOG_KEY)),
				   usePolygonOffset(TCUserDefaults::longForKey(POLYGON_OFFSET_KEY)),
				   useLighting(TCUserDefaults::longForKey(LIGHTING_KEY, 1)),
				   subduedLighting(TCUserDefaults::longForKey(SUBDUED_LIGHTING_KEY)),
				   fullScreenRefresh(TCUserDefaults::
					   longForKey(FULLSCREEN_REFRESH_KEY)),
				   useStipple(TCUserDefaults::longForKey(STIPPLE_KEY)),
				   sortTransparent(TCUserDefaults::longForKey(SORT_KEY))
{
	int i;
	int r, g, b;

	if (useSeams == -1)
	{
		if (seamWidth)
		{
			useSeams = 1;
		}
		else
		{
			useSeams = 0;
		}
	}
	getBackgroundRGB(r, g, b);
	if (modelViewer)
	{
		modelViewer->setQualityLighting(qualityLighting);
		modelViewer->setShowsHighlightLines(showsHighlightLines);
		modelViewer->setQualityStuds(qualityStuds);
		modelViewer->setAllowPrimitiveSubstitution(allowPrimitiveSubstitution);
		modelViewer->setUsesFlatShading(usesFlatShading);
		modelViewer->setUsesSpecular(usesSpecular);
		modelViewer->setBackgroundRGB(r, g, b);
		if (useSeams)
		{
			modelViewer->setSeamWidth(seamWidth / 100.0f);
		}
		else
		{
			modelViewer->setSeamWidth(0.0f);
		}
		modelViewer->setDrawWireframe(drawWireframe);
		modelViewer->setUseWireframeFog(useWireframeFog);
		modelViewer->setUsePolygonOffset(usePolygonOffset);
		modelViewer->setUseLighting(useLighting);
		modelViewer->setSubduedLighting(subduedLighting);
		modelViewer->setUseStipple(useStipple);
		modelViewer->setSortTransparent(sortTransparent);
	}
	for (i = 0; i < 16; i++)
	{
		char key[128];

		sprintf(key, "%s/Color%02d", CUSTOM_COLORS_KEY, i);
		customColors[i] = TCUserDefaults::longForKey(key);
	}
}

void LDViewPreferences::dealloc(void)
{
	if (hBackgroundColorBitmap)
	{
		DeleteObject(hBackgroundColorBitmap);
	}
	if (hBackgroundColorDC)
	{
		DeleteDC(hBackgroundColorDC);
	}
	if (modelViewer)
	{
		modelViewer->release();
	}
}

void LDViewPreferences::setup(HWND dialog)
{
	if (dialog)
	{
		hDialog = dialog;
		hApplyButton = GetDlgItem(hDialog, IDC_APPLY);
		hFullScreenRateField = GetDlgItem(hDialog, IDC_FS_RATE);
		hSeamSpin = GetDlgItem(hDialog, IDC_SEAM_SPIN);
		SendDlgItemMessage(hDialog, IDC_FS_RATE, EM_LIMITTEXT, 3, 0);
		SendDlgItemMessage(hDialog, IDC_LIGHTING_QUALITY, BM_SETCHECK,
			qualityLighting, 0);
		SendDlgItemMessage(hDialog, IDC_FRAME_RATE, BM_SETCHECK, showsFPS,
			0);
		SendDlgItemMessage(hDialog, IDC_HIGHLIGHTS, BM_SETCHECK,
			showsHighlightLines, 0);
		SendDlgItemMessage(hDialog, IDC_STUD_QUALITY, BM_SETCHECK,
			!qualityStuds, 0);
		SendDlgItemMessage(hDialog, IDC_PRIMITIVE_SUBSTITUTION, BM_SETCHECK,
			allowPrimitiveSubstitution, 0);
		SendDlgItemMessage(hDialog, IDC_FLAT_SHADING, BM_SETCHECK,
			usesFlatShading, 0);
		SendDlgItemMessage(hDialog, IDC_SHOW_ERRORS, BM_SETCHECK,
			showErrors, 0);
		SendDlgItemMessage(hDialog, IDC_STIPPLE, BM_SETCHECK,
			useStipple, 0);
		SendDlgItemMessage(hDialog, IDC_SORT, BM_SETCHECK,
			sortTransparent, 0);
		SendDlgItemMessage(hDialog, IDC_SPECULAR, BM_SETCHECK,
			usesSpecular, 0);
		SendDlgItemMessage(hDialog, IDC_WIREFRAME, BM_SETCHECK,
			drawWireframe, 0);
		SendDlgItemMessage(hDialog, IDC_WIREFRAME_FOG, BM_SETCHECK,
			useWireframeFog, 0);
		SendDlgItemMessage(hDialog, IDC_QUALITY_LINES, BM_SETCHECK,
			usePolygonOffset, 0);
		SendDlgItemMessage(hDialog, IDC_LIGHTING, BM_SETCHECK,
			useLighting, 0);
		SendDlgItemMessage(hDialog, IDC_LIGHTING_SUBDUED, BM_SETCHECK,
			subduedLighting, 0);
		SendDlgItemMessage(hDialog, IDC_SEAMS, BM_SETCHECK,
			useSeams, 0);
		setupBackgroundColorButton();
		setupSeamWidth();
		setupFullScreenRefresh();
		EnableWindow(hApplyButton, FALSE);
	}
	else
	{
		hDialog = NULL;
		hApplyButton = NULL;
		hFullScreenRateField = NULL;
		hSeamSpin = NULL;
	}
}

void LDViewPreferences::setupBackgroundColorButton(void)
{
	RECT clientRect;
	int imageWidth;
	int imageHeight;
	HDC hdc;

	hBackgroundColorButton = GetDlgItem(hDialog, IDC_BACKGROUND_COLOR);
	GetClientRect(hBackgroundColorButton, &clientRect);
	imageWidth = clientRect.right - clientRect.left - 10;
	imageHeight = clientRect.bottom - clientRect.top - 10;
//	hBackgroundColorDC = CreateDC("DISPLAY", NULL, NULL, NULL);
//	hBackgroundColorDC = CreateCompatibleDC(GetDC(hDialog));
//	hdc = CreateDC("DISPLAY", NULL, NULL, NULL);
	hdc = GetDC(hDialog);
	hBackgroundColorDC = CreateCompatibleDC(hdc);
	hBackgroundColorBitmap = CreateCompatibleBitmap(hdc, imageWidth,
		imageHeight);
	ReleaseDC(hDialog, hdc);
//	DeleteDC(hdc);
	SetBitmapDimensionEx(hBackgroundColorBitmap, imageWidth, imageHeight, NULL);
	redrawBackgroundColorBitmap();
	SendDlgItemMessage(hDialog, IDC_BACKGROUND_COLOR, BM_SETIMAGE,
		(WPARAM)IMAGE_BITMAP, (LPARAM)hBackgroundColorBitmap);
}

void LDViewPreferences::setupSeamWidth(void)
{
	char seamWidthString[128];
	UDACCEL accels[] = {{0, 1}, {1, 10}, {2, 20}};

	sprintf(seamWidthString, "%0.2f", seamWidth / 100.0f);
	SendDlgItemMessage(hDialog, IDC_SEAM_WIDTH_FIELD, WM_SETTEXT, 0,
		(LPARAM)seamWidthString);
	SendDlgItemMessage(hDialog, IDC_SEAM_SPIN, UDM_SETRANGE, 0,
		MAKELONG((short)500, (short)0)); 
	SendDlgItemMessage(hDialog, IDC_SEAM_SPIN, UDM_SETPOS, 0, seamWidth);
	SendDlgItemMessage(hDialog, IDC_SEAM_SPIN, UDM_SETACCEL,
		sizeof(accels) / sizeof(UDACCEL), (long)accels);
}

void LDViewPreferences::setupFullScreenRefresh(void)
{
	if (fullScreenRefresh)
	{
		char buf[128];

		sprintf(buf, "%d", fullScreenRefresh);
		buf[3] = 0;
		SendDlgItemMessage(hDialog, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)buf);
		SendDlgItemMessage(hDialog, IDC_FS_REFRESH, BM_SETCHECK, 1, 0);
		EnableWindow(hFullScreenRateField, TRUE);
	}
	else
	{
		SendDlgItemMessage(hDialog, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)"");
		SendDlgItemMessage(hDialog, IDC_FS_REFRESH, BM_SETCHECK, 0, 0);
		EnableWindow(hFullScreenRateField, FALSE);
	}
}

void LDViewPreferences::drawButtonBorder(HDC hdc, COLORREF color1,
										 COLORREF color2, RECT rect)
{
	HPEN hPen1 = CreatePen(PS_SOLID, 1, color1);
	HPEN hPen2 = CreatePen(PS_SOLID, 1, color2);
	HPEN hOldPen;

	hOldPen = (HPEN)SelectObject(hdc, hPen1);
	MoveToEx(hBackgroundColorDC, rect.left, rect.bottom, NULL);
	LineTo(hBackgroundColorDC, rect.left, rect.top);
	LineTo(hBackgroundColorDC, rect.right, rect.top);
	SelectObject(hBackgroundColorDC, hPen2);
	LineTo(hBackgroundColorDC, rect.right, rect.bottom);
	LineTo(hBackgroundColorDC, rect.left, rect.bottom);
	MoveToEx(hBackgroundColorDC, rect.left, rect.bottom, NULL);
	LineTo(hBackgroundColorDC, rect.right, rect.bottom);
	SelectObject(hdc, hOldPen);
	DeleteObject(hPen1);
	DeleteObject(hPen2);
}

void LDViewPreferences::redrawBackgroundColorBitmap(void)
{
	HBRUSH hBrush = CreateSolidBrush(backgroundColor);
	RECT bitmapRect;
	SIZE bitmapSize;
	HBITMAP hOldBitmap;
	HBRUSH hOldBrush;

	GetBitmapDimensionEx(hBackgroundColorBitmap, &bitmapSize);
	bitmapRect.left = 0;
	bitmapRect.top = 0;
	bitmapRect.right = bitmapSize.cx - 1;
	bitmapRect.bottom = bitmapSize.cy - 1;
	hOldBrush = (HBRUSH)SelectObject(hBackgroundColorDC, hBrush);
	hOldBitmap = (HBITMAP)SelectObject(hBackgroundColorDC,
		hBackgroundColorBitmap);
	FillRect(hBackgroundColorDC, &bitmapRect, hBrush);
	drawButtonBorder(hBackgroundColorDC, GetSysColor(COLOR_3DSHADOW),
		GetSysColor(COLOR_3DHILIGHT), bitmapRect);
	InflateRect(&bitmapRect, -1, -1);
	drawButtonBorder(hBackgroundColorDC, GetSysColor(COLOR_3DDKSHADOW),
		GetSysColor(COLOR_3DLIGHT), bitmapRect);
	SelectObject(hBackgroundColorDC, hOldBrush);
	SelectObject(hBackgroundColorDC, hOldBitmap);
	DeleteObject(hBrush);
	InvalidateRect(hBackgroundColorButton, NULL, 1);
}

DWORD LDViewPreferences::doApply(void)
{
	int temp;
//	DWORD retVal = 0;

	if ((temp = SendDlgItemMessage(hDialog, IDC_LIGHTING_QUALITY,
		BM_GETCHECK, 0, 0)) != qualityLighting)
	{
		qualityLighting = temp;
		if (modelViewer)
		{
			modelViewer->setQualityLighting(qualityLighting);
		}
		TCUserDefaults::setLongForKey(qualityLighting, QUALITY_LIGHTING_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_FRAME_RATE,
		BM_GETCHECK, 0, 0)) != showsFPS)
	{
		showsFPS = temp;
		TCUserDefaults::setLongForKey(showsFPS, SHOW_FPS_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_HIGHLIGHTS,
		BM_GETCHECK, 0, 0)) != showsHighlightLines)
	{
		showsHighlightLines = temp;
		if (modelViewer)
		{
			modelViewer->setShowsHighlightLines(showsHighlightLines);
		}
//		retVal |= LDP_NEEDS_RELOAD;
		TCUserDefaults::setLongForKey(showsHighlightLines,
			SHOWS_HIGHLIGHT_LINES_KEY);
	}
	if ((temp = !SendDlgItemMessage(hDialog, IDC_STUD_QUALITY,
		BM_GETCHECK, 0, 0)) != qualityStuds)
	{
		qualityStuds = temp;
		if (modelViewer)
		{
			modelViewer->setQualityStuds(qualityStuds);
		}
//		retVal |= LDP_NEEDS_RELOAD;
		TCUserDefaults::setLongForKey(qualityStuds, QUALITY_STUDS_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_PRIMITIVE_SUBSTITUTION,
		BM_GETCHECK, 0, 0)) != allowPrimitiveSubstitution)
	{
		allowPrimitiveSubstitution = temp;
		if (modelViewer)
		{
			modelViewer->setAllowPrimitiveSubstitution(allowPrimitiveSubstitution);
		}
//		retVal |= LDP_NEEDS_RELOAD;
		TCUserDefaults::setLongForKey(allowPrimitiveSubstitution,
			PRIMITIVE_SUBSTITUTION_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_FLAT_SHADING,
		BM_GETCHECK, 0, 0)) != usesFlatShading)
	{
		usesFlatShading = temp;
		if (modelViewer)
		{
			modelViewer->setUsesFlatShading(usesFlatShading);
		}
		TCUserDefaults::setLongForKey(usesFlatShading, FLAT_SHADING_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_SHOW_ERRORS,
		BM_GETCHECK, 0, 0)) != showErrors)
	{
		showErrors = temp;
		TCUserDefaults::setLongForKey(showErrors, SHOW_ERRORS_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_STIPPLE,
		BM_GETCHECK, 0, 0)) != useStipple)
	{
		useStipple = temp;
		if (modelViewer)
		{
			modelViewer->setUseStipple(useStipple);
		}
/*
		if (sortTransparent)
		{
			retVal |= LDP_NEEDS_RELOAD;
		}
		else
		{
			retVal |= LDP_NEEDS_RECOMPILE;
		}
*/
		TCUserDefaults::setLongForKey(useStipple, STIPPLE_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_SORT,
		BM_GETCHECK, 0, 0)) != sortTransparent)
	{
		sortTransparent = temp;
		if (modelViewer)
		{
			modelViewer->setSortTransparent(sortTransparent);
		}
//		retVal |= LDP_NEEDS_RELOAD;
		TCUserDefaults::setLongForKey(sortTransparent, SORT_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_SPECULAR,
		BM_GETCHECK, 0, 0)) != usesSpecular)
	{
		usesSpecular = temp;
		if (modelViewer)
		{
			modelViewer->setUsesSpecular(usesSpecular);
		}
//		retVal |= LDP_NEEDS_SETUP;
		TCUserDefaults::setLongForKey(usesSpecular, SPECULAR_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_LIGHTING,
		BM_GETCHECK, 0, 0)) != useLighting)
	{
		useLighting = temp;
		if (modelViewer)
		{
			modelViewer->setUseLighting(useLighting);
		}
//		retVal |= LDP_NEEDS_RECOMPILE;
		TCUserDefaults::setLongForKey(useLighting, LIGHTING_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_WIREFRAME,
		BM_GETCHECK, 0, 0)) != drawWireframe)
	{
		drawWireframe = temp;
		if (modelViewer)
		{
			modelViewer->setDrawWireframe(drawWireframe);
		}
		TCUserDefaults::setLongForKey(drawWireframe, WIREFRAME_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_WIREFRAME_FOG,
		BM_GETCHECK, 0, 0)) != useWireframeFog)
	{
		useWireframeFog = temp;
		if (modelViewer)
		{
			modelViewer->setUseWireframeFog(useWireframeFog);
		}
		TCUserDefaults::setLongForKey(useWireframeFog, WIREFRAME_FOG_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_QUALITY_LINES,
		BM_GETCHECK, 0, 0)) != usePolygonOffset)
	{
		usePolygonOffset = temp;
		if (modelViewer)
		{
			modelViewer->setUsePolygonOffset(usePolygonOffset);
		}
//		retVal |= LDP_NEEDS_RECOMPILE;
		TCUserDefaults::setLongForKey(usePolygonOffset, POLYGON_OFFSET_KEY);
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_LIGHTING_SUBDUED,
		BM_GETCHECK, 0, 0)) != subduedLighting)
	{
		subduedLighting = temp;
		if (modelViewer)
		{
			modelViewer->setSubduedLighting(subduedLighting);
		}
		TCUserDefaults::setLongForKey(subduedLighting, SUBDUED_LIGHTING_KEY);
	}
	temp = SendDlgItemMessage(hDialog, IDC_FS_REFRESH, BM_GETCHECK, 0, 0);
	if (temp)
	{
		char buf[4];

		SendDlgItemMessage(hDialog, IDC_FS_RATE, WM_GETTEXT, 4,
			(LPARAM)buf);
		if (sscanf(buf, "%d", &temp) == 1)
		{
			if (!temp)
			{
				temp = -1;
			}
		}
		else
		{
			temp = -1;
		}
	}
	else
	{
		temp = 0;
	}
	if (temp != fullScreenRefresh && temp >= 0)
	{
		fullScreenRefresh = temp;
		TCUserDefaults::setLongForKey(fullScreenRefresh,
			FULLSCREEN_REFRESH_KEY);
	}
	if ((COLORREF)TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY) !=
		backgroundColor)
	{
		int r, g, b;

		getBackgroundRGB(r, g, b);
		TCUserDefaults::setLongForKey((long)backgroundColor,
			BACKGROUND_COLOR_KEY);
		if (modelViewer)
		{
			modelViewer->setBackgroundRGB(r, g, b);
		}
	}
	if (TCUserDefaults::longForKey(SEAM_WIDTH_KEY) != seamWidth)
	{
		TCUserDefaults::setLongForKey(seamWidth, SEAM_WIDTH_KEY);
		if (useSeams)
		{
			if (modelViewer)
			{
				modelViewer->setSeamWidth(seamWidth / 100.0f);
			}
//			retVal |= LDP_NEEDS_RELOAD;
		}
	}
	if ((temp = SendDlgItemMessage(hDialog, IDC_SEAMS,
		BM_GETCHECK, 0, 0)) != useSeams)
	{
		useSeams = temp;
		if (modelViewer)
		{
			if (useSeams)
			{
				modelViewer->setSeamWidth(seamWidth / 100.0f);
			}
			else
			{
				modelViewer->setSeamWidth(0.0f);
			}
		}
//		retVal |= LDP_NEEDS_RELOAD;
		TCUserDefaults::setLongForKey(useSeams, SEAMS_KEY);
	}
	EnableWindow(hApplyButton, FALSE);
	return 0;
//	return retVal;
}

void LDViewPreferences::getBackgroundRGB(int& r, int& g, int& b)
{
	r = backgroundColor & 0xFF;
	g = (backgroundColor >> 8) & 0xFF;
	b = (backgroundColor >> 16) & 0xFF;
}

void LDViewPreferences::chooseBackgroundColor(void)
{
	CHOOSECOLOR chooseColor;
	int i;

	memset(&chooseColor, 0, sizeof CHOOSECOLOR);
	chooseColor.lStructSize = sizeof CHOOSECOLOR;
	chooseColor.hwndOwner = hDialog;
	chooseColor.rgbResult = backgroundColor;
	chooseColor.lpCustColors = customColors;
	chooseColor.Flags = CC_ANYCOLOR | CC_RGBINIT;
	EnableWindow(hDialog, FALSE);
	if (ChooseColor(&chooseColor))
	{
//		int r, g, b;
//
		backgroundColor = chooseColor.rgbResult;
//		getBackgroundRGB(r, g, b);
//		TCUserDefaults::setLongForKey((long)backgroundColor,
//			BACKGROUND_COLOR_KEY);
//		modelViewer->setBackgroundRGB(r, g, b);
//		forceRedraw();
		redrawBackgroundColorBitmap();
	}
	for (i = 0; i < 16; i++)
	{
		if (customColors[i])
		{
			char key[128];

			sprintf(key, "%s/Color%02d", CUSTOM_COLORS_KEY, i);
			TCUserDefaults::setLongForKey((long)customColors[i], key);
		}
	}
	EnableWindow(hDialog, TRUE);
}

DWORD LDViewPreferences::doClick(int controlId, HWND controlHWnd)
{
	switch (controlId)
	{
		case IDCANCEL:
			return doCancel();
			break;
		case IDOK:
			return doOK();
			break;
		case IDC_APPLY:
			return doApply();
			break;
		case IDC_FS_REFRESH:
			doFSRefresh();
			break;
		case IDC_STIPPLE:
			doStipple();
			EnableWindow(hApplyButton, TRUE);
			break;
		case IDC_SORT:
			doSort();
			EnableWindow(hApplyButton, TRUE);
			break;
		case IDC_HIGHLIGHTS:
			doHighlights();
			EnableWindow(hApplyButton, TRUE);
			break;
		case IDC_QUALITY_LINES:
			doQualityLines();
			EnableWindow(hApplyButton, TRUE);
			break;
		case IDC_BACKGROUND_COLOR:
			chooseBackgroundColor();
			EnableWindow(hApplyButton, TRUE);
			break;
		default:
			EnableWindow(hApplyButton, TRUE);
			break;
	}
	return 0;
}

DWORD LDViewPreferences::doOK(void)
{
	DWORD retVal = doApply();
	return retVal | LDP_DONE;
}

void LDViewPreferences::doFSRefresh(void)
{
	if (SendDlgItemMessage(hDialog, IDC_FS_REFRESH, BM_GETCHECK, 0, 0))
	{
		EnableWindow(hFullScreenRateField, TRUE);
	}
	else
	{
		SendDlgItemMessage(hDialog, IDC_FS_RATE, WM_SETTEXT, 0,
			(LPARAM)"");
		EnableWindow(hFullScreenRateField, FALSE);
	}
}

void LDViewPreferences::doStipple(void)
{
	if (SendDlgItemMessage(hDialog, IDC_STIPPLE, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hDialog, IDC_SORT, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doSort(void)
{
	if (SendDlgItemMessage(hDialog, IDC_SORT, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hDialog, IDC_STIPPLE, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doHighlights(void)
{
	if (!SendDlgItemMessage(hDialog, IDC_HIGHLIGHTS, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hDialog, IDC_QUALITY_LINES, BM_SETCHECK, 0, 0);
	}
}

void LDViewPreferences::doQualityLines(void)
{
	if (SendDlgItemMessage(hDialog, IDC_QUALITY_LINES, BM_GETCHECK, 0, 0))
	{
		SendDlgItemMessage(hDialog, IDC_HIGHLIGHTS, BM_SETCHECK, 1, 0);
	}
}

DWORD LDViewPreferences::doCancel(void)
{
	backgroundColor =
		(COLORREF)TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY);
	seamWidth = TCUserDefaults::longForKey(SEAM_WIDTH_KEY);
//	fullScreenRefresh = TCUserDefaults::longForKey(FULLSCREEN_REFRESH_KEY);

	return LDP_DONE;
//	doDialogClose(hDialog);
}

BOOL LDViewPreferences::doVScroll(HWND hDlg, int scrollCode, int position,
								  HWND hScrollBar)
{
	if (scrollCode == SB_THUMBPOSITION && hScrollBar == hSeamSpin)
	{
		char seamWidthString[128];

		seamWidth += (position - seamWidth);
		sprintf(seamWidthString, "%0.2f", seamWidth / 100.0f);
		SendDlgItemMessage(hDlg, IDC_SEAM_WIDTH_FIELD, WM_SETTEXT, 0,
			(LPARAM)seamWidthString);
		SendDlgItemMessage(hDlg, IDC_SEAM_SPIN, UDM_SETPOS, 0, seamWidth);
		if (seamWidth)
		{
			SendDlgItemMessage(hDialog, IDC_SEAMS, BM_SETCHECK,
				1, 0);
		}
		else
		{
			SendDlgItemMessage(hDialog, IDC_SEAMS, BM_SETCHECK,
				0, 0);
		}
		EnableWindow(hApplyButton, TRUE);
	}
	return TRUE;
}

DWORD LDViewPreferences::doCommand(HWND hDlg, int controlId, int notifyCode,
								   HWND controlHWnd)
{
	if (notifyCode == BN_CLICKED)
	{
		if (hDlg == hDialog)
		{
			return doClick(controlId, controlHWnd);
		}
	}
	else if (notifyCode == EN_CHANGE)
	{
		if (controlId == IDC_FS_RATE)
		{
			EnableWindow(hApplyButton, TRUE);
			return 0;
		}
	}
	return LDP_UNKNOWN_COMMAND;
}