#ifndef __LDVPREFERENCES_H__
#define __LDVPREFERENCES_H__

#include <TCFoundation/TCObject.h>
#include <windows.h>

class LDrawModelViewer;

#define LDP_DONE 0x01
#define LDP_UNKNOWN_COMMAND 0x02
/*
#define LDP_NEEDS_SETUP 0x04
#define LDP_NEEDS_RELOAD 0x08
#define LDP_NEEDS_RECOMPILE 0x10
*/

class LDViewPreferences: public TCObject
{
public:
	LDViewPreferences(LDrawModelViewer *modelViewer = NULL);
	void setup(HWND dialog);

	BOOL getQualityLighting(void) { return qualityLighting; }
	BOOL getShowsFPS(void) { return showsFPS; }
	BOOL getShowsHighlightLines(void) { return showsHighlightLines; }
	BOOL getQualityStuds(void) { return qualityStuds; }
	BOOL getAllowPrimitiveSubstitution(void)
	{
		return allowPrimitiveSubstitution;
	}
	BOOL getUsesFlatShading(void) { return usesFlatShading; }
	BOOL getUsesSpecular(void) { return usesSpecular; }
	COLORREF getBackgroundColor(void) { return backgroundColor; }
	COLORREF* getCustomColors(void) { return customColors; }
	int getUseSeams(void) { return useSeams; }
	int getSeamWidth(void) { return seamWidth; }
	BOOL getDrawWireframe(void) { return drawWireframe; }
	BOOL getUseWireframeFog(void) { return useWireframeFog; }
	BOOL getUsePolygonOffset(void) { return usePolygonOffset; }
	BOOL getUseLighting(void) { return useLighting; }
	BOOL getSubduedLighting(void) { return subduedLighting; }
	int getFullScreenRefresh(void) { return fullScreenRefresh; }
	BOOL getUseStipple(void) { return useStipple; }
	BOOL getSortTransparent(void) { return sortTransparent; }
	virtual DWORD doApply(void);

	virtual BOOL doVScroll(HWND hDlg, int scrollCode, int position,
		HWND hScrollBar);
	virtual DWORD doCommand(HWND hDlg, int controlId, int notifyCode,
		HWND controlHWnd);
protected:
	virtual void dealloc(void);
	virtual void setupBackgroundColorButton(void);
	virtual void setupSeamWidth(void);
	virtual void setupFullScreenRefresh(void);
	virtual void redrawBackgroundColorBitmap(void);
	virtual void getBackgroundRGB(int&, int&, int&);
	virtual void chooseBackgroundColor(void);
	virtual DWORD doOK(void);
	virtual void doFSRefresh(void);
	virtual void doStipple(void);
	virtual void doSort(void);
	virtual void doHighlights(void);
	virtual void doQualityLines(void);
	virtual DWORD doCancel(void);
	virtual DWORD doClick(int controlId, HWND controlHWnd);
	virtual void drawButtonBorder(HDC hdc, COLORREF color1, COLORREF color2,
		RECT rect);

	LDrawModelViewer* modelViewer;
	HWND hDialog;
	HWND hApplyButton;
	HBITMAP hBackgroundColorBitmap;
	HWND hBackgroundColorButton;
	HDC hBackgroundColorDC;
	HWND hFullScreenRateField;
	HWND hSeamSpin;

	BOOL qualityLighting;
	BOOL showsFPS;
	BOOL showsHighlightLines;
	BOOL qualityStuds;
	BOOL allowPrimitiveSubstitution;
	BOOL usesFlatShading;
	BOOL showErrors;
	BOOL usesSpecular;
	COLORREF backgroundColor;
	COLORREF customColors[16];
	int useSeams;
	int seamWidth;
	BOOL drawWireframe;
	BOOL useWireframeFog;
	BOOL usePolygonOffset;
	BOOL useLighting;
	BOOL subduedLighting;
	int fullScreenRefresh;
	BOOL useStipple;
	BOOL sortTransparent;
};

#endif // __LDVPREFERENCES_H__
