#ifndef __LDVPREFERENCES_H__
#define __LDVPREFERENCES_H__

//#include <TCFoundation/TCObject.h>
#include <CUI/CUIPropertySheet.h>
#include <windows.h>

#include <LDLib/LDrawModelViewer.h>

//class LDrawModelViewer;

#define LDP_DONE 0x01
#define LDP_UNKNOWN_COMMAND 0x02
/*
#define LDP_NEEDS_SETUP 0x04
#define LDP_NEEDS_RELOAD 0x08
#define LDP_NEEDS_RECOMPILE 0x10
*/

class LDViewPreferences: public CUIPropertySheet
{
public:
	LDViewPreferences(HINSTANCE hInstance,
		LDrawModelViewer *modelViewer = NULL);

	bool getQualityLighting(void) { return qualityLighting; }
	bool getShowsFPS(void) { return showsFPS; }
	bool getShowsHighlightLines(void) { return showsHighlightLines; }
	bool getDrawConditionalHighlights(void)
	{
		return drawConditionalHighlights;
	}
	bool getPerformSmoothing(void) { return performSmoothing; }
	bool getLineSmoothing(void) { return lineSmoothing; }
	bool getQualityStuds(void) { return qualityStuds; }
	bool getAllowPrimitiveSubstitution(void)
	{
		return allowPrimitiveSubstitution;
	}
	bool getUsesFlatShading(void) { return usesFlatShading; }
	bool getUsesSpecular(void) { return usesSpecular; }
	COLORREF getBackgroundColor(void) { return backgroundColor; }
	COLORREF* getCustomColors(void) { return customColors; }
	int getUseSeams(void) { return useSeams; }
	int getSeamWidth(void) { return seamWidth; }
	bool getDrawWireframe(void) { return drawWireframe; }
	bool getUseWireframeFog(void) { return useWireframeFog; }
	bool getUsePolygonOffset(void) { return usePolygonOffset; }
	bool getUseLighting(void) { return useLighting; }
	bool getSubduedLighting(void) { return subduedLighting; }
	int getFullScreenRefresh(void) { return fullScreenRefresh; }
	bool getUseStipple(void) { return useStipple; }
	bool getSortTransparent(void) { return sortTransparent; }
	virtual void applyChanges(void);
	virtual int run(void);

	virtual BOOL doDialogVScroll(HWND hDlg, int scrollCode, int position,
		HWND hScrollBar);
	int getGeneralPageNumber(void) { return generalPageNumber; }
	int getGeometryPageNumber(void) { return geometryPageNumber; }
	int getEffectsPageNumber(void) { return effectsPageNumber; }

	static char* getLDViewPath(char* helpFilename, bool useQuotes = false);
	static char* getLDViewPath(bool useQuotes = false);
protected:
	virtual ~LDViewPreferences(void);
	virtual void dealloc(void);
	virtual void setupColorButton(HWND hPage, HWND &hColorButton,
		int controlID, HBITMAP &hButtonBitmap, COLORREF color);
	virtual void setupBackgroundColorButton(void);
	virtual void setupDefaultColorButton(void);
	virtual void setupSeamWidth(void);
	virtual void setupFullScreenRefresh(void);
	virtual void redrawColorBitmap(HWND hColorButton, HBITMAP hButtonBitmap,
		COLORREF color);
	virtual void getRGB(int color, int &r, int &g, int &b);
	virtual void chooseColor(HWND hColorButton, HBITMAP hColorBitmap,
		COLORREF &color);
	virtual void chooseBackgroundColor(void);
	virtual void chooseDefaultColor(void);
	virtual void doFSRefresh(void);
	virtual void doStipple(void);
	virtual void doSort(void);
	virtual void doFlatShading(void);
	virtual void doSmoothCurves(void);
	virtual void doHighlights(void);
	virtual void doWireframe(void);
	virtual void doLighting(void);
	virtual void doStereo(void);
	virtual void doCutaway(void);
	virtual void doFSAA(void);
	virtual void doSeams(void);
	virtual void doPrimitives(void);
	virtual void doTextureStuds(void);
	virtual bool doApply(void);
	virtual DWORD doClick(HWND hPage, int controlId, HWND controlHWnd);
	virtual void doGeneralClick(int controlId, HWND controlHWnd);
	virtual void doGeometryClick(int controlId, HWND controlHWnd);
	virtual void doEffectsClick(int controlId, HWND controlHWnd);
	virtual void doPrimitivesClick(int controlId, HWND controlHWnd);
//	virtual void drawButtonBorder(HDC hdc, COLORREF color1, COLORREF color2,
//		RECT rect);
	virtual BOOL doDialogNotify(HWND hDlg, int controlId, LPNMHDR notification);
	virtual BOOL doDialogCommand(HWND hDlg, int controlId, int notifyCode,
		HWND controlHWnd);
	virtual BOOL doDialogHelp(HWND hDlg, LPHELPINFO helpInfo);
	virtual void doReset(void);
	virtual void setupPage(int pageNumber);
	virtual void setupGeneralPage(void);
	virtual void setupGeometryPage(void);
	virtual void setupEffectsPage(void);
	virtual void setupPrimitivesPage(void);
	virtual void setupFSAA(void);
	virtual void setupModelGeometry(void);
	virtual void setupWireframe(void);
	virtual void setupSubstitution(void);
	virtual void enableWireframe(void);
	virtual void disableWireframe(void);
	virtual void setupEdgeLines(void);
	virtual void enableEdges(void);
	virtual void disableEdges(void);
	virtual void enableFSAA(void);
	virtual void disableFSAA(void);
	virtual void applyGeneralChanges(void);
	virtual void applyGeometryChanges(void);
	virtual void applyEffectsChanges(void);
	virtual void applyPrimitivesChanges(void);
	virtual void clear(void);
	virtual void setupStereo(void);
	virtual void enableStereo(void);
	virtual void disableStereo(void);
	virtual void setupCutaway(void);
	virtual void enableCutaway(void);
	virtual void disableCutaway(void);
	virtual void setupLighting(void);
	virtual void enableLighting(void);
	virtual void disableLighting(void);
	virtual void enableSeams(void);
	virtual void disableSeams(void);
	virtual void enablePrimitives(void);
	virtual void disablePrimitives(void);
	virtual void enableTextureFiltering(void);
	virtual void disableTextureFiltering(void);
	virtual void setupOpacitySlider(void);

	void applySettings(void);
	void applyGeneralSettings(void);
	void applyGeometrySettings(void);
	void applyEffectsSettings(void);
	void applyPrimitivesSettings(void);
	void loadSettings(void);
	void loadDefaultGeneralSettings(void);
	void loadDefaultGeometrySettings(void);
	void loadDefaultEffectsSettings(void);
	void loadDefaultPrimitivesSettings(void);
	void loadGeneralSettings(void);
	void loadGeometrySettings(void);
	void loadEffectsSettings(void);
	void loadPrimitivesSettings(void);

	LDrawModelViewer* modelViewer;

	int fsaaMode;
	bool showsFPS;
	bool showErrors;
	int fullScreenRefresh;
	COLORREF backgroundColor;
	COLORREF defaultColor;
	int defaultColorNumber;

	int useSeams;
	int seamWidth;
	bool drawWireframe;
	bool useWireframeFog;
	long wireframeThickness;
	bool showsHighlightLines;
	bool usePolygonOffset;
	bool blackHighlights;
	long edgeThickness;

	bool useLighting;
	bool qualityLighting;
	bool subduedLighting;
	bool usesSpecular;
	bool oneLight;
	LDVStereoMode stereoMode;
	long stereoEyeSpacing;
	LDVCutawayMode cutawayMode;
	long cutawayAlpha;
	long cutawayThickness;
	bool sortTransparent;
	bool useStipple;
	bool usesFlatShading;

	bool allowPrimitiveSubstitution;
	bool textureStuds;
	int textureFilterType;
	long curveQuality;
	bool qualityStuds;
	bool drawConditionalHighlights;
	bool performSmoothing;
	bool lineSmoothing;

	COLORREF customColors[16];
	float zoomMax;
	bool showAllConditionalLines;
	bool showConditionalControlPoints;

	int generalPageNumber;
	int geometryPageNumber;
	int effectsPageNumber;
	int primitivesPageNumber;

	HWND hGeneralPage;
	HWND hFullScreenRateField;
	HWND hFSAA2xButton;
	HWND hFSAA2xEnhancedButton;
	HWND hFSAA4xButton;
	HWND hFSAA4xEnhancedButton;
	HBITMAP hBackgroundColorBitmap;
	HWND hBackgroundColorButton;
	HBITMAP hDefaultColorBitmap;
	HWND hDefaultColorButton;
	HDC hButtonColorDC;

	HWND hGeometryPage;
	HWND hSeamSpin;
	HWND hWireframeFogButton;
	HWND hWireframeThicknessLabel;
	HWND hWireframeThicknessSlider;
	HWND hConditionalHighlightsButton;
	HWND hQualityLinesButton;
	HWND hAlwaysBlackButton;
	HWND hEdgeThicknessLabel;
	HWND hEdgeThicknessSlider;

	HWND hEffectsPage;
	HWND hLightQualityButton;
	HWND hLightSubduedButton;
	HWND hLightSpecularButton;
	HWND hLightAlternateButton;
	HWND hHardwareStereoButton;
	HWND hCrossEyedStereoButton;
	HWND hParallelStereoButton;
	HWND hStereoSpacingSlider;
	HWND hStereoSpacingLabel;
	HWND hCutawayColorButton;
	HWND hCutawayMonochromButton;
	HWND hCutawayOpacityLabel;
	HWND hCutawayOpacitySlider;
	HWND hCutawayThicknessLabel;
	HWND hCutawayThicknessSlider;

	HWND hPrimitivesPage;
	HWND hTextureStudsButton;
	HWND hTextureNearestButton;
	HWND hTextureBilinearButton;
	HWND hTextureTrilinearButton;
	HWND hCurveQualityLabel;
	HWND hCurveQualitySlider;
};

#endif // __LDVPREFERENCES_H__
