#include <shlobj.h>
#include <shlwapi.h>
#include "LDViewWindow.h"
#include "LDVExtensionsSetup.h"
#include "LDViewPreferences.h"
#include "SSModelWindow.h"
#include "AppResources.h"
#include "UserDefaultsKeys.h"
#include <LDLib/LDrawModel.h>
#include <TCFoundation/TCUserDefaults.h>
#include <TCFoundation/mystring.h>
#include <TCFoundation/TCAutoreleasePool.h>
#include <TCFoundation/TCStringArray.h>
#include <CUI/CUIWindowResizer.h>
#include <afxres.h>

#include "ModelWindow.h"
#include <LDLib/ModelMacros.h>

TCStringArray* LDViewWindow::recentFiles = NULL;

LDViewWindow::LDViewWindowCleanup LDViewWindow::ldViewWindowCleanup;

void debugOut(char *fmt, ...);

LDViewWindow::LDViewWindowCleanup::~LDViewWindowCleanup(void)
{
	if (LDViewWindow::recentFiles)
	{
		LDViewWindow::recentFiles->release();
		LDViewWindow::recentFiles = NULL;
	}
}

LDViewWindow::LDViewWindow(char* windowTitle, HINSTANCE hInstance, int x,
						   int y, int width, int height)
			  :CUIWindow(windowTitle, hInstance, x, y, width, height),
			   modelWindow(NULL),
			   hAboutWindow(NULL),
			   hLDrawDirWindow(NULL),
			   hOpenGLInfoWindow(NULL),
			   userLDrawDir(NULL),
			   fullScreen(NO),
			   fullScreenActive(NO),
			   switchingModes(NO),
			   videoModes(NULL),
			   numVideoModes(0),
			   currentVideoModeIndex(-1),
			   fsWidth(TCUserDefaults::longForKey(FULLSCREEN_WIDTH_KEY, 640)),
			   fsHeight(TCUserDefaults::longForKey(FULLSCREEN_HEIGHT_KEY, 480)),
			   fsDepth(TCUserDefaults::longForKey(FULLSCREEN_DEPTH_KEY, 32)),
			   showStatusBar(TCUserDefaults::longForKey(STATUS_BAR_KEY, 1) != 0),
			   topmost(TCUserDefaults::longForKey(TOPMOST_KEY) != 0),
			   showStatusBarOverride(false),
			   skipMinimize(NO),
			   screenSaver(NO),
			   originalMouseX(-999999),
			   originalMouseY(-999999),
			   hFileMenu(NULL),
			   hViewMenu(NULL),
			   loading(FALSE),
			   openGLInfoWindoResizer(NULL),
			   hOpenGLStatusBar(NULL)
//			   modelWindowShown(false)
{
//	DWORD backgroundColor = TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY);

	standardWindowStyle = windowStyle;
	if (!recentFiles)
	{
		recentFiles = new TCStringArray(10);
		populateRecentFiles();
	}
//	DeleteObject(hBackgroundBrush);
// 	hBackgroundBrush = CreateSolidBrush(RGB(backgroundColor & 0xFF,
//		(backgroundColor >> 8) & 0xFF, (backgroundColor >> 16) & 0xFF));
}

LDViewWindow::~LDViewWindow(void)
{
}

void LDViewWindow::dealloc(void)
{
	delete userLDrawDir;
	userLDrawDir = NULL;
	delete videoModes;
	videoModes = NULL;
	if (hOpenGLInfoWindow)
	{
		DestroyWindow(hOpenGLInfoWindow);
	}
	if (openGLInfoWindoResizer)
	{
		openGLInfoWindoResizer->release();
	}
	CUIWindow::dealloc();
}

HBRUSH LDViewWindow::getBackgroundBrush(void)
{
//	return CUIWindow::getBackgroundBrush();
	return NULL;
}

void LDViewWindow::showWindow(int nCmdShow)
{
	if (screenSaver)
	{
		x = 0;
		y = 0;
		width = GetSystemMetrics(SM_CXSCREEN);
		height = GetSystemMetrics(SM_CYSCREEN);
		CUIWindow::showWindow(SW_SHOW);
	}
	else
	{
		CUIWindow::showWindow(nCmdShow);
	}
	modelWindow->finalSetup();
}

void LDViewWindow::setScreenSaver(BOOL flag)
{
	screenSaver = flag;
	if (screenSaver)
	{
		windowStyle |= WS_CLIPCHILDREN;
	}
	else
	{
		windowStyle &= ~WS_CLIPCHILDREN;
	}
}

char* LDViewWindow::windowClassName(void)
{
	if (fullScreen || screenSaver)
	{
		return "LDViewFullScreenWindow";
	}
	else
	{
		return "LDViewWindow";
	}
}

LRESULT LDViewWindow::doEraseBackground(RECT* updateRect)
{
	BOOL noRect = FALSE;

	if (updateRect == NULL)
	{
		if (paintStruct)
		{
			return 0;
		}
		updateRect = new RECT;
		GetUpdateRect(hWindow, updateRect, FALSE);
		noRect = TRUE;
	}
	DWORD backgroundColor = TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY);
 	HBRUSH hBrush = CreateSolidBrush(RGB(backgroundColor & 0xFF,
		(backgroundColor >> 8) & 0xFF, (backgroundColor >> 16) & 0xFF));

//	printf("updateRect size: %d, %d\n", updateRect->right - updateRect->left,
//		updateRect->bottom - updateRect->top);
	if (!fullScreen)
	{
		updateRect->left += 2;
		updateRect->top += 2;
		updateRect->right -= 2;
		if (showStatusBar || showStatusBarOverride)
		{
			updateRect->bottom -= 22;
		}
		else
		{
			updateRect->bottom -= 2;
		}
	}
	FillRect(hdc, updateRect, hBrush);
	DeleteObject(hBrush);
	if (noRect)
	{
		delete updateRect;
		updateRect = NULL;
	}
	CUIWindow::doEraseBackground(updateRect);
	return 1;
}

void LDViewWindow::forceShowStatusBar(bool value)
{
	if (value != showStatusBarOverride)
	{
		showStatusBarOverride = value;
		if (!showStatusBar)
		{
			if (showStatusBarOverride)
			{
				addStatusBar();
			}
			else
			{
				removeStatusBar();
			}
		}
	}
}

void LDViewWindow::createStatusBar(void)
{
	if (showStatusBar || showStatusBarOverride)
	{
		int parts[] = {100, -1};
		HWND hProgressBar;

		ModelWindow::initCommonControls(ICC_TREEVIEW_CLASSES | ICC_BAR_CLASSES);
		hStatusBar = CreateStatusWindow(WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP,
			"", hWindow, 42000);
		SendMessage(hStatusBar, SB_SETPARTS, 2, (LPARAM)parts);
//		SendMessage(hStatusBar, SB_SETTEXT, 1, (LPARAM)"");
		SendMessage(hStatusBar, SB_SETTEXT, 0 | SBT_OWNERDRAW, (LPARAM)"");
		hProgressBar = CreateWindowEx(0, PROGRESS_CLASS, "",
			WS_CHILD | WS_VISIBLE | PBS_SMOOTH, 0, 2, 100, 18,
			hStatusBar, NULL, hInstance, NULL);
//		SendMessage(hProgressBar, PBM_SETPOS, 50, 0);
		if (modelWindow)
		{
			modelWindow->setStatusBar(hStatusBar);
			modelWindow->setProgressBar(hProgressBar);
		}
		RedrawWindow(hStatusBar, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
	}
}

BOOL LDViewWindow::initWindow(void)
{
	BOOL retValue;

	if (fullScreen || screenSaver)
	{
		windowStyle = WS_POPUP | WS_MAXIMIZE;
		if (screenSaver)
		{
			windowStyle |= WS_CLIPCHILDREN;
		}
#ifndef _DEBUG
		exWindowStyle |= WS_EX_TOPMOST;
#endif
	}
	else
	{
		windowStyle = standardWindowStyle;
		if (TCUserDefaults::longForKey(WINDOW_MAXIMIZED_KEY))
		{
			windowStyle |= WS_MAXIMIZE;
		}
		if (topmost)
		{
			exWindowStyle |= WS_EX_TOPMOST;
		}
		else
		{
			exWindowStyle &= ~WS_EX_TOPMOST;
		}
	}
	retValue = CUIWindow::initWindow();
	hFileMenu = GetSubMenu(GetMenu(hWindow), 0);
	hViewMenu = GetSubMenu(GetMenu(hWindow), 2);
	switch (TCUserDefaults::longForKey(VIEW_MODE_KEY))
	{
	case LDVViewExamine:
		switchToExamineMode();
		break;
	case LDVViewFlythrough:
		switchToFlythroughMode();
		break;
	default:
		switchToExamineMode();
		break;
	}
	populateRecentFileMenuItems();
	updateModelMenuItems();
	if (!fullScreen)
	{
		int heightDelta = 0;

		setMenuCheck(hViewMenu, ID_VIEW_ALWAYSONTOP, topmost);
		if (showStatusBar || showStatusBarOverride)
		{
			heightDelta = 20;
			setMenuCheck(hViewMenu, ID_VIEW_STATUSBAR, true);
		}
		else
		{
			setMenuCheck(hViewMenu, ID_VIEW_STATUSBAR, false);
		}
/*
		hFrameWindow = CreateWindowEx(WS_EX_CLIENTEDGE, "STATIC",
			"LDView Frame", WS_CHILD | WS_CLIPSIBLINGS | SS_OWNERDRAW, 0, 0,
			width, height - heightDelta, hWindow, NULL, hInstance, NULL);
//		ShowWindow(hFrameWindow, SW_SHOW);
		ShowWindow(modelWindow->getHWindow(), SW_SHOW);
		SetWindowPos(hFrameWindow, HWND_BOTTOM, 0, 0, width,
			height - heightDelta, SWP_SHOWWINDOW);
//		SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 0, 0, width,
//			height - heightDelta, SWP_SHOWWINDOW);
*/
		if (showStatusBar || showStatusBarOverride)
		{
			addStatusBar();
		}
//		createStatusBar();
	}
	return retValue;
}

BOOL LDViewWindow::showAboutBox(void)
{
	if (!hAboutWindow)
	{
		createAboutBox();
	}
	if (hAboutWindow)
	{
		runDialogModal(hAboutWindow);
		return TRUE;
	}
	return FALSE;
}

void LDViewWindow::createAboutBox(void)
{
	hAboutWindow = createDialog(IDD_ABOUT_BOX);
}

BOOL LDViewWindow::doLDrawDirOK(HWND hDlg)
{
	int length = SendDlgItemMessage(hDlg, IDC_LDRAWDIR, WM_GETTEXTLENGTH, 0, 0);

	if (length)
	{
		delete userLDrawDir;
		userLDrawDir = new char[length + 1];
		SendDlgItemMessage(hDlg, IDC_LDRAWDIR, WM_GETTEXT, (WPARAM)(length + 1),
			(LPARAM)userLDrawDir);
		doDialogClose(hDlg);
	}
	else
	{
		MessageBeep(MB_ICONEXCLAMATION);
	}
	return TRUE;
}

LRESULT LDViewWindow::doMouseWheel(short keyFlags, short zDelta, int /*xPos*/,
								  int /*yPos*/)
{
	if (modelWindow)
	{
		if (keyFlags & MK_CONTROL)
		{
			modelWindow->setClipZoom(YES);
		}
		else
		{
			modelWindow->setClipZoom(NO);
		}
		modelWindow->zoom((float)zDelta * -0.5f);
		return 0;
	}
	return 1;
}


UINT CALLBACK lDrawDirBrowseHook(HWND /*hDlg*/, UINT message, WPARAM /*wParam*/,
								 LPARAM lParam)
{
#ifdef _DEBUG
	_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "hook message: 0x%X\n", message);
#endif // _DEBUG
	if (message == WM_NOTIFY)
	{
		LPOFNOTIFY notification = (LPOFNOTIFY)lParam;

		switch (notification->hdr.code)
		{
			case CDN_FILEOK:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "OK Pressed\n");
#endif // _DEBUG
				break;
			case CDN_FOLDERCHANGE:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL,
					"Folder Change: 0x%X\n", notification->hdr.hwndFrom);
#endif // _DEBUG
				break;
			case CDN_HELP:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "Help Pressed\n");
#endif // _DEBUG
				break;
			case CDN_INITDONE:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "Init done\n");
#endif // _DEBUG
				break;
			case CDN_SELCHANGE:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL,
					"Selection Change: 0x%X\n", notification->hdr.hwndFrom);
#endif // _DEBUG
				break;
			case CDN_SHAREVIOLATION:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "Share violation\n");
#endif // _DEBUG
				break;
			case CDN_TYPECHANGE:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "Type change\n");
#endif // _DEBUG
				break;
			default:
#ifdef _DEBUG
				_CrtDbgReport(_CRT_WARN, NULL, 0, NULL,
					"Unknown notification: 0x%X\n", notification->hdr.code);
#endif // _DEBUG
				break;
		}
	}
	return 0;
}

BOOL LDViewWindow::tryVideoMode(VideoModeT* videoMode, int refreshRate)
{
	if (videoMode)
	{
		DEVMODE deviceMode;
		long result;

		fullScreenActive = YES;

		memset(&deviceMode, 0, sizeof DEVMODE);
		deviceMode.dmSize = sizeof DEVMODE;
		deviceMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		deviceMode.dmPelsWidth = videoMode->width;
		deviceMode.dmPelsHeight = videoMode->height;
		deviceMode.dmBitsPerPel = videoMode->depth;
		if (refreshRate)
		{
			deviceMode.dmFields |= DM_DISPLAYFREQUENCY;
			deviceMode.dmDisplayFrequency = refreshRate;
		}
		result = ChangeDisplaySettings(&deviceMode, CDS_FULLSCREEN);
		switch (result)
		{
			case DISP_CHANGE_SUCCESSFUL:
				return YES;
				break;
			case DISP_CHANGE_RESTART:
				fullScreenActive = NO;
				return NO;
				break;
			case DISP_CHANGE_BADFLAGS:
				fullScreenActive = NO;
				return NO;
				break;
			case DISP_CHANGE_BADPARAM:
				fullScreenActive = NO;
				return NO;
				break;
			case DISP_CHANGE_FAILED:
				fullScreenActive = NO;
				return NO;
				break;
			case DISP_CHANGE_BADMODE:
				fullScreenActive = NO;
				return NO;
				break;
			case DISP_CHANGE_NOTUPDATED:
				fullScreenActive = NO;
				return NO;
				break;
		}
	}
	return NO;
}

void LDViewWindow::setFullScreenDisplayMode(void)
{
	VideoModeT* videoMode = getCurrentVideoMode();

	if (!fullScreenActive && videoMode)
	{
		int refreshRate = TCUserDefaults::longForKey(FULLSCREEN_REFRESH_KEY);

		if (!tryVideoMode(videoMode, refreshRate) && refreshRate)
		{
			tryVideoMode(videoMode, 0);
		}
	}
}

void LDViewWindow::restoreDisplayMode(void)
{
	if (fullScreenActive)
	{
		long result = ChangeDisplaySettings(NULL, 0);

		switch (result)
		{
			case DISP_CHANGE_SUCCESSFUL:
				break;
			case DISP_CHANGE_RESTART:
				break;
			case DISP_CHANGE_BADFLAGS:
				break;
			case DISP_CHANGE_BADPARAM:
				break;
			case DISP_CHANGE_FAILED:
				break;
			case DISP_CHANGE_BADMODE:
				break;
			case DISP_CHANGE_NOTUPDATED:
				break;
		}
		fullScreenActive = NO;
	}
}

void LDViewWindow::setModelWindow(ModelWindow *value)
{
	value->retain();
	if (modelWindow)
	{
		modelWindow->release();
	}
	modelWindow = value;
}

void LDViewWindow::activateFullScreenMode(void)
{
	if (!fullScreenActive)
	{
		skipMinimize = YES;
		if (!getCurrentVideoMode())
		{
			MessageBeep(MB_ICONEXCLAMATION);
			return;
		}
//		modelWindow->uncompile();
		switchingModes = YES;
		if (modelWindow)
		{
			modelWindow->closeWindow();
//			modelWindowShown = false;
		}
		DestroyWindow(hWindow);
		switchingModes = NO;
		setFullScreenDisplayMode();
		if (initWindow())
		{
			if (modelWindow->initWindow())
			{
//				modelWindow->showWindow(SW_SHOW);
//				modelWindowShown = true;
				showWindow(SW_SHOW);
				modelWindow->setNeedsRecompile();
				modelWindow->forceRedraw(1);
			}
		}
		skipMinimize = NO;
	}
}

void LDViewWindow::deactivateFullScreenMode(void)
{
	modelWindow->uncompile();
	restoreDisplayMode();
	if (skipMinimize)
	{
//		modelWindow->forceRedraw(1);
	}
	else
	{
		ShowWindow(hWindow, SW_MINIMIZE);
		modelWindow->closeWindow();
//		modelWindowShown = false;
	}
}

LRESULT LDViewWindow::doActivate(int activateFlag, BOOL /*minimized*/,
								  HWND /*previousHWindow*/)
{
	if (modelWindow)
	{
		if (!modelWindowIsShown())
		{
//			modelWindowIsShown();
			modelWindow->showWindow(SW_NORMAL);
//			modelWindowShown = true;
		}
		if (activateFlag == WA_ACTIVE || activateFlag == WA_CLICKACTIVE)
		{
			modelWindow->startPolling();
		}
		else
		{
			if (modelWindow->getPollSetting() != POLL_BACKGROUND)
			{
				modelWindow->stopPolling();
			}
		}
	}
	return 0;
}

LRESULT LDViewWindow::doActivateApp(BOOL activateFlag, DWORD /*threadId*/)
{
	VideoModeT* videoMode = getCurrentVideoMode();

	if (fullScreen && videoMode)
	{
		if (activateFlag)
		{
			activateFullScreenMode();
		}
		else
		{
			deactivateFullScreenMode();
		}
		return 0;
	}
	else
	{
		return 1;
	}
}

BOOL LDViewWindow::doLDrawDirBrowse(HWND hDlg)
{
	OPENFILENAME openStruct;
	char fileTypes[1024];
	char filename[1024] = "";
	char initialDir[1024];

	SendDlgItemMessage(hDlg, IDC_LDRAWDIR, WM_GETTEXT, (WPARAM)(1024),
		(LPARAM)initialDir);
	memset(fileTypes, 0, 2);
	addFileType(fileTypes, "All Files (*.*)", "*.*");
	memset(&openStruct, 0, sizeof(OPENFILENAME));
	openStruct.lStructSize = sizeof(OPENFILENAME);
	openStruct.lpstrFilter = fileTypes;
	openStruct.nFilterIndex = 1;
	openStruct.lpstrFile = filename;
	openStruct.nMaxFile = 1024;
	openStruct.lpstrInitialDir = initialDir;
	openStruct.lpstrTitle = "Select the LDraw directory";
	openStruct.lpfnHook = lDrawDirBrowseHook;
	openStruct.Flags = OFN_EXPLORER | OFN_NOTESTFILECREATE | OFN_HIDEREADONLY |
		OFN_ALLOWMULTISELECT | OFN_ENABLEHOOK |
		OFN_ENABLESIZING;
	if (GetSaveFileName(&openStruct))
	{
		SendDlgItemMessage(hDlg, IDC_LDRAWDIR, WM_SETTEXT, 0,
			(LPARAM)filename);
	}
	return TRUE;
}

BOOL LDViewWindow::doDialogCommand(HWND hDlg, int controlId, int notifyCode,
									HWND /*controlHWnd*/)
{
	if (notifyCode == BN_CLICKED)
	{
		switch (controlId)
		{
			case IDOK:
				if (hDlg == hLDrawDirWindow)
				{
					return doLDrawDirOK(hDlg);
				}
				else
				{
					doDialogClose(hDlg);
//					return FALSE;
				}
				break;
			case IDCANCEL:
				doDialogClose(hDlg);
				break;
//			case IDC_LDRAWDIR_BROWSE:
//				return doLDrawDirBrowse(hDlg);
//				break;
			default:
				return FALSE;
				break;
		}
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

BOOL LDViewWindow::checkMouseMove(HWND hWnd, LPARAM lParam)
{
	POINT mousePoint = { LOWORD(lParam), HIWORD(lParam)};

	ClientToScreen(hWnd, &mousePoint);
	if (originalMouseX == -999999)
	{
		originalMouseX = mousePoint.x;
		originalMouseY = mousePoint.y;
	}
	if (originalMouseX == mousePoint.x && originalMouseY == mousePoint.y)
	{
		return FALSE;
	}
	return TRUE;
}

void LDViewWindow::initMouseMove(void)
{
	originalMouseX = -999999;
}

LRESULT LDViewWindow::doEnterMenuLoop(bool /*isTrackPopupMenu*/)
{
	if (loading)
	{
		return 0;
	}
	if (modelWindow)
	{
		LDrawModelViewer *modelViewer = modelWindow->getModelViewer();

		if (modelViewer)
		{
			modelViewer->pause();
		}
	}
	return 1;
}

LRESULT LDViewWindow::doExitMenuLoop(bool /*isTrackPopupMenu*/)
{
	if (modelWindow)
	{
		LDrawModelViewer *modelViewer = modelWindow->getModelViewer();

		if (modelViewer)
		{
			modelViewer->unpause();
		}
	}
	return 1;
}

LRESULT LDViewWindow::windowProc(HWND hWnd, UINT message, WPARAM wParam,
								 LPARAM lParam)
{
	if (screenSaver && modelWindow &&
		((SSModelWindow*)modelWindow)->checkForExit(hWnd, message, wParam,
		lParam))
	{
		shutdown();
		return 1;
	}
	return CUIWindow::windowProc(hWnd, message, wParam, lParam);
}

void LDViewWindow::switchModes(void)
{
	skipMinimize = YES;
	if (!getCurrentVideoMode())
	{
		MessageBeep(MB_ICONEXCLAMATION);
		return;
	}
	modelWindow->uncompile();
	switchingModes = YES;
	if (modelWindow)
	{
		modelWindow->closeWindow();
//		modelWindowShown = false;
	}
	DestroyWindow(hWindow);
//	modelWindow->forceRedraw(3);
	switchingModes = NO;
	fullScreen = !fullScreen;
	if (fullScreen)
	{
//		printf("switching to fullscreen...\n");
		setFullScreenDisplayMode();
//		printf("done.\n");
	}
	else
	{
		int newWidth = TCUserDefaults::longForKey(WINDOW_WIDTH_KEY);
		int newHeight = TCUserDefaults::longForKey(WINDOW_HEIGHT_KEY);
		int dWidth = newWidth - width;
		int dHeight = newHeight - height;

//		printf("switching from fullscreen...\n");
		restoreDisplayMode();
//		printf("done\n");
		width = newWidth;
		height = newHeight;
		newWidth = modelWindow->getWidth() + dWidth;
		newHeight = modelWindow->getHeight() + dHeight;
		modelWindow->resize(newWidth, newHeight);
	}
	if (initWindow())
	{
		if (modelWindow->initWindow())
		{
			if (!fullScreen)
			{
				if (showStatusBar || showStatusBarOverride)
				{
					SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 0, 0,
						width, height - 20, 0);
/*
					SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 2, 2,
						width - 4, height - 24, 0);
*/
				}
				else
				{
					SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 0, 0,
						width, height, 0);
/*
					SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 2, 2,
						width - 4, height - 4, 0);
*/
				}
			}
			else
			{
				SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 0, 0,
					width, height, 0);
			}
//			modelWindow->showWindow(SW_SHOW);
//			modelWindowShown = true;
			showWindow(SW_SHOW);
			if (!fullScreen)
			{
//				printf("normal.\n");
			}
			modelWindow->setNeedsRecompile();
//			modelWindow->forceRedraw(1);
			skipMinimize = NO;
		}
		else
		{
			skipMinimize = NO;
			if (fullScreen)
			{
				switchModes();
			}
			else
			{
				MessageBox(hWindow, "Error switching back.  Aborting.", "Error",
					MB_OK);
				shutdown();
			}
		}
	}
	modelWindow->forceRedraw();
//	modelWindow->forceRedraw(3);
}

void LDViewWindow::shutdown(void)
{
	skipMinimize = YES;
	if (modelWindow)
	{
		ModelWindow *tmpModelWindow = modelWindow;

		modelWindow = NULL;
		tmpModelWindow->closeWindow();
		tmpModelWindow->release();
//		modelWindowShown = false;
	}
	DestroyWindow(hWindow);
}

LRESULT LDViewWindow::doChar(TCHAR characterCode, LPARAM /*keyData*/)
{
	switch (characterCode)
	{
		case 10: // Ctrl+Enter (Linefeed)
			switchModes();
			return 0;
			break;
		case 27: // Escape
			if (fullScreen)
			{
				switchModes();
//				shutdown();
				return 0;
			}
			break;
	}
	return 1;
}

void LDViewWindow::selectFSVideoModeMenuItem(int index)
{
	VideoModeT* videoMode = getCurrentVideoMode();
	HMENU bitDepthMenu;
	MENUITEMINFO itemInfo;
	int menuIndex;
	HMENU viewMenu = GetSubMenu(GetMenu(hWindow), 2);

	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_STATE;
	if (videoMode)
	{
		bitDepthMenu = menuForBitDepth(hWindow, videoMode->depth, &menuIndex);
		itemInfo.fState = MFS_UNCHECKED;
		SetMenuItemInfo(bitDepthMenu, 30000 + currentVideoModeIndex, FALSE,
			&itemInfo);
		SetMenuItemInfo(viewMenu, menuIndex, TRUE, &itemInfo);
	}
	currentVideoModeIndex = index;
	videoMode = getCurrentVideoMode();
	fsWidth = videoMode->width;
	fsHeight = videoMode->height;
	fsDepth = videoMode->depth;
	TCUserDefaults::setLongForKey(fsWidth, FULLSCREEN_WIDTH_KEY);
	TCUserDefaults::setLongForKey(fsHeight, FULLSCREEN_HEIGHT_KEY);
	TCUserDefaults::setLongForKey(fsDepth, FULLSCREEN_DEPTH_KEY);
	if (videoMode)
	{
		bitDepthMenu = menuForBitDepth(hWindow, videoMode->depth, &menuIndex);
		itemInfo.fState = MFS_CHECKED;
		SetMenuItemInfo(bitDepthMenu, 30000 + currentVideoModeIndex, FALSE,
			&itemInfo);
		SetMenuItemInfo(viewMenu, menuIndex, TRUE, &itemInfo);
	}
}

void LDViewWindow::initPollingMenu(void)
{
	int pollSetting = TCUserDefaults::longForKey(POLL_KEY);
	HMENU pollingMenu = getPollingMenu();
	int i;
	int count = GetMenuItemCount(pollingMenu);
	MENUITEMINFO itemInfo;
	char title[256];

	memset(&itemInfo, 0, sizeof(MENUITEMINFO));
	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_TYPE | MIIM_STATE;
	itemInfo.dwTypeData = title;
	for (i = 0; i < count; i++)
	{
		itemInfo.cch = 256;
		GetMenuItemInfo(pollingMenu, i, TRUE, &itemInfo);
		itemInfo.fType = MFT_STRING | MFT_RADIOCHECK;
		if (i == pollSetting)
		{
			itemInfo.fState = MFS_CHECKED;
		}
		else
		{
			itemInfo.fState = MFS_UNCHECKED;
		}
		SetMenuItemInfo(pollingMenu, i, TRUE, &itemInfo);
	}
}

void LDViewWindow::selectPollingMenuItem(int index)
{
	int pollSetting = TCUserDefaults::longForKey(POLL_KEY);
	HMENU pollingMenu = getPollingMenu();
	MENUITEMINFO itemInfo;

	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_STATE;
	itemInfo.fState = MFS_UNCHECKED;
	SetMenuItemInfo(pollingMenu, ID_FILE_POLLING_DISABLED + pollSetting, FALSE,
		&itemInfo);
	itemInfo.fState = MFS_CHECKED;
	SetMenuItemInfo(pollingMenu, index, FALSE, &itemInfo);
	TCUserDefaults::setLongForKey(index - ID_FILE_POLLING_DISABLED, POLL_KEY);
	if (modelWindow)
	{
		modelWindow->setPollSetting(index - ID_FILE_POLLING_DISABLED);
	}
}

HMENU LDViewWindow::getPollingMenu(void)
{
	HMENU fileMenu = GetSubMenu(GetMenu(hWindow), 0);
	int i;
	int count = GetMenuItemCount(fileMenu);
	HMENU pollingMenu = 0;
	MENUITEMINFO itemInfo;

	memset(&itemInfo, 0, sizeof(MENUITEMINFO));
	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_SUBMENU | MIIM_ID;
	for (i = 0; i < count && !pollingMenu; i++)
	{
		GetMenuItemInfo(fileMenu, i, TRUE, &itemInfo);
		if (itemInfo.hSubMenu)
		{
			pollingMenu = itemInfo.hSubMenu;
//			pollingMenu = GetSubMenu(viewMenu, i);
			GetMenuItemInfo(pollingMenu, 0, TRUE, &itemInfo);
			if (itemInfo.wID != ID_FILE_POLLING_DISABLED)
			{
				pollingMenu = 0;
			}
		}
	}
	return pollingMenu;
}

void LDViewWindow::showHelp(void)
{
	HINSTANCE executeHandle;
	char* helpPath = LDViewPreferences::getLDViewPath("Help.html", true);

	setWaitCursor();
	executeHandle = ShellExecute(hWindow, NULL, helpPath, NULL, ".",
		SW_SHOWNORMAL);
	delete helpPath;
	setArrowCursor();
	if ((int)executeHandle <= 32)
	{
		char* errorString;

		switch ((int)executeHandle)
		{
			case 0:
			case SE_ERR_OOM:
				errorString = "Not enough memory.";
				break;
			case ERROR_FILE_NOT_FOUND:
				errorString = "Help.html not found.";
				break;
			case ERROR_PATH_NOT_FOUND:
				errorString = "Couldn't determine path to Help.html.";
				break;
			case SE_ERR_ACCESSDENIED:
				errorString = "Access denied opening Help.html.";
				break;
			case SE_ERR_SHARE:
				errorString = "Sharing violation opening Help.html.";
				break;
			default:
				errorString = "Error opening Help.html";
				break;
		}
		MessageBox(hWindow, errorString, "Error", MB_OK | MB_ICONEXCLAMATION);
	}
}

void LDViewWindow::openRecentFile(int index)
{
	char* filename = recentFiles->stringAtIndex(index);

	if (filename)
	{
		openModel(filename);
	}
}

void LDViewWindow::setMenuEnabled(HMENU hParentMenu, int itemID, bool enabled,
								  BOOL byPosition)
{
	MENUITEMINFO itemInfo;

	itemInfo.cbSize = sizeof(itemInfo);
	itemInfo.fMask = MIIM_STATE;
	GetMenuItemInfo(hParentMenu, itemID, byPosition, &itemInfo);
	if (enabled)
	{
		itemInfo.fState &= ~MFS_DISABLED;
	}
	else
	{
		itemInfo.fState |= MFS_DISABLED;
	}
	itemInfo.fMask = MIIM_STATE;
	SetMenuItemInfo(hParentMenu, itemID, byPosition, &itemInfo);
}

void LDViewWindow::updateModelMenuItems(void)
{
	bool haveModel = modelIsLoaded();

	setMenuEnabled(hFileMenu, ID_FILE_SAVE, haveModel);
	setMenuEnabled(hFileMenu, ID_FILE_PRINT, haveModel);
	setMenuEnabled(hFileMenu, ID_FILE_PAGESETUP, haveModel);
}

bool LDViewWindow::modelIsLoaded(void)
{
	if (modelWindow)
	{
		LDrawModelViewer* modelViewer = modelWindow->getModelViewer();

		if (modelViewer && modelViewer->getLDrawModel())
		{
			return true;
		}
	}
	return false;
}

LRESULT LDViewWindow::doMenuSelect(UINT menuID, UINT /*flags*/, HMENU hMenu)
{
//	printf("LDViewWindow::doMenuSelect(%d, 0x%04X, 0x%04X)\n", menuID, flags,
//		hMenu);
	if (hMenu == GetMenu(hWindow) && menuID == 0)
	{
		// This shouldn't ever be necessary, but it can't hurt.
		updateModelMenuItems();
	}
	return 1;
}

void LDViewWindow::showLDrawCommandLine(void)
{
	if (modelWindow)
	{
		LDrawModelViewer* modelViewer = modelWindow->getModelViewer();

		if (modelViewer)
		{
			HGLOBAL hBuf = GlobalAlloc(GMEM_DDESHARE, 1024);
			char *buf = (char*)GlobalLock(hBuf);
//			char shortFilename[1024];

//			GetShortPathName(modelViewer->getFilename(), shortFilename, 1024);
//			modelViewer->getLDrawCommandLine(shortFilename, buf, 1024);
			modelViewer->getLDGLiteCommandLine(buf, 1024);
			MessageBox(hWindow, buf, "LDraw Command Line", MB_OK);
//			printf("%s\n", buf);
			GlobalUnlock(hBuf);
			if (OpenClipboard(hWindow))
			{
				EmptyClipboard();
				SetClipboardData(CF_TEXT, hBuf);
				CloseClipboard();
//				GlobalFree(hBuf);
			}
		}
	}
}

BOOL LDViewWindow::doDialogSize(HWND hDlg, WPARAM sizeType, int newWidth,
							   int newHeight)
{
//	printf("LDViewWindow::doDialogSize(%d, %d, %d)\n", sizeType, newWidth,
//		newHeight);
	if (hDlg == hOpenGLInfoWindow)
	{
		SendMessage(hOpenGLStatusBar, WM_SIZE, sizeType,
			MAKELPARAM(newWidth, newHeight));
		openGLInfoWindoResizer->resize(newWidth, newHeight);
	}
	return FALSE;
}

LRESULT LDViewWindow::showOpenGLDriverInfo(void)
{
	if (!hOpenGLInfoWindow)
	{
		char *vendorString = (char*)glGetString(GL_VENDOR);
		char *rendererString = (char*)glGetString(GL_RENDERER);
		char *versionString = (char*)glGetString(GL_VERSION);
		char *extensionsString = (char*)glGetString(GL_EXTENSIONS);
		char *wglExtensionsString = LDVExtensionsSetup::getWglExtensions();
		int len;
		char *message;
		int parts[2] = {100, -1};
		char buf[128];

		if (!vendorString)
		{
			vendorString = "*Unknown*";
		}
		if (!rendererString)
		{
			rendererString = "*Unknown*";
		}
		if (!versionString)
		{
			versionString = "*Unknown*";
		}
		if (!extensionsString)
		{
			extensionsString = "*None*";
		}
		if (!wglExtensionsString)
		{
			wglExtensionsString = "*None*";
		}
		len = strlen(extensionsString);
		if (len && (extensionsString[len - 1] == ' '))
		{
			extensionsString[len - 1] = 0;
		}
		len = strlen(wglExtensionsString);
		if (len && (wglExtensionsString[len - 1] == ' '))
		{
			wglExtensionsString[len - 1] = 0;
		}
		extensionsString = stringByReplacingSubstring(extensionsString, " ",
			"\r\n");
		wglExtensionsString = stringByReplacingSubstring(wglExtensionsString,
			" ", "\r\n");
		len = strlen(vendorString) + strlen(rendererString) +
			strlen(versionString) + strlen(extensionsString) +
			strlen(wglExtensionsString) + 128;
		message = new char[len];
		sprintf(message, "Vendor: %s\r\n"
			"Renderer: %s\r\n"
			"Version: %s\r\n\r\n"
			"Extenstions:\r\n%s\r\n\r\n"
			"WGL Extensions:\r\n%s",
			vendorString, rendererString, versionString, extensionsString,
			wglExtensionsString);
		hOpenGLInfoWindow = createDialog(IDD_OPENGL_INFO);
		SendDlgItemMessage(hOpenGLInfoWindow, IDC_OPENGL_INFO, WM_SETTEXT, 0,
			(LPARAM)message);
		hOpenGLStatusBar = CreateStatusWindow(WS_CHILD | WS_VISIBLE |
			SBARS_SIZEGRIP, "", hOpenGLInfoWindow, 42001);
		SendMessage(hOpenGLStatusBar, SB_SETPARTS, 2, (LPARAM)parts);
		sprintf(buf, "%d Extension", countStringLines(extensionsString));
		SendMessage(hOpenGLStatusBar, SB_SETTEXT, 0, (LPARAM)buf);
		sprintf(buf, "%d WGL Extension", countStringLines(wglExtensionsString));
		SendMessage(hOpenGLStatusBar, SB_SETTEXT, 1, (LPARAM)buf);
		calcSystemSizes();
		if (openGLInfoWindoResizer)
		{
			openGLInfoWindoResizer->release();
		}
		openGLInfoWindoResizer = new CUIWindowResizer;
		openGLInfoWindoResizer->setHWindow(hOpenGLInfoWindow);
		openGLInfoWindoResizer->addSubWindow(IDC_OPENGL_INFO,
			CUISizeHorizontal | CUISizeVertical);
		openGLInfoWindoResizer->addSubWindow(IDOK, CUIFloatLeft | CUIFloatTop);
		delete message;
		delete extensionsString;
		delete wglExtensionsString;
	}
	ShowWindow(hOpenGLInfoWindow, SW_SHOW);
	return 0;
}
/*
{
	int pollSetting = TCUserDefaults::longForKey(POLL_KEY);
	HMENU pollingMenu = getPollingMenu();
	int i;
	int count = GetMenuItemCount(pollingMenu);
	MENUITEMINFO itemInfo;
	char title[256];

	memset(&itemInfo, 0, sizeof(MENUITEMINFO));
	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_TYPE | MIIM_STATE;
	itemInfo.dwTypeData = title;
	for (i = 0; i < count; i++)
	{
		itemInfo.cch = 256;
		GetMenuItemInfo(pollingMenu, i, TRUE, &itemInfo);
		itemInfo.fType = MFT_STRING | MFT_RADIOCHECK;
		if (i == pollSetting)
		{
			itemInfo.fState = MFS_CHECKED;
		}
		else
		{
			itemInfo.fState = MFS_UNCHECKED;
		}
		SetMenuItemInfo(pollingMenu, i, TRUE, &itemInfo);
	}
}
*/

void LDViewWindow::setMenuRadioCheck(HMENU hParentMenu, UINT uItem, bool checked)
{
	setMenuCheck(hParentMenu, uItem, checked, true);
}

void LDViewWindow::setMenuCheck(HMENU hParentMenu, UINT uItem, bool checked,
								bool radio)
{
	MENUITEMINFO itemInfo;
	char title[256];

	memset(&itemInfo, 0, sizeof(MENUITEMINFO));
	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_STATE | MIIM_TYPE;
	itemInfo.dwTypeData = title;
	itemInfo.cch = 256;
	GetMenuItemInfo(hParentMenu, uItem, FALSE, &itemInfo);
	if (checked)
	{
		itemInfo.fState = MFS_CHECKED;
	}
	else
	{
		itemInfo.fState = MFS_UNCHECKED;
	}
	itemInfo.fType = MFT_STRING;
	if (radio)
	{
		itemInfo.fType |= MFT_RADIOCHECK;
	}
	SetMenuItemInfo(hParentMenu, uItem, FALSE, &itemInfo);
}

LRESULT LDViewWindow::switchToExamineMode(void)
{
	setMenuRadioCheck(hViewMenu, ID_VIEW_EXAMINE, true);
	setMenuRadioCheck(hViewMenu, ID_VIEW_FLYTHROUGH, false);
	modelWindow->setViewMode(LDVViewExamine);
	return 0;
}

LRESULT LDViewWindow::switchToFlythroughMode(void)
{
	setMenuRadioCheck(hViewMenu, ID_VIEW_EXAMINE, false);
	setMenuRadioCheck(hViewMenu, ID_VIEW_FLYTHROUGH, true);
	modelWindow->setViewMode(LDVViewFlythrough);
	return 0;
}

void LDViewWindow::chooseNewLDrawDir(void)
{
	char *oldDir = getLDrawDir();

	if (!verifyLDrawDir(true))
	{
		if (oldDir)
		{
			TCUserDefaults::setStringForKey(oldDir, LDRAWDIR_KEY);
			LDrawModel::setLDrawDir(oldDir);
		}
	}
	delete oldDir;
}

void LDViewWindow::resizeModelWindow(int newWidth, int newHeight)
{
	if (modelWindow)
	{
		SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 0, 0, newWidth,
			newHeight, 0);
/*
		SetWindowPos(modelWindow->getHWindow(), HWND_TOP, 2, 2, newWidth - 4,
			newHeight - 4, 0);
		SetWindowPos(hFrameWindow, HWND_BOTTOM, 0, 0, newWidth, newHeight,
			SWP_DRAWFRAME | SWP_NOZORDER);
//		MoveWindow(hFrameWindow, 0, 0, newWidth, newHeight, TRUE);
		RedrawWindow(hFrameWindow, NULL, NULL, RDW_FRAME | RDW_INVALIDATE |
			RDW_INTERNALPAINT | RDW_UPDATENOW);
*/
	}
}

void LDViewWindow::removeStatusBar(void)
{
	resizeModelWindow(width, height);
	if (modelWindow)
	{
		modelWindow->setStatusBar(NULL);
		modelWindow->setProgressBar(NULL);
	}
	DestroyWindow(hStatusBar);
	hStatusBar = NULL;
}

void LDViewWindow::addStatusBar(void)
{
	resizeModelWindow(width, height - 20);
	createStatusBar();
}

LRESULT LDViewWindow::switchStatusBar(void)
{
	showStatusBar = !showStatusBar;
	if (showStatusBar || showStatusBarOverride)
	{
		addStatusBar();
		setMenuCheck(hViewMenu, ID_VIEW_STATUSBAR, true);
	}
	else
	{
		removeStatusBar();
		setMenuCheck(hViewMenu, ID_VIEW_STATUSBAR, false);
	}
	TCUserDefaults::setLongForKey(showStatusBar ? 1 : 0, STATUS_BAR_KEY);
	return 0;
}

LRESULT LDViewWindow::switchTopmost(void)
{
	HWND hPlaceFlag = HWND_NOTOPMOST;
	RECT rect;

	topmost = !topmost;
	GetWindowRect(hWindow, &rect);
	if (topmost)
	{
		setMenuCheck(hViewMenu, ID_VIEW_ALWAYSONTOP, true);
		hPlaceFlag = HWND_TOPMOST;
	}
	else
	{
		setMenuCheck(hViewMenu, ID_VIEW_ALWAYSONTOP, false);
	}
	SetWindowPos(hWindow, hPlaceFlag, rect.left, rect.top,
		rect.right - rect.left, rect.bottom - rect.top, 0);
	TCUserDefaults::setLongForKey(topmost ? 1 : 0, TOPMOST_KEY);
	return 0;
}

LRESULT LDViewWindow::doCommand(int itemId, int notifyCode, HWND controlHWnd)
{
	char* message = NULL;

	if (modelWindow && controlHWnd == modelWindow->getHPrefsWindow())
	{
		SendMessage(modelWindow->getHWindow(), WM_COMMAND,
			MAKELONG(BN_CLICKED, notifyCode),
			(LPARAM)modelWindow->getHPrefsWindow());
	}
//	printf("LDViewWindow::doCommand(%d, %0x%08X, 0x%04X)\n", itemId,
//		notifyCode, controlHWnd);
	switch (itemId)
	{
		case ID_EDIT_PREFERENCES:
			modelWindow->showPreferences();
			return 0;
			break;
		case ID_FILE_OPEN:
			openModel();
			return 0;
			break;
		case ID_FILE_PRINT:
			printModel();
			return 0;
			break;
		case ID_FILE_PAGESETUP:
			pageSetup();
			return 0;
			break;
		case ID_FILE_SAVE:
			saveSnapshot();
			return 0;
			break;
		case ID_FILE_LDRAWDIR:
			chooseNewLDrawDir();
			return 0;
			break;
		case ID_FILE_CANCELLOAD:
			if (modelWindow)
			{
				modelWindow->setCancelLoad();
			}
			return 0;
			break;
		case ID_FILE_EXIT:
			shutdown();
			return 0;
			break;
		case ID_VIEW_FULLSCREEN:
			switchModes();
			return 0;
			break;
		case ID_VIEW_RESET:
			resetView();
			return 0;
			break;
		case ID_VIEW_ERRORS:
			modelWindow->showErrors();
			return 0;
			break;
		case ID_VIEW_STATUSBAR:
			return switchStatusBar();
			break;
		case ID_VIEW_ALWAYSONTOP:
			return switchTopmost();
			break;
		case ID_VIEW_EXAMINE:
			return switchToExamineMode();
			break;
		case ID_VIEW_FLYTHROUGH:
			return switchToFlythroughMode();
			break;
/*
		case ID_VIEW_LDRAWCOMMANDLINE:
			showLDrawCommandLine();
			return 0;
			break;
*/
		case ID_FILE_RELOAD:
			modelWindow->reload();
			modelWindow->update();
			return 0;
			break;
		case ID_HELP_ABOUT:
			showAboutBox();
			return 0;
		case ID_HELP_CONTENTS:
			if (!fullScreenActive)
			{
				showHelp();
				return 0;
			}
			break;
		case ID_HELP_OPENGL_INFO:
			return showOpenGLDriverInfo();
			break;
		case ID_HELP_OPENGLINFO_VENDOR:
			message = (char*)glGetString(GL_VENDOR);
			break;
		case ID_HELP_OPENGLINFO_RENDERER:
			message = (char*)glGetString(GL_RENDERER);
			break;
		case ID_HELP_OPENGLINFO_VERSION:
			message = (char*)glGetString(GL_VERSION);
			break;
		case ID_HELP_OPENGLINFO_EXTENSIONS:
			message = (char*)glGetString(GL_EXTENSIONS);
			break;
		case ID_HELP_OPENGLINFO_WGLEXTENSIONS:
			message = LDVExtensionsSetup::getWglExtensions();
			if (!message)
			{
				message = "None";
			}
			break;
	}
	if (itemId >= 30000 && itemId < 30000 + numVideoModes)
	{
		selectFSVideoModeMenuItem(itemId - 30000);
	}
	else if (itemId >= 31000 && itemId < 31100)
	{
		openRecentFile(itemId - 31000);
		return 0;
	}
	if (itemId >= ID_FILE_POLLING_DISABLED &&
		itemId <= ID_FILE_POLLING_BACKGROUND)
	{
		selectPollingMenuItem(itemId);
	}
	if (message)
	{
		float rotationSpeed = modelWindow->getRotationSpeed();
		float zoomSpeed = modelWindow->getZoomSpeed();

		modelWindow->setRotationSpeed(0.0f);
		modelWindow->setZoomSpeed(0.0f);
		MessageBox(NULL, message, "Open GL Info",
			MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		modelWindow->setRotationSpeed(rotationSpeed);
		modelWindow->setZoomSpeed(zoomSpeed);
		return 0;
	}
	return CUIWindow::doCommand(itemId, notifyCode, controlHWnd);
}

WNDCLASSEX LDViewWindow::getWindowClass(void)
{
	WNDCLASSEX windowClass = CUIWindow::getWindowClass();

	windowClass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APP_ICON));
	if (fullScreen || screenSaver)
	{
		windowClass.lpszMenuName = NULL;
	}
	else
	{
		windowClass.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN_MENU);
	}
	return windowClass;
}

LRESULT LDViewWindow::doSize(WPARAM sizeType, int newWidth, int newHeight)
{
	if (!fullScreen && !screenSaver)
	{
		int heightDelta = 0;

		if (showStatusBar || showStatusBarOverride)
		{
			heightDelta = 20;
		}
		if (sizeType == SIZE_MAXIMIZED)
		{
			TCUserDefaults::setLongForKey(1, WINDOW_MAXIMIZED_KEY);
		}
		else
		{
			TCUserDefaults::setLongForKey(0, WINDOW_MAXIMIZED_KEY);
			if (sizeType == SIZE_RESTORED)
			{
				TCUserDefaults::setLongForKey(newWidth, WINDOW_WIDTH_KEY);
				TCUserDefaults::setLongForKey(newHeight - heightDelta,
					WINDOW_HEIGHT_KEY);
/*
				TCUserDefaults::setLongForKey(newWidth - 4, WINDOW_WIDTH_KEY);
				TCUserDefaults::setLongForKey(newHeight - 4 - heightDelta,
					WINDOW_HEIGHT_KEY);
*/
			}
		}
/*
		SetWindowPos(hFrameWindow, HWND_BOTTOM, 0, 0, newWidth,
			newHeight - heightDelta,
			SWP_DRAWFRAME | SWP_NOZORDER | SWP_DEFERERASE);
*/
		SendMessage(hStatusBar, WM_SIZE, SIZE_RESTORED,
			MAKELPARAM(newWidth, newHeight));
	}
	return CUIWindow::doSize(sizeType, newWidth, newHeight);
}

/*
int LDViewWindow::getDecorationHeight(void)
{
	int menuHeight = GetSystemMetrics(SM_CYMENU);

	return CUIWindow::getDecorationHeight() + menuHeight;
}
*/

LRESULT LDViewWindow::doClose(void)
{
#ifdef _DEBUG
//	_CrtDbgReport(_CRT_WARN, NULL, 0, NULL, "doClose\n");
#endif // _DEBUG
	skipMinimize = YES;
	if (modelWindow)
	{
		modelWindow->closeWindow();
//		modelWindowShown = false;
	}
	return CUIWindow::doClose();
}

LRESULT LDViewWindow::doDestroy(void)
{
	if (!switchingModes)
	{
		if (fullScreen)
		{
			restoreDisplayMode();
		}
		if (modelWindow)
		{
			modelWindow->release();
		}
		CUIWindow::doDestroy();
		TCAutoreleasePool::processReleases();
		debugOut("LDViewWindow::doDestroy\n");
		PostQuitMessage(0);
	}
	return 1;
}


HMENU LDViewWindow::menuForBitDepth(HWND hWnd, int bitDepth, int* index)
{
	HMENU viewMenu = GetSubMenu(GetMenu(hWnd), 2);
	int i;
	int count = GetMenuItemCount(viewMenu);
	HMENU bitDepthMenu = 0;
	MENUITEMINFO itemInfo;

	memset(&itemInfo, 0, sizeof(MENUITEMINFO));
	itemInfo.cbSize = sizeof(MENUITEMINFO);
	itemInfo.fMask = MIIM_DATA | MIIM_SUBMENU;
	for (i = 0; i < count && !bitDepthMenu; i++)
	{
		GetMenuItemInfo(viewMenu, i, TRUE, &itemInfo);
		if (itemInfo.hSubMenu && itemInfo.dwItemData == (unsigned)bitDepth)
		{
			bitDepthMenu = GetSubMenu(viewMenu, i);
			if (index)
			{
				*index = i;
			}
		}
	}
	return bitDepthMenu;
}

VideoModeT* LDViewWindow::getCurrentVideoMode(void)
{
	if (currentVideoModeIndex >= 0 && currentVideoModeIndex < numVideoModes)
	{
		return videoModes + currentVideoModeIndex;
	}
	else
	{
		return NULL;
	}
}

int LDViewWindow::getMenuItemIndex(HMENU hMenu, UINT itemID)
{
	int count = GetMenuItemCount(hMenu);
	int i;

	for (i = 0; i < count; i++)
	{
		if (GetMenuItemID(hMenu, i) == itemID)
		{
			return i;
		}
	}
	return -1;
}

int LDViewWindow::clearRecentFileMenuItems(void)
{
	int firstIndex = getMenuItemIndex(hFileMenu, ID_FILE_PRINT);
	int lastIndex = getMenuItemIndex(hFileMenu, ID_FILE_EXIT);

	if (firstIndex >= 0 && lastIndex >= 0)
	{
		int i;

		for (i = firstIndex + 1; i < lastIndex - 1; i++)
		{
			RemoveMenu(hFileMenu, firstIndex + 2, MF_BYPOSITION);
		}
		return firstIndex + 1;
	}
	return -1;
}

void LDViewWindow::populateRecentFiles(void)
{
	int i;
	long maxRecentFiles = TCUserDefaults::longForKey(MAX_RECENT_FILES_KEY, 10);

	recentFiles->removeAll();
	for (i = 1; i <= maxRecentFiles; i++)
	{
		char key[128];
		char *filename;

		sprintf(key, "%s/File%02d", RECENT_FILES_KEY, i);
		filename = TCUserDefaults::stringForKey(key);
		if (filename)
		{
			recentFiles->addString(filename);
			delete filename;
		}
		else
		{
			recentFiles->addString(NULL);
		}
	}
}

void LDViewWindow::recordRecentFiles(void)
{
	int i;
	long maxRecentFiles = TCUserDefaults::longForKey(MAX_RECENT_FILES_KEY, 10);

	for (i = 1; i <= maxRecentFiles; i++)
	{
		char key[128];
		char *filename;

		sprintf(key, "%s/File%02d", RECENT_FILES_KEY, i);
		filename = recentFiles->stringAtIndex(i - 1);
		if (filename)
		{
			TCUserDefaults::setStringForKey(filename, key);
		}
		else
		{
			TCUserDefaults::removeValue(key);
		}
	}
}

void LDViewWindow::populateRecentFileMenuItems(void)
{
	int index = clearRecentFileMenuItems();

	if (index >= 0 && recentFiles->stringAtIndex(0))
	{
		int i;
		MENUITEMINFO itemInfo;
		long maxRecentFiles = TCUserDefaults::longForKey(MAX_RECENT_FILES_KEY,
			10);

		memset(&itemInfo, 0, sizeof(MENUITEMINFO));
		itemInfo.cbSize = sizeof(MENUITEMINFO);
		itemInfo.fMask = MIIM_TYPE;
		itemInfo.fType = MFT_SEPARATOR;
		InsertMenuItem(hFileMenu, index, TRUE, &itemInfo);
//		SelectObject(hdc, GetStockObject(DEFAULT_GUI_FONT));
		for (i = 0; i < maxRecentFiles; i++)
		{
			char *filename = recentFiles->stringAtIndex(i);

			if (filename)
			{
				char title[2048];
				char *partialFilename = copyString(filename);

				PathCompactPath(hdc, partialFilename, 250);
				if (i < 10)
				{
					sprintf(title, "%s&%d %s", i == 9 ? "1" : "", (i + 1) % 10,
						partialFilename);
				}
				else
				{
					strcpy(title, partialFilename);
				}
				itemInfo.fMask = MIIM_TYPE | MIIM_ID;
				itemInfo.fType = MFT_STRING;
				itemInfo.dwTypeData = title;
				itemInfo.wID = 31000 + i;
				InsertMenuItem(hFileMenu, index + i + 1, TRUE, &itemInfo);
				delete partialFilename;
			}
			else
			{
				break;
			}
		}
	}
}

void LDViewWindow::populateDisplayModeMenuItems(void)
{
	int i;
	HMENU viewMenu = GetSubMenu(GetMenu(hWindow), 2);
//	VideoModeT* currentVideoMode = getCurrentVideoMode();
	
	for (i = 0; i < numVideoModes; i++)
	{
		int count = GetMenuItemCount(viewMenu);
		VideoModeT videoMode = videoModes[i];
		HMENU bitDepthMenu = menuForBitDepth(hWindow, videoMode.depth);
		char title[128];
		MENUITEMINFO itemInfo;

		if (!bitDepthMenu)
		{
			memset(&itemInfo, 0, sizeof(MENUITEMINFO));
			sprintf(title, "%d-Bit Modes", videoMode.depth);
			bitDepthMenu = CreatePopupMenu();
			itemInfo.cbSize = sizeof(MENUITEMINFO);
			itemInfo.fMask = MIIM_TYPE | MIIM_SUBMENU | MIIM_DATA;
			itemInfo.fType = MFT_STRING | MFT_RADIOCHECK;
			itemInfo.dwTypeData = title;
			itemInfo.dwItemData = videoMode.depth;
			itemInfo.hSubMenu = bitDepthMenu;
			InsertMenuItem(viewMenu, count, TRUE, &itemInfo);
		}
		sprintf(title, "%dx%d", videoMode.width, videoMode.height);
		memset(&itemInfo, 0, sizeof(MENUITEMINFO));
		itemInfo.cbSize = sizeof(MENUITEMINFO);
		itemInfo.fMask = MIIM_TYPE /*| MIIM_STATE*/ | MIIM_ID;
		itemInfo.fType = MFT_STRING | MFT_RADIOCHECK;
		itemInfo.dwTypeData = title;
		itemInfo.wID = 30000 + i;
/*
		if (currentVideoMode && videoMode.width == currentVideoMode->width &&
			videoMode.height == currentVideoMode->height &&
			videoMode.depth == currentVideoMode->depth)
		{
			itemInfo.fState = MFS_CHECKED;
		}
		else
		{
			itemInfo.fState = MFS_UNCHECKED;
		}
*/
		InsertMenuItem(bitDepthMenu, GetMenuItemCount(bitDepthMenu), TRUE,
			&itemInfo);
	}
}

void LDViewWindow::checkVideoMode(int width, int height, int depth)
{
	DEVMODE deviceMode;
	long result;

	memset(&deviceMode, 0, sizeof DEVMODE);
	deviceMode.dmSize = sizeof DEVMODE;
	deviceMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	deviceMode.dmPelsWidth = width;
	deviceMode.dmPelsHeight = height;
	deviceMode.dmBitsPerPel = depth;
	result = ChangeDisplaySettings(&deviceMode, CDS_TEST | CDS_FULLSCREEN);
	if (result == DISP_CHANGE_SUCCESSFUL)
	{
		if (videoModes)
		{
			videoModes[numVideoModes].width = width;
			videoModes[numVideoModes].height = height;
			videoModes[numVideoModes].depth = depth;
			if (videoModes[numVideoModes].width == fsWidth &&
				videoModes[numVideoModes].height == fsHeight)
			{
				if (currentVideoModeIndex < 0 ||
					videoModes[numVideoModes].depth == fsDepth)
				{
					currentVideoModeIndex = numVideoModes;
				}
			}
		}
		numVideoModes++;
	}
}

void LDViewWindow::checkLowResModes(void)
{
	checkVideoMode(320, 240, 16);
	checkVideoMode(400, 300, 16);
	checkVideoMode(480, 360, 16);
	checkVideoMode(512, 384, 16);
	checkVideoMode(320, 240, 32);
	checkVideoMode(400, 300, 32);
	checkVideoMode(480, 360, 32);
	checkVideoMode(512, 384, 32);
}

void LDViewWindow::getAllDisplayModes(void)
{
	int i;
	int count;
	DEVMODE deviceMode;

	memset(&deviceMode, 0, sizeof DEVMODE);
	deviceMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	numVideoModes = 0;
	videoModes = NULL;
	checkLowResModes();
	for (i = 0; EnumDisplaySettings(NULL, i, &deviceMode); i++)
	{
		if (deviceMode.dmBitsPerPel >= 15)
		{
			numVideoModes++;
		}
	}
	count = i;
	videoModes = new VideoModeT[numVideoModes];
	numVideoModes = 0;
	checkLowResModes();
	for (i = 0; i < count; i++)
	{
		EnumDisplaySettings(NULL, i, &deviceMode);
		if (deviceMode.dmBitsPerPel >= 15)
		{
			int j;
			BOOL found = NO;

			for (j = 0; j < numVideoModes && !found; j++)
			{
				VideoModeT videoMode = videoModes[j];

				if (videoMode.width == (int)deviceMode.dmPelsWidth &&
					videoMode.height == (int)deviceMode.dmPelsHeight &&
					videoMode.depth == (int)deviceMode.dmBitsPerPel)
				{
					found = YES;
				}
			}
			if (!found)
			{
				videoModes[numVideoModes].width = deviceMode.dmPelsWidth;
				videoModes[numVideoModes].height = deviceMode.dmPelsHeight;
				videoModes[numVideoModes].depth = deviceMode.dmBitsPerPel;
				if (videoModes[numVideoModes].width == fsWidth &&
					videoModes[numVideoModes].height == fsHeight)
				{
					if (currentVideoModeIndex < 0 ||
						videoModes[numVideoModes].depth == fsDepth)
					{
						currentVideoModeIndex = numVideoModes;
					}
				}
				numVideoModes++;
			}
		}
	}
}

LRESULT LDViewWindow::doCreate(HWND hWnd, LPCREATESTRUCT lpcs)
{
	LRESULT retVal;

	retVal = CUIWindow::doCreate(hWnd, lpcs);
	if (!videoModes)
	{
		getAllDisplayModes();
	}
	populateDisplayModeMenuItems();
	initPollingMenu();
	if (currentVideoModeIndex >= 0)
	{
		selectFSVideoModeMenuItem(currentVideoModeIndex);
	}
	return retVal;
}

void LDViewWindow::resetView(void)
{
	modelWindow->resetView();
}

void LDViewWindow::printModel(void)
{
	if (modelIsLoaded())
	{
		modelWindow->print();
	}
}

void LDViewWindow::pageSetup(void)
{
	if (modelIsLoaded())
	{
		modelWindow->pageSetup();
	}
}

void LDViewWindow::saveSnapshot(void)
{
	if (modelIsLoaded())
	{
		modelWindow->saveSnapshot();
	}
}

bool LDViewWindow::saveSnapshot(char *saveFilename)
{
	if (modelIsLoaded())
	{
		return modelWindow->saveSnapshot(saveFilename);
	}
	return false;
}

void LDViewWindow::openModel(const char* filename)
{
	if (filename && strlen(filename) > 0)
	{
		char* newFilename = NULL;

		if (!verifyLDrawDir())
		{
			return;
		}
		if (filename[0] == '"')
		{
			newFilename = copyString(filename + 1);
			int length = strlen(newFilename);

			if (newFilename[length - 1] == '"')
			{
				newFilename[length - 1] = 0;
			}
		}
		else
		{
			newFilename = copyString(filename);
		}
		if (modelWindow->chDirFromFilename(newFilename))
		{
			modelWindow->setFilename(newFilename);
			if (modelWindow->loadModel())
			{
				updateModelMenuItems();
				setLastOpenFile(newFilename);
			}
		}
		delete newFilename;
	}
	else
	{
		OPENFILENAME openStruct;
		char fileTypes[1024];
		char openFilename[1024] = "";
		char* initialDir = lastOpenPath();

		if (initialDir)
		{
			memset(fileTypes, 0, 2);
			addFileType(fileTypes, "LDraw Files (*.ldr,*.dat,*.mpd)",
				"*.ldr;*.dat;*.mpd");
			addFileType(fileTypes, "LDraw Model Files (*.ldr,*.dat)",
				"*.ldr;*.dat");
			addFileType(fileTypes, "LDraw Multi-Part Files (*.mpd)", "*.mpd");
			addFileType(fileTypes, "All Files (*.*)", "*.*");
			memset(&openStruct, 0, sizeof(OPENFILENAME));
			openStruct.lStructSize = sizeof(OPENFILENAME);
			openStruct.hwndOwner = hWindow;
			openStruct.lpstrFilter = fileTypes;
			openStruct.nFilterIndex = 1;
			openStruct.lpstrFile = openFilename;
			openStruct.nMaxFile = 1024;
			openStruct.lpstrInitialDir = initialDir;
			openStruct.lpstrTitle = "Select a Model file";
			openStruct.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST |
				OFN_HIDEREADONLY;
			openStruct.lpstrDefExt = "ldr";
			if (GetOpenFileName(&openStruct))
			{
				modelWindow->setFilename(openStruct.lpstrFile);
				setLastOpenFile(openStruct.lpstrFile);
				modelWindow->loadModel();
				updateModelMenuItems();
			}
			delete initialDir;
		}
	}
	populateRecentFileMenuItems();
}

void LDViewWindow::addFileType(char* fileTypes, char* description, char* filter)
{
	char* spot = fileTypes;

	for (spot = fileTypes; spot[0] != 0 || spot[1] != 0; spot++)
		;
	if (spot != fileTypes)
	{
		spot++;
	}
	strcpy(spot, description);
	spot += strlen(description) + 1;
	strcpy(spot, filter);
	spot += strlen(filter) + 1;
	*spot = 0;
}

char* LDViewWindow::getLDrawDir(void)
{
	char* lDrawDir = TCUserDefaults::stringForKey(LDRAWDIR_KEY);

	if (!lDrawDir)
	{
		char buf[1024];

		if (GetPrivateProfileString("LDraw", "BaseDirectory",
			LDrawModel::lDrawDir(), buf, 1024, "ldraw.ini"))
		{
			buf[1023] = 0;
			lDrawDir = copyString(buf);
		}
		else
		{
			// We shouldn't be able to get here, but just to be safe...
			lDrawDir = copyString(LDrawModel::lDrawDir());
		}
	}
	stripTrailingSlash(lDrawDir);
	LDrawModel::setLDrawDir(lDrawDir);
	return lDrawDir;
}

void LDViewWindow::createLDrawDirWindow(void)
{
	hLDrawDirWindow = createDialog(IDD_LDRAWDIR);
}

int CALLBACK LDViewWindow::pathBrowserCallback(HWND hwnd, UINT uMsg,
											   LPARAM lParam, LPARAM /*lpData*/)
{
	if (uMsg == BFFM_SELCHANGED)
	{
		char path[MAX_PATH+10];
		ITEMIDLIST* itemIdList = (ITEMIDLIST*)lParam;

		// For some reason, computers on the network are considered directories,
		// and the ok button is enabled for them.  We don't want to allow that,
		// and if one is selected, the following method will fail.
		if (!SHGetPathFromIDList(itemIdList, path))
		{
			SendMessage(hwnd, BFFM_ENABLEOK, 0, 0);
		}
	}
	else if (uMsg == BFFM_INITIALIZED)
	{
		char *initialDir = getLDrawDir();

		if (initialDir)
		{
			SendMessage(hwnd, BFFM_SETSELECTION, TRUE, (LPARAM)initialDir);
			delete initialDir;
		}
	}
	return 0;
}

BOOL LDViewWindow::promptForLDrawDir(void)
{
	BROWSEINFO browseInfo;
	char displayName[MAX_PATH];
	ITEMIDLIST* itemIdList;

	browseInfo.hwndOwner = NULL; //hWindow;
	browseInfo.pidlRoot = NULL;
	browseInfo.pszDisplayName = displayName;
	browseInfo.lpszTitle = "Please select the directory in which you installed"
		" LDraw.";
	browseInfo.ulFlags = BIF_RETURNONLYFSDIRS;
	browseInfo.lpfn = pathBrowserCallback;
	browseInfo.lParam = 0;
	browseInfo.iImage = 0;
	if ((itemIdList = SHBrowseForFolder(&browseInfo)) != NULL)
	{
		char path[MAX_PATH+10];

		if (SHGetPathFromIDList(itemIdList, path))
		{
			stripTrailingSlash(path);
			TCUserDefaults::setStringForKey(path, LDRAWDIR_KEY);
			LDrawModel::setLDrawDir(path);
			return TRUE;
		}
		MessageBox(NULL/*hWindow*/, "Invalid directory selected.", "Error", MB_OK);
	}
	return FALSE;
/*
	if (!hLDrawDirWindow)
	{
		createLDrawDirWindow();
	}
	if (hLDrawDirWindow)
	{
		runDialogModal(hLDrawDirWindow);
		if (userLDrawDir)
		{
			stripTrailingSlash(userLDrawDir);
			TCUserDefaults::setStringForKey(userLDrawDir, LDRAWDIR_KEY);
			LDrawModel::setLDrawDir(userLDrawDir);
			delete userLDrawDir;
			userLDrawDir = NULL;
			return TRUE;
		}
	}
	return FALSE;
*/
}

void LDViewWindow::stripTrailingSlash(char* value)
{
	int end = strlen(value) - 1;

	if (end >= 0 && (value[end] == '/' || value[end] == '\\'))
	{
		value[end] = 0;
	}
}

BOOL LDViewWindow::verifyLDrawDir(char* value)
{
//	int length = strlen(value);
	char currentDir[1024];
	char newDir[1024];
	BOOL found = FALSE;

//	stripTrailingSlash(value);
	sprintf(newDir, "%s\\parts", value);
	GetCurrentDirectory(1024, currentDir);
	if (SetCurrentDirectory(newDir))
	{
		sprintf(newDir, "%s\\p", value);
		if (SetCurrentDirectory(newDir))
		{
			found = TRUE;
		}
		SetCurrentDirectory(currentDir);
	}
	return found;
}

BOOL LDViewWindow::verifyLDrawDir(bool forceChoose)
{
	char* lDrawDir = getLDrawDir();
	BOOL found = FALSE;

	if (!forceChoose && verifyLDrawDir(lDrawDir))
	{
		delete lDrawDir;
		found = TRUE;
	}
	else
	{
		delete lDrawDir;
		while (!found)
		{
			if (promptForLDrawDir())
			{
				lDrawDir = getLDrawDir();
				if (verifyLDrawDir(lDrawDir))
				{
					found = TRUE;
				}
				else
				{
					MessageBox(NULL, "The directory you selected does not "
						"contain LDraw.", "Invalid directory",
						MB_OK | MB_ICONWARNING | MB_TASKMODAL);
				}
				delete lDrawDir;
			}
			else
			{
				break;
			}
		}
	}
	return found;
}

char* LDViewWindow::lastOpenPath(char* pathKey)
{
	if (!pathKey)
	{
		pathKey = LAST_OPEN_PATH_KEY;
	}
	if (verifyLDrawDir())
	{
		char* path = TCUserDefaults::stringForKey(pathKey);

		if (!path)
		{
			path = getLDrawDir();
		}
		return path;
	}
	else
	{
		return NULL;
	}
}

void LDViewWindow::setLastOpenFile(const char* filename, char* pathKey)
{
	if (filename)
	{
		char* spot = strrchr(filename, '\\');
		int index;

		if (!pathKey)
		{
			pathKey = LAST_OPEN_PATH_KEY;
		}
		if (spot)
		{
			int length = spot - filename;
			char* path = new char[length + 1];

			strncpy(path, filename, length);
			path[length] = 0;
			TCUserDefaults::setStringForKey(path, pathKey);
			delete path;
		}
		if (recentFiles)
		{
			index = recentFiles->indexOfString(filename);
			recentFiles->insertString(filename);
			if (index >= 0)
			{
				// Insert before removal.  Since the one being removed could have the
				// same pointer value as the string in the array, we could otherwise
				// access a pointer after it had been deleted.
				recentFiles->removeString(index + 1);
			}
			recordRecentFiles();
		}
	}
}

LRESULT LDViewWindow::doShowWindow(BOOL showFlag, LPARAM status)
{
	printf("LDViewWindow::doShowWindow\n");
	if (modelWindow && showFlag && !modelWindowIsShown())
	{
//		modelWindowIsShown();
		modelWindow->showWindow(SW_NORMAL);
//		modelWindowShown = true;
	}
	return CUIWindow::doShowWindow(showFlag, status);
}

bool LDViewWindow::modelWindowIsShown(void)
{
	if (modelWindow)
	{
		HWND hModelWindow = modelWindow->getHWindow();
		return IsWindowVisible(hModelWindow) != FALSE;
/*
		WINDOWPLACEMENT windowPlacement;

		windowPlacement.length = sizeof(WINDOWPLACEMENT);
		if (GetWindowPlacement(hModelWindow, &windowPlacement))
		{
			return windowPlacement.showCmd != SW_HIDE;
		}
*/
	}
	return false;
}

LRESULT LDViewWindow::doKeyDown(int keyCode, long keyData)
{
	if (modelWindow)
	{
		return modelWindow->processKeyDown(keyCode, keyData);
	}
	else
	{
		return 1;
	}
}

LRESULT LDViewWindow::doKeyUp(int keyCode, long keyData)
{
	if (modelWindow)
	{
		return modelWindow->processKeyUp(keyCode, keyData);
	}
	else
	{
		return 1;
	}
}

LRESULT LDViewWindow::doDrawItem(HWND /*hControlWnd*/,
								 LPDRAWITEMSTRUCT drawItemStruct)
{
	if (drawItemStruct->hwndItem == hStatusBar)
	{
		if (drawItemStruct->itemID == 0)
		{
			DWORD backgroundColor =
				TCUserDefaults::longForKey(BACKGROUND_COLOR_KEY);
 			HBRUSH hBrush = CreateSolidBrush(RGB(backgroundColor & 0xFF,
				(backgroundColor >> 8) & 0xFF, (backgroundColor >> 16) & 0xFF));

			FillRect(drawItemStruct->hDC, &drawItemStruct->rcItem, hBrush);
			DeleteObject(hBrush);
			return TRUE;
		}
	}
	return FALSE;
}

void LDViewWindow::startLoading(void)
{
}

void LDViewWindow::stopLoading(void)
{
}

void LDViewWindow::setLoading(bool value)
{
	loading = value;
	if (loading)
	{
		startLoading();
	}
	else
	{
		stopLoading();
	}
}

LRESULT LDViewWindow::doInitMenuPopup(HMENU hPopupMenu, UINT /*uPos*/,
									  BOOL /*fSystemMenu*/)
{
	int i;
	int count = GetMenuItemCount(hPopupMenu);

	for (i = 0; i < count; i++)
	{
		setMenuEnabled(hPopupMenu, i, !loading, TRUE);
	}
	if (hPopupMenu == hFileMenu)
	{
		if (loading)
		{
			setMenuEnabled(hFileMenu, ID_FILE_CANCELLOAD, true);
		}
		else
		{
			setMenuEnabled(hFileMenu, ID_FILE_CANCELLOAD, false);
			updateModelMenuItems();
		}
	}
	return 1;
}

BOOL LDViewWindow::doDialogGetMinMaxInfo(HWND hDlg, LPMINMAXINFO minMaxInfo)
{
	if (hDlg == hOpenGLInfoWindow)
	{
		calcSystemSizes();
		minMaxInfo->ptMaxSize.x = systemMaxWidth;
		minMaxInfo->ptMaxSize.y = systemMaxHeight;
		minMaxInfo->ptMinTrackSize.x = 250;
		minMaxInfo->ptMinTrackSize.y = 200;
		minMaxInfo->ptMaxTrackSize.x = systemMaxTrackWidth;
		minMaxInfo->ptMaxTrackSize.y = systemMaxTrackHeight;
		return TRUE;
	}
	return FALSE;
}

