#include "ModelLoader.h"
#include "ModelWindow.h"
#include "SSModelWindow.h"
#include "LDViewWindow.h"
#include "UserDefaultsKeys.h"
#include "LDVExtensionsSetup.h"

#include <LDLib/LDrawModel.h>

#include <TCFoundation/mystring.h>
#include <TCFoundation/TCUserDefaults.h>
#include <TCFoundation/TCStringArray.h>

#define TITLE "LDView"

#define WIN_WIDTH 640
#define WIN_HEIGHT 480

ModelLoader::ModelLoader(HINSTANCE hInstance, int nCmdShow, BOOL screenSaver)
			:modelWindow(NULL),
			 parentWindow(NULL),
			 hInstance(hInstance),
			 nCmdShow(nCmdShow),
			 screenSaver(screenSaver)
{
	init();
}

ModelLoader::~ModelLoader(void)
{
}

void ModelLoader::dealloc(void)
{
	if (modelWindow)
	{
		modelWindow->release();
	}
	if (parentWindow)
	{
		parentWindow->release();
	}
	TCObject::dealloc();
}

void ModelLoader::init(void)
{
	startup();
}

void ModelLoader::update(void)
{
	modelWindow->update();
}

void ModelLoader::startup(void)
{
	int width;
	int height;
	int widthDelta = 0;
	int heightDelta = 0;
	int xOffset = 0;
	int yOffset = 0;
	BOOL maximized;

	if (screenSaver)
	{
		TCUserDefaults::setAppName("Travis Cobbs/LDView Screen Saver");
	}
	else
	{
		TCUserDefaults::setAppName("Travis Cobbs/LDView");
//		xOffset = 2;
//		yOffset = 2;
//		widthDelta = 4;
		if (TCUserDefaults::longForKey(STATUS_BAR_KEY, 1))
		{
			heightDelta = 20;
//			heightDelta = 24;
		}
		else
		{
//			heightDelta = 4;
		}
	}
	width = TCUserDefaults::longForKey(WINDOW_WIDTH_KEY, WIN_WIDTH);
	height = TCUserDefaults::longForKey(WINDOW_HEIGHT_KEY, WIN_HEIGHT);
	maximized = TCUserDefaults::longForKey(WINDOW_MAXIMIZED_KEY);
	parentWindow = new LDViewWindow(TITLE, hInstance, CW_USEDEFAULT,
		CW_USEDEFAULT, width + widthDelta, height + heightDelta);
	parentWindow->setMinWidth(320);
	parentWindow->setMinHeight(240);
	parentWindow->setScreenSaver(screenSaver);
	if (screenSaver)
	{
		modelWindow = new SSModelWindow(parentWindow, 0, 0, width, height);
	}
	else
	{
		modelWindow = new ModelWindow(parentWindow, xOffset, yOffset, width,
			height);
	}
	parentWindow->setModelWindow(modelWindow);

	if (parentWindow->initWindow())
	{
		char *commandLineFilename = getCommandLineFilename();
		char *snapshotFilename = TCUserDefaults::stringForKey("SaveSnapshot");
		bool savedSnapshot = false;

		modelWindow->initWindow();
		if (!screenSaver && commandLineFilename && snapshotFilename &&
			LDVExtensionsSetup::havePixelBufferExtension())
		{
			parentWindow->openModel(commandLineFilename);
			if (parentWindow->saveSnapshot(snapshotFilename))
			{
				parentWindow->shutdown();
				savedSnapshot = true;
			}
		}
		if (!savedSnapshot)
		{
//			printf("About to show main window...\n");
			parentWindow->showWindow(nCmdShow);
//			printf("Main window shown.\n");
			if (maximized)
			{
				parentWindow->maximize();
			}
			if (!screenSaver)
			{
				parentWindow->openModel(commandLineFilename);
				if (commandLineFilename && snapshotFilename)
				{
					if (parentWindow->saveSnapshot(snapshotFilename))
					{
						parentWindow->shutdown();
					}
				}
			}
		}
		delete commandLineFilename;
		delete snapshotFilename;
	}
}

char *ModelLoader::getCommandLineFilename(void)
{
	char *commandLineFilename = NULL;
	TCStringArray *commandLine = TCUserDefaults::getProcessedCommandLine();

	if (commandLine)
	{
		int i;
		int count = commandLine->getCount();

		for (i = 0; i < count && !commandLineFilename; i++)
		{
			char *arg = commandLine->stringAtIndex(i);

			if (arg[0] != '-')
			{
				commandLineFilename = arg;
			}
		}
	}
	return copyString(commandLineFilename);
}
