#include "SSConfigure.h"
#include "LDViewPreferences.h"
#include "LDViewWindow.h"
#include "AppResources.h"
#include <TCFoundation/TCUserDefaults.h>
#include <TCFoundation/myString.h>
#include <commctrl.h>

#include <stdio.h>

#define SS_SIZE_KEY "SSSize"
#define SS_SPEED_KEY "SSSpeed"
#define SS_ROTATION_SPEED_KEY "SSRotationSpeed"
#define SS_FILENAME_KEY "SSFilename"
#define SS_LAST_OPEN_PATH_KEY "SSLastOpenPath"

#define QUALITY_LIGHTING_KEY "UseQualityLighting"

SSConfigure* SSConfigure::globalSSConfigure = NULL;

SSConfigure::SSConfigure(HINSTANCE hInstance)
			:CUIWindow("blah", hInstance, 0, 0, 100, 100),
			 hPropSheet(NULL),
			 hSSPage(NULL),
			 hLDViewPage(NULL),
			 numPropSheetPages(0),
			 prefs(NULL),
			 initializingPrefs(FALSE)
{
	globalSSConfigure = this;
}

void SSConfigure::dealloc(void)
{
	if (prefs)
	{
		prefs->release();
	}
	CUIWindow::dealloc();
}

void SSConfigure::run(void)
{
	createConfigureWindow();
//	runDialogModal(hConfigureWindow);
}

void SSConfigure::createConfigureWindow(void)
{
//	hConfigureWindow = createDialog(IDD_SS_SETTINGS);
//	hPropSheet = GetDlgItem(hConfigureWindow, IDC_SS_PROP_SHEET);
	setupPropSheet();
/*
	hApplyButton = GetDlgItem(hPrefsWindow, IDC_APPLY);
	hFullScreenRateField = GetDlgItem(hPrefsWindow, IDC_FS_RATE);
	hSeamSpin = GetDlgItem(hPrefsWindow, IDC_SEAM_SPIN);
	SendDlgItemMessage(hPrefsWindow, IDC_FS_RATE, EM_LIMITTEXT, 3, 0);
*/
}

void SSConfigure::setupPropSheet(void)
{
	addPropSheetPage(IDD_SS_SETTINGS, "Screensaver Preferences");
	addPropSheetPage(IDD_SS_PREFERENCES, "LDView Preferences");
	createPropSheet();
}

void SSConfigure::addPropSheetPage(int resourceId, char* title)
{
//	HPROPSHEETPAGE hPage;

	memset(propSheetPages + numPropSheetPages, 0, sizeof(PROPSHEETPAGE));
	propSheetPages[numPropSheetPages].dwSize = sizeof(PROPSHEETPAGE);
	propSheetPages[numPropSheetPages].dwFlags = PSP_USETITLE | PSP_USECALLBACK;
	propSheetPages[numPropSheetPages].hInstance = hInstance;
	propSheetPages[numPropSheetPages].pszTemplate = MAKEINTRESOURCE(resourceId);
//	propSheetPages[numPropSheetPages].hIcon = ???
	propSheetPages[numPropSheetPages].pszTitle = title;
	propSheetPages[numPropSheetPages].pfnDlgProc = staticPropDialogProc;
	propSheetPages[numPropSheetPages].lParam = (LPARAM)this;
	propSheetPages[numPropSheetPages].pfnCallback = staticPropSheetPageProc;
	propSheetPageHandles[numPropSheetPages] =
		CreatePropertySheetPage(&propSheetPages[numPropSheetPages]);
	numPropSheetPages++;
}

void SSConfigure::createPropSheet(void)
{
	PROPSHEETHEADER psHeader;

	memset(&psHeader, 0, sizeof(PROPSHEETHEADER));
	psHeader.dwSize = sizeof(PROPSHEETHEADER);
	psHeader.dwFlags = /*PSH_MODELESS | PSH_PROPSHEETPAGE*/PSH_USECALLBACK;
	psHeader.hwndParent = hWindow;//hConfigureWindow;
	psHeader.hInstance = hInstance;
	psHeader.pszCaption = "LDView Screensaver Settings";
	psHeader.nPages = numPropSheetPages;
	psHeader.nStartPage = 0;
	psHeader.phpage = propSheetPageHandles;
	psHeader.pfnCallback = staticPropSheetProc;
	PropertySheet(&psHeader);
}

BOOL CALLBACK SSConfigure::staticPropDialogProc(HWND hDlg, UINT message,
												WPARAM wParam, LPARAM lParam)
{
	char className[1024];
	HWND hParentDlg = (HWND)GetWindowLong(hDlg, GWL_HWNDPARENT);
	HWND hLastParent = hDlg;

	GetClassName(hDlg, className, 1024);
	if (strcmp(className, "#32770") == 0)
	{
		while (hParentDlg)
		{
			hLastParent = hParentDlg;
			hParentDlg = (HWND)GetWindowLong(hParentDlg, GWL_HWNDPARENT);
		}
		SSConfigure *ssConfigure = (SSConfigure *)GetWindowLong(hLastParent,
			GWL_USERDATA);

		if (ssConfigure)
		{
			return ssConfigure->dialogProc(hDlg, message, wParam, lParam);
		}
		return TRUE;
//		return globalSSConfigure->dialogProc(hDlg, message, wParam, lParam);
	}
	else
	{
		return staticDialogProc(hDlg, message, wParam, lParam);
	}
}

BOOL SSConfigure::dialogProc(HWND hDlg, UINT message, WPARAM wParam,
							 LPARAM lParam)
{
	return CUIWindow::dialogProc(hDlg, message, wParam, lParam);
}

BOOL SSConfigure::doDialogCommand(HWND hDlg, int controlId, int notifyCode,
								  HWND controlHWnd)
{
//	printf("SSConfigure::doDialogCommand: 0x%04X, %d, 0x%X\n", hDlg, controlId,
//		notifyCode);
	if (hDlg == hLDViewPage)
	{
		if (!(prefs->doCommand(hDlg, controlId, notifyCode, controlHWnd) &
			LDP_UNKNOWN_COMMAND) && !initializingPrefs)
		{
			PropSheet_Changed(hPropSheet, hLDViewPage);
		}
		return TRUE;
	}
	else if (hDlg == hSSPage)
	{
		if (controlId == IDC_BROWSE_BUTTON)
		{
			doBrowse();
		}
		return TRUE;
	}
	return FALSE;
}

void SSConfigure::doBrowse(void)
{
	OPENFILENAME openStruct;
	char fileTypes[1024];
	char openFilename[1024] = "";
	char* initialDir = LDViewWindow::lastOpenPath(SS_LAST_OPEN_PATH_KEY);

	if (initialDir)
	{
		memset(fileTypes, 0, 2);
		LDViewWindow::addFileType(fileTypes, "LDraw Files (.dat;.mpd)",
			"*.dat;*.mpd");
		LDViewWindow::addFileType(fileTypes, "LDraw Model Files (.dat)",
			"*.dat");
		LDViewWindow::addFileType(fileTypes, "LDraw Multi-Part Files (.mpd)",
			"*.mpd");
		LDViewWindow::addFileType(fileTypes, "All Files (*.*)", "*.*");
		memset(&openStruct, 0, sizeof(OPENFILENAME));
		openStruct.lStructSize = sizeof(OPENFILENAME);
		openStruct.hwndOwner = hWindow;
		openStruct.lpstrFilter = fileTypes;
		openStruct.nFilterIndex = 1;
		openStruct.lpstrFile = openFilename;
		openStruct.nMaxFile = 1024;
		openStruct.lpstrInitialDir = initialDir;
		openStruct.lpstrTitle = "Select a Model file";
		openStruct.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST |
			OFN_HIDEREADONLY;
		openStruct.lpstrDefExt = "tri";
		if (GetOpenFileName(&openStruct))
		{
			LDViewWindow::setLastOpenFile(openStruct.lpstrFile,
				SS_LAST_OPEN_PATH_KEY);
			SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_SETTEXT, 0,
				(LPARAM)openStruct.lpstrFile);
			PropSheet_Changed(hPropSheet, hLDViewPage);
		}
		delete initialDir;
	}
}

void SSConfigure::applyChanges(void)
{
	char filename[1024];

	TCUserDefaults::setLongForKey(
		SendDlgItemMessage(hSSPage, IDC_SIZE_SLIDER, TBM_GETPOS, 0, 0),
		SS_SIZE_KEY);
	TCUserDefaults::setLongForKey(
		SendDlgItemMessage(hSSPage, IDC_SPEED_SLIDER, TBM_GETPOS, 0, 0),
		SS_SPEED_KEY);
	TCUserDefaults::setLongForKey(
		SendDlgItemMessage(hSSPage, IDC_ROTATION_SPEED_SLIDER, TBM_GETPOS, 0, 0),
		SS_ROTATION_SPEED_KEY);
	SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_GETTEXT, 1024,
		(LPARAM)filename);
	TCUserDefaults::setStringForKey(filename, SS_FILENAME_KEY);
	if (prefs)
	{
		prefs->doApply();
	}
}

BOOL SSConfigure::doDialogNotify(HWND hDlg, int controlId, LPNMHDR notification)
{
//	printf("SSConfigure::doDialogNotify: 0x%04X, %d, 0x%X\n", hDlg, controlId,
//		notification->code);
	if (hPropSheet == NULL)
	{
		hPropSheet = notification->hwndFrom;
		long userData = GetWindowLong(hPropSheet, GWL_USERDATA);
		long dlgData = GetWindowLong(hPropSheet, DWL_USER);
	}
	switch (notification->code)
	{
	case PSN_SETACTIVE:
		{
			TCITEM *item;
			HWND hTabControl = PropSheet_GetTabControl(hPropSheet);

			item = (TCITEM*)malloc(sizeof(TCITEM) + 4);
			item->mask = TCIF_PARAM;
			if (!hSSPage)
			{
				if (TabCtrl_GetItem(hTabControl, 0, item))
				{
					hSSPage = (HWND)item->lParam;
				}
				setupSSPage();
			}
			if (!hLDViewPage)
			{
				if (TabCtrl_GetItem(hTabControl, 1, item))
				{
					hLDViewPage = (HWND)item->lParam;
				}
				if (hLDViewPage)
				{
					setupLDViewPage();
				}
			}
			free(item);
		}
		break;
	case PSN_APPLY:
		applyChanges();
		break;
	case NM_RELEASEDCAPTURE:
		if (hDlg == hSSPage)
		{
			PropSheet_Changed(hPropSheet, hSSPage);
		}
		break;
	}
	return FALSE;
}

int CALLBACK SSConfigure::staticPropSheetProc(HWND hDlg, UINT uMsg,
											  LPARAM lParam)
{
	if (hDlg)
	{
		if (globalSSConfigure)
		{
			SetWindowLong(hDlg, GWL_USERDATA, (long)globalSSConfigure);
			SSConfigure::clearGlocalSSConfigure();

		}
	}
	return 0;
}

UINT CALLBACK SSConfigure::staticPropSheetPageProc(HWND hwnd, UINT uMsg,
												   LPPROPSHEETPAGE ppsp)
{
	return ((SSConfigure*)ppsp->lParam)->propSheetPageProc(hwnd, uMsg, ppsp);
}

UINT SSConfigure::propSheetPageProc(HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp)
{
	if (uMsg == PSPCB_CREATE)
	{
/*
		if (ppsp->pszTemplate == MAKEINTRESOURCE(IDD_SS_SETTINGS))
		{
			setupSSPage();
		}
		else
		{
			setupLDViewPage();
		}
*/
	}
	return 1;
}

void SSConfigure::setupSlider(int controlId, short min, short max,
							  WORD frequency, int value)
{
	SendDlgItemMessage(hSSPage, controlId, TBM_SETRANGE, TRUE,
		(LPARAM)MAKELONG(min, max));
	SendDlgItemMessage(hSSPage, controlId, TBM_SETTICFREQ, frequency, 0);
	SendDlgItemMessage(hSSPage, controlId, TBM_SETPOS, 1, value);
}

char* SSConfigure::defaultFilename(void)
{
	char *path = LDViewWindow::getLDrawDir();
	char buf[2048];

	sprintf(buf, "%s\\%s", path, "models\\car.dat");
	delete path;
	return copyString(buf);
}

void SSConfigure::setupSSPage(void)
{
	char *filename = TCUserDefaults::stringForKey(SS_FILENAME_KEY, NULL);

	if (!filename)
	{
		filename = defaultFilename();
	}
	SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_SETTEXT, 0,
		(LPARAM)filename);
	setupSlider(IDC_SIZE_SLIDER, 100, 1000, 50,
		TCUserDefaults::longForKey(SS_SIZE_KEY));
	setupSlider(IDC_SPEED_SLIDER, 1, 20, 1,
		TCUserDefaults::longForKey(SS_SPEED_KEY));
	setupSlider(IDC_ROTATION_SPEED_SLIDER, 0, 1000, 50,
		TCUserDefaults::longForKey(SS_ROTATION_SPEED_KEY));
	delete filename;
}

void SSConfigure::setupLDViewPage(void)
{
	prefs = new LDViewPreferences();
	initializingPrefs = TRUE;
	prefs->setup(hLDViewPage);
	initializingPrefs = FALSE;
//	SendDlgItemMessage(hLDViewPage, IDC_LIGHTING_QUALITY, BM_SETCHECK,
//		TCUserDefaults::longForKey(QUALITY_LIGHTING_KEY), 0);
}

BOOL SSConfigure::doDialogVScroll(HWND hDlg, int scrollCode, int position,
								  HWND hScrollBar)
{
	if (hDlg == hLDViewPage)
	{
		return prefs->doVScroll(hDlg, scrollCode, position, hScrollBar);
	}
	return FALSE;
}
