#include "SSConfigure.h"
#include "LDViewPreferences.h"
#include "LDViewWindow.h"
#include "LDVExtensionsSetup.h"
#include "AppResources.h"
#include <TCFoundation/TCUserDefaults.h>
#include <TCFoundation/myString.h>
#include <commctrl.h>
#include "UserDefaultsKeys.h"

#include <stdio.h>

SSConfigure::SSConfigure(HINSTANCE hInstance):
	LDViewPreferences(hInstance),
//	CUIPropertySheet("LDView Screensaver Settings", hInstance),
	hSSPage(NULL)
{
	generalPageNumber++;
	geometryPageNumber++;
	effectsPageNumber++;
	primitivesPageNumber++;
	loadSSSettings();
}

SSConfigure::~SSConfigure(void)
{
}

void SSConfigure::dealloc(void)
{
	delete ssFilename;
	CUIPropertySheet::dealloc();
}

void SSConfigure::loadDefaultSSSettings(void)
{
	ssFilename = NULL;
	ssSize = DEFAULT_SS_SIZE;
	ssSpeed = DEFAULT_SS_SPEED;
	ssRotationSpeed = DEFAULT_SS_ROT_SPEED;
}

void SSConfigure::loadSSSettings(void)
{
	loadDefaultSSSettings();
	ssFilename = TCUserDefaults::stringForKey(SS_FILENAME_KEY, ssFilename);
	if (!ssFilename)
	{
		ssFilename = defaultFilename();
	}
	ssSize = TCUserDefaults::longForKey(SS_SIZE_KEY, ssSize);
	ssSpeed = TCUserDefaults::longForKey(SS_SPEED_KEY, ssSpeed);
	ssRotationSpeed = TCUserDefaults::longForKey(SS_ROTATION_SPEED_KEY,
		ssRotationSpeed);
}

int SSConfigure::run(void)
{
	LDVExtensionsSetup::setup(hInstance);
	addPage(IDD_SS_SETTINGS);//, "Screensaver Preferences");
	return LDViewPreferences::run();
}

BOOL SSConfigure::doDialogCommand(HWND hDlg, int controlId, int notifyCode,
								  HWND controlHWnd)
{
	if (hDlg == hSSPage)
	{
		switch (controlId)
		{
			case IDC_BROWSE_BUTTON:
				doBrowse();
				break;
			case IDC_SS_RESET:
				loadDefaultSSSettings();
				setupSSPage();
				break;
		}
		return TRUE;
	}
	return LDViewPreferences::doDialogCommand(hDlg, controlId, notifyCode,
		controlHWnd);
}

void SSConfigure::doBrowse(void)
{
	OPENFILENAME openStruct;
	char fileTypes[1024];
	char openFilename[1024] = "";
	char* initialDir = LDViewWindow::lastOpenPath(SS_LAST_OPEN_PATH_KEY);

	if (initialDir)
	{
		memset(fileTypes, 0, 2);
		LDViewWindow::addFileType(fileTypes, "LDraw Files (.dat;.mpd)",
			"*.dat;*.mpd");
		LDViewWindow::addFileType(fileTypes, "LDraw Model Files (.dat)",
			"*.dat");
		LDViewWindow::addFileType(fileTypes, "LDraw Multi-Part Files (.mpd)",
			"*.mpd");
		LDViewWindow::addFileType(fileTypes, "All Files (*.*)", "*.*");
		memset(&openStruct, 0, sizeof(OPENFILENAME));
		openStruct.lStructSize = sizeof(OPENFILENAME);
		openStruct.hwndOwner = hWindow;
		openStruct.lpstrFilter = fileTypes;
		openStruct.nFilterIndex = 1;
		openStruct.lpstrFile = openFilename;
		openStruct.nMaxFile = 1024;
		openStruct.lpstrInitialDir = initialDir;
		openStruct.lpstrTitle = "Select a Model file";
		openStruct.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST |
			OFN_HIDEREADONLY;
		openStruct.lpstrDefExt = "tri";
		if (GetOpenFileName(&openStruct))
		{
			LDViewWindow::setLastOpenFile(openStruct.lpstrFile,
				SS_LAST_OPEN_PATH_KEY);
			SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_SETTEXT, 0,
				(LPARAM)openStruct.lpstrFile);
			enableApply(hSSPage);
		}
		delete initialDir;
	}
}

void SSConfigure::applyChanges(void)
{
	char filename[1024];

	ssSize = SendDlgItemMessage(hSSPage, IDC_SIZE_SLIDER, TBM_GETPOS, 0, 0);
	ssSpeed = SendDlgItemMessage(hSSPage, IDC_SPEED_SLIDER, TBM_GETPOS, 0, 0);
	ssRotationSpeed = SendDlgItemMessage(hSSPage, IDC_ROTATION_SPEED_SLIDER,
		TBM_GETPOS, 0, 0);
	SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_GETTEXT, 1024,
		(LPARAM)filename);
	delete ssFilename;
	ssFilename = copyString(filename);
	TCUserDefaults::setLongForKey(ssSize, SS_SIZE_KEY);
	TCUserDefaults::setLongForKey(ssSpeed, SS_SPEED_KEY);
	TCUserDefaults::setLongForKey(ssRotationSpeed, SS_ROTATION_SPEED_KEY);
	TCUserDefaults::setStringForKey(ssFilename, SS_FILENAME_KEY);
	LDViewPreferences::applyChanges();
}

BOOL SSConfigure::doDialogNotify(HWND hDlg, int controlId, LPNMHDR notification)
{
	if (hDlg == hSSPage && notification->code == NM_RELEASEDCAPTURE)
	{
		enableApply(hSSPage);
		return FALSE;
	}
	return LDViewPreferences::doDialogNotify(hDlg, controlId, notification);
}

char* SSConfigure::defaultFilename(void)
{
	char *path = LDViewWindow::getLDrawDir();
	char buf[2048];

	sprintf(buf, "%s\\%s", path, "models\\car.dat");
	delete path;
	return copyString(buf);
}

void SSConfigure::setupSSPage(void)
{
	hSSPage = hwndArray->pointerAtIndex(0);
	SendDlgItemMessage(hSSPage, IDC_FILENAME_FIELD, WM_SETTEXT, 0,
		(LPARAM)ssFilename);
	setupDialogSlider(hSSPage, IDC_SIZE_SLIDER, 100, 1000, 50, ssSize);
	setupDialogSlider(hSSPage, IDC_SPEED_SLIDER, 1, 20, 1, ssSpeed);
	setupDialogSlider(hSSPage, IDC_ROTATION_SPEED_SLIDER, 1, 1000, 50,
		ssRotationSpeed);
}

void SSConfigure::setupPage(int pageNumber)
{
	switch (pageNumber)
	{
	case 0:
		setupSSPage();
		break;
	default:
		LDViewPreferences::setupPage(pageNumber);
		break;
	}
}

void SSConfigure::setupGeneralPage(void)
{
	LDViewPreferences::setupGeneralPage();
	EnableWindow(GetDlgItem(hGeneralPage, IDC_FRAME_RATE), FALSE);
	EnableWindow(GetDlgItem(hGeneralPage, IDC_SHOW_ERRORS), FALSE);
	EnableWindow(GetDlgItem(hGeneralPage, IDC_FS_REFRESH), FALSE);
}

void SSConfigure::enableCutaway(void)
{
	disableCutaway();
}

void SSConfigure::disableCutaway(void)
{
	LDViewPreferences::disableCutaway();
	EnableWindow(GetDlgItem(hEffectsPage, IDC_CUTAWAY), FALSE);
}