#include "SSModelWindow.h"
#include "SSConfigure.h"
#include "SSPassword.h"
#include "LDViewWindow.h"
#include "UserDefaultsKeys.h"

#include <TCFoundation/myString.h>
#include <TCFoundation/TCUserDefaults.h>
#include <math.h>

void debugOut(char *fmt, ...);

SSModelWindow::SSModelWindow(CUIWindow* parentWindow, int x, int y,
							 int width, int height)
			  :ModelWindow(parentWindow, x, y, width, height),
			   shownOnce(FALSE),
			   screenWidth(GetSystemMetrics(SM_CXFULLSCREEN)),
			   screenHeight(GetSystemMetrics(SM_CYFULLSCREEN) +
			   GetSystemMetrics(SM_CYCAPTION)),
			   ssSize(TCUserDefaults::longForKey(SS_SIZE_KEY, DEFAULT_SS_SIZE)),
			   ssSpeed(TCUserDefaults::longForKey(SS_SPEED_KEY,
				   DEFAULT_SS_SPEED)),
			   ssRotationSpeed(TCUserDefaults::longForKey(SS_ROTATION_SPEED_KEY,
				   DEFAULT_SS_ROT_SPEED)),
			   ssFilename(TCUserDefaults::stringForKey(SS_FILENAME_KEY))
{
	char *programPath;

	TCUserDefaults::setAppName("Travis Cobbs/LDView");
	programPath = TCUserDefaults::stringForKey(INSTALL_PATH_KEY);
	if (programPath)
	{
		modelViewer->setProgramPath(programPath);
		delete programPath;
	}
	TCUserDefaults::setAppName("Travis Cobbs/LDView Screen Saver");
	modelViewer->setDistanceMultiplier(1.35f);
	ssPassword = new SSPassword;
}

SSModelWindow::~SSModelWindow(void)
{
}

void SSModelWindow::dealloc(void)
{
	delete ssFilename;
	if (ssPassword)
	{
		ssPassword->release();
	}
	ModelWindow::dealloc();
}

int SSModelWindow::loadModel(void)
{
	int retVal = ModelWindow::loadModel();

	if (retVal)
	{
		setXRotate(1.0f);
		setYRotate(1.0f);
		setRotationSpeed(ssRotationSpeed / 100.0f);
	}
	return retVal;
}

void SSModelWindow::verifySsFilename(void)
{
	if (ssFilename == NULL)
	{
		ssFilename = SSConfigure::defaultFilename();
	}
}

/*
WNDCLASSEX SSModelWindow::getWindowClass(void)
{
	WNDCLASSEX windowClass = CUIOGLWindow::getWindowClass();

	windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);
	windowClass.hCursor = NULL;
	return windowClass;
}
*/

void SSModelWindow::finalSetup(void)
{
	if (!shownOnce)
	{
		shownOnce = TRUE;
		resize(ssSize, ssSize);
		verifySsFilename();
		((LDViewWindow*)parentWindow)->openModel(ssFilename);
		ShowCursor(FALSE);
	}
}

/*
LRESULT SSModelWindow::doShowWindow(BOOL showFlag, LPARAM status)
{
	LRESULT retVal = ModelWindow::doShowWindow(showFlag, status);

	if (!shownOnce)
	{
		shownOnce = TRUE;
		parentWindow->maximize();
		resize(ssSize, ssSize);
	}
	return retVal;
}
*/

void SSModelWindow::doPaint(void)
{
	RECT rect;
	double xScale = (screenWidth - width) / 2.0;
	double yScale = (screenHeight - height) / 2.0;
	double newX;
	double newY;
	DWORD t = GetTickCount();

	rect.left = x;
	rect.right = x + width;
	rect.top = y;
	rect.bottom = y + height;
	newX = (sin(t * (ssSpeed + 3) / 15000.0) + 1.0) * xScale;
	newY = (sin(t * (ssSpeed + 3) / 13456.0) + 1.0) * yScale;
	setXRotate((GLfloat)sin(t / 1000.0) + 1.0f);
	setYRotate((GLfloat)cos(t / 1000.0) + 1.0f);
	MoveWindow(hWindow, (int)newX, (int)newY, width, height, FALSE);
	ModelWindow::doPaint();
	reallySwapBuffers();
	InvalidateRect(parentWindow->getHWindow(), &rect, TRUE);
}

BOOL SSModelWindow::showProgress(void)
{
	loading = TRUE;
	cancelLoad = FALSE;
	return TRUE;
}

void SSModelWindow::hideProgress(void)
{
	loading = FALSE;
}

int SSModelWindow::progressCallback(char* /*message*/, float /*progress*/)
{
	return 1;
}

void SSModelWindow::swapBuffers(void)
{
}

void SSModelWindow::reallySwapBuffers(void)
{
	ModelWindow::swapBuffers();
}

void SSModelWindow::showErrorsIfNeeded(BOOL /*onlyIfNeeded*/)
{
}

bool SSModelWindow::checkForExit(HWND hWnd, UINT message, WPARAM wParam,
								 LPARAM lParam)
{
	if ((message >= WM_MOUSEFIRST && message <= WM_MOUSELAST) ||
		(message >= WM_KEYFIRST && message <= WM_KEYLAST))
	{
		debugOut("SSModelWindow::checkForExit 0x%04X 0x%04X 0x%04X 0x%04X\n",
			hWnd, message, wParam, lParam);
		if (message == WM_MOUSEMOVE &&
			!((LDViewWindow*)parentWindow)->checkMouseMove(hWnd, lParam))
		{
			return false;
		}

		if (ssPassword->verifyPassword(hWindow))
		{
			return true;
			if (initializedGL && hglrc)
			{
				wglMakeCurrent(NULL, NULL);
				wglDeleteContext(hglrc);
				hglrc = NULL;
				initializedGL = FALSE;
			}
			((LDViewWindow*)parentWindow)->shutdown();
			ssPassword->release();
			ssPassword = NULL;
		}
		else
		{
			((LDViewWindow*)parentWindow)->initMouseMove();
		}
	}
	return false;
}

LRESULT SSModelWindow::windowProc(HWND hWnd, UINT message, WPARAM wParam,
								  LPARAM lParam)
{
	if (checkForExit(hWnd, message, wParam, lParam))
	{
/*
		if (initializedGL && hglrc)
		{
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(hglrc);
			hglrc = NULL;
			initializedGL = FALSE;
		}
*/
		((LDViewWindow*)parentWindow)->shutdown();
		return 1;
	}
	return CUIOGLWindow::windowProc(hWnd, message, wParam, lParam);
}

void SSModelWindow::closeWindow(void)
{
//	ShowWindow(hWindow, SW_HIDE);
	ModelWindow::closeWindow();
}
