#include <stdio.h>

#include "TCUserDefaults.h"
#include "TCStringArray.h"
#include "mystring.h"

TCUserDefaults* TCUserDefaults::currentUserDefaults = NULL;

TCUserDefaults::TCUserDefaultsCleanup TCUserDefaults::userDefaultsCleanup;

TCUserDefaults::TCUserDefaultsCleanup::~TCUserDefaultsCleanup(void)
{
	if (currentUserDefaults)
	{
		currentUserDefaults->release();
	}
}

TCUserDefaults::TCUserDefaults(void)
			   :appName(NULL),
			   commandLine(NULL)
#ifdef WIN32
			   ,hAppDefaultsKey(NULL)
#endif // WIN32
{
#ifdef _LEAK_DEBUG
	strcpy(className, "TCUserDefaults");
#endif
#ifdef _QT
	qSettings = new QSettings;
#endif // _QT
}

void TCUserDefaults::dealloc(void)
{
#ifdef _QT
	delete qSettings;
#endif // _QT
	delete appName;
	appName = NULL;
	if (commandLine)
	{
		commandLine->release();
		commandLine = NULL;
	}
#ifdef WIN32
	if (hAppDefaultsKey)
	{
		RegCloseKey(hAppDefaultsKey);
	}
#endif // WIN32
	TCObject::dealloc();
}

TCUserDefaults::~TCUserDefaults(void)
{
}

void TCUserDefaults::deleteResult(void* value)
{
	delete (char*)value;
}

TCUserDefaults* TCUserDefaults::getCurrentUserDefaults(void)
{
	if (!currentUserDefaults)
	{
		currentUserDefaults = new TCUserDefaults;
	}
	return currentUserDefaults;
}

TCStringArray* TCUserDefaults::getProcessedCommandLine(void)
{
	return getCurrentUserDefaults()->defGetProcessedCommandLine();
}

void TCUserDefaults::setCommandLine(char *argv[])
{
	TCStringArray *argArray = new TCStringArray;
	int i;

	for (i = 0; argv[i]; i++)
	{
		if (i > 0)
		{
			argArray->addString(argv[i]);
		}
	}
	getCurrentUserDefaults()->defSetCommandLine(argArray);
	argArray->release();
}

void TCUserDefaults::setCommandLine(const char *args)
{
	TCStringArray *argArray = new TCStringArray;
	const char *tmpString = args;
	char tmpBuf[1024];

	while (tmpString[0])
	{
		char *end = NULL;

		if (tmpString[0] == '"')
		{
			end = strchr(tmpString + 1, '"');
			tmpString++;
		}
		else
		{
			end = strchr(tmpString, ' ');
		}
		if (end)
		{
			int length = end - tmpString;

			strncpy(tmpBuf, tmpString, length);
			tmpBuf[length] = 0;
			argArray->addString(tmpBuf);
			tmpString += length + 1;
			if (end[0] == '"' && tmpString[0])
			{
				tmpString++;
			}
		}
		else
		{
			argArray->addString(tmpString);
			tmpString += strlen(tmpString);
		}
	}
	getCurrentUserDefaults()->defSetCommandLine(argArray);
	argArray->release();
}

void TCUserDefaults::setAppName(const char* value)
{
	getCurrentUserDefaults()->defSetAppName(value);
}

const char* TCUserDefaults::getAppName(void)
{
	return getCurrentUserDefaults()->defGetAppName();
}

void TCUserDefaults::setStringForKey(const char* value, const char* key)
{
	getCurrentUserDefaults()->defSetStringForKey(value, key);
}

char* TCUserDefaults::stringForKey(const char* key, const char* defaultValue)
{
	return getCurrentUserDefaults()->defStringForKey(key, defaultValue);
}

void TCUserDefaults::setLongForKey(long value, const char* key)
{
	getCurrentUserDefaults()->defSetLongForKey(value, key);
}

long TCUserDefaults::longForKey(const char* key, long defaultValue)
{
	return getCurrentUserDefaults()->defLongForKey(key, defaultValue);
}

void TCUserDefaults::removeValue(const char* key)
{
	getCurrentUserDefaults()->defRemoveValue(key);
}

void TCUserDefaults::defSetStringForKey(const char* value, const char* key)
{
#ifdef WIN32
	defSetValueForKey((LPBYTE)value, strlen(value) + 1, REG_SZ, key);
#else // WIN32
#ifdef _QT
	qSettings->writeEntry(qKeyForKey(key), value);
#endif // _QT
#endif // !WIN32
}

int TCUserDefaults::defCommandLineIndexForKey(const char *key)
{
	if (commandLine)
	{
		int i;
		int count = commandLine->getCount();
		int keyLength = strlen(key);
		char *keyEquals = new char[keyLength + 3];
		int returnValue = -1;

		strcpy(keyEquals, "-");
		strcat(keyEquals, key);
		strcat(keyEquals, "=");
		for (i = 0; i < count && returnValue == -1; i++)
		{
			char *arg = commandLine->stringAtIndex(i);

			if (stringHasCaseInsensitivePrefix(arg, keyEquals))
			{
				returnValue = i;
			}
		}
		delete keyEquals;
		return returnValue;
	}
	return -1;
}

char* TCUserDefaults::defCommandLineStringForKey(const char* key)
{
	if (commandLine)
	{
		int index = defCommandLineIndexForKey(key);

		if (index >= 0)
		{
			char *arg = commandLine->stringAtIndex(index);
			int keyLength = strlen(key);

			if (strlen(arg + keyLength + 1) > 0)
			{
				return copyString(arg + keyLength + 2);
			}
		}
	}
	return NULL;
}

char* TCUserDefaults::defStringForKey(const char* key, const char* defaultValue)
{
	char *commandLineValue = defCommandLineStringForKey(key);

	if (commandLineValue)
	{
		return commandLineValue;
	}
#ifdef WIN32
	DWORD size;
	LPBYTE value = defValueForKey(size, REG_SZ, key);

	if (value)
	{
		return (char*)value;
	}
	else
	{
		return copyString(defaultValue);
	}
#else // WIN32
#ifdef _QT
	QString string = qSettings->readEntry(qKeyForKey(key), defaultValue);

	if (string == QString::null)
	{
		return NULL;
	}
	else
	{
		char *returnValue = new char[string.length() + 1];

		strcpy(returnValue, string);
		return returnValue;
	}
#else // _QT
	return copyString(defaultValue);
#endif // !_QT
#endif // !WIN32
}

void TCUserDefaults::defSetLongForKey(long value, const char* key)
{
#ifdef WIN32
	defSetValueForKey((LPBYTE)&value, sizeof value, REG_DWORD, key);
#else // WIN32
#ifdef _QT
	qSettings->writeEntry(qKeyForKey(key), (int)value);
#endif // _QT
#endif // !WIN32
}

long TCUserDefaults::defLongForKey(const char* key, long defaultValue)
{
	char *commandLineValue = defCommandLineStringForKey(key);

	if (commandLineValue)
	{
		long returnValue;

		if (sscanf(commandLineValue, "%li", &returnValue) == 1)
		{
			delete commandLineValue;
			return returnValue;
		}
		delete commandLineValue;
	}
#ifdef _QT
	return qSettings->readNumEntry(qKeyForKey(key), defaultValue);
#endif // _QT
#ifdef WIN32
	DWORD size;
	LPBYTE value = defValueForKey(size, REG_DWORD, key);

	if (value)
	{
		long returnValue;

		memcpy(&returnValue, value, sizeof returnValue);
		delete value;
		return returnValue;
	}
	else
	{
		return defaultValue;
	}
#endif // WIN32
	return defaultValue;
}

void TCUserDefaults::defRemoveValue(const char* key)
{
#ifdef _QT
	qSettings->removeEntry(key);
#endif // _QT
#ifdef WIN32
	if (hAppDefaultsKey)
	{
		HKEY parentKey = hAppDefaultsKey;
		const char* spot;

		if ((spot = strrchr(key, '/')) != NULL)
		{
			int subKeyLength = spot - key;
			char* keyPath = strncpy(new char[subKeyLength + 1], key, subKeyLength);

			keyPath[subKeyLength] = 0;
			parentKey = openKeyPathUnderKey(parentKey, keyPath, 1);
			delete keyPath;
			spot++;
		}
		else
		{
			spot = key;
		}
		RegDeleteValue(parentKey, spot);
	}
#else // WIN32
#endif // WIN32
}

TCStringArray* TCUserDefaults::defGetProcessedCommandLine(void)
{
	return commandLine;
}

void TCUserDefaults::defSetCommandLine(TCStringArray *argArray)
{
	int i, j;
	int count = argArray->getCount();

	if (commandLine)
	{
		commandLine->release();
	}
	commandLine = new TCStringArray;
	for (i = 0, j = 0; i < count; i++)
	{
		char *arg = argArray->stringAtIndex(i);

		if (i > 0 && strcmp(arg, "=") == 0)
		{
			commandLine->appendString("=", j - 1);
		}
		else if (i > 0 && stringHasSuffix(commandLine->stringAtIndex(j - 1),
			"="))
		{
			commandLine->appendString(arg, j - 1);
		}
		else
		{
			commandLine->addString(arg);
			j++;
		}
	}
}

void TCUserDefaults::defSetAppName(const char* value)
{
	delete appName;
	appName = copyString(value);
#ifdef WIN32
	if (hAppDefaultsKey)
	{
		RegCloseKey(hAppDefaultsKey);
	}
	hAppDefaultsKey = openAppDefaultsKey();
#endif // WIN32
}

#ifdef WIN32

void TCUserDefaults::defSetValueForKey(const LPBYTE value, int length,
									   DWORD type, const char* key)
{
	int index = defCommandLineIndexForKey(key);

	if (index >= 0)
	{
		commandLine->removeString(index);
	}
	if (hAppDefaultsKey)
	{
		HKEY parentKey = hAppDefaultsKey;
		const char* spot;

		if ((spot = strrchr(key, '/')) != NULL)
		{
			int subKeyLength = spot - key;
			char* keyPath = strncpy(new char[subKeyLength + 1], key, subKeyLength);

			keyPath[subKeyLength] = 0;
			parentKey = openKeyPathUnderKey(parentKey, keyPath, 1);
			spot++;
			delete keyPath;
		}
		else
		{
			spot = key;
		}
		RegSetValueEx(parentKey, spot, 0, type, value, length);
	}
}

LPBYTE TCUserDefaults::defValueForKey(DWORD& size, DWORD type, const char* key)
{
	if (hAppDefaultsKey)
	{
		DWORD valueType;
		HKEY parentKey = hAppDefaultsKey;
		const char* spot;
		LPBYTE value = NULL;

		if ((spot = strrchr(key, '/')) != NULL)
		{
			int subKeyLength = spot - key;
			char* keyPath = strncpy(new char[subKeyLength + 1], key, subKeyLength);

			keyPath[subKeyLength] = 0;
			parentKey = openKeyPathUnderKey(parentKey, keyPath, 1);
			delete keyPath;
			spot++;
		}
		else
		{
			spot = key;
		}
		size = 0;
		if (RegQueryValueEx(parentKey, spot, 0, &valueType, NULL,
			&size) == ERROR_SUCCESS)
		{
			if (valueType == type)
			{
				value = new BYTE[size];
				if (RegQueryValueEx(parentKey, spot, 0, &valueType,
					value, &size) != ERROR_SUCCESS)
				{
					delete value;
					value = NULL;
				}
			}
		}
		if (parentKey != hAppDefaultsKey)
		{
			RegCloseKey(parentKey);
		}
		return value;
	}
	return NULL;
}

HKEY TCUserDefaults::openKeyPathUnderKey(HKEY parentKey, const char* keyPath,
										 int create)
{
	int count;
	char** components = componentsSeparatedByString(keyPath, "/", count);
	HKEY currentKey = parentKey;
	DWORD disposition;
	int i;
	BOOL failed = NO;

	for (i = 0; i < count && !failed; i++)
	{
		char* component = components[i];

		if (strlen(component) > 0)
		{
			HKEY newKey;

			if (create)
			{
				if (RegCreateKeyEx(currentKey, component, 0, NULL, 0,
					KEY_WRITE | KEY_READ, NULL, &newKey, &disposition) !=
					ERROR_SUCCESS)
				{
					failed = YES;
				}
			}
			else
			{
				if (RegOpenKeyEx(currentKey, component, 0, KEY_WRITE | KEY_READ,
					&newKey) != ERROR_SUCCESS)
				{
					failed = YES;
				}
			}
			if (currentKey != parentKey)
			{
				RegCloseKey(currentKey);
			}
			currentKey = newKey;
		}
	}
	deleteStringArray(components, count);
	if (failed)
	{
		return NULL;
	}
	else
	{
		return currentKey;
	}
}

HKEY TCUserDefaults::openAppDefaultsKey(void)
{
//	HKEY softwareKey;
//	DWORD disposition;
//	HKEY returnValue = NULL;
	char* keyPath = new char[strlen(appName) + 128];
	HKEY retValue;

	sprintf(keyPath, "Software/%s", appName);
	retValue = openKeyPathUnderKey(HKEY_CURRENT_USER, keyPath, 1);
	delete keyPath;
	return retValue;
/*
	if (hAppDefaultsKey)
	{
		RegCloseKey(hAppDefaultsKey);
	}
	if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software", 0, NULL, 0,
		0, NULL, &softwareKey, &disposition) == ERROR_SUCCESS)
	{
		HKEY appKey;

		if (RegCreateKeyEx(softwareKey, appName, 0, NULL, 0, KEY_ALL_ACCESS,
			NULL, &appKey, &disposition) == ERROR_SUCCESS)
		{
			if (RegCreateKeyEx(appKey, "UserDefaults", 0, NULL, 0,
				KEY_ALL_ACCESS, NULL, &returnValue, &disposition)
				!= ERROR_SUCCESS)
			{
				returnValue = NULL;
			}
			RegCloseKey(appKey);
		}
		RegCloseKey(softwareKey);
	}
	return returnValue;
*/
}

#endif // WIN32


#ifdef _QT

char *TCUserDefaults::qKeyForKey(const char *key)
{
	sprintf(qKey, "/%s/%s", appName, key);
	return qKey;
}

#endif // _QT
