#ifndef __TCUSERDEFAULTS_H__
#define __TCUSERDEFAULTS_H__

#include <TCFoundation/TCObject.h>

#ifdef WIN32
#include <windows.h>
#else // WIN32
#ifdef _QT
#include <qsettings.h>
#endif // _QT
#endif // !WIN32

class TCStringArray;

class TCExport TCUserDefaults: public TCObject
{
	public:
		TCUserDefaults(void);
		static void setStringForKey(const char* value,
			const char* key);
		static char* stringForKey(const char* key,
			const char* defaultValue = NULL);
		static void setLongForKey(long value, const char* key);
		static long longForKey(const char* key, long defaultValue = 0);
		static void removeValue(const char* key);
		static void setAppName(const char* value);
		static const char* getAppName(void);
		static void deleteResult(void* value);
		static void setCommandLine(char *argv[]);
		static void setCommandLine(const char *args);
		static TCStringArray* getProcessedCommandLine(void);
	protected:
		virtual void dealloc(void);
		virtual ~TCUserDefaults(void);
		static TCUserDefaults* getCurrentUserDefaults(void);
		void defSetStringForKey(const char* value, const char* key);
		char* defStringForKey(const char* key,
			const char* defaultValue = NULL);
		void defSetLongForKey(long value, const char* key);
		long defLongForKey(const char* key, long defaultValue = 0);
		void defRemoveValue(const char* key);
		void defSetAppName(const char* value);
		const char* defGetAppName(void) { return appName; }
		void defSetCommandLine(TCStringArray *argArray);
		char* defCommandLineStringForKey(const char* key);
		TCStringArray* defGetProcessedCommandLine(void);
		int defCommandLineIndexForKey(const char *key);

#ifdef WIN32
		HKEY openAppDefaultsKey(void);
		HKEY openKeyPathUnderKey(HKEY parentKey, const char* keyPath,
			int create = 0);
		void defSetValueForKey(const LPBYTE value, int length,
			DWORD type, const char* key);
		LPBYTE defValueForKey(DWORD& size, DWORD type,
			const char* key);

		HKEY hAppDefaultsKey;
#else // WIN32
#ifdef _QT
		QSettings *qSettings;
		char qKey[1024];

		char *qKeyForKey(const char *key);
#endif // _QT
#endif // !WIN32

		char* appName;
		static TCUserDefaults* currentUserDefaults;
		TCStringArray* commandLine;

		static class TCUserDefaultsCleanup
		{
		public:
			~TCUserDefaultsCleanup(void);
		} userDefaultsCleanup;
		friend TCUserDefaultsCleanup;
};

#endif // __TCUSERDEFAULTS_H__
